function [gamma,alpha]=gen_ang_nomin_tan_plane(s_func,s_par,dip_ind,dz_dhx,dz_dmx,gamma_0,alpha_0,i_gamma_probl,i_alpha_probl,flag)

% From the offset and midpoint dips, computes the PHASE aperture and dip angles
[n_gamma,n_alpha]=size(gamma_0);

for i_alpha=1:n_alpha
  for i_gamma=1:n_gamma
  
    % We first start with an initial guess (usually the offset and midpoint dips)
    gam_0=gamma_0(i_gamma,i_alpha);
    alp_0=alpha_0(i_gamma,i_alpha);
    
    if(dip_ind)
      gamma(i_gamma,i_alpha)=atand(dz_dhx(i_gamma,i_alpha));
      alpha(i_gamma,i_alpha)=atand(dz_dmx(i_gamma,i_alpha));
    
    else
      S_s=s_func(s_par,alp_0-gam_0,'phase','s');
      S_r=s_func(s_par,alp_0+gam_0,'phase','r');
      Delta=(S_r-S_s)/(S_r+S_s);
      
      % We compute the coefficients of the quadratic equation for gamma 
      A_gam=(dz_dmx(i_gamma,i_alpha)*Delta - dz_dhx(i_gamma,i_alpha)*Delta^2);
      B_gam=(1. -Delta^2);
      C_gam=dz_dmx(i_gamma,i_alpha)*Delta - dz_dhx(i_gamma,i_alpha);

      if (A_gam ~= 0.) 
        if (i_alpha ~= i_alpha_probl) 
          rad=(B_gam^2 - 4*A_gam*C_gam);
          if(rad >= 0)
            gamma(i_gamma,i_alpha) = atand((-B_gam  + sqrt(rad))/(2.*A_gam));
          else
            gamma(i_gamma,i_alpha) = atand(dz_dhx(i_gamma,i_alpha));
          end
        else
          gamma(i_gamma,i_alpha) = atand(dz_dhx(i_gamma,i_alpha));
        end
      else
        if (B_gam ~= 0.) 
          gamma(i_gamma,i_alpha) = atand(-C_gam/B_gam);
        else
          disp('Problems in finding gamma in gen_ang_nomin_tan_plane')
        end
      end

      % We compute the coefficients of the quadratic equation for alpha
      A_alp=(dz_dhx(i_gamma,i_alpha)*Delta - dz_dmx(i_gamma,i_alpha)*Delta^2);
      B_alp=(1. -Delta^2);
      C_alp=dz_dhx(i_gamma,i_alpha)*Delta - dz_dmx(i_gamma,i_alpha);

      if (A_alp ~= 0.) 
        if (i_gamma ~= i_gamma_probl) 
          rad=(B_alp^2 - 4*A_alp*C_alp);
          if(rad >= 0)
            alpha(i_gamma,i_alpha) = atand((-B_alp  + sqrt(rad))/(2.*A_alp));
          else
            alpha(i_gamma,i_alpha) = atand(dz_dmx(i_gamma,i_alpha));
          end
        else
          alpha(i_gamma,i_alpha) = atand(dz_dmx(i_gamma,i_alpha));
        end
      else
        if (B_alp ~= 0.) 
          alpha(i_gamma,i_alpha)=atand(-C_alp/B_alp);
        else
          disp('Problems in finding alpha in gen_ang_nomin_tan_plane')
        end
      end
    end
  end
end
