function [z_plot,gamma_plot]=mig_flat_event_new(true_aniso_mig_iso,use_default,plot_res,slowness_type,rock_type,rho_vmod,scale_pert,n_gamma,gamma_0,d_gamma,n_alpha,alpha_0,d_alpha)

% ----------------------------------------------------------
% The output are the following:
% 1. The RMO with use of the correct migration velocity
% 2. The theoretical RMO with the wrong migration velocity
% 3. The predicted RMO with the group angles
% 4. The predicted RMO with the phase angles
% ----------------------------------------------------------  

% Parameters
verb=0;
skip_mig_wrong=0;

n_iter=0; % Number of iterations to estimate the values of the phase
          % angles, solving the two quadratic equations

fix_rays=0; % Model assumptions (initially fix_rays=0)
aniso=1;

n_vel=4;

flag=-1;
color_surf='gray';
show_surf=0;
show_cig=1;
show_err=0;
show_points=1;

mygreen=[0,1,0];
myred=[1,0,0];
myblue=[0,0,1];
mycyan=[0,1,1];
mymagenta=[1,0,1];
myyellow=[1,1,0];

z_ref=.4;
mx_ref=0.;
tilt_ang=0;

% -------------------------------------------------------------------------------
% Definition of the velocity model
% -------------------------------------------------------------------------------

[gen_s_func,gen_s_der_func,gen_ds_drho,gen_phase_match_func,gen_phase_match_der_func]=slowness_func(slowness_type);

[name,v_mod,epsilon,delta]=rock(rock_type);

% -------------------------------------------------------------------------------
% Definition of the velocity parameters (right ones and wrong ones)
% -------------------------------------------------------------------------------

eta=(epsilon-delta)/(1.+2.*delta);

if (rho_vmod == 1) % If we use the right velocity, rho_vmod == 1
  if(true_aniso_mig_iso)
    delta_delta=-delta*scale_pert;
    delta_eta=-eta*scale_pert;
  else
    delta_delta=delta*scale_pert;
    delta_eta=eta*scale_pert;
    epsilon=0;
    delta=0;
    eta=0;
  end
else
    delta_delta=0.;
    delta_eta=0.;
end
  

a_v=1;

% If we don't input velocity parameters, then we have to enter them.

if(use_default ~= 1)
  fprintf('Default a_v =%f\n',a_v);
  a_v_inp=input('Vertical velocity mult? ');
  if(a_v_inp ~= ' ')
    a_v=a_v_inp;
  end
  
  if(ischar(rock_type) == 0)
    fprintf('Default epsilon =%f\n',epsilon);
    epsilon_inp=input('Epsilon? ');
    if(epsilon_inp ~= ' ')
      epsilon=epsilon_inp;
    end
    
    fprintf('Default delta =%f\n',delta);
    delta_inp=input('Delta ? ');
    if(delta_inp ~= ' ')
      delta=delta_inp;
    end
  end

  fprintf('Default tilt_ang =%f\n',tilt_ang);
  tilt_ang_inp=input('Tilt of symmetry axis in degrees? ');
  if(tilt_ang_inp ~= ' ')
    tilt_ang=tilt_ang_inp;
  end

end

if(verb ==1) % Display the parameters of the perturbed velocity model 
  fprintf('Running with: a_v =%f\n',a_v);
  fprintf('Running with: epsilon =%f\n',epsilon);
  fprintf('Running with: delta =%f\n',delta);
  fprintf('Running with: eta =%f\n',eta);
  
  fprintf('Running with: tilt_ang =%f\n',tilt_ang);
end

a_h=sqrt(1.+2.*epsilon)*a_v;
a_n=sqrt(1.+2.*delta)*a_v;

gen_s_par_inp(1)=1/(a_v*v_mod);
gen_s_par_inp(2)=1/(a_h*v_mod);
gen_s_par_inp(3)=tilt_ang;
gen_s_par_inp(4)=1/(a_n*v_mod);

if(use_default ~= 1) 
  rho_vv=rho_vmod;
  fprintf('Default rho_vmod =%f\n',rho_vmod);
  rho_vmod_inp=input('Scaling of slowness parameters ? ');
  if(rho_vmod_inp ~= ' ')
    rho_vmod=rho_vmod_inp;
  end
  
  fprintf('Default rho_vv =%f\n',rho_vv);
  rho_vv_inp=input('Scaling of vertical slowness parameters ? ');
  if(rho_vv_inp ~= ' ')
    rho_vv=rho_vv_inp;
  end

  fprintf('Default delta_delta =%f\n',delta_delta);
  delta_delta_inp=input('Delta perturbation ? ');
  if(delta_delta_inp ~= ' ')
    delta_delta=delta_delta_inp;
  end

  fprintf('Default delta_eta =%f\n',delta_eta);
  delta_eta_inp=input('Eta perturbation ? ');
  if(delta_eta_inp ~= ' ')
    delta_eta=delta_eta_inp;
  end
end

if (verb ==1)
  for i_s_par=1:n_s_par(2)
    fprintf('Running with: gen_s_par_inp(%d) =%f\n',i_s_par,gen_s_par_inp(i_s_par));
  end
  fprintf('Running with: rho_vmod =%f\n',rho_vmod);
  fprintf('Running with: rho_vv =%f\n',rho_vv);
  fprintf('Running with: delta_delta =%f\n',delta_delta);
  fprintf('Running with: delta_eta =%f\n',delta_eta);
  fprintf(' New delta =%f\n',delta+delta_delta);
  fprintf(' New eta =%f\n',eta+delta_eta);
end

rho_inp=[1,1,1];
rho_vv=rho_vmod;
rho_vn=sqrt((1 + 2*(delta+delta_delta))/(1 + 2*delta))*rho_vv;
rho_vh=sqrt((1 + 2*(eta+delta_eta))/(1 + 2*eta))*rho_vn;

rho=[rho_vv,rho_vh,rho_vn];

i_gamma_probl=(-gamma_0/d_gamma)+1;
i_alpha_probl=(-alpha_0/d_alpha)+1;

% ----------------------------------------------------------------------
% Creation of the matrix of the alpha and gamma values (GROUP angles)
% ----------------------------------------------------------------------

for i_alpha=1:n_alpha
  for i_gamma=1:n_gamma
    alpha(i_gamma,i_alpha)=alpha_0+(i_alpha-1)*d_alpha;
    gamma(i_gamma,i_alpha)=gamma_0+(i_gamma-1)*d_gamma;
%    fprintf('Group gamma: %4.3f\n',gamma(i_gamma,i_alpha));
  end
end

v_v_mod=1/gen_s_par_inp(1);
v_h_mod=1/gen_s_par_inp(2);
v_n_mod=1/gen_s_par_inp(4);

fprintf('Modeling with epsi=%3.3f, delta= %3.3f, eta= %3.3f, and v_v=%3.2f, v_h=%3.2f, v_n=%3.2f\n',epsilon,delta,eta,v_v_mod,v_h_mod,v_n_mod);

% ----------------------------------------------------------------------
% Modeling of the events with the right migration velocity
% ----------------------------------------------------------------------

% Computation of the migration impulse response
% Note that the input angles are GROUP angles (comes from the geometric
% interpretation of the rays)

% The prestack image is a function of t, m and h (the subsurface offset)...
disp('Modeling events')
[t_data,mx_data,hx_data]=gen_model_ang(gen_s_func,gen_s_par_inp,z_ref,mx_ref,gamma,alpha,flag);

% Plot of the impulse response 
if(show_surf)
disp('Show data event surf')

  [min_t_data,max_t_data,min_mx_data,max_mx_data,min_hx_data,max_hx_data]=max_min_zmx(t_data,mx_data,hx_data,flag);
  axis_data=[min_mx_data,max_mx_data,min_hx_data,max_hx_data,min_t_data,1.3*max_t_data]
  az=90;
  el=0;
  [hfig_data,haxis_data]=basic_mesh(mx_data,hx_data,t_data,axis_data,'Mx (km)','Hx (km)','Time (km)','Data',color_surf,1,0,az,el);
end



v_v_mig=1/gen_s_par_inp(1);
v_h_mig=1/gen_s_par_inp(2);
v_n_mig=1/gen_s_par_inp(4);

% -----------------------------------------------------------------------
% Computation of the new surface angles 
% -----------------------------------------------------------------------

% Change due to the change in the
% velocity models. When the two velocity models are the same, the
% difference in ang_in and ang_out is only due to the fact that the
% relationships to go from group to phase angles are only approximate)
% The angles we compute are still GROUP angles

[gamma_match_corr,alpha_match_corr]=match_surf_data(gen_s_func, ...
                                                  gen_s_der_func, ...
                                                  gen_phase_match_func, ...
                                                  gen_s_par_inp, ...
                                                  gen_s_par_inp,gamma, ...
                                                  alpha,aniso,fix_rays); % gen_phase_match_func
                                                                         % comes
                                                                         % from PHASE_MATCH_GEN.M

%for i_alpha=1:n_alpha
%  for i_gamma=1:n_gamma
%    fprintf('Delta gamma: %4.3f\n',gamma_match_corr(i_gamma,i_alpha)-gamma(i_gamma,i_alpha));
%  end
%end


% --------------------------------------------------------------------------
% Migration of the impulse response with the correct migration velocities
% --------------------------------------------------------------------------

% Note that the use of the angles gamma_match_corr and alpha_match_corr
% instead of gamma and alpha actually doesn't change the results for the
% right migration velocity 

fprintf('Kinematic migration with right velocities epsi=%5.4f, delta= %5.4f, eta= %5.4f, and v_v=%5.4f, v_h=%5.4f, v_n=%5.4f\n',epsilon,delta,eta,v_v_mig,v_h_mig,v_n_mig);

% Computation of the useful quantities for the transformation to the
% subsurface-offset domain (note that the required angles are GROUP angles)
disp('Migrating events')
[z_mig_corr,mx_mig_corr,hx_mig_corr]=gen_surf_ang(gen_s_func,gen_s_par_inp,t_data,mx_data,hx_data,gamma_match_corr,alpha_match_corr,flag);
%[z_mig_corr,mx_mig_corr,hx_mig_corr]=gen_surf_ang(gen_s_func,gen_s_par_inp,t_data,mx_data,hx_data,gamma,alpha,flag);

% Computation of the tangent planes of the migration response 
% Assuming that the rays are STATIONNARY (1 - in the function)
disp('Computing tangent planes of events')
[dz_dhx_corr,dz_dmx_corr]=gen_tan_plane_ang(gen_s_func,gen_s_der_func,gen_s_par_inp,t_data,mx_data,hx_data,gamma_match_corr,alpha_match_corr,1,flag);
%[dz_dhx_corr,dz_dmx_corr]=gen_tan_plane_ang(gen_s_func,gen_s_der_func,gen_s_par_inp,t_data,mx_data,hx_data,gamma,alpha,1,flag);

% Computation of the migrated image in the angle-domain (gamma_image is
% the PHASE aperture angle) 
[z_image,gamma_image]=image_cig(gen_s_func,gen_s_par_inp,z_mig_corr,mx_mig_corr,hx_mig_corr,dz_dhx_corr,dz_dmx_corr,n_iter,i_gamma_probl,i_alpha_probl,flag);
%[z_image,gamma_image]=image_cig(gen_s_func,gen_s_par_inp,z_mig_corr,mx_mig_corr,hx_mig_corr,dz_dhx_corr,dz_dmx_corr,n_iter,i_gamma_probl,i_alpha_probl,flag);

%for i_alpha=1:n_alpha
%  for i_gamma=1:n_gamma
%    fprintf('gamma: %4.3f\n',gamma_image(i_gamma,i_alpha));
%  end
%end

% ------------------------------------------------------------------
% Plot of the RMO curve with the correct migration velocities (cyan)
% ------------------------------------------------------------------

z_plot(:,1)=z_image;
gamma_plot(:,1)=gamma_image; % !!! PHASE aperture angle

% ------------------------------------------------------------------
% Computation of the residual moveout with wrong velocity model
% ------------------------------------------------------------------

% Definition of the wrong velocity parameters 
if(skip_mig_wrong ~= 1)
v_v_mig=rho_vv/gen_s_par_inp(1);
v_h_mig=rho_vh/gen_s_par_inp(2);
v_n_mig=rho_vn/gen_s_par_inp(4);

gen_s_par(1)=1./v_v_mig;
gen_s_par(2)=1./v_h_mig;
gen_s_par(4)=1./v_n_mig;

epsilon_mig=.5*((v_h_mig/v_v_mig)^2 -1);
delta_mig=.5*((v_n_mig/v_v_mig)^2 -1);
eta_mig=(epsilon_mig-delta_mig)/(1.+2.*delta_mig);


% --------------------------------------------------------------------------
% Migration of the synthetic data with the perturbed migration velocities
% --------------------------------------------------------------------------

fprintf('Kinematic migration with wrong velocities epsi=%5.4f, delta= %5.4f, eta= %5.4f, and v_v=%5.4f, v_h=%5.4f, v_n=%5.4f\n',epsilon_mig,delta_mig,eta_mig,v_v_mig,v_h_mig,v_n_mig);

% Computation of the new surface angles assuming that the rays are stationnary
[gamma_match,alpha_match]=match_surf_data(gen_s_func,gen_s_der_func,gen_phase_match_func,gen_s_par_inp,gen_s_par,gamma,alpha,aniso,fix_rays);

%for i_alpha=1:n_alpha
%  for i_gamma=1:n_gamma
%    fprintf('Delta gamma: %4.3f\n',gamma_match_corr(i_gamma,i_alpha)-gamma(i_gamma,i_alpha));
%  end
%end

% Kinematic migration of the data with the wrong velocity model
disp('Migrating events')
[z_mig,mx_mig,hx_mig]=gen_surf_ang(gen_s_func,gen_s_par,t_data,mx_data,hx_data,gamma_match,alpha_match,flag);
%[z_mig,mx_mig,hx_mig]=gen_surf_ang(gen_s_func,gen_s_par,t_data,mx_data,hx_data,gamma,alpha,flag);

% Computation of the required quantities to do the transformation to the
% angle-domain - Assuming that the rays are stationnary (1 parameter - in the
% function gen_tan_plane)
disp('Computing tangent planes of events')
[dz_dhx,dz_dmx]=gen_tan_plane_ang(gen_s_func,gen_s_der_func,gen_s_par,t_data,mx_data,hx_data,gamma_match,alpha_match,1,flag);
%[dz_dhx,dz_dmx]=gen_tan_plane_ang(gen_s_func,gen_s_der_func,gen_s_par,t_data,mx_data,hx_data,gamma,alpha,1,flag);

% Computation of the transformation from the susbsurface-offset domain to
% the angle-domain.
% Note that the output angle, gamma_image, is the PHASE aperture angle
[z_image,gamma_image]=image_cig(gen_s_func,gen_s_par,z_mig,mx_mig,hx_mig,dz_dhx,dz_dmx,n_iter,i_gamma_probl,i_alpha_probl,flag);
%[z_image,gamma_image]=image_cig(gen_s_func,gen_s_par,z_mig,mx_mig,hx_mig,dz_dhx,dz_dmx,n_iter,i_gamma_probl,i_alpha_probl,flag);

% ----------------------------------------------------------------------
% Plot of the RMO curve with the perturbed migration velocities (green)
% ----------------------------------------------------------------------
z_plot(:,2)=z_image; % RMO obtained by kinematic migration 
gamma_plot(:,2)=gamma_image; % Phase aperture angle

[min_z_plot,max_z_plot,min_gamma_plot,max_gamma_plot,min_gamma_plot,max_gamma_plot]=max_min_zmx(z_plot(:,2),gamma_plot(:,2),gamma_plot(:,2),flag);

else
z_plot(:,2)=z_image;
gamma_plot(:,2)=gamma_image;
min_z_plot=.2;
max_z_plot=.5;
min_gamma_plot=-70;
max_gamma_plot=70;
end

axis_xy=[min_gamma_plot,max_gamma_plot,min_z_plot,max_z_plot];

if (show_points ==1 )
  color='gx';
else
  color='g';
end

[stitle,err]=sprintf('%s - epsi=%3.3f, delta= %3.3f and rho_v=%3.2f, rho_h=%3.2f, rho_n=%3.2f',name,epsilon,delta,rho_vv,rho_vh,rho_vn);

[hfig,haxis,hplot]=basic_plot(gamma_plot(:,2),z_plot(:,2),'Aperture Angle (deg)','Depth (km)',stitle,axis_xy,color,0,1,plot_res);

if (show_points ==1 )
  color='cx';
else
  color='c';
end
[hfig,haxis,hplot]=basic_plot(gamma_plot(:,1),z_plot(:,1),'Aperture Angle (deg)','Depth (km)',stitle,axis_xy,color,1,1,plot_res);

% ---------------------------------------------------------------------------
% Computation of the RMO curve (given the equation / using the group angles)
% ---------------------------------------------------------------------------

fprintf('Residual migration with rho_v_v=%4.3f, rho_v_h= %4.3f, rho_v_n= %4.3f\n',rho(1),rho(2),rho(3));

approx=0; % Implies that we actually use the group angles to compute the
          % RMO curve

% The input angles (gamma_match_corr and alpha_match_corr) are GROUP
% angles. Note that the use of gamma_match_corr or gamma_match doesn't
% have any large impact on the predicted RMO curve. 
[z_image,mx_image,gamma_image,alpha_image]=analy_rmo_aniso_bend(gen_s_func,gen_ds_drho,gen_phase_match_der_func,gen_s_par_inp,rho_inp,rho,z_ref,mx_ref,dz_dhx_corr,dz_dmx_corr,gamma_match_corr,alpha_match_corr,approx,flag);
%[z_image,mx_image,gamma_image,alpha_image]=analy_rmo_aniso_bend(gen_s_func,gen_ds_drho,gen_phase_match_der_func,gen_s_par_inp,rho_inp,rho,z_ref,mx_ref,dz_dhx,dz_dmx,gamma_match,alpha_match,approx,flag);


% ----------------------------------------------------------------------
% Plot of the predicted RMO curve / using the group angles (blue)
% ----------------------------------------------------------------------
z_plot(:,3)=z_image; % Depth of the event when using the equation (group
                     % angles are accurately estimated)
gamma_plot(:,3)=gamma_image; % Gamma is the PHASE aperture angle

if (show_points ==1 )
  color='bx';
else
  color='b';
end

[hfig,haxis,hplot]=basic_plot(gamma_plot(:,3),z_plot(:,3),'Aperture Angle (deg)','Depth (km)',stitle,axis_xy,color,1,1,plot_res);

% ---------------------------------------------------------------------------
% Computation of the RMO curve (given the equation / using the phase angles)
% ---------------------------------------------------------------------------
fprintf('Approx Residual migration with rho_v_v=%4.3f, rho_v_h= %4.3f, rho_v_n= %4.3f\n',rho(1),rho(2),rho(3));

approx=1; % Implies that we actually use the phase angles to estimate the
          % group angles

% The input angles (gamma_match_corr and alpha_match_corr) are GROUP
% angles. Note that the use of gamma_match_corr or gamma_match doesn't
% have any large impact on the predicted RMO curve. 
[z_image,mx_image,gamma_image,alpha_image]=analy_rmo_aniso_bend(gen_s_func,gen_ds_drho,gen_phase_match_der_func,gen_s_par_inp,rho_inp,rho,z_ref,mx_ref,dz_dhx_corr,dz_dmx_corr,gamma_match_corr,alpha_match_corr,approx,flag);
%[z_image,mx_image,gamma_image,alpha_image]=analy_rmo_aniso_bend(gen_s_func,gen_ds_drho,gen_phase_match_der_func,gen_s_par_inp,rho_inp,rho,z_ref,mx_ref,dz_dhx,dz_dmx,gamma_match,alpha_match,approx,flag);


% ----------------------------------------------------------------------
% Plot of the predicted RMO curve / using the phase angles (red)
% ----------------------------------------------------------------------

z_plot(:,4)=z_image; % Depth of the event in ADCIGs when using the
                     % equation (group angles are approximated with PHASE
                     % angles)
gamma_plot(:,4)=gamma_image; % Phase APERTURE angle

if (show_points ==1 )
  color='rx';
else
  color='r';
end

[hfig,haxis,hplot]=basic_plot(gamma_plot(:,4),z_plot(:,4),'Aperture Angle (deg)','Depth (km)',stitle,axis_xy,color,1,1,plot_res);

return

%
%[z_image,mx_image]=analy_rmo(gen_s_func,gen_s_par,rho_vh*rho_vmod,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);


for i_vel=1:n_vel

switch (i_vel)
  case (1)
%    disp('Migration with correct velocities')
    figure;
    if (show_points ==1 )
      color='co';
    else
      color='c';
    end
    gen_s_par=gen_s_par_inp;
    gamma_match=gamma;
    alpha_match=alpha;
  case (2)
%    disp('Migration with wrong velocities')
    hold on;
    if (show_points ==1 )
      color='cx';
    else
      color='c';
    end
%    gen_s_par=gen_s_par_inp;
    gen_s_par(1)=(1./(rho_vv))*gen_s_par_inp(1);
    gen_s_par(2)=(1./(rho_vh))*gen_s_par_inp(2);
    gen_s_par(4)=(1./(rho_vn))*gen_s_par_inp(4);
%    gen_s_par=rho_vmod*gen_s_par_inp;
%  disp('Matching surface data')
  [gamma_match,alpha_match]=match_surf_data(gen_s_func,gen_s_der_func,gen_phase_match_func,gen_s_par_inp,gen_s_par,gamma,alpha,aniso,fix_rays);

%min(gamma_match)
%max(gamma_match)

%[gamma_match_min,gamma_match_max,gamma_min,gamma_max,j5,j6]=max_min_zmx(gamma_match,gamma,gamma,flag)
  case (3)
    disp('Anlytical residual moveout (aniso) (gren)')
    hold on;
    if (show_points ==1 )
      color='gv';
    else
      color='g';
    end
    gamma_match=gamma;
    alpha_match=alpha;
    gamma_image=gamma_image_corr;
  case (4)
%    disp('Analytical moveout with wrong velocities')
    disp('Anlytical residual moveout (no aniso) (blue)')
    hold on;
    if (show_points ==1 )
      color='b*';
    else
      color='b';
    end
    gamma_match=gamma;
    alpha_match=alpha;
   gamma_image=gamma_image_corr;
%    gamma_image=gamma_match;
  case (5)
%    disp('Analytical moveout with wrong velocities')
    disp('Anlytical residual moveout (no aniso) (blue)')
    hold on;
    if (show_points ==1 )
      color='c*';
    else
      color='c';
    end
    gen_s_par(1)=(1./(rho_vv))*gen_s_par_inp(1);
    gen_s_par(2)=(1./(rho_vh))*gen_s_par_inp(2);
    gen_s_par(4)=(1./(rho_vn))*gen_s_par_inp(4);
  otherwise 
    disp('wrong i_vel')
end

if((i_vel == 1) | (i_vel ==2))


%disp('Migrating events')
[z_mig,mx_mig,hx_mig]=gen_surf_ang(gen_s_func,gen_s_par,t_data,mx_data,hx_data,gamma_match,alpha_match,flag);

[min_z_mig,max_z_mig,min_mx_mig,max_mx_mig,min_hx_mig,max_hx_mig]=max_min_zmx(z_mig,mx_mig,hx_mig,flag);

%fprintf('min_hx_mig =%f, max_hx_mig =%f \n',min_hx_mig,max_hx_mig);


%disp('Compting tangent planes of events')
[dz_dhx,dz_dmx]=gen_tan_plane_ang(gen_s_func,gen_s_der_func,gen_s_par,t_data,mx_data,hx_data,gamma_match,alpha_match,1,flag);

[min_dz_dhx,max_dz_dhx,min_mx_mig,max_mx_mig,min_hx_mig,max_hx_mig]=max_min_zmx(dz_dhx,mx_mig,hx_mig,flag);

%fprintf('min_dz_dhx =%f, max_dz_dhx =%f \n',min_dz_dhx,max_dz_dhx);
%fprintf('min_dz_dmx =%f, max_dz_dmx =%f \n',min_dz_dmx,max_dz_dmx);

%disp('Imaging cigs')

[z_image,gamma_image]=image_cig(gen_s_func,gen_s_par,z_mig,mx_mig,hx_mig,dz_dhx,dz_dmx,n_iter,i_gamma_probl,i_alpha_probl,flag);

%if(i_vel ==1)
%    figure;
%%    p1=plot(gamma,hx_mig,'c', 'LineWidth',[.1]);
%%    hold on;

%    p1=plot(gamma_match,z_mig,'r', 'LineWidth',[.1]);
%    hold on;
%    p1=plot(gamma_match,z_image,'b', 'LineWidth',[.1]);

%%    figure ;
%%    p1=plot(gamma_match,dz_dhx,'g', 'LineWidth',[.1]);
%end


switch (i_vel)
  case (1)
    dz_dhx_corr=dz_dhx;
    dz_dmx_corr=dz_dmx;
    gamma_image_corr=gamma_image;
    z_image_corr=z_image;
   [min_z,max_z,min_gamma,max_gamma,min_gamma_image,max_gamma_image]=max_min_zmx(z_image,gamma,gamma_image,flag);
  case (2)

%figure
%p1m=plot(gamma_match,z_mig,'g', 'LineWidth',[.1]);
%hold on
%p1m=plot(gamma_match,hx_mig,'r', 'LineWidth',[.1]);
%hold on
%p1m=plot(gamma_match,dz_dhx,'b', 'LineWidth',[.1]);

    z_image_rho=z_image;
    mx_image_rho=mx_mig;

    gamma_image_rho=gamma_image;
    alpha_image_rho=gamma_image*0.;

    gamma_match_rho=gamma_match;
    alpha_match_rho=gamma_match*0.;

   [min_z,max_z,min_gamma,max_gamma,min_gamma_image,max_gamma_image]=max_min_zmx(z_image_rho,gamma,gamma_image,flag);
%    dz_dhx_corr=dz_dhx;
%    dz_dmx_corr=dz_dmx;
%    paxis=gca;
%    set(paxis,'YDir','reverse');
  otherwise 
    disp('wrong i_vel')
end

else

switch (i_vel)
  case (3)
%[z_image,mx_image]=analy_rmo_iso_bend(gen_s_func,gen_s_par,rho_vh*rho_vmod,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);

[z_image,mx_image,gamma_pert,alpha_pert]=analy_rmo_iso_bend(gen_s_func,gen_phase_match_der_func,gen_s_par_inp,rho_inp(1),rho_vh,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,z_ref,flag);

%  if(show_err)
%    figure
%%    z_image=100*(z_image-z_image_rho)/z_image_rho(i_gamma_probl);
%    z_image=100*(z_image-z_image_rho)/z_image_rho;
%  end
  case (4)
%[z_image,mx_image]=analy_moveout(gen_s_func,gen_s_par,rho_vmod,z_image_corr,mx_mig,dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);
%[z_image,mx_image]=analy_rmo(gen_s_func,gen_s_par,rho_vh*rho_vmod,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);

%[z_image,mx_image,gamma_pert,alpha_pert]=analy_rmo_bend(gen_s_func,gen_s_par,rho_vh*rho_vmod,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);

[z_image,mx_image,gamma_pert,alpha_pert]=analy_rmo_aniso_bend(gen_s_func,gen_ds_drho,gen_phase_match_der_func,gen_s_par_inp,rho_inp,rho,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,z_ref,flag);

gamma_image=gamma_pert;
%
%[z_image,mx_image]=analy_rmo(gen_s_func,gen_s_par,rho_vh*rho_vmod,z_image_rho(i_gamma_probl),mx_mig(i_gamma_probl),dz_dhx_corr,dz_dmx_corr,gamma_match,alpha_match,flag);


%  if(show_err)
%    figure
%  if(show_err)
%%    z_image=100*(z_image-z_image_rho)/z_image_rho(i_gamma_probl);
%    z_image=100*(z_image-z_image_rho)/z_image_rho;
%  end
  case (5)

[z_image,mx_image,gamma_pert,alpha_pert]=analy_moveout_aniso_bend(gen_s_func,gen_phase_match_der_func,gen_s_par,rho_vh,z_image_rho,mx_image_rho,gamma_image_rho,alpha_image_rho,gamma_match_rho,alpha_match_rho,z_ref,i_gamma_probl,i_alpha_probl,flag);

gamma_image=gamma_image_rho;

  otherwise 
    disp('wrong i_vel')
end

end

%if ((show_cig) & (i_vel ~= 1))
%if ((show_cig) & (i_vel ~= 1) & (i_vel ~=3) & (i_vel ~=5))
if ((show_cig) & (i_vel ~= 1) &  (i_vel ~=5))
%[gamma_image_min,gamma_image_max,gamma_min,gamma_max,j5,j6]=max_min_zmx(gamma_image,gamma,gamma,flag)
  p1=plot(gamma_image,z_image,color, 'LineWidth',[.1]);
  paxis=gca;
  set(paxis,'YDir','reverse');
%  axis([min_gamma_image,max_gamma_image,min(1,min_z),max(1,max_z)])
  grid on
  title(paxis,stitle);

end
end
