module filter_data_mod
  use gendown3d_types
  use data_types
  use down_process_types

  implicit none
  real, pointer, dimension (:), private :: time_filter
contains
  subroutine filter_data(i_block_w,gen_par,data_par,down_par,data_block)
    logical filter_time
    logical high_bound,low_bound
    integer i_block_w

    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par

    complex, dimension (:,:,:,:,:,:) :: data_block

    low_bound =  (down_par%w_min_1_dis .gt. down_par%w_min_block) .or. &
    ((down_par%w_min_1_dis .ge. down_par%w_min_block) .and. &
    (down_par%w_min_1_dis .eq. down_par%w_min_0_dis))
    high_bound = (down_par%w_max_1_dis .lt. down_par%w_max_block) .or. &
    ((down_par%w_max_1_dis .le. down_par%w_max_block) .and. &
    (down_par%w_max_1_dis .eq. down_par%w_max_0_dis))
    filter_time= low_bound .or. high_bound

    down_par%i_w_min=1
    down_par%i_w_max=gen_par%n_w_block
    if(gen_par%debug2) write(0,*)'beginning i_w_min=',down_par%i_w_min
    if(gen_par%debug2) write(0,*)'beginning i_w_max=',down_par%i_w_max
    if(filter_time) then
       if(gen_par%debug1) write(0,*)'apply time filtering'
       call filter_data_time(i_block_w,gen_par,data_par,down_par,data_block)
    else
       if(gen_par%rho_filt) then
          if(gen_par%debug1) write(0,*)'apply time filtering (rho)'
          call filter_data_time(i_block_w,gen_par,data_par,down_par,data_block)
       else
          if(gen_par%debug1) write(0,*)'NOT apply time filtering'
       end if
    end if

    down_par%n_w_extrap_act=down_par%n_w_extrap_act+ &
    ((down_par%i_w_max-down_par%i_w_min)+1)*1
    if(gen_par%debug1) write(0,*)'n_w_extrap_act=',down_par%n_w_extrap_act
    return
  end subroutine filter_data

  subroutine filter_data_time(i_block_w,gen_par,data_par,down_par,data_block)
    integer i_block_w
    integer i_w,i_w_block_0,i_w_block_par,i_w_block_ser
    logical low_zero,high_zero,zero_filter
    real filter

    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par

    complex, dimension (:,:,:,:,:,:) :: data_block

    !	time filter data
    down_par%i_w_min=1
    down_par%i_w_max=gen_par%n_w_block

    zero_filter=.true.
    low_zero=((down_par%w_min_0 .ge. down_par%w_min_block).and. &
    (down_par%w_min_0 .le. down_par%w_max_block))
    high_zero=.true.
!$omp parallel
!$omp do
    do i_w_block_par=1,down_par%n_w_block_par
       !{
       do i_w_block_ser=1,down_par%n_w_block_ser
          !{
          i_w=down_par%n_w_block_par*(i_w_block_ser-1) + i_w_block_par
          i_w_block_0=(i_block_w-1)*gen_par%n_w_block+ i_w

          filter=time_filter(i_w_block_0)
          if (filter .eq. 0.) then
             zero_filter=.true.
          else
             zero_filter=.false.
          end if

          if (zero_filter) then
             if (low_zero) then
                down_par%i_w_min=i_w+1
             else
                if(high_zero) then
                   down_par%i_w_max=i_w-1
                   high_zero=.false.
                end if
             end if
          else
             low_zero=.false.
          end if

          data_block(:,:,:,:,i_w_block_ser,i_w_block_par)= &
          data_block(:,:,:,:,i_w_block_ser,i_w_block_par)*filter
          !}
       end do
       !}
    end do
!$omp end parallel
    return

  end subroutine filter_data_time

  subroutine design_time_filter(gen_par,data_par,down_par)
    implicit none

    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par

    integer ierr
    integer i_w_mig,i_w_min_0,i_w_min_1,i_w_max_0,i_w_max_1
    real dw_min,dw_max,nint_w_min_0,rho_filt_w,max_data_w

    ! first allocate vector of weights
    allocate (time_filter(data_par%n_w),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of time_filter")

    nint_w_min_0=nint(down_par%w_min_0/data_par%d_w)*data_par%d_w

    i_w_min_0=nint(((down_par%w_min_0-nint_w_min_0)/data_par%d_w+1))
    i_w_min_1=nint(((down_par%w_min_1-nint_w_min_0)/data_par%d_w+1))
    i_w_max_0=nint(((down_par%w_max_0-nint_w_min_0)/data_par%d_w+1))
    i_w_max_1=nint(((down_par%w_max_1-nint_w_min_0)/data_par%d_w+1))

    down_par%w_min_0_dis=nint_w_min_0+(i_w_min_0-1)*data_par%d_w
    down_par%w_min_1_dis=nint_w_min_0+(i_w_min_1-1)*data_par%d_w
    down_par%w_max_0_dis=nint_w_min_0+(i_w_max_0-1)*data_par%d_w
    down_par%w_max_1_dis=nint_w_min_0+(i_w_max_1-1)*data_par%d_w

    if(i_w_min_1 .ne. i_w_min_0) dw_min=1./(i_w_min_1-i_w_min_0)
    if(i_w_max_0 .ne. i_w_max_1) dw_max=1./(i_w_max_0-i_w_max_1)

    if((down_par%w_min_0 < data_par%w_0) .or. &
    (down_par%w_min_1 < data_par%w_0) .or. &
    (down_par%w_max_0 < data_par%w_0) .or. &
    (down_par%w_max_0 < data_par%w_0)) then
       write(0,*)'data_par%w_0 =',data_par%w_0 
       write(0,*)'down_par%w_min_0 =',down_par%w_min_0 
       write(0,*)'down_par%w_min_1 =',down_par%w_min_1 
       write(0,*)'down_par%w_max_0 =',down_par%w_max_0 
       write(0,*)'down_par%w_max_1 =',down_par%w_max_1 
       call seperr('Frequency for filtering too low')
    end if

    max_data_w= data_par%w_0 + (data_par%n_w-1)*data_par%d_w
    if((down_par%w_min_0 > max_data_w) .or. &
       (down_par%w_min_1 > max_data_w) .or. &
       (down_par%w_max_1 > max_data_w) .or. &
       (down_par%w_max_0 > max_data_w)) then
       write(0,*)'max_data_w =',max_data_w
       write(0,*)'down_par%w_min_0 =',down_par%w_min_0 
       write(0,*)'down_par%w_min_1 =',down_par%w_min_1 
       write(0,*)'down_par%w_max_0 =',down_par%w_max_0 
       write(0,*)'down_par%w_max_1 =',down_par%w_max_1 
       call seperr('Frequency for filtering too high')
    end if
    rho_filt_w=1.
    if((i_w_min_0 .eq. 1).and.(i_w_min_1 .eq. 1)) then
       time_filter(1)=1.
    else
       do i_w_mig=1,i_w_min_0,1
          time_filter(i_w_mig)=0.
       end do
    end if

    do i_w_mig=max(1,i_w_min_0+1),i_w_min_1,1
       if(gen_par%rho_filt) then
          rho_filt_w=down_par%w_min_0_dis+(i_w_mig-1)*data_par%d_w
       end if
       time_filter(i_w_mig)=(i_w_mig-i_w_min_0)*dw_min*rho_filt_w
    end do

    do i_w_mig=max(1,i_w_min_1+1),i_w_max_1,1
       if(gen_par%rho_filt) then
          rho_filt_w=down_par%w_min_0_dis+(i_w_mig-1)*data_par%d_w
       end if
       time_filter(i_w_mig)=rho_filt_w
    end do

    do i_w_mig=max(1,i_w_max_1+1),i_w_max_0,1
       if(gen_par%rho_filt) then
          rho_filt_w=down_par%w_min_0_dis+(i_w_mig-1)*data_par%d_w
       end if
       time_filter(i_w_mig)=(i_w_max_0-i_w_mig)*dw_max*rho_filt_w
    end do

    do i_w_mig=max(1,i_w_max_0),data_par%n_w,1
       time_filter(i_w_mig)=0.
    end do

    if(gen_par%verb) then
       if(gen_par%debug1) write(0,*)'Total number of frequency slices=',(i_w_max_0-i_w_min_0)-1
       do i_w_mig=1,data_par%n_w
          if(gen_par%debug1) write(0,*)'time_filter(',i_w_mig,')=',time_filter(i_w_mig)
       end do
    end if

    down_par%n_w_extrap_act=0
    return
  end subroutine design_time_filter

  subroutine clean_time_filter()
    integer ierr
    if(associated(time_filter)) then
       deallocate (time_filter,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating time_filter')
    end if
    return
  end subroutine clean_time_filter
end module filter_data_mod











