module Pre2D_salt_ffdpi_mod

  use Pre2D_salt_ffdpi_types

  use gendown3d_types
  use down_process_types
  use data_types
  use image_types
  use slow_types
  use fft_types

  use fft_data_mod

! Boundary condition module
use bound_mod

  implicit none

  type (Pre2D_salt_ffdpi_param_type), private :: Pre2D_salt_ffdpi_par_loc
  type (fft_type), private :: fft_3d_id
  logical, private :: time_mig
  logical, private,save :: apply_any_bound
  integer, private :: n_interp,n_comp_slow,n_total_slow_ref,nz_total
  real, private :: slow_const,slow_salt,epsi,min_prop_cos_sq,log_min_ratio
  real, private :: opt_ang_cos,opt_ang_sin,opt_ang_sin_sq,max_acc_sin
  real, private :: opt_azim_cos_sq,opt_azim_sin_sq
  real, private :: d_k_norm,max_disp_rel,adj_slow
  real, allocatable, dimension (:), private :: magn,phase
  real, allocatable, dimension (:), private :: k_mx,k_my,k_hx
  real, allocatable, dimension (:,:,:,:), private :: k_sr_sq
   
!  integer, external ::  omp_get_thread_num

contains

  subroutine init_Pre2D_salt_ffdpi_par(Pre2D_salt_ffdpi_par)
    type (Pre2D_salt_ffdpi_param_type) Pre2D_salt_ffdpi_par

    Pre2D_salt_ffdpi_par_loc=Pre2D_salt_ffdpi_par

  end subroutine init_Pre2D_salt_ffdpi_par

  function Pre2D_salt_ffdpi_init(gen_par,data_par,slow_par)  result(info)
    type (gendown3d_info_type) info
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par

    integer i_k_mx,i_k_my,i_k_hx,ierr,i_k_norm
    real kmx,kmy,khx,pi2,k_norm_sq,k_norm
    real n_cycles_ksq,k_start,k_cutoff,k_decay,sigma_magn_sq,edge_gauss,gauss_magn
    real delta_k
    real k_start_lin,k_norm_sq_start_lin,kk0

    epsi=epsilon(1.)

!    info=info_default

     apply_any_bound=gen_par%apply_bound

    if(Pre2D_salt_ffdpi_par_loc%ffdpi_order >= 1) then
      if(data_par%d_mx /= data_par%d_hx) then
        call seperr('For the momrnt d_mx must be equal to d_hx')
      end if
    end if

    if(slow_par%n_comp_slow /= 2) then
       write(0,*)'Reset n_comp_slow to 2'
       slow_par%n_comp_slow = 2
    end if
    n_comp_slow=slow_par%n_comp_slow


    select case (Pre2D_salt_ffdpi_par_loc%ref_slow(1:3)) 
    case ('min')
      adj_slow=1.
    case ('max')
      adj_slow=0.
      call seperr(" I think ref_slow should be min for ffd")
    case ('avg')
      adj_slow=.5
      call seperr(" I think ref_slow should be min for ffd")
    case default
      call seperr('Not supported reference_slowness ')
    end select

    slow_const=gen_par%slow_const
    pi2=2.*acos(-1.)

    n_total_slow_ref=0
    nz_total=0
write(0,*)'max_prop_angle=',gen_par%max_propag_angle
write(0,*)'max_prop_angle (local)=',Pre2D_salt_ffdpi_par_loc%max_acc_propag_angle
write(0,*)'pi2=',pi2
    if(Pre2D_salt_ffdpi_par_loc%max_acc_propag_angle > 0) then
      max_acc_sin=Pre2D_salt_ffdpi_par_loc%max_acc_propag_angle*pi2/360.
write(0,*)'#1 max_ac_sin=',max_acc_sin
      max_acc_sin=sin(max_acc_sin)
    else
      max_acc_sin=-1.
    end if
write(0,*)'max_ac_sin=',max_acc_sin

    if(Pre2D_salt_ffdpi_par_loc%min_delta_slow > 0.) then
      write(0,*)'ATTENTION: the use of min_delta_slow is deprecated'
      write(0,*)'Use max_acc_propag_angle'
      if(Pre2D_salt_ffdpi_par_loc%min_delta_slow >= 1) then
        call seperr('min_delta_slow must be < 1')
      end if
      max_acc_sin=1.-Pre2D_salt_ffdpi_par_loc%min_delta_slow
    else
      if(max_acc_sin < 0.) then
        call seperr('You need to set either max_acc_propag_angle or min_delta_slow')
      end if
    end if
    write(0,*)'#1 Max ratio=',1./max_acc_sin
    log_min_ratio=log(1./max_acc_sin)
    write(0,*)'Max ratio=',1./max_acc_sin
  

write(0,*)'max_prop_angle=',gen_par%max_propag_angle
write(0,*)'pi2=',pi2
    min_prop_cos_sq=cos(gen_par%max_propag_angle*pi2/360.)**2
write(0,*)'min_prop_cos_sq=',min_prop_cos_sq
    opt_ang_cos=cos(Pre2D_salt_ffdpi_par_loc%opt_ang*pi2/360.)

    opt_ang_sin=sin(Pre2D_salt_ffdpi_par_loc%opt_ang*pi2/360.)
    opt_ang_sin_sq=opt_ang_sin**2

   write(0,*)'opt_ang_cos=',opt_ang_cos
   write(0,*)'opt_ang_sin_sq=',opt_ang_sin_sq

    opt_azim_cos_sq=cos(Pre2D_salt_ffdpi_par_loc%opt_azim*pi2/360.)**2
    opt_azim_sin_sq=sin(Pre2D_salt_ffdpi_par_loc%opt_azim*pi2/360.)**2
    
   write(0,*)'opt_azim_cos_sq=',opt_azim_cos_sq
   write(0,*)'opt_azim_sin_sq=',opt_azim_sin_sq
    
    k_start_lin=Pre2D_salt_ffdpi_par_loc%k_start_lin
    if(k_start_lin > 0.) then
      k_norm_sq_start_lin=(1.-k_start_lin**2)
    else 
      k_norm_sq_start_lin=-huge(k_norm_sq_start_lin)
    end if
    
    if(opt_ang_sin > max_acc_sin) then
      write(0,*)'opt_ang=',Pre2D_salt_ffdpi_par_loc%opt_ang
      write(0,*)'Effective max_acc_propag_angle=',(360/pi2)*asin(max_acc_sin)
      call seperr ('opt_ang must be < max_acc_propag_angle')
    end if

    if(Pre2D_salt_ffdpi_par_loc%n_slow_ref < 2) then
      n_interp=1
      write(0,*)'ATT: Set n_interp =1'
    else
      n_interp=Pre2D_salt_ffdpi_par_loc%n_interp
      write(0,*)'ATT: Set n_interp =',n_interp
    end if
  
    if (slow_par%slowness) then
       if((Pre2D_salt_ffdpi_par_loc%slow_salt > 0.).and.(slow_const <= 0.)) then
          slow_salt=Pre2D_salt_ffdpi_par_loc%range_salt* &
                    (Pre2D_salt_ffdpi_par_loc%slow_salt)
          if(Pre2D_salt_ffdpi_par_loc%n_slow_ref < 2) call seperr('n_slow_ref <2')
       else 
          slow_salt=-1.
       end if
    else
       if((Pre2D_salt_ffdpi_par_loc%vel_salt > 0.).and.(slow_const <= 0.)) then
          slow_salt=Pre2D_salt_ffdpi_par_loc%range_salt* &
                    (1./Pre2D_salt_ffdpi_par_loc%vel_salt)
          if(Pre2D_salt_ffdpi_par_loc%n_slow_ref < 2) call seperr('n_slow_ref <2')
       else 
          slow_salt=-1.
       end if
    end if

    if(gen_par%debug1) write(0,*)'Init slow_salt=',slow_salt
    if(gen_par%debug1) write(0,*)'min_prop_cos_sq=',min_prop_cos_sq
    if(gen_par%debug1) write(0,*)'n_mx_pad=',data_par%n_mx_pad
    if(gen_par%debug1) write(0,*)'n_my_pad=',data_par%n_my_pad
    if(gen_par%debug1) write(0,*)'n_hx_pad=',data_par%n_hx_pad

!    fft_3d_id=init_fft_3d(data_par%n_mx_pad,data_par%n_my_pad, &
!                          data_par%n_hx_pad,fast_init=.false.)
    call init_fft_3d(fft_3d_id,data_par%n_mx_pad,data_par%n_my_pad, &
                           data_par%n_hx_pad,fast_init=.false.)
    info%status=fft_3d_id%ierr

    if(gen_par%debug1) write(0,*)'new scale=',fft_3d_id%scale

    if(data_par%n_hy_pad /= 1 ) then
       call seperr(" Pre2D: Trying to use Prestack migration with n_hy > 1")
    end if

    allocate (phase(Pre2D_salt_ffdpi_par_loc%n_disp_rel),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of phase ")
    allocate (magn(Pre2D_salt_ffdpi_par_loc%n_disp_rel),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of magn ")

    allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_my(data_par%n_my_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
    allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")

    allocate (k_sr_sq(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2), &
    stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_sr_sq ")

!    write(0,*)'d_kmx=',data_par%d_kmx
!    write(0,*)'d_kmy=',data_par%d_kmy
!    write(0,*)'kmx_0=',data_par%kmx_0
!    write(0,*)'kmy_0=',data_par%kmy_0

    do i_k_mx=1,data_par%n_mx_pad
       kmx  = i_k_mx - 1.
       if(kmx <= float(data_par%n_mx_pad/2)) then
          k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
       else
          k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
       end if
    end do

    do i_k_my=1,data_par%n_my_pad
       kmy = i_k_my - 1.
       if(kmy <= float(data_par%n_my_pad/2)) then
          k_my(i_k_my) =  data_par%kmy_0+data_par%d_kmy* kmy
       else
          k_my(i_k_my) = -data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad)
       end if
    end do

    if(gen_par%single_hx_dip) then
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
       end do
    else
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          if(khx <= float(data_par%n_hx_pad/2)) then
             k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
          else
             k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
          end if
       end do
    end if
    !write(0,*)'k_hx=',k_hx

    do i_k_hx=1,data_par%n_hx_pad
       do i_k_my=1,data_par%n_my_pad
          do i_k_mx=1,data_par%n_mx_pad
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,SOU_COMP)= &
             .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,REC_COMP)= &
             .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)
             !write(0,*)'k_s_sq(',i_k_hx,') =',k_s_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_r_sq(',i_k_hx,') =',k_r_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_my(',i_k_my,') =',k_mx(i_k_my) 
             !write(0,*)'k_mx(',i_k_mx,') =',k_mx(i_k_mx) 
          end do
       end do
    end do

    n_cycles_ksq=5.
    k_cutoff =0.
    delta_k=sqrt(min_prop_cos_sq)
    k_start=delta_k
    k_decay=-1.*delta_k
    if(k_decay**2 > n_cycles_ksq) call seperr('k_decay^2 > n_cycles_ksq')
    edge_gauss = exp(k_decay)
    sigma_magn_sq=-(k_decay-k_start)**2/(2.*log(edge_gauss))

    !write(0,*)'edge_gauss=',edge_gauss
    !write(0,*)'sigma_magn_sq=',sigma_magn_sq

    d_k_norm=n_cycles_ksq/(Pre2D_salt_ffdpi_par_loc%n_disp_rel-1)
    do i_k_norm=1,Pre2D_salt_ffdpi_par_loc%n_disp_rel
       k_norm_sq=1.-(i_k_norm-1)*d_k_norm
       if( k_norm_sq > 0.) then
          k_norm=sqrt(k_norm_sq)
       else
          k_norm=-sqrt(-k_norm_sq)
       end if

       if(k_norm_sq < k_norm_sq_start_lin) then
         kk0=sqrt((i_k_norm-1)*d_k_norm)
         phase(i_k_norm)=sqrt(k_norm_sq_start_lin) - &
                 (kk0-k_start_lin)   /sqrt(k_norm_sq_start_lin) - &
              .5*(kk0-k_start_lin)**2/(sqrt(k_norm_sq_start_lin)*k_norm_sq_start_lin)
       else
         phase(i_k_norm)=k_norm
       end if

!  if(k_norm > k_start) then
       if(k_norm > -k_start) then !new
          magn(i_k_norm)=0.
       else
          !    gauss_magn=exp(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
          !    magn(i_k_norm)=log(gauss_magn)
          magn(i_k_norm)=(-((k_norm+k_start)**2)/(2.*sigma_magn_sq)) !new
       end if
    end do
    max_disp_rel=float(Pre2D_salt_ffdpi_par_loc%n_disp_rel)

    !    write(0,*)'n_disp_rel=',Pre2D_salt_ffdpi_par_loc%n_disp_rel
    call snap('Phase.h',Pre2D_salt_ffdpi_par_loc%n_disp_rel,1,phase)
    call snap('Magn.h',Pre2D_salt_ffdpi_par_loc%n_disp_rel,1,magn)
  end function Pre2D_salt_ffdpi_init

  function Pre2D_salt_ffdpi_down_oper(adj,add, &
  n_w_data,w_0_data,d_w_data, &
  z_down,d_z_down,data_par, &
  data_slice,ddata_slice,slow_slice) &
  result(stat)

    integer stat,n_w_data
    logical add,adj,found_salt
    real w_0_data,d_w_data,z_down,d_z_down
    type (data_param_type) data_par
    real, dimension (:,:,:,:,:) :: slow_slice
    complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
    optional data_slice,ddata_slice,slow_slice

    logical prop_data,prop_ddata,var_slow
    integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx
    integer i_mx,i_my,i_hx,i_ksq_norm,i_w,i_hx_p1,i_hx_m1
    integer i_slow,i_slow_m1,i_slow_p1,i_surf_comp,i_t_surf_comp
    integer i_interp,beg_i_mx_ffd,end_i_mx_ffd,beg_i_hx_ffd,end_i_hx_ffd
    integer inc_i_mx,i_mx_inc,i_mx_inc_m1,i_mx_inc_p1
    integer len_trid_ffd

    real k0_sq,rphase,rmagn,dphase,dphase_m1,dphase_p1,w_data
    real slow,slow_p1,slow_m1
    real slow_0,slow_0_p1,slow_0_m1
    real slow_sq,slow_sq_m1,slow_sq_p1
    real slow_0_sq,slow_0_sq_m1,slow_0_sq_p1
    real delta_slow,delta_slow_m1,delta_slow_p1
    real inv_delta_slow,inv_delta_slow_m1,inv_delta_slow_p1
    real prod_slow,prod_slow_m1,prod_slow_p1
    real inv_prod_slow_sq,inv_prod_slow_sq_m1,inv_prod_slow_sq_p1
    real slow_cube,slow_cube_m1,slow_cube_p1
    real slow_0_cube,slow_0_cube_m1,slow_0_cube_p1
    real sum_sq_slow,sum_sq_slow_m1,sum_sq_slow_p1
    real sum_sq_slow_2,sum_sq_slow_m1_2,sum_sq_slow_p1_2

    real vel,vel_m1,vel_p1,vel_0,vel_0_m1,vel_0_p1
    real delta_vel,delta_vel_m1,delta_vel_p1
    real sum_sq_vel,sum_sq_vel_m1,sum_sq_vel_p1,vel_sq
    real w_slow
    
    real inv_d_k_norm,k_0_opt,k_0_opt_p1
    real ph_opt_slow,ph_opt_slow_0,ph_opt_slow_0_p1
    real delta_ph_opt
    real p_rist_0,p_rist_0_p1,q_rist_0,q_rist_0_p1,p_opt_sq
    real inv_w_data,d_mx_hx,inv_d_mx_hx,k_exact,k_appr
    real opt_ang_sin_sq_appr_x,opt_ang_sin_sq_appr_y
    
    
    real sigma,rat,dscat,dz_w,pi2,norm_ksq,k0,k0_dz
    real prod_slow_w_data_sq_8,w_data_dx_sq_16
    real d_mx_hx_sq
    complex p_1,p_n,p_1_int,p_n_int
    

    integer, allocatable,  dimension (:,:,:,:) :: map_slow
    complex, allocatable,  dimension (:,:,:,:) :: data_interp
    complex, allocatable,  dimension (:,:,:) :: temp_data
    complex, allocatable,  dimension (:,:,:) :: temp_ddata
    complex, allocatable, dimension (:,:) :: diag_ffd
    complex, allocatable, dimension (:) :: temp_ffd

    integer,  dimension(n_comp_slow)  :: n_slow_ref_eff
    real,  dimension(Pre2D_salt_ffdpi_par_loc%n_slow_ref,n_comp_slow)  :: slow_prop
    complex,  dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: ph_shift

    n_mx=data_par%n_mx_pad
    n_my=data_par%n_my_pad
    n_hx=data_par%n_hx_pad
    pi2=2.*acos(-1.)
    inv_d_k_norm=1./d_k_norm

    d_mx_hx=data_par%d_mx*sqrt(2.)
    d_mx_hx_sq=d_mx_hx**2
    inv_d_mx_hx=1./d_mx_hx

    p_1=cmplx(Pre2D_salt_ffdpi_par_loc%p_1_real,Pre2D_salt_ffdpi_par_loc%p_1_imag)
    p_n=cmplx(Pre2D_salt_ffdpi_par_loc%p_n_real,Pre2D_salt_ffdpi_par_loc%p_n_imag)


    prop_data= present(data_slice)
    prop_ddata= present(ddata_slice)
    var_slow= present(slow_slice)


    if(var_slow) then
      allocate (map_slow(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
      stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of map_slow ")

      if(Pre2D_salt_ffdpi_par_loc%ffdpi_order >= 1) then
        allocate (diag_ffd(data_par%n_hx_pad,3),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of diag_ffd")
        allocate (temp_ffd(data_par%n_hx_pad),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp_ffd")
      end if
  
    end if
    if(prop_data) then
       allocate(temp_data(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad),&
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of temp_data")
       allocate(data_interp(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
           stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of data_interp")
    end if
    if(prop_ddata) then
       allocate(temp_ddata(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad),&
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of temp_ddata")
    end if


    if(var_slow) then
       n_slow_ref_eff=Pre2D_salt_slow_prop(n_mx,n_my,n_hx,  &
       slow_slice, slow_prop, map_slow) 
!      if((omp_get_thread_num() == 0) .and. adj) then
!      if((omp_get_thread_num() == 0)) then
        write(0,*)'n_slow_ref_eff(SOU)=',n_slow_ref_eff(SOU_COMP)
        write(0,*)'n_slow_ref_eff(REC)=',n_slow_ref_eff(REC_COMP)
        write(0,*)'vel_prop(sou)=',1./slow_prop(1:n_slow_ref_eff(SOU_COMP),SOU_COMP)
        write(0,*)'vel_prop(rec)=',1./slow_prop(1:n_slow_ref_eff(REC_COMP),REC_COMP)
!      end if
    else
       if(Pre2D_salt_ffdpi_par_loc%n_slow_ref /= 1) call seperr('n_slow_ref /= 1')
       n_slow_ref_eff(1:2)= 1 
       slow_prop(1,SOU_COMP)=slow_const
       slow_prop(1,REC_COMP)=slow_const
    end if

    do i_w=1,n_w_data
    !{
      w_data=w_0_data+(i_w-1)*d_w_data
      inv_w_data=1./w_data

      if( w_data > epsi) then
      !{
      do i_surf_comp=1,2
      !{
        if(adj) then
        !{
          if (i_surf_comp == 1) then
            i_t_surf_comp=REC_COMP
          else
            i_t_surf_comp=SOU_COMP
          end if
        !}
        else
        !{
          if (i_surf_comp == 1) then
            i_t_surf_comp=SOU_COMP
          else
            i_t_surf_comp=REC_COMP
          end if
        !}
        end if

! for each reference slowness apply SSR and merge into n_interp wavefields
        do i_slow=1,n_slow_ref_eff(i_t_surf_comp)
        !{
          slow_0=slow_prop(i_slow,i_t_surf_comp)
!          slow_0_m1=slow_prop(i_slow,i_t_surf_comp)
!          slow_0_p1=slow_prop(i_slow,i_t_surf_comp)
        
          k0= (w_data*slow_0)
          k0_sq= k0*k0
          k0_dz=k0*d_z_down
          dz_w=d_z_down*w_data
          w_data_dx_sq_16=16*(w_data**2)*d_mx_hx_sq
          
          if(k0 > epsi) then
            norm_ksq=1./(d_k_norm*k0_sq)
          else
            write(0,*)'I should not be here k0_sq=',k0_sq
          write(0,*)'k0_sq=',k0_sq
          write(0,*)'slow_0=',slow_0
          write(0,*)'w_data=',w_data
           call seperr('Pre2D_ffdpi: Zero frequency')
            norm_ksq=0.
          end if
        
          do i_k_hx=1,n_hx
          !{
            do i_k_my=1,n_my
            !{
              do i_k_mx=1,n_mx
              !{
                i_ksq_norm=nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,i_t_surf_comp)* &
                           norm_ksq+1,max_disp_rel))
                rphase=phase(i_ksq_norm)*k0_dz
                rmagn=abs(k0_dz)*magn(i_ksq_norm)
          
                ph_shift(i_k_mx,i_k_my,i_k_hx)=cexp(cmplx(rmagn,rphase))
          
              !}
              end do
            !}
            end do
          !}
          end do
          if(prop_data) then
            temp_data(:,:,:) = data_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
            call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data(:,:,:))
          end if
          if(prop_ddata) then
            temp_ddata(:,:,:) = ddata_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
            call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_ddata(:,:,:))
          end if
        
          do i_hx=1,n_hx
          !{
            do i_my=1,n_my
            !{
              do i_mx=1,n_mx
              !{
                if(map_slow(i_mx,i_my,i_hx,i_t_surf_comp) == i_slow) then 
                  data_interp(i_mx,i_my,i_hx,1)= temp_data(i_mx,i_my,i_hx)
                end if
        
                if(n_interp >1) then
                  if(map_slow(i_mx,i_my,i_hx,i_t_surf_comp) == i_slow-1) then 
                    data_interp(i_mx,i_my,i_hx,2)= temp_data(i_mx,i_my,i_hx)
                  end if
                end if
              !}
              end do
            !}
            end do
          !}
          end do
        
        !} end slowness loop
        end do

! now correct the merged wavefileds
        do i_interp=1,n_interp
         !{
           ! write coeficients for boundaries (i_interp=1 => delta_slow >0)
           if(i_interp == 1) then
             p_1_int=conjg(p_1)
             p_n_int=conjg(p_n)
           else
             p_1_int=p_1
             p_n_int=p_n
           end if
           if(n_interp == 1) then
             p_1_int=p_1
             p_n_int=p_n
           end if
      
! compute split step correction
           do i_hx=1,n_hx
           !{
             do i_my=1,n_my
             !{
               do i_mx=1,n_mx
               !{
                 i_slow=map_slow(i_mx,i_my,i_hx,i_t_surf_comp)+i_interp-1 
                 slow_0=slow_prop(i_slow,i_t_surf_comp)
                 dphase=(slow_slice(i_mx,i_my,i_hx,1,i_t_surf_comp)-slow_0)* dz_w
                 ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
               !}
               end do
             !}
             end do
           !}
           end do
!if((omp_get_thread_num() == 0) .and. (z_down== 0.) .and. i_interp == 2) write(0,*)'slow_0',slow_0
           if(prop_data) then
           !{
! apply split step corection
             data_interp(:,:,:,i_interp) = data_interp(:,:,:,i_interp)* &
                        ph_shift(:,:,:)
          
             if(Pre2D_salt_ffdpi_par_loc%ffdpi_order >= 1) then
             !{

if(i_t_surf_comp == REC_COMP) then
  inc_i_mx=1
  beg_i_mx_ffd=2-n_hx
  end_i_mx_ffd=n_mx
else
  inc_i_mx=-1
  beg_i_mx_ffd=1
  end_i_mx_ffd=n_mx+n_hx-1
end if
!
!if(omp_get_thread_num() == 0)write(0,*)'inc_i_mx',inc_i_mx
!if(omp_get_thread_num() == 0)write(0,*)'beg_i_mx_ffd',beg_i_mx_ffd
!if(omp_get_thread_num() == 0)write(0,*)'end_i_mx_ffd',end_i_mx_ffd

do i_mx=beg_i_mx_ffd,end_i_mx_ffd
!{
  if(inc_i_mx > 0) then
    beg_i_hx_ffd=max(1,-i_mx+2)
    end_i_hx_ffd=min(n_hx,n_mx+1-i_mx)
  else
    beg_i_hx_ffd=max(1,i_mx-n_mx+1)
    end_i_hx_ffd=min(n_hx,i_mx)
  end if
!if(omp_get_thread_num() == 0)write(0,*)'i_mx',i_mx
!if(omp_get_thread_num() == 0)write(0,*)'beg_i_hx_ffd',beg_i_hx_ffd
!if(omp_get_thread_num() == 0)write(0,*)'end_i_hx_ffd',end_i_hx_ffd

  len_trid_ffd=end_i_hx_ffd - beg_i_hx_ffd+1

  if(len_trid_ffd > 5) then
  !{

! copy data and slow
    do i_hx=beg_i_hx_ffd,end_i_hx_ffd
    !{
      i_mx_inc=i_mx+inc_i_mx*(i_hx-1)
  
!  if((omp_get_thread_num() == 0) .and. &
!        ((i_mx == beg_i_mx_ffd) .or.  (i_mx == end_i_mx_ffd)) .and. &
!        ((i_hx == beg_i_hx_ffd) .or.  (i_hx == end_i_hx_ffd))) then
!    write(0,*)'beg_i_mx_ffd',beg_i_mx_ffd
!    write(0,*)'end_i_mx_ffd',end_i_mx_ffd
!    write(0,*)'beg_i_hx_ffd',beg_i_hx_ffd
!    write(0,*)'end_i_hx_ffd',end_i_hx_ffd
!    write(0,*)'i_mx',i_mx
!    write(0,*)'i_hx',i_hx
!    write(0,*)'i_mx_inc',i_mx_inc
!  end if
  
      i_hx_m1=max(beg_i_hx_ffd,i_hx-1)
      i_hx_p1=min(end_i_hx_ffd,i_hx+1)
      i_mx_inc_m1=i_mx+inc_i_mx*(i_hx_m1-1)
      i_mx_inc_p1=i_mx+inc_i_mx*(i_hx_p1-1)
  
      i_slow=map_slow(i_mx_inc,1,i_hx,i_t_surf_comp)+i_interp-1 
      slow_0=slow_prop(i_slow,i_t_surf_comp)
  
      slow=slow_slice(i_mx_inc,1,i_hx,1,i_t_surf_comp)
  
      if(Pre2D_salt_ffdpi_par_loc%stable) then

        i_slow_m1=map_slow(i_mx_inc_m1,1,i_hx_m1,i_t_surf_comp)+i_interp-1 
        i_slow_p1=map_slow(i_mx_inc_p1,1,i_hx_p1,i_t_surf_comp)+i_interp-1 
        slow_0_m1=slow_prop(i_slow_m1,i_t_surf_comp)
        slow_0_p1=slow_prop(i_slow_p1,i_t_surf_comp)
  
        slow_m1=slow_slice(i_mx_inc_m1,1,i_hx_m1,1,i_t_surf_comp)
        slow_p1=slow_slice(i_mx_inc_p1,1,i_hx_p1,1,i_t_surf_comp)
      else
        slow_0_m1=slow_0
        slow_0_p1=slow_0
  
        slow_m1=slow
        slow_p1=slow
      end if
  
      vel=1/slow
      vel_m1=1/slow_m1
      vel_p1=1/slow_p1
  
      vel_0=1/slow_0
      vel_0_m1=1/slow_0_m1
      vel_0_p1=1/slow_0_p1
  
      delta_vel=vel-vel_0
      delta_vel_m1=vel_m1-vel_0_m1
      delta_vel_p1=vel_p1-vel_0_p1
  
      sum_sq_vel=vel**2 + vel_0**2 + vel*vel_0
      sum_sq_vel_m1=vel_m1**2 + vel_0_m1**2 + vel_m1*vel_0_m1
      sum_sq_vel_p1=vel_p1**2 + vel_0_p1**2 + vel_p1*vel_0_p1
  
! symmetric stable (XDDX) with change of variables

      diag_ffd(i_hx,1)=-(sqrt(sum_sq_vel/sum_sq_vel_m1)* &
                      cmplx( &
                      sum_sq_vel_m1, &
                      dz_w*delta_vel_m1))
  
      diag_ffd(i_hx,2)=-w_data_dx_sq_16 + &
                        2.*cmplx(sum_sq_vel,dz_w*delta_vel)
  
      diag_ffd(i_hx,3)=-(sqrt(sum_sq_vel/sum_sq_vel_p1)* &
                      cmplx( &
                      sum_sq_vel_p1, &
                      dz_w*delta_vel_p1))
  
!left boundary
      if((i_hx-beg_i_hx_ffd+1) <= Pre2D_salt_ffdpi_par_loc%bound_width_1) then
        diag_ffd(i_hx,2)=- w_data_dx_sq_16 + &
                       (1.+p_1_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
        if(i_hx == beg_i_hx_ffd ) then
          diag_ffd(i_hx,1)=cmplx(0.,0.)
        end if
      end if
! right boundary
      if(i_hx >= end_i_hx_ffd-Pre2D_salt_ffdpi_par_loc%bound_width_n+1) then
        diag_ffd(i_hx,2)=- w_data_dx_sq_16 + &
                       (1.+p_n_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
        if(i_hx == end_i_hx_ffd ) then
          diag_ffd(i_hx,3)=cmplx(0.,0.)
        end if
      end if
  
      temp_ffd(i_hx) = &
            data_interp(i_mx_inc_m1,1,i_hx_m1,i_interp) * diag_ffd(i_hx,1) + &
            data_interp(i_mx_inc   ,1,i_hx   ,i_interp) * diag_ffd(i_hx,2) + &
            data_interp(i_mx_inc_p1,1,i_hx_p1,i_interp) * diag_ffd(i_hx,3) 
  
  !    temp_ffd(i_hx) = &
  !          data_interp(i_mx_inc,1,i_hx   ,i_interp) 
  
! conjugate diagonals

      diag_ffd(i_hx,1)=conjg(diag_ffd(i_hx,1))
      diag_ffd(i_hx,2)=conjg(diag_ffd(i_hx,2))
      diag_ffd(i_hx,3)=conjg(diag_ffd(i_hx,3))
  
! left boundary
      if((i_hx-beg_i_hx_ffd+1) <= Pre2D_salt_ffdpi_par_loc%bound_width_1) then
      !{
        diag_ffd(i_hx,2)=(-w_data_dx_sq_16 + &
                       (1.+p_1_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
      !}
      end if
! right boundary
      if(i_hx >= end_i_hx_ffd-Pre2D_salt_ffdpi_par_loc%bound_width_n+1) then
      !{
        diag_ffd(i_hx,2)=(-w_data_dx_sq_16 + &
                       (1.+p_n_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
      !}
      end if
    !}
    end do


! solve tridiagonal
!    call cgtsl(len_trid_ffd,diag_ffd(beg_i_hx_ffd,1), &
!                            diag_ffd(beg_i_hx_ffd,2), &
!                            diag_ffd(beg_i_hx_ffd,3), &
!                            temp_ffd(beg_i_hx_ffd)  ,ierr)
    if (ierr /= 0) then
      write(0,*)'At i_mx= ',i_mx
      write(0,*)'The ',ierr,'-th diagonal element becomes zero'
      write(0,*)'WARNING: Problem in tridiagonal solver (x dir)'
  !          call seperr('Problem in tridiagonal solver (x dir)')
    end if
  
! copy solution 
    do i_hx=beg_i_hx_ffd,end_i_hx_ffd
    !{
      i_mx_inc=i_mx+inc_i_mx*(i_hx-1)
      data_interp(i_mx_inc,1,i_hx,i_interp) = temp_ffd(i_hx)
    !}
    end do
  !}
  end if
!}
end do
             !}
            end if
           !}
          end if
         !}
        end do
       
! Interpolate between the corrected n_interp wavefields
!(reset indentation)

do i_hx=1,n_hx
!{
  do i_my=1,n_my
  !{
    do i_mx=1,n_mx
    !{
      if(n_interp == 1) then
      !{
        data_slice(i_mx,i_my,i_hx,1,i_w) = data_interp(i_mx,i_my,i_hx,1)
      !}
      else
      !{
        slow=slow_slice(i_mx,i_my,i_hx,1,i_t_surf_comp)
        vel=1./slow
        vel_sq=vel*vel

        i_slow=map_slow(i_mx,i_my,i_hx,i_t_surf_comp)
        slow_0=slow_prop(i_slow,i_t_surf_comp)
        slow_0_p1=slow_prop(i_slow+1,i_t_surf_comp)

        vel_0=1./slow_0
        vel_0_p1=1./slow_0_p1

        ph_opt_slow=slow*opt_ang_cos
        p_opt_sq=opt_ang_sin_sq*slow*slow

        if(Pre2D_salt_ffdpi_par_loc%ffdpi_order >= 1) then
        !{
          if(Pre2D_salt_ffdpi_par_loc%dispers_interp) then
            k_exact=opt_ang_sin*slow*w_data
            k_appr=(2.*inv_d_mx_hx)*sin(.5*k_exact*d_mx_hx)
            opt_ang_sin_sq_appr_x=((k_appr*inv_w_data*vel)**2)*opt_azim_cos_sq
            opt_ang_sin_sq_appr_y=((k_appr*inv_w_data*vel)**2)*opt_azim_sin_sq
          else
            opt_ang_sin_sq_appr_x=opt_ang_sin_sq*opt_azim_cos_sq
            opt_ang_sin_sq_appr_y=opt_ang_sin_sq*opt_azim_sin_sq
          end if

          p_rist_0=vel_0/vel
          q_rist_0=.5*(p_rist_0**2 + p_rist_0 + 1)
          p_rist_0_p1=vel_0_p1/vel
          q_rist_0_p1=.5*(p_rist_0_p1**2 + p_rist_0_p1 + 1)

          k_0_opt=(p_opt_sq*vel_0*vel_0)*inv_d_k_norm+1
          k_0_opt_p1=(p_opt_sq*vel_0_p1*vel_0_p1)*inv_d_k_norm+1

          ph_opt_slow_0= &
            slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
            (slow-slow_0) - & 
            ((vel-vel_0)/vel_sq)*(opt_ang_sin_sq_appr_x/ &
                                  (2.-q_rist_0*opt_ang_sin_sq_appr_x) + &
                                  opt_ang_sin_sq_appr_y/ &
                                  (2.-q_rist_0*opt_ang_sin_sq_appr_y))

          ph_opt_slow_0_p1= &
            slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
            (slow-slow_0_p1) - &
            ((vel-vel_0_p1)/vel_sq)*(opt_ang_sin_sq_appr_x/ &
                                  (2.-q_rist_0_p1*opt_ang_sin_sq_appr_x) + &
                                  opt_ang_sin_sq_appr_y/ &
                                  (2.-q_rist_0_p1*opt_ang_sin_sq_appr_y))

!            ph_opt_slow_0= &
!              slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
!              (slow-slow_0) - & 
!              ((vel-vel_0)/vel_sq)*opt_ang_sin_sq_appr/ &
!                                  (2.-q_rist_0*opt_ang_sin_sq_appr)
!  
!            ph_opt_slow_0_p1= &
!              slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
!              (slow-slow_0_p1) - &
!              ((vel-vel_0_p1)/vel_sq)*opt_ang_sin_sq_appr/ &
!                                  (2.-q_rist_0_p1*opt_ang_sin_sq_appr)

        !}
        else 
        !{
          k_0_opt=(p_opt_sq*vel_0*vel_0)*inv_d_k_norm+1
          k_0_opt_p1=(p_opt_sq*vel_0_p1*vel_0_p1)*inv_d_k_norm+1

          ph_opt_slow_0= &
            slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
            (slow-slow_0) 

          ph_opt_slow_0_p1= &
            slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
            (slow-slow_0_p1) 
        !}
        end if
         

        delta_ph_opt=ph_opt_slow_0-ph_opt_slow_0_p1
        if(abs(delta_ph_opt) >= epsi) then
          if((phase(nint(min(k_0_opt,max_disp_rel))) >= 0.)) then
            w_slow=abs((ph_opt_slow-ph_opt_slow_0_p1)/delta_ph_opt)
          else
            w_slow=0.
          end if
        else
          w_slow=1.
        end if

!          w_slow=(slow-slow_0)/(slow_prop(i_slow+1)-slow_0)
!          w_slow=delta_vel/delta_vel_p1
!
!          if(w_slow > .5) then
!            w_slow=1.
!          else
!            w_slow=(2.)*delta_vel/delta_vel_p1
!          end if 


!if((omp_get_thread_num() == 0)  .and. &
!   (i_mx == n_mx/2) ) then
!      write(0,*)'slow',slow
!      write(0,*)'slow_0',slow_0
!      write(0,*)'slow_0_p1',slow_0_p1
!      write(0,*)'ph_opt_slow',ph_opt_slow
!      write(0,*)'ph_opt_slow_0',ph_opt_slow_0
!      write(0,*)'ph_opt_slow_0_p1',ph_opt_slow_0_p1
!!      write(0,*)'p_opt_sq',p_opt_sq
!!      write(0,*)'k_0_opt',k_0_opt
!!      write(0,*)'k_0_opt_p1',k_0_opt_p1
!      write(0,*)'w_slow',w_slow
!end if

!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow',slow
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow_0',slow_0
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'vel',vel
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'vel_p1',vel_p1
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'delta_vel',delta_vel
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow',ph_opt_slow
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow_0',ph_opt_slow_0
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow_0_p1',ph_opt_slow_0_p1
!
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'w_slow',w_slow
!write(0,*)'w_slow',w_slow

        data_slice(i_mx,i_my,i_hx,1,i_w) =  &
           (   w_slow)*data_interp(i_mx,i_my,i_hx,1) +  &
           (1.-w_slow)*data_interp(i_mx,i_my,i_hx,2)


      !}
      end if
    !}
    end do
  !}
  end do
!}
end do

if((var_slow) .and. (.not. adj)) then
!{
  if(prop_data) then
  !{
    if(apply_any_bound) then
      call apply_bound(data_par,data_slice(:,:,:,:,i_w))
    end if
    call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
  !}
  end if
  if(prop_ddata) then
  !{
    call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,ddata_slice(:,:,:,1,i_w))
  !}
  end if
!}
end if

  
      !} i_surf_comp loop
      end do
      !}
      else
      !{
        if(prop_data) then
          data_slice(:,:,:,1,i_w)=0.
        end if
        if(prop_ddata) then
          ddata_slice(:,:,:,1,i_w)=0.
        end if
      !}
      end if
    !} i_w loop
    end do







!
!          if((var_slow) .and. (adj)) then
!             if(prop_data) then
!                call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data(:,:,:,2))
!             end if
!             if(prop_ddata) then
!                call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_ddata(:,:,:,2))
!             end if
!          end if
!
!          do i_slow=1,n_slow_ref_eff(surf_comp)
!             !{
!             slow_0=slow_prop(i_slow,surf_comp)
!
!             k0= (w_data*slow_0)
!             k0_sq= k0*k0
!             if(time_mig) then
!                k0_dz=.5*w_data*d_z_down
!             else
!                k0_dz=k0*d_z_down
!                dz_w=d_z_down*w_data
!             end if
!
!             if(k0_sq > epsi) then
!                norm_ksq=1./(d_k_norm*k0_sq)
!             else
!                norm_ksq=0.
!             end if
!
!             if((var_slow) .and. (adj)) then
!                !{
!                if(.not. time_mig) then
!                   !{
!                   do i_hx=1,n_hx
!                      !{
!                      do i_my=1,n_my
!                         !{
!                         do i_mx=1,n_mx
!                            !{
!                            dphase=(slow_slice(i_mx,i_my,i_hx,1,surf_comp)-slow_0)*dz_w
!                            ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
!                            !}
!                         end do
!                         !}
!                      end do
!                      !}
!                   end do
!                   !}
!                end if
!
!                if(prop_data) then
!                   !{
!                   do i_hx=1,n_hx
!                      !{
!                      do i_my=1,n_my
!                         !{
!                         do i_mx=1,n_mx
!                            !{
!                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
!                               temp_data(i_mx,i_my,i_hx,1)=temp_data(i_mx,i_my,i_hx,2)
!                            end if
!                            !}
!                         end do
!                         !}
!                      end do
!                      !}
!                   end do
!                   if(.not. time_mig) then
!                      !{
!                      temp_data(:,:,:,1) = temp_data(:,:,:,1) * ph_shift(:,:,:)
!                      !}
!                   end if
!                   call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,temp_data(:,:,:,1))
!                     !} 
!                  end if
!                  if(prop_ddata) then
!                     !{
!                     do i_hx=1,n_hx
!                        !{
!                        do i_my=1,n_my
!                           !{
!                           do i_mx=1,n_mx
!                              !{
!                              if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
!                                 temp_ddata(i_mx,i_my,i_hx,1)=temp_ddata(i_mx,i_my,i_hx,2)
!                              end if
!                              !}
!                           end do
!                           !}
!                        end do
!                        !}
!                     end do
!                     if(.not. time_mig) then
!                        !{
!                        temp_ddata(:,:,:,1) = temp_ddata(:,:,:,1) * ph_shift(:,:,:)
!                        !}
!                     end if
!                     call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,temp_ddata(:,:,:,1))
!                     !} 
!                  end if
!                  !}
!               else
!                  !{
!                  if(prop_data) then
!                     !{
!                     temp_data(:,:,:,1) = temp_data(:,:,:,2)
!                     !}
!                  end if
!                  if(prop_ddata) then
!                     !{
!                     temp_ddata(:,:,:,1) = temp_ddata(:,:,:,2)
!                     !}
!                  end if
!                  !} 
!               end if
!
!             do i_k_hx=1,n_hx
!                !{
!                do i_k_my=1,n_my
!                   !{
!                   do i_k_mx=1,n_mx
!                      !{
!                      i_ksq_norm=nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)* &
!                      norm_ksq+1,max_disp_rel))
!                      rphase=phase(i_ksq_norm)*k0_dz
!                      rmagn=abs(k0_dz)*magn(i_ksq_norm)
!
!                      ph_shift(i_k_mx,i_k_my,i_k_hx)=cexp(cmplx(rmagn,rphase))
!
!                      !}
!                   end do
!                   !}
!                end do
!                !}
!             end do
!
!             if(prop_data) then
!                temp_data(:,:,:,1) = temp_data(:,:,:,1) * ph_shift(:,:,:)
!             end if
!             if(prop_ddata) then
!                temp_ddata(:,:,:,1) = temp_ddata(:,:,:,1) * ph_shift(:,:,:)
!             end if
!
!             if((var_slow) .and. (.not. adj)) then
!                !{
!                if(.not. time_mig) then
!                   !{
!                   do i_hx=1,n_hx
!                      !{
!                      do i_my=1,n_my
!                         !{
!                         do i_mx=1,n_mx
!                            !{
!                            dphase=(slow_slice(i_mx,i_my,i_hx,1,surf_comp)-slow_0)* dz_w
!                            ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
!                            !}
!                         end do
!                         !}
!                      end do
!                      !}
!                   end do
!                   !}
!                end if
!
!                if(prop_data) then
!                   !{
!                   call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data(:,:,:,1))
!                   if(.not. time_mig) then
!                      !{
!                      temp_data(:,:,:,1) = temp_data(:,:,:,1) * ph_shift(:,:,:)
!                      !}
!                   end if
!                   do i_hx=1,n_hx
!                      !{
!                      do i_my=1,n_my
!                         !{
!                         do i_mx=1,n_mx
!                            !{
!                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
!                               data_slice(i_mx,i_my,i_hx,1,i_w)=temp_data(i_mx,i_my,i_hx,1)
!                            end if
!                            !}
!                         end do
!                         !}
!                      end do
!                      !}
!                   end do
!                   !} 
!                end if
!                if(prop_ddata) then
!                   !{
!                   call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_ddata(:,:,:,1))
!                   if(.not. time_mig) then
!                      !{
!                      temp_ddata(:,:,:,1) = temp_ddata(:,:,:,1) * ph_shift(:,:,:)
!                      !}
!                   end if
!                   do i_hx=1,n_hx
!                      !{
!                      do i_my=1,n_my
!                         !{
!                         do i_mx=1,n_mx
!                            !{
!                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
!                               ddata_slice(i_mx,i_my,i_hx,1,i_w)=temp_ddata(i_mx,i_my,i_hx,1)
!                            end if
!                            !}
!                         end do
!                         !}
!                      end do
!                      !}
!                   end do
!                   !} 
!                end if
!                !}
!             else
!                !{
!                if(prop_data) then
!                   data_slice(:,:,:,1,i_w)=data_slice(:,:,:,1,i_w)+temp_data(:,:,:,1)
!                   temp_data(:,:,:,1)=cmplx(0.,0.)
!                end if
!                if(prop_ddata) then
!                   ddata_slice(:,:,:,1,i_w)=ddata_slice(:,:,:,1,i_w)+temp_ddata(:,:,:,1)
!                   temp_ddata(:,:,:,1)=cmplx(0.,0.)
!                end if
!                !} 
!             end if
!
!             !} sou slowness loop
!          end do
!
!          if((var_slow) .and. (.not. adj)) then
!             if(prop_data) then
!                call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
!             end if
!             if(prop_ddata) then
!                call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,ddata_slice(:,:,:,1,i_w))
!             end if
!          end if
!          !} i_surf_comp loop
!       end do
!       !}
!    else
!       !{
!       if(prop_data) then
!         data_slice(:,:,:,1,i_w)=0.
!       end if
!       if(prop_ddata) then
!         ddata_slice(:,:,:,1,i_w)=0.
!       end if
!       !}
!    end if
!       !} i_w loop
!    end do

    if(var_slow) then
       deallocate (map_slow,stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with deallocation of map_slow")

      if(Pre2D_salt_ffdpi_par_loc%ffdpi_order >= 1) then
        deallocate (diag_ffd,stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with deallocation of diag_ffd")
        deallocate (temp_ffd,stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with deallocation of temp_ffd")
      end if
    end if
    if(prop_data) then
       deallocate (temp_data,stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with deallocation of temp_data")
       deallocate (data_interp,stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with deallocation of data_interp")
    end if
    if(prop_ddata) then
       deallocate (temp_ddata,stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with deallocation of temp_ddata")
    end if

    stat=0
    return	
  end function Pre2D_salt_ffdpi_down_oper

  function Pre2D_salt_ffdpi_scatter_oper(adj,add, &
  n_w_data,w_0_data,d_w_data, &
  z_down,d_z_down,data_par, &
  data_slice,ddata_slice,dslow_slice,slow_slice) &
  result(stat)

    integer stat,n_w_data
    logical add,adj
    real w_0_data,d_w_data,z_down,d_z_down
    type (data_param_type) data_par
    real, dimension (:,:,:,:,:) :: slow_slice
    complex, dimension (:,:,:,:,:) :: dslow_slice
    complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
    optional slow_slice

    logical var_slow
    integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx,i_mx,i_my,i_hx,i_w
    integer i_surf_comp,surf_comp,i_slow
    real slow_0,k0_sq,w_data
    real sigma,rat,dscat,dz_w,k0
    complex im_dz_w,ddscat

    integer,  dimension(n_comp_slow)  :: n_slow_ref_eff
    integer, allocatable,  dimension (:,:,:,:) :: map_slow

    real,  dimension(Pre2D_salt_ffdpi_par_loc%n_slow_ref,2)   :: slow_prop
    complex, dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: data_scat,ddata_scat

    data_scat=cmplx(0.,0.)
    ddata_scat=cmplx(0.,0.)

!    n_mx=data_par%n_mx_pad
!    n_my=data_par%n_my_pad
!    n_hx=data_par%n_hx_pad
!
!    var_slow= present(slow_slice)
!
!    if(var_slow) then
!       allocate (map_slow(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
!       stat=ierr)  
!       if(ierr .ne. 0) call seperr("Problems with allocation of map_slow ")
!    end if
!
!    if(var_slow) then
!       n_slow_ref_eff=Pre2D_salt_slow_prop(n_mx,n_my,n_hx,  &
!       slow_slice, slow_prop, map_slow) 
!    else
!       if(Pre2D_salt_ffdpi_par_loc%n_slow_ref /= 1) call seperr('n_slow_ref /= 1')
!       n_slow_ref_eff(1:2)= 1 
!       slow_prop(1,SOU_COMP)=slow_const
!       slow_prop(1,REC_COMP)=slow_const
!    end if
!
!    if(any(n_slow_ref_eff /= 1)) call seperr('Pre2D_salt_ffdpi_scatter_oper cannot handle n_slow_ref_eff >1')
!
!    do i_w=1,n_w_data
!       !{
!       w_data=w_0_data+(i_w-1)*d_w_data
!
!       do i_surf_comp=1,2
!          !{
!          if(adj) then
!             !{
!             if (i_surf_comp == 1) then
!                surf_comp=REC_COMP
!             else
!                surf_comp=SOU_COMP
!             end if
!             !}
!          else
!             !{
!             if (i_surf_comp == 1) then
!                surf_comp=SOU_COMP
!             else
!                surf_comp=REC_COMP
!             end if
!             !}
!          end if
!
!          do i_slow=1,n_slow_ref_eff(surf_comp)
!             !{
!             slow_0=slow_prop(i_slow,surf_comp)
!
!             k0= (w_data*slow_0)
!             k0_sq= k0*k0
!             dz_w=d_z_down*w_data
!             im_dz_w=cmplx(0.,dz_w)
!
!             if(time_mig) then
!                call seperr('Scattering for time mig not supported yet!')
!             else
!                dz_w=d_z_down*w_data
!             end if
!             im_dz_w=cmplx(0.,dz_w)
!
!             ! compute and apply weights of scattered wavefield in wavenumber domain
!             data_scat=data_slice(:,:,:,1,i_w)
!             if(Pre2D_salt_ffdpi_par_loc%ffdpi_order > 1) then
!                !{
!                do i_k_hx=1,n_hx
!                   !{
!                   do i_k_my=1,n_my
!                      !{
!                      do i_k_mx=1,n_mx
!                         !{
!                         if((k0_sq - k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp) ) > 0.) then
!                            rat=(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)/k0_sq)
!                            sigma=1+.5*rat + .375*rat*rat
!                         else
!                            sigma=1
!                         end if
!                         data_scat(i_k_mx,i_k_my,i_k_hx)=data_scat(i_k_mx,i_k_my,i_k_hx)*sigma
!                         !}
!                      end do
!                      !}
!                   end do
!                   !}
!                end do
!                !}
!             end if
!
!             ! transform in space domain scattered wavefield
!             call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,data_scat)
!
!             if(adj) then
!                !{
!                ! transform in space domain total scattered wavefield
!                ddata_scat=ddata_slice(:,:,:,1,i_w)
!                call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,ddata_scat)
!                !}
!             end if
!
!             do i_hx=1,n_hx
!                !{
!                do i_my=1,n_my
!                   !{
!                   do i_mx=1,n_mx
!                      !{
!                      if(adj) then
!                         !{
!                         ! add contribution to slowness perturbations 
!                         dslow_slice(i_mx,i_my,i_hx,1,surf_comp) =  &
!                         dslow_slice(i_mx,i_my,i_hx,1,surf_comp) + &
!                         (im_dz_w*conjg(data_scat(i_mx,i_my,i_hx)))*ddata_scat(i_mx,i_my,i_hx)
!                         !}
!                      else
!                         !{
!                         ! compute scatered wavefield
!                         ddscat=dslow_slice(i_mx,i_my,i_hx,1,surf_comp)*im_dz_w
!                         if(cabs(ddscat) /= 0.) then
!                            data_scat(i_mx,i_my,i_hx) = &
!                            data_scat(i_mx,i_my,i_hx) * ddscat
!                         else
!                            data_scat(i_mx,i_my,i_hx) = cmplx(0.,0.)
!                         end if
!                         !}
!                      end if
!                      !}
!                   end do
!                   !}
!                end do
!                !}
!             end do
!
!             if(.not. adj) then
!                ! transform scattered wavefield back to wavenumber 
!                call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,data_scat)
!                ! sum to total scattered wavefield
!                ddata_slice(:,:,:,1,i_w) = ddata_slice(:,:,:,1,i_w) + data_scat
!             end if
!             !} i_slow loop
!          end do
!          !} i_surf_comp loop
!       end do
!       !} i_w loop
!    end do

    stat=0
    return	
  end function Pre2D_salt_ffdpi_scatter_oper

  function Pre2D_salt_ffdpi_clean()  result(stat)
    integer stat,ierr
    stat=0
    if(allocated(phase)) deallocate (phase,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(magn)) deallocate (magn,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_mx)) deallocate (k_mx,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_my)) deallocate (k_my,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_hx)) deallocate (k_hx,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_sr_sq)) deallocate (k_sr_sq,stat=ierr)  
    if(ierr /=0) stat=stat+1

    return	
  end function Pre2D_salt_ffdpi_clean

  function Pre2D_salt_slow_prop(n_mx,n_my,n_hx,  &
  slow_slice, slow_prop, map_slow) &
  result (n_slow_ref_eff)

  integer n_slow_ref_eff(n_comp_slow)
  integer n_mx,n_my,n_hx
  integer,  dimension (:,:,:,:), intent(inout) :: map_slow
  real, dimension (:,:,:,:,:), intent(in)      :: slow_slice
  real,  dimension(:,:), intent(inout)           :: slow_prop

  logical found_salt,not_found
  integer i_mx,i_my,i_hx,i_surf_comp,ierr
  integer i_mx_p1,i_mx_m1,i_my_p1,i_my_m1,i_hx_p1,i_hx_m1
  integer n_slow_ref_no_salt,i_slow,i_slow_max_0
  real min_slow,max_slow,range_slow,delta_slow
  real mean_slow,inv_delta_slow
  real temp_slow_salt,thresh_slow_salt
  real min_slow_salt

  logical, allocatable, dimension (:,:,:,:) :: mask_slow 

  do i_surf_comp=1,n_comp_slow
  !{
    if(slow_salt > 0.) then
    !{
       found_salt=any(slow_slice(:,:,:,:,i_surf_comp) <= slow_salt)
       min_slow_salt=minval(slow_slice(:,:,:,:,i_surf_comp))

       temp_slow_salt=min(slow_salt,minval(slow_slice(:,:,:,:,i_surf_comp)))
       thresh_slow_salt=min(slow_salt, &
         Pre2D_salt_ffdpi_par_loc%range_salt*minval(slow_slice(:,:,:,:,i_surf_comp)))
!if((omp_get_thread_num() == 0)) write(0,*)'temp_vel_salt=',1./temp_slow_salt
!if((omp_get_thread_num() == 0)) write(0,*)'thresh_vel_salt=',1./thresh_slow_salt

       if(found_salt) then
       !{
!if((omp_get_thread_num() == 0)) write(0,*)'Found salt'
         allocate (mask_slow(n_mx,n_my,n_hx,1),stat=ierr)  
         if(ierr .ne. 0) call seperr("Pre2D_salt_ssf: Problems with allocation of mask_slow ")
!if((omp_get_thread_num() == 0)) write(0,*)'shape(slow_slice)',shape(slow_slice)
!if((omp_get_thread_num() == 0)) write(0,*)'shape(mask_slow)',shape(mask_slow)
         mask_slow=.true.
         do i_hx=1,n_hx
           do i_my=1,n_my
             do i_mx=1,n_mx
               if(slow_slice(i_mx,i_my,i_hx,1,i_surf_comp) < thresh_slow_salt) then
                 i_mx_p1=min(n_mx,i_mx+1)
                 i_mx_m1=max(1,i_mx-1)
                 i_my_p1=min(n_my,i_my+1)
                 i_my_m1=max(1,i_my-1)
                 i_hx_p1=min(n_hx,i_hx+1)
                 i_hx_m1=max(1,i_hx-1)
 mask_slow(i_mx_m1:i_mx_p1,i_my_m1:i_my_p1,i_hx_m1:i_hx_p1,1)=.false.
                end if
             end do
           end do
         end do
         if(any(mask_slow))then
           min_slow=minval(slow_slice(:,:,:,:,i_surf_comp),mask=mask_slow)
           max_slow=maxval(slow_slice(:,:,:,:,i_surf_comp))
           n_slow_ref_no_salt=Pre2D_salt_ffdpi_par_loc%n_slow_ref-1

!if((omp_get_thread_num() == 0)) write(0,*)'inv min_slow=',1./min_slow
!           if((omp_get_thread_num() == 0)) write(0,*)'not constant'
!           if((omp_get_thread_num() == 0)) write(0,*)'max_vel=',1./min_slow

         else
!           if((omp_get_thread_num() == 0)) write(0,*)'constant salt'
           n_slow_ref_no_salt=0
         end if
       !}
       else
       !{
         min_slow=minval(slow_slice(:,:,:,:,i_surf_comp))
         max_slow=maxval(slow_slice(:,:,:,:,i_surf_comp))
         n_slow_ref_no_salt=Pre2D_salt_ffdpi_par_loc%n_slow_ref
       !}
       end if
    !}
    else
    !{
      found_salt=.false.
      n_slow_ref_no_salt=Pre2D_salt_ffdpi_par_loc%n_slow_ref
      if(Pre2D_salt_ffdpi_par_loc%min_slow_const >0) then
        min_slow=Pre2D_salt_ffdpi_par_loc%min_slow_const
      else
        if(Pre2D_salt_ffdpi_par_loc%max_vel_const >0) then
          min_slow=1./Pre2D_salt_ffdpi_par_loc%max_vel_const
        else
          min_slow=minval(slow_slice(:,:,:,:,i_surf_comp))
        end if
      end if

      if(Pre2D_salt_ffdpi_par_loc%max_slow_const >0) then
        max_slow=Pre2D_salt_ffdpi_par_loc%max_slow_const
      else
        if(Pre2D_salt_ffdpi_par_loc%min_vel_const >0) then
          max_slow=1./Pre2D_salt_ffdpi_par_loc%min_vel_const
        else
          max_slow=maxval(slow_slice(:,:,:,:,i_surf_comp))
        end if
      end if
    !}
    end if

    range_slow=max_slow/min_slow
    delta_slow=(1.00001*range_slow)**(1./max(1,n_slow_ref_no_salt-1))
    mean_slow=.5*(max_slow+min_slow)

! if((omp_get_thread_num() == 0)) then
!   write(0,*)'min_vel=',1./max_slow
!   write(0,*)'max_vel=',1./min_slow
!   write(0,*)'range_slow=',range_slow
!   write(0,*)'delta_slow=',delta_slow
!   write(0,*)'inv of mean_slow=',1./mean_slow
! end if

    if(delta_slow < (1./max_acc_sin)) then
    !{
      if(range_slow < (1./max_acc_sin)) then
        n_slow_ref_no_salt= 2 
        delta_slow=1.00001*range_slow
      else
        n_slow_ref_no_salt= & 
           min(n_slow_ref_no_salt, &
           max(1,1+ceiling(log(range_slow)/log_min_ratio)))
        delta_slow=(1.00001*range_slow)**(1./max(1,n_slow_ref_no_salt-1))
  !write(0,*)'Second: n_slow_ref_no_salt=',n_slow_ref_no_salt
  !write(0,*)'Second: delta_slow=',delta_slow
      end if
    !}
    else
    !{
!      if((omp_get_thread_num() == 0)) then
        write(0,*)'ATTENTION: ratio between velocity exceeds limits:' 
        write(0,*)'Desired limit=',(1./max_acc_sin)
        write(0,*)'Actual ratio=',delta_slow
!      end if
    !}
    end if

    ! set reference velocities 
    if(found_salt) then
      n_slow_ref_eff(i_surf_comp)= n_slow_ref_no_salt + 1
      slow_prop(1,i_surf_comp)=min_slow_salt
      i_slow_max_0=2
!if((omp_get_thread_num() == 0)) write(0,*)'vel salt',1./slow_prop(1)
    else
      n_slow_ref_eff(i_surf_comp)= n_slow_ref_no_salt 
      i_slow_max_0=1
    end if

! This is a bug when commented out the omp_ call
!    if((omp_get_thread_num() == 0)) then
      n_total_slow_ref=n_total_slow_ref+n_slow_ref_eff(i_surf_comp)
      nz_total=nz_total+1
!    end if


    if(n_slow_ref_no_salt > 1) then
      do i_slow=i_slow_max_0,n_slow_ref_eff(i_surf_comp)
         slow_prop(i_slow,i_surf_comp)=min_slow*(delta_slow**(i_slow-adj_slow-i_slow_max_0+1))

!  if((omp_get_thread_num() == 0)) write(0,*)'min_slow=',min_slow
!  if((omp_get_thread_num() == 0)) write(0,*)'delta_slow=',delta_slow
!  if((omp_get_thread_num() == 0)) write(0,*)'adj_slow=',adj_slow
!  if((omp_get_thread_num() == 0)) write(0,*)'i_slow_max_0=',i_slow_max_0
!  if((omp_get_thread_num() == 0)) write(0,*)'tot=',(i_slow-adj_slow-i_slow_max_0+1)
!  if((omp_get_thread_num() == 0)) write(0,*)'tot1=',delta_slow**(i_slow-adj_slow-i_slow_max_0+1)
!  if((omp_get_thread_num() == 0)) write(0,*)'i_slow=',i_slow,'vel_prop(i_slow)',1./slow_prop(i_slow,i_surf_comp)

      end do
    else
      if(slow_const > 0.) then  
        slow_prop(i_slow_max_0,i_surf_comp)=slow_const
      else
        slow_prop(i_slow_max_0,i_surf_comp)=min_slow+(1.-.5-i_slow_max_0+1)*delta_slow
      end if

!if((omp_get_thread_num() == 0)) write(0,*)'i_slow=only vel_prop(i_slow_max_0)',1./slow_prop(i_slow_max_0)

    end if

    if (n_slow_ref_eff(i_surf_comp) >1) then
    !{
!if((omp_get_thread_num() == 0)) write(0,*)'Here #1'
      do i_hx=1,n_hx
      !{
        do i_my=1,n_my
        !{
          do i_mx=1,n_mx
          !{
            if(found_salt) then
            !{ 
!              if(.not.mask_slow(i_mx,i_my,i_hx,1)) then
              if(slow_slice(i_mx,i_my,i_hx,1,i_surf_comp) < thresh_slow_salt) then
              !{ 
!if((omp_get_thread_num() == 0)) write(0,*)'Here #2: In salt'
                map_slow(i_mx,i_my,i_hx,i_surf_comp)= i_slow_max_0-1
!if((omp_get_thread_num() == 0)) write(0,*)'map_slow',map_slow(i_mx,i_my,i_hx,i_surf_comp)
                not_found=.false.
              !} 
              else
              !{ 
!if((omp_get_thread_num() == 0)) write(0,*)'Here #3: outside salt'
                not_found=.true.
              !} 
              end if
            !}
            else 
            !{
!if((omp_get_thread_num() == 0)) write(0,*)'Here #4: salt not found'
              not_found=.true.
            !}
            end if
!            i_slow=i_slow_max_0
            i_slow=1
            do while (not_found .and. i_slow <n_slow_ref_eff(i_surf_comp))
            !{ 
              if((slow_slice(i_mx,i_my,i_hx,1,i_surf_comp) >= &
                  slow_prop(i_slow,i_surf_comp)) .and. &
                 (slow_slice(i_mx,i_my,i_hx,1,i_surf_comp) <= &
                  slow_prop(i_slow+1,i_surf_comp))  &
                ) then
                map_slow(i_mx,i_my,i_hx,i_surf_comp)= i_slow
                not_found=.false.
              else
                i_slow=i_slow+1
              end if
              if(i_slow >= n_slow_ref_eff(i_surf_comp) ) then
    write(0,*)'i_surf_comp=',i_surf_comp
    write(0,*)'i_slow=',i_slow
    write(0,*)'slow_slice=',slow_slice(i_mx,i_my,i_hx,1,i_surf_comp)
    write(0,*)'slow_prop(:,i_surf_comp)=',slow_prop(:,i_surf_comp)
                call seperr('Problems in defining velocity map')
              end if
            !} 
            end do
          !}
          end do
        !}
        end do
      !}
      end do
    !}
    else
    !{
      map_slow(:,:,:,i_surf_comp)=1
    !}
    end if

    if(found_salt) then
    !{
       deallocate (mask_slow,stat=ierr)  
       if(ierr .ne. 0) call seperr("Pre2D_salt_ssf: Problems with deallocation of mask_slow ")
    !}
    end if
  !}
  end do
!if((omp_get_thread_num() == 0)) write(0,*)'minval(map_slow)',minval(map_slow)
!if((omp_get_thread_num() == 0)) write(0,*)'maxval(map_slow)',maxval(map_slow)

  end function Pre2D_salt_slow_prop

end module Pre2D_salt_ffdpi_mod

