module Ristow_salt_mod

use Ristow_salt_types

use gendown3d_types
use down_process_types
use data_types
use image_types
use slow_types

use fft_data_mod

implicit none

type (Ristow_salt_param_type), private :: Ristow_salt_par_loc
type (fft_type), private :: fft_3d_id
integer, private :: n_interp,n_comp_slow,n_total_slow_ref,nz_total
real, private :: slow_const,slow_salt,epsi,min_prop_cos_sq,log_min_ratio
real, private :: opt_ang_cos,opt_ang_sin,opt_ang_sin_sq,max_acc_sin
real, private :: opt_azim_cos_sq,opt_azim_sin_sq
real, private :: d_k_norm,max_disp_rel,adj_slow
real, allocatable, dimension (:), private :: magn,phase
real, allocatable, dimension (:), private :: k_mx,k_my,k_hx
real, allocatable, dimension (:,:,:,:), private :: k_sr_sq

contains

subroutine init_Ristow_salt_par(Ristow_salt_par)
type (Ristow_salt_param_type) Ristow_salt_par

Ristow_salt_par_loc=Ristow_salt_par

end subroutine init_Ristow_salt_par

function Ristow_salt_init(gen_par,data_par,slow_par)  result(info)
type (gendown3d_info_type) info
type (gendown3d_param_type) gen_par
type (data_param_type) data_par
type (slow_param_type) slow_par
 
integer i_k_mx,i_k_my,i_k_hx,ierr,i_k_norm
real kmx,kmy,khx,pi2,k_norm_sq,k_norm
real n_cycles_ksq,k_start,k_cutoff,k_decay,sigma_magn_sq,edge_gauss,gauss_magn
real delta_k
real k_start_lin,k_norm_sq_start_lin,kk0

epsi=epsilon(1.)

!info=info_default

if(slow_par%n_comp_slow /= 2) then
  write(0,*)'Reset n_comp_slow to 2 (It should be changed!)'
  slow_par%n_comp_slow = 2
end if
n_comp_slow=slow_par%n_comp_slow

select case (Ristow_salt_par_loc%ref_slow(1:3)) 
case ('min')
  adj_slow=1.
case ('max')
  adj_slow=0.
case ('avg')
  adj_slow=.5
case default
  call seperr('Not supported reference_slowness ')
end select

slow_const=gen_par%slow_const
pi2=2.*acos(-1.)

n_total_slow_ref=0
nz_total=0
if(Ristow_salt_par_loc%max_acc_propag_angle > 0) then
  max_acc_sin=sin(Ristow_salt_par_loc%max_acc_propag_angle*pi2/360.)
else
  max_acc_sin=-1.
end if

if(Ristow_salt_par_loc%min_delta_slow > 0.) then
  write(0,*)'ATTENTION: the use of min_delta_slow is deprecated'
  write(0,*)'Use max_acc_propag_angle'
  if(Ristow_salt_par_loc%min_delta_slow >= 1) then
    call seperr('min_delta_slow must be < 1')
  end if
  max_acc_sin=1.-Ristow_salt_par_loc%min_delta_slow
else
  if(max_acc_sin < 0.) then
    call seperr('You need to set either max_acc_propag_angle or min_delta_slow')
  end if
end if
log_min_ratio=log(1./max_acc_sin)
write(0,*)'Max ratio=',1./max_acc_sin


min_prop_cos_sq=cos(gen_par%max_propag_angle*pi2/360.)**2
opt_ang_cos=cos(Ristow_salt_par_loc%opt_ang*pi2/360.)

opt_ang_sin=sin(Ristow_salt_par_loc%opt_ang*pi2/360.)
opt_ang_sin_sq=opt_ang_sin**2
opt_azim_cos_sq=cos(Ristow_salt_par_loc%opt_azim*pi2/360.)**2
opt_azim_sin_sq=sin(Ristow_salt_par_loc%opt_azim*pi2/360.)**2

!write(0,*)'opt_azim_cos_sq=',opt_azim_cos_sq
!write(0,*)'opt_azim_sin_sq=',opt_azim_sin_sq

k_start_lin=Ristow_salt_par_loc%k_start_lin
if(k_start_lin > 0.) then
  k_norm_sq_start_lin=(1.-k_start_lin**2)
else 
  k_norm_sq_start_lin=-huge(k_norm_sq_start_lin)
end if

if(opt_ang_sin > max_acc_sin) then
  write(0,*)'opt_ang=',Ristow_salt_par_loc%opt_ang
  write(0,*)'Effective max_acc_propag_angle=',(360/pi2)*asin(max_acc_sin)
  call seperr ('opt_ang must be < max_acc_propag_angle')
end if

if(Ristow_salt_par_loc%n_slow_split < 2) then
  n_interp=1
  write(0,*)'ATT: Set n_interp =1'
else
  n_interp=Ristow_salt_par_loc%n_interp
  write(0,*)'ATT: Set n_interp =',n_interp
end if

if (slow_par%slowness) then
  if(Ristow_salt_par_loc%slow_salt > 0.) then
    slow_salt=(Ristow_salt_par_loc%slow_salt)
    if(Ristow_salt_par_loc%n_slow_split < 2)  then
      call seperr('slow_salt >0 and n_slow_split <2')
    end if
  else 
    slow_salt=-1.
  end if
else
  if(Ristow_salt_par_loc%vel_salt > 0.) then
    slow_salt=(1./Ristow_salt_par_loc%vel_salt)
    if(Ristow_salt_par_loc%n_slow_split < 2) then
      call seperr('vel_salt >0 and n_slow_split <2')
    end if
  else 
    slow_salt=-1.
  end if
end if

!write(0,*)'Init slow_salt=',slow_salt
!write(0,*)'min_prop_cos_sq=',min_prop_cos_sq
!write(0,*)'opt_ang_cos=',opt_ang_cos
!write(0,*)'opt_ang_sin_sq=',opt_ang_sin_sq
!write(0,*)'n_mx_pad=',data_par%n_mx_pad
!write(0,*)'n_my_pad=',data_par%n_my_pad
!write(0,*)'n_hx_pad=',data_par%n_hx_pad

!fft_3d_id=init_fft_3d(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad)
call init_fft_3d(fft_3d_id,data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad)
info%status=fft_3d_id%ierr

!write(0,*)'shape(fft_coeff)=',shape(fft_3d_id%fft_coeff)
!write(0,*)'scale=',fft_3d_id%scale

if(data_par%n_hy_pad /= 1 ) then
  call seperr(" Ristow_salting to use Prestack migration with n_hy > 1")
end if


allocate (phase(Ristow_salt_par_loc%n_disp_rel),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of phase ")
allocate (magn(Ristow_salt_par_loc%n_disp_rel),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of magn ")

allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
allocate (k_my(data_par%n_my_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")

allocate (k_sr_sq(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2), &
         stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_sr_sq ")

!write(0,*)'d_kmx=',data_par%d_kmx
!write(0,*)'d_kmy=',data_par%d_kmy
!write(0,*)'kmx_0=',data_par%kmx_0
!write(0,*)'kmy_0=',data_par%kmy_0

do i_k_mx=1,data_par%n_mx_pad
  kmx  = i_k_mx - 1.
  if(kmx <= float(data_par%n_mx_pad/2)) then
    k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
  else
    k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
  end if
end do

do i_k_my=1,data_par%n_my_pad
  kmy = i_k_my - 1.
  if(kmy <= float(data_par%n_my_pad/2)) then
    k_my(i_k_my) =  data_par%kmy_0+data_par%d_kmy* kmy
  else
    k_my(i_k_my) = -data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad)
  end if
end do

!write(0,*)'k_mx=',k_mx
!write(0,*)'k_my=',k_my

if(gen_par%single_hx_dip) then
  do i_k_hx=1,data_par%n_hx_pad
    khx = i_k_hx - 1.
    k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
  end do
else
  do i_k_hx=1,data_par%n_hx_pad
    khx = i_k_hx - 1.
    if(khx <= float(data_par%n_hx_pad/2)) then
      k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
    else
      k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
    end if
  end do
end if
!write(0,*)'k_hx=',k_hx

if(Ristow_salt_par_loc%inline_2d) then
!write(0,*)'Inline-2d is TRUE'
  do i_k_hx=1,data_par%n_hx_pad
    do i_k_my=1,data_par%n_my_pad
      do i_k_mx=1,data_par%n_mx_pad
        k_sr_sq(i_k_mx,i_k_my,i_k_hx,SOU_COMP)= &
              .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
        k_sr_sq(i_k_mx,i_k_my,i_k_hx,REC_COMP)= &
              .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)
  !write(0,*)'k_s_sq(',i_k_hx,') =',k_s_sq(i_k_mx,i_k_my,i_k_hx) 
  !write(0,*)'k_r_sq(',i_k_hx,') =',k_r_sq(i_k_mx,i_k_my,i_k_hx) 
  !write(0,*)'k_my(',i_k_my,') =',k_mx(i_k_my) 
  !write(0,*)'k_mx(',i_k_mx,') =',k_mx(i_k_mx) 
      end do
    end do
  end do
else
!write(0,*)'Inline-2d is FALSE'
  do i_k_hx=1,data_par%n_hx_pad
    do i_k_my=1,data_par%n_my_pad
      do i_k_mx=1,data_par%n_mx_pad
        k_sr_sq(i_k_mx,i_k_my,i_k_hx,SOU_COMP)= &
              .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2 + k_my(i_k_my)**2)
        k_sr_sq(i_k_mx,i_k_my,i_k_hx,REC_COMP)= &
              .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2 + k_my(i_k_my)**2)
  !write(0,*)'k_s_sq(',i_k_hx,') =',k_s_sq(i_k_mx,i_k_my,i_k_hx) 
  !write(0,*)'k_r_sq(',i_k_hx,') =',k_r_sq(i_k_mx,i_k_my,i_k_hx) 
  !write(0,*)'k_my(',i_k_my,') =',k_my(i_k_my) 
      end do
    end do
  end do
end if



n_cycles_ksq=5.
k_cutoff =0.
delta_k=sqrt(min_prop_cos_sq)
k_start=delta_k
k_decay=-1.*delta_k
if(k_decay**2 > n_cycles_ksq) call seperr('k_decay^2 > n_cycles_ksq')
edge_gauss = exp(k_decay)
!sigma_magn_sq=-.25*(k_decay-k_start)**2/(2.*log(edge_gauss))
sigma_magn_sq=-.025*(k_decay-k_start)**2/(2.*log(edge_gauss)) !new

!write(0,*)'edge_gauss=',edge_gauss
!write(0,*)'sigma_magn_sq=',sigma_magn_sq

d_k_norm=n_cycles_ksq/(Ristow_salt_par_loc%n_disp_rel-1)
do i_k_norm=1,Ristow_salt_par_loc%n_disp_rel
  k_norm_sq=1.-(i_k_norm-1)*d_k_norm
  if( k_norm_sq > 0.) then
    k_norm=sqrt(k_norm_sq)
  else
    k_norm=-sqrt(-k_norm_sq)
  end if

  if(k_norm_sq < k_norm_sq_start_lin) then
    kk0=sqrt((i_k_norm-1)*d_k_norm)
    phase(i_k_norm)=sqrt(k_norm_sq_start_lin) - &
            (kk0-k_start_lin)   /sqrt(k_norm_sq_start_lin) - &
         .5*(kk0-k_start_lin)**2/(sqrt(k_norm_sq_start_lin)*k_norm_sq_start_lin)
  else
    phase(i_k_norm)=k_norm
  end if

!  if(k_norm > k_start) then
  if(k_norm > -k_start) then !new
    magn(i_k_norm)=0.
  else
!    gauss_magn=exp(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
!    magn(i_k_norm)=log(gauss_magn)
    magn(i_k_norm)=(-((k_norm+k_start)**2)/(2.*sigma_magn_sq)) !new
  end if
end do
max_disp_rel=float(Ristow_salt_par_loc%n_disp_rel)

!write(0,*)'k_norm_sq_start_lin=',k_norm_sq_start_lin
call snap('Phase.h',Ristow_salt_par_loc%n_disp_rel,1,phase)
call snap('Magn.h',Ristow_salt_par_loc%n_disp_rel,1,magn)
end function Ristow_salt_init

function Ristow_salt_down_oper(adj,add, &
                     n_w_data,w_0_data,d_w_data, &
                     z_down,d_z_down,data_par, &
                     data_slice,ddata_slice,slow_slice) &
         result(stat)

integer stat,n_w_data
logical add,adj,found_salt
real w_0_data,d_w_data,z_down,d_z_down
type (data_param_type) data_par
real, dimension (:,:,:,:,:) :: slow_slice
complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
optional data_slice,ddata_slice,slow_slice

logical prop_data,prop_ddata,var_slow
integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx
integer i_mx,i_my,i_hx,i_ksq_norm,i_w,i_mx_p1,i_mx_m1,i_my_p1,i_my_m1
integer i_slow,i_slow_m1,i_slow_p1,n_slow_split_eff,i_surf_comp,surf_comp
integer i_interp

real k0_sq,rphase,rmagn,dphase,dphase_m1,dphase_p1,w_data
real slow,slow_p1,slow_m1
real slow_0,slow_0_p1,slow_0_m1
real slow_sq,slow_sq_m1,slow_sq_p1
real slow_0_sq,slow_0_sq_m1,slow_0_sq_p1
real delta_slow,delta_slow_m1,delta_slow_p1
real inv_delta_slow,inv_delta_slow_m1,inv_delta_slow_p1
real prod_slow,prod_slow_m1,prod_slow_p1
real inv_prod_slow_sq,inv_prod_slow_sq_m1,inv_prod_slow_sq_p1
real slow_cube,slow_cube_m1,slow_cube_p1
real slow_0_cube,slow_0_cube_m1,slow_0_cube_p1
real sum_sq_slow,sum_sq_slow_m1,sum_sq_slow_p1
real sum_sq_slow_2,sum_sq_slow_m1_2,sum_sq_slow_p1_2

real vel,vel_m1,vel_p1,vel_0,vel_0_m1,vel_0_p1
real delta_vel,delta_vel_m1,delta_vel_p1
real sum_sq_vel,sum_sq_vel_m1,sum_sq_vel_p1,vel_sq
real w_slow

real inv_d_k_norm,k_0_opt,k_0_opt_p1,ph_opt_slow,ph_opt_slow_0,ph_opt_slow_0_p1
real delta_ph_opt
real p_rist_0,p_rist_0_p1,q_rist_0,q_rist_0_p1,p_opt_sq
real inv_w_data,d_mx,inv_d_mx,k_exact,k_appr
real opt_ang_sin_sq_appr_x,opt_ang_sin_sq_appr_y


real sigma,rat,dscat,dz_w,pi2,norm_ksq,k0,k0_dz
real prod_slow_w_data_sq_8,w_data_dx_sq_16,w_data_dy_sq_16
real d_mx_sq,d_my_sq
complex p_1,p_n,p_1_int,p_n_int

integer, allocatable,  dimension (:,:,:,:) :: map_slow
complex, allocatable,  dimension (:,:,:,:) :: data_interp
complex, allocatable,  dimension (:,:,:) :: temp_data
complex, allocatable,  dimension (:,:,:) :: temp_ddata
complex, allocatable, dimension (:,:) :: diagx,diagy
complex, allocatable, dimension (:) :: tempx,tempy

real,  dimension(Ristow_salt_par_loc%n_slow_split,n_comp_slow)  :: slow_prop
complex,  dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: ph_shift

n_mx=data_par%n_mx_pad
n_my=data_par%n_my_pad
n_hx=data_par%n_hx_pad
pi2=2.*acos(-1.)
inv_d_k_norm=1./d_k_norm

d_mx=data_par%d_mx
d_mx_sq=data_par%d_mx**2
d_my_sq=data_par%d_my**2
inv_d_mx=1./d_mx

p_1=cmplx(Ristow_salt_par_loc%p_1_real,Ristow_salt_par_loc%p_1_imag)
p_n=cmplx(Ristow_salt_par_loc%p_n_real,Ristow_salt_par_loc%p_n_imag)

allocate (diagx(data_par%n_mx_pad,3),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of diagx")
allocate (diagy(data_par%n_my_pad,3),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of diagy")
allocate (tempx(data_par%n_mx_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of tempx")
allocate (tempy(data_par%n_my_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of tempy")

prop_data= present(data_slice)
prop_ddata= present(ddata_slice)
var_slow= present(slow_slice)

if(var_slow) then
  allocate (map_slow(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
           stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of map_slow ")
end if
if(prop_data) then
  allocate(data_interp(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
           stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of data_interp")
  allocate(temp_data(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad),&
           stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of temp_data")
end if
if(prop_ddata) then
  allocate(temp_ddata(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad),&
           stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of temp_ddata")
end if



if(var_slow) then
!{
  n_slow_split_eff=Ristow_salt_slow_prop(n_mx,n_my,n_hx,  &
                                        slow_slice, slow_prop, map_slow) 
!}
else
!{
  if(Ristow_salt_par_loc%n_slow_split /= 1) call seperr('n_slow_split /= 1')
  n_slow_split_eff= 1 
  slow_prop(1,:)=slow_const
!}
end if
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow_prop(:,SOU_COMP)',slow_prop(:,SOU_COMP)
!write(0,*)'slow_prop(:,SOU_COMP)',slow_prop(:,SOU_COMP)
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow_prop(:,REC_COMP)',slow_prop(:,REC_COMP)
!write(0,*)'slow_prop(:,REC_COMP)',slow_prop(:,REC_COMP)

!write(0,*)'var_slow',var_slow
!write(0,*)'prop_data',prop_data
!write(0,*)'prop_ddata',prop_ddata
!write(0,*)'shape(temp_data)',shape(temp_data)

!write(0,*)'n_slow_split_eff',n_slow_split_eff
!write(0,*)'slow_prop',slow_prop


do i_w=1,n_w_data
!{

  w_data=w_0_data+(i_w-1)*d_w_data
  inv_w_data=1./w_data

do i_surf_comp=1,n_comp_slow
!{
  if(adj) then
  !{
    if (i_surf_comp == 1) then
      surf_comp=REC_COMP
    else
      surf_comp=SOU_COMP
    end if
  !}
  else
  !{
    if (i_surf_comp == 1) then
      surf_comp=SOU_COMP
    else
      surf_comp=REC_COMP
    end if
  !}
  end if


do i_slow=1,n_slow_split_eff
!{
  slow_0=slow_prop(i_slow,surf_comp)
  slow_0_m1=slow_prop(i_slow,surf_comp)
  slow_0_p1=slow_prop(i_slow,surf_comp)

  k0= (w_data*slow_0)
  k0_sq= k0*k0
  k0_dz=k0*d_z_down
  dz_w=d_z_down*w_data
  w_data_dx_sq_16=16*(w_data**2)*d_mx_sq
  w_data_dy_sq_16=16*(w_data**2)*d_my_sq
  
  if(k0 > epsi) then
    norm_ksq=1./(d_k_norm*k0_sq)
  else
    write(0,*)'I should not be here k0_sq=',k0_sq
  write(0,*)'k0_sq=',k0_sq
  write(0,*)'slow_0=',slow_0
  write(0,*)'w_data=',w_data
   call seperr('Pre3d_ssf: Zero frequency')
!   norm_ksq=2.*Ristow_salt_par_loc%n_disp_rel/min_lap_sq
    norm_ksq=0.
  end if

  do i_k_hx=1,n_hx
  !{
    do i_k_my=1,n_my
    !{
      do i_k_mx=1,n_mx
      !{
        i_ksq_norm=nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)* &
                   norm_ksq+1,max_disp_rel))
        rphase=phase(i_ksq_norm)*k0_dz
        rmagn=abs(k0_dz)*magn(i_ksq_norm)
  
        ph_shift(i_k_mx,i_k_my,i_k_hx)=cexp(cmplx(rmagn,rphase))
  
      !}
      end do
    !}
    end do
  !}
  end do
  if(prop_data) then
    temp_data(:,:,:) = data_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
    call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data(:,:,:))
  end if
  if(prop_ddata) then
    temp_ddata(:,:,:) = ddata_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
    call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_ddata(:,:,:))
  end if

  do i_hx=1,n_hx
  !{
    do i_my=1,n_my
    !{
      do i_mx=1,n_mx
      !{
        if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
          data_interp(i_mx,i_my,i_hx,1)= temp_data(i_mx,i_my,i_hx)
        end if

        if(n_interp >1) then
          if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow-1) then 
            data_interp(i_mx,i_my,i_hx,2)= temp_data(i_mx,i_my,i_hx)
          end if
        end if
      !}
      end do
    !}
    end do
  !}
  end do

!} end slowness loop
end do

if((var_slow) .and. (.not. adj)) then
!{
  do i_interp=1,n_interp
  !{
    ! write coeficients for boundaries (i_interp=1 => delta_slow >0)
    if(i_interp == 1) then
      p_1_int=conjg(p_1)
      p_n_int=conjg(p_n)
    else
      p_1_int=p_1
      p_n_int=p_n
    end if
    if(n_interp == 1) then
      p_1_int=p_1
      p_n_int=p_n
    end if

    do i_hx=1,n_hx
    !{
      do i_my=1,n_my
      !{
        do i_mx=1,n_mx
        !{
          i_slow=map_slow(i_mx,i_my,i_hx,surf_comp)+i_interp-1 
          slow_0=slow_prop(i_slow,surf_comp)
          dphase=(slow_slice(i_mx,i_my,i_hx,1,surf_comp)-slow_0)* dz_w
          ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
        !}
        end do
      !}
      end do
    !}
    end do
!if((omp_get_thread_num() == 0) .and. (z_down== 0.) .and. i_interp == 2) write(0,*)'slow_0',slow_0
    if(prop_data) then
    !{
   ! apply split step corection
      data_interp(:,:,:,i_interp) = data_interp(:,:,:,i_interp)*ph_shift(:,:,:)
    
      if(Ristow_salt_par_loc%ristow_order >= 1) then
      !{
        i_hx=1

        ! apply finite difference along x
        if(n_mx /= 1) then
        !{
        do i_my=1,n_my
        !{
        ! set diagonals
          do i_mx=1,n_mx
          !{
            i_mx_m1=max(   1,i_mx-1)
            i_mx_p1=min(n_mx,i_mx+1)
  
            i_slow=map_slow(i_mx,i_my,i_hx,surf_comp)+i_interp-1 
  
            slow_0=slow_prop(i_slow,surf_comp)
            slow=slow_slice(i_mx,i_my,i_hx,1,surf_comp)

            if(Ristow_salt_par_loc%stable) then

              i_slow_m1=map_slow(i_mx_m1,i_my,i_hx,surf_comp)+i_interp-1 
              i_slow_p1=map_slow(i_mx_p1,i_my,i_hx,surf_comp)+i_interp-1 

              slow_0_m1=slow_prop(i_slow_m1,surf_comp)
              slow_0_p1=slow_prop(i_slow_p1,surf_comp)
    
              slow_m1=slow_slice(i_mx_m1,i_my,i_hx,1,surf_comp)
              slow_p1=slow_slice(i_mx_p1,i_my,i_hx,1,surf_comp)
            else
              slow_0_m1=slow_0
              slow_0_p1=slow_0
    
              slow_m1=slow
              slow_p1=slow
            end if
  
            vel=1/slow
            vel_m1=1/slow_m1
            vel_p1=1/slow_p1
  
            vel_0=1/slow_0
            vel_0_m1=1/slow_0_m1
            vel_0_p1=1/slow_0_p1
  
            delta_vel=vel-vel_0
            delta_vel_m1=vel_m1-vel_0_m1
            delta_vel_p1=vel_p1-vel_0_p1
  
            sum_sq_vel=vel**2 + vel_0**2 + vel*vel_0
            sum_sq_vel_m1=vel_m1**2 + vel_0_m1**2 + vel_m1*vel_0_m1
            sum_sq_vel_p1=vel_p1**2 + vel_0_p1**2 + vel_p1*vel_0_p1
  
!  ! symmetric stable (XDDX) without change of variables
!  
!            diagx(i_mx,1)=-(sqrt(sum_sq_vel*sum_sq_vel_m1)/delta_vel_m1)* &
!                            cmplx( &
!                            sum_sq_vel_m1, &
!                            dz_w*delta_vel_m1)
!  
!            diagx(i_mx,2)=(sum_sq_vel/delta_vel)*(-w_data_dx_sq_16 + &
!                              2.*cmplx(sum_sq_vel,dz_w*delta_vel))
!  
!            diagx(i_mx,3)=-(sqrt(sum_sq_vel*sum_sq_vel_p1)/delta_vel_p1)* &
!                            cmplx( &
!                            sum_sq_vel_p1, &
!                            dz_w*delta_vel_p1)

  ! symmetric stable (XDDX) with change of variables
  
            diagx(i_mx,1)=-(sqrt(sum_sq_vel/sum_sq_vel_m1)* &
                            cmplx( &
                            sum_sq_vel_m1, &
                            dz_w*delta_vel_m1))
  
            diagx(i_mx,2)=-w_data_dx_sq_16 + &
                              2.*cmplx(sum_sq_vel,dz_w*delta_vel)

            diagx(i_mx,3)=-(sqrt(sum_sq_vel/sum_sq_vel_p1)* &
                            cmplx( &
                            sum_sq_vel_p1, &
                            dz_w*delta_vel_p1))
  
  ! left boundary
            if(i_mx <= Ristow_salt_par_loc%bound_width_1) then
!              diagx(i_mx,2)=(sum_sq_vel/delta_vel)*(-w_data_dx_sq_16 + &
!                             (1.+p_1_int)*cmplx(sum_sq_vel,dz_w*delta_vel))
              diagx(i_mx,2)=- w_data_dx_sq_16 + &
                             (1.+p_1_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
              if(i_mx == 1 ) then
                diagx(i_mx,1)=cmplx(0.,0.)
              end if
            end if
  ! right boundary
            if(i_mx >= n_mx-Ristow_salt_par_loc%bound_width_n+1) then
!              diagx(i_mx,2)=(sum_sq_vel/delta_vel)*(-w_data_dx_sq_16 + &
!                             (1.+p_n_int)*cmplx(sum_sq_vel,dz_w*delta_vel))
              diagx(i_mx,2)=- w_data_dx_sq_16 + &
                             (1.+p_n_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
              if(i_mx == n_mx ) then
                diagx(i_mx,3)=cmplx(0.,0.)
              end if
            end if

            tempx(i_mx) = &
                  data_interp(i_mx_m1,i_my,i_hx,i_interp) * diagx(i_mx,1) + &
                  data_interp(i_mx   ,i_my,i_hx,i_interp) * diagx(i_mx,2) + &
                  data_interp(i_mx_p1,i_my,i_hx,i_interp) * diagx(i_mx,3) 
  ! change variable

           
  
  ! conjugate diagonals
  
            diagx(i_mx,1)=conjg(diagx(i_mx,1))
            diagx(i_mx,2)=conjg(diagx(i_mx,2))
            diagx(i_mx,3)=conjg(diagx(i_mx,3))
  
  ! left boundary
            if(i_mx <= Ristow_salt_par_loc%bound_width_1) then
            !{
!              diagx(i_mx,2)=- w_data_dx_sq_16 + &
!                             (1.+p_1_int)*cmplx(sum_sq_vel,-dz_w*delta_vel)
!              diagx(i_mx,2)=(sum_sq_vel/delta_vel)*(-w_data_dx_sq_16 + &
!                             (1.+p_1_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
              diagx(i_mx,2)=(-w_data_dx_sq_16 + &
                             (1.+p_1_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
            !}
            end if
  ! right boundary
            if(i_mx >= n_mx-Ristow_salt_par_loc%bound_width_n+1) then
            !{
              diagx(i_mx,2)=(-w_data_dx_sq_16 + &
                             (1.+p_n_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
!              diagx(i_mx,2)=(sum_sq_vel/delta_vel)*(-w_data_dx_sq_16 + &
!                             (1.+p_n_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
!              diagx(i_mx,2)=- w_data_dx_sq_16 + &
!                             (1.+p_n_int)*cmplx(sum_sq_vel,-dz_w*delta_vel)
            !}
            end if
          !}
          end do
  
  
          ! solve tridiagonal
          call cgtsl(n_mx,diagx(1,1),diagx(1,2),diagx(1,3),tempx,ierr)
          if (ierr /= 0) then
            write(0,*)'At i_my= ',i_my
            write(0,*)'The ',ierr,'-th diagonal element becomes zero'
            write(0,*)'WARNING: Problem in tridiagonal solver (x dir)'
  !             call seperr('Problem in tridiagonal solver (x dir)')
          end if
    
          ! copy solution 
          data_interp(:,i_my,i_hx,i_interp) = tempx(:)
        !}
        end do
        !}
        end if

        ! apply finite difference along y
        if(n_my /= 1) then
        !{
        do i_mx=1,n_mx
        !{
        ! set diagonals
          do i_my=1,n_my
          !{
            i_my_m1=max(   1,i_my-1)
            i_my_p1=min(n_my,i_my+1)
  
            i_slow=map_slow(i_mx,i_my,i_hx,surf_comp)+i_interp-1 
  
            slow_0=slow_prop(i_slow,surf_comp)
            slow=slow_slice(i_mx,i_my,i_hx,1,surf_comp)

            if(Ristow_salt_par_loc%stable) then

              i_slow_m1=map_slow(i_mx,i_my_m1,i_hx,surf_comp)+i_interp-1 
              i_slow_p1=map_slow(i_mx,i_my_p1,i_hx,surf_comp)+i_interp-1 

              slow_0_m1=slow_prop(i_slow_m1,surf_comp)
              slow_0_p1=slow_prop(i_slow_p1,surf_comp)
    
              slow_m1=slow_slice(i_mx,i_my_m1,i_hx,1,surf_comp)
              slow_p1=slow_slice(i_mx,i_my_p1,i_hx,1,surf_comp)
            else
              slow_0_m1=slow_0
              slow_0_p1=slow_0
    
              slow_m1=slow
              slow_p1=slow
            end if
  
            vel=1/slow
            vel_m1=1/slow_m1
            vel_p1=1/slow_p1
  
            vel_0=1/slow_0
            vel_0_m1=1/slow_0_m1
            vel_0_p1=1/slow_0_p1
  
            delta_vel=vel-vel_0
            delta_vel_m1=vel_m1-vel_0_m1
            delta_vel_p1=vel_p1-vel_0_p1
  
            sum_sq_vel=vel**2 + vel_0**2 + vel*vel_0
            sum_sq_vel_m1=vel_m1**2 + vel_0_m1**2 + vel_m1*vel_0_m1
            sum_sq_vel_p1=vel_p1**2 + vel_0_p1**2 + vel_p1*vel_0_p1
  
  ! symmetric stable (XDDX) with change of variables
  
            diagy(i_my,1)=-(sqrt(sum_sq_vel/sum_sq_vel_m1)* &
                            cmplx( &
                            sum_sq_vel_m1, &
                            dz_w*delta_vel_m1))
  
            diagy(i_my,2)=-w_data_dy_sq_16 + &
                              2.*cmplx(sum_sq_vel,dz_w*delta_vel)

            diagy(i_my,3)=-(sqrt(sum_sq_vel/sum_sq_vel_p1)* &
                            cmplx( &
                            sum_sq_vel_p1, &
                            dz_w*delta_vel_p1))
  
  ! left boundary
            if(i_my <= Ristow_salt_par_loc%bound_width_1) then
              diagy(i_my,2)=- w_data_dy_sq_16 + &
                             (1.+p_1_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
              if(i_my == 1 ) then
                diagy(i_my,1)=cmplx(0.,0.)
              end if
            end if
  ! right boundary
            if(i_my >= n_my-Ristow_salt_par_loc%bound_width_n+1) then
              diagy(i_my,2)=- w_data_dy_sq_16 + &
                             (1.+p_n_int)*cmplx(sum_sq_vel,dz_w*delta_vel)
              if(i_my == n_my ) then
                diagy(i_my,3)=cmplx(0.,0.)
              end if
            end if

            tempy(i_my) = &
                  data_interp(i_mx,i_my_m1,i_hx,i_interp) * diagy(i_my,1) + &
                  data_interp(i_mx,i_my   ,i_hx,i_interp) * diagy(i_my,2) + &
                  data_interp(i_mx,i_my_p1,i_hx,i_interp) * diagy(i_my,3) 
  ! conjugate diagonals
  
            diagy(i_my,1)=conjg(diagy(i_my,1))
            diagy(i_my,2)=conjg(diagy(i_my,2))
            diagy(i_my,3)=conjg(diagy(i_my,3))
  
  ! left boundary
            if(i_my <= Ristow_salt_par_loc%bound_width_1) then
            !{
              diagy(i_my,2)=(-w_data_dy_sq_16 + &
                             (1.+p_1_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
            !}
            end if
  ! right boundary
            if(i_my >= n_my-Ristow_salt_par_loc%bound_width_n+1) then
            !{
              diagy(i_my,2)=(-w_data_dy_sq_16 + &
                             (1.+p_n_int)*cmplx(sum_sq_vel,-dz_w*delta_vel))
            !}
            end if
          !}
          end do
  
          ! solve tridiagonal
          call cgtsl(n_my,diagy(1,1),diagy(1,2),diagy(1,3),tempy,ierr)
          if (ierr /= 0) then
            write(0,*)'At i_mx= ',i_mx
            write(0,*)'The ',ierr,'-th diagonal element becomes zero'
            write(0,*)'WARNING: Problem in tridiagonal solver (y dir)'
  !             call seperr('Problem in tridiagonal solver (y dir)')
          end if
    
          ! copy solution 
          data_interp(i_mx,:,i_hx,i_interp) = tempy(:)
        !}
        end do
        !}
        end if

      !}
      end if
    !}
    end if
  !}
  end do

  do i_hx=1,n_hx
  !{
    do i_my=1,n_my
    !{
      do i_mx=1,n_mx
      !{
        if(n_interp == 1) then
        !{
          data_slice(i_mx,i_my,i_hx,1,i_w) = data_interp(i_mx,i_my,i_hx,1)
        !}
        else
        !{
          slow=slow_slice(i_mx,i_my,i_hx,1,surf_comp)
          vel=1./slow
          vel_sq=vel*vel

          i_slow=map_slow(i_mx,i_my,i_hx,surf_comp)
          slow_0=slow_prop(i_slow,surf_comp)
          slow_0_p1=slow_prop(i_slow+1,surf_comp)

          vel_0=1./slow_0
          vel_0_p1=1./slow_0_p1

          ph_opt_slow=slow*opt_ang_cos
          p_opt_sq=opt_ang_sin_sq*slow*slow

          if(Ristow_salt_par_loc%ristow_order >= 1) then
          !{
            if(Ristow_salt_par_loc%dispers_interp) then
              k_exact=opt_ang_sin*slow*w_data
              k_appr=(2.*inv_d_mx)*sin(.5*k_exact*d_mx)
              opt_ang_sin_sq_appr_x=((k_appr*inv_w_data*vel)**2)*opt_azim_cos_sq
              opt_ang_sin_sq_appr_y=((k_appr*inv_w_data*vel)**2)*opt_azim_sin_sq
            else
              opt_ang_sin_sq_appr_x=opt_ang_sin_sq*opt_azim_cos_sq
              opt_ang_sin_sq_appr_y=opt_ang_sin_sq*opt_azim_sin_sq
            end if
  
            p_rist_0=vel_0/vel
            q_rist_0=.5*(p_rist_0**2 + p_rist_0 + 1)
            p_rist_0_p1=vel_0_p1/vel
            q_rist_0_p1=.5*(p_rist_0_p1**2 + p_rist_0_p1 + 1)
  
            k_0_opt=(p_opt_sq*vel_0*vel_0)*inv_d_k_norm+1
            k_0_opt_p1=(p_opt_sq*vel_0_p1*vel_0_p1)*inv_d_k_norm+1

            ph_opt_slow_0= &
              slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
              (slow-slow_0) - & 
              ((vel-vel_0)/vel_sq)*(opt_ang_sin_sq_appr_x/ &
                                    (2.-q_rist_0*opt_ang_sin_sq_appr_x) + &
                                    opt_ang_sin_sq_appr_y/ &
                                    (2.-q_rist_0*opt_ang_sin_sq_appr_y))
  
            ph_opt_slow_0_p1= &
              slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
              (slow-slow_0_p1) - &
              ((vel-vel_0_p1)/vel_sq)*(opt_ang_sin_sq_appr_x/ &
                                    (2.-q_rist_0_p1*opt_ang_sin_sq_appr_x) + &
                                    opt_ang_sin_sq_appr_y/ &
                                    (2.-q_rist_0_p1*opt_ang_sin_sq_appr_y))
  
!            ph_opt_slow_0= &
!              slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
!              (slow-slow_0) - & 
!              ((vel-vel_0)/vel_sq)*opt_ang_sin_sq_appr/ &
!                                  (2.-q_rist_0*opt_ang_sin_sq_appr)
!  
!            ph_opt_slow_0_p1= &
!              slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
!              (slow-slow_0_p1) - &
!              ((vel-vel_0_p1)/vel_sq)*opt_ang_sin_sq_appr/ &
!                                  (2.-q_rist_0_p1*opt_ang_sin_sq_appr)

          !}
          else 
          !{
            k_0_opt=(p_opt_sq*vel_0*vel_0)*inv_d_k_norm+1
            k_0_opt_p1=(p_opt_sq*vel_0_p1*vel_0_p1)*inv_d_k_norm+1

            ph_opt_slow_0= &
              slow_0*phase(nint(min(k_0_opt,max_disp_rel)))+&
              (slow-slow_0) 
  
            ph_opt_slow_0_p1= &
              slow_0_p1*phase(nint(min(k_0_opt_p1,max_disp_rel)))+&
              (slow-slow_0_p1) 
          !}
          end if
           

          delta_ph_opt=ph_opt_slow_0-ph_opt_slow_0_p1
          if(abs(delta_ph_opt) >= epsi) then
            if((phase(nint(min(k_0_opt,max_disp_rel))) >= 0.)) then
              w_slow=(ph_opt_slow-ph_opt_slow_0_p1)/delta_ph_opt
            else
              w_slow=0.
            end if
          else
            w_slow=1.
          end if


!          w_slow=(slow-slow_0)/(slow_prop(i_slow+1)-slow_0)
!          w_slow=delta_vel/delta_vel_p1
!
!          if(w_slow > .5) then
!            w_slow=1.
!          else
!            w_slow=(2.)*delta_vel/delta_vel_p1
!          end if 


!if((omp_get_thread_num() == 0) .and. (z_down== 0.) .and. &
!   (i_mx == n_mx/2) .and. (i_my == n_my/2)) then
!      write(0,*)'slow',slow
!      write(0,*)'ph_opt_slow',ph_opt_slow
!      write(0,*)'ph_opt_slow_0',ph_opt_slow_0
!      write(0,*)'ph_opt_slow_0_p1',ph_opt_slow_0_p1
!      write(0,*)'w_slow',w_slow
!end if

!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow',slow
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'slow_0',slow_0
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'vel',vel
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'vel_p1',vel_p1
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'delta_vel',delta_vel
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow',ph_opt_slow
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow_0',ph_opt_slow_0
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'ph_opt_slow_0_p1',ph_opt_slow_0_p1
!
!if((omp_get_thread_num() == 0) .and. (z_down== 0.)) write(0,*)'w_slow',w_slow
!write(0,*)'w_slow',w_slow

          data_slice(i_mx,i_my,i_hx,1,i_w) =  &
             (   w_slow)*data_interp(i_mx,i_my,i_hx,1) +  &
             (1.-w_slow)*data_interp(i_mx,i_my,i_hx,2)
!stop

!          data_slice(i_mx,i_my,i_hx,1,i_w) =  &
!                 data_interp(i_mx,i_my,i_hx,2)

        !}
        end if
      !}
      end do
    !}
    end do
  !}
  end do
!} 
end if
  
if((var_slow) .and. (.not. adj)) then
!{
  if(prop_data) then
  !{
    call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
  !}
  end if
  if(prop_ddata) then
  !{
    call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,ddata_slice(:,:,:,1,i_w))
  !}
  end if
!}
end if

!} i_surf_comp loop
end do

!} i_w loop
end do

! deallocate temp allocatables
if(var_slow) then
  deallocate (map_slow,stat=ierr)  
end if
if(prop_data) then
  deallocate (data_interp,stat=ierr)  
  deallocate (temp_data,stat=ierr)  
end if
if(prop_ddata) then
  deallocate (temp_ddata,stat=ierr)  
end if

deallocate (diagx,stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with deallocation of diagx")
deallocate (diagy,stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with deallocation of diagy")
deallocate (tempx,stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with deallocation of tempx")
deallocate (tempy,stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with deallocation of tempy")

stat=0

return	
end function Ristow_salt_down_oper


function Ristow_salt_scatter_oper(adj,add, &
                     n_w_data,w_0_data,d_w_data, &
                     z_down,d_z_down,data_par, &
                     data_slice,ddata_slice,dslow_slice,slow_slice) &
         result(stat)

integer stat,n_w_data
logical add,adj
real w_0_data,d_w_data,z_down,d_z_down
type (data_param_type) data_par
real, dimension (:,:,:,:,:) :: slow_slice
complex, dimension (:,:,:,:,:) :: dslow_slice
complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
optional slow_slice

logical var_slow
integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx,i_mx,i_my,i_hx,i_w
integer i_surf_comp,surf_comp,i_slow,n_slow_split_eff
real slow_0,k0_sq,w_data
real sigma,rat,dscat,dz_w,k0
complex im_dz_w,ddscat

integer, allocatable,  dimension (:,:,:,:) :: map_slow

real,  dimension(Ristow_salt_par_loc%n_slow_split,n_comp_slow)   :: slow_prop
complex, dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: data_scat,ddata_scat


stat=0

return	
end function Ristow_salt_scatter_oper

function Ristow_salt_clean()  result(stat)
  integer stat,ierr
  stat=0
  if(allocated(phase)) deallocate (phase,stat=ierr)  
  if(ierr /=0) stat=stat+1
  if(allocated(magn)) deallocate (magn,stat=ierr)  
  if(ierr /=0) stat=stat+1
  if(allocated(k_mx)) deallocate (k_mx,stat=ierr)  
  if(ierr /=0) stat=stat+1
  if(allocated(k_my)) deallocate (k_my,stat=ierr)  
  if(ierr /=0) stat=stat+1
  if(allocated(k_hx)) deallocate (k_hx,stat=ierr)  
  if(ierr /=0) stat=stat+1
  if(allocated(k_sr_sq)) deallocate (k_sr_sq,stat=ierr)  
  if(ierr /=0) stat=stat+1

  return	
end function Ristow_salt_clean


function Ristow_salt_slow_prop(n_mx,n_my,n_hx,  &
                              slow_slice, slow_prop, map_slow) &
                              result (n_slow_split_eff)

integer n_slow_split_eff,ierr
integer n_mx,n_my,n_hx
integer,  dimension (:,:,:,:), intent(inout) :: map_slow
real, dimension (:,:,:,:,:), intent(in)      :: slow_slice
real,  dimension(:,:), intent(inout)           :: slow_prop

logical found_salt,not_found
integer i_mx,i_my,i_hx,i_surf_comp
integer n_slow_split_no_salt,i_slow,max_map_slow,i_slow_max_0
real min_slow,max_slow,range_slow,delta_slow,mean_slow
real min_slow_salt

do i_surf_comp=1,n_comp_slow
!{
if(slow_salt > 0.) then
!{
  min_slow=minval(slow_slice(:,:,:,:,i_surf_comp), &
                 mask=(slow_slice(:,:,:,:,i_surf_comp) > slow_salt))
  max_slow=maxval(slow_slice(:,:,:,:,i_surf_comp))
  found_salt=any(slow_slice(:,:,:,:,i_surf_comp) <= slow_salt)
!    write(0,*)'min_slow=',min_slow
!    write(0,*)'found_salt=',found_salt
!    write(0,*)'slow_salt=',slow_salt
  if(found_salt) then
    min_slow_salt=minval(slow_slice(:,:,:,:,i_surf_comp))
    n_slow_split_no_salt=Ristow_salt_par_loc%n_slow_split-1
  else
    n_slow_split_no_salt=Ristow_salt_par_loc%n_slow_split
  end if
!}
else
!{
  found_salt=.false.
  min_slow=minval(slow_slice(:,:,:,:,i_surf_comp))
  n_slow_split_no_salt=Ristow_salt_par_loc%n_slow_split
  if(Ristow_salt_par_loc%min_slow_const >0) then
    min_slow=Ristow_salt_par_loc%min_slow_const
  else
    if(Ristow_salt_par_loc%max_vel_const >0) then
      min_slow=1./Ristow_salt_par_loc%max_vel_const
    else
      min_slow=minval(slow_slice(:,:,:,:,i_surf_comp))
    end if
  end if
  if(Ristow_salt_par_loc%max_slow_const >0) then
    max_slow=Ristow_salt_par_loc%max_slow_const
  else
    if(Ristow_salt_par_loc%min_vel_const >0) then
      max_slow=1./Ristow_salt_par_loc%min_vel_const
    else
      max_slow=maxval(slow_slice(:,:,:,:,i_surf_comp))
    end if
  end if
!}
end if

!    write(0,*)'found_salt=',found_salt
!    write(0,*)'min_slow=',min_slow
!    write(0,*)'slow_salt=',slow_salt
!    write(0,*)'min_slow_salt=',min_slow_salt
!    write(0,*)'n_slow_split_no_salt=',n_slow_split_no_salt
!
!    write(0,*)'max_slow=',max_slow

!range_slow=max_slow-min_slow
!delta_slow=range_slow/max(1,n_slow_split_no_salt-1)

range_slow=max_slow/min_slow

delta_slow=(1.00001*range_slow)**(1./max(1,n_slow_split_no_salt-1))

!range_slow=(min_slow/max_slow)
!delta_slow=range_slow**(-1./max(1,n_slow_split_no_salt-1))
!delta_slow=range_slow**(max(1,n_slow_split_no_salt-1))

!write(0,*)'expon=',1./max(1,n_slow_split_no_salt-1)
!write(0,*)'range_slow=',range_slow

!write(0,*)'First: range_slow=',range_slow
!write(0,*)'First: delta_slow=',delta_slow

mean_slow=.5*(max_slow+min_slow)
if(delta_slow < (1./max_acc_sin)) then
  if(range_slow < (1./max_acc_sin)) then
    n_slow_split_no_salt= 2 
    delta_slow=1.00001*range_slow
  else
    n_slow_split_no_salt= & 
       min(n_slow_split_no_salt, &
       max(1,1+ceiling(log(range_slow)/log_min_ratio)))
    delta_slow=(1.00001*range_slow)**(1./max(1,n_slow_split_no_salt-1))
!write(0,*)'Second: n_slow_split_no_salt=',n_slow_split_no_salt
!write(0,*)'Second: delta_slow=',delta_slow
  end if
else
!  if((omp_get_thread_num() == 0)) then
    write(0,*)'ATTENTION: ratio between velocity exceeds limits:' 
    write(0,*)'Desired limit=',(1./max_acc_sin)
    write(0,*)'Actual ratio=',delta_slow
!  end if
end if


! set reference velocities 
if(found_salt) then
  n_slow_split_eff= n_slow_split_no_salt + 1
  slow_prop(1,i_surf_comp)=min_slow_salt
  i_slow_max_0=2
else
  n_slow_split_eff= n_slow_split_no_salt 
  i_slow_max_0=1
end if

! write(0,*)'n_slow_split_eff=',n_slow_split_eff

! This is a bug when commented out the omp_ call
!if((omp_get_thread_num() == 0)) then
  n_total_slow_ref=n_total_slow_ref+n_slow_split_eff
  nz_total=nz_total+1
!end if

if(n_slow_split_no_salt > 1) then
  do i_slow=i_slow_max_0,n_slow_split_eff
     slow_prop(i_slow,i_surf_comp)=min_slow*(delta_slow**(i_slow-adj_slow-i_slow_max_0+1))
  end do
else
  if(slow_const > 0.) then  
    slow_prop(i_slow_max_0,i_surf_comp)=slow_const
  else
    slow_prop(i_slow_max_0,i_surf_comp)=min_slow+(1.-adj_slow-i_slow_max_0+1)*delta_slow
  end if
end if


!if((omp_get_thread_num() == 0)) then
!  write(0,*)'minval(slow_slice)',minval(slow_slice),'  maxval(slow_slice)',maxval(slow_slice)
! write(0,*)'n_slow_split_eff=',n_slow_split_eff
!  write(0,*)'n_slow_split_no_salt',n_slow_split_no_salt
!  write(0,*)'delta_slow=',delta_slow
!  write(0,*)'min_slow=',min_slow
!  write(0,*)'max_slow=',max_slow
!  write(0,*)'mean_slow=',mean_slow
!  write(0,*)'range_slow=',range_slow
!  write(0,*)'ratio of logs=',log(range_slow)/log_min_ratio
!  write(0,*)'slow_prop=',slow_prop
!  write(0,*)'n_total_slow_ref=',n_total_slow_ref
!  write(0,*)'nz_total=',nz_total
!
!  write(0,*)'Running Average number of reference velocities=',float(n_total_slow_ref)/nz_total
!
!end if


if (n_slow_split_eff >1) then
!{
  do i_hx=1,n_hx
  !{
    do i_my=1,n_my
    !{
      do i_mx=1,n_mx
      !{
        not_found=.true.
        i_slow=1
        do while (not_found .and. i_slow <=n_slow_split_eff-1)
          if((slow_prop(i_slow,i_surf_comp)  <= &
              slow_slice(i_mx,i_my,i_hx,1,i_surf_comp)).and.&
             (slow_prop(i_slow+1,i_surf_comp)>= &
              slow_slice(i_mx,i_my,i_hx,1,i_surf_comp)) )then
            map_slow(i_mx,i_my,i_hx,i_surf_comp)= i_slow
            not_found=.false.
          else
            i_slow=i_slow+1
          end if
          if(i_slow >= n_slow_split_eff ) then
            write(0,*)'i_surf_comp=',i_surf_comp
            write(0,*)'i_slow=',i_slow
            write(0,*)'slow_slice=',slow_slice(i_mx,i_my,i_hx,1,i_surf_comp)
            write(0,*)'slow_prop(:,i_surf_comp)=',slow_prop(:,i_surf_comp)
            call seperr('Problems in defining velocity map')
          end if
       end do
      !}
      end do
    !}
    end do
  !}
  end do
!}
else
!{
  map_slow(:,:,:,i_surf_comp)=1
!}
end if

!} i_surf_comp loop
end do

!  write(0,*)'minval(map_slow(SOU_COMP))',minval(map_slow(:,:,:,SOU_COMP)),'  maxval(map_slow)',maxval(map_slow(:,:,:,SOU_COMP))
!  write(0,*)'minval(map_slow(REC_COMP))',minval(map_slow(:,:,:,REC_COMP)),'  maxval(map_slow)',maxval(map_slow(:,:,:,REC_COMP))
!!write(0,*)'shape(slow_slice)',shape(slow_slice)
!write(0,*)'n_slow_split_eff',n_slow_split_eff



end function Ristow_salt_slow_prop

end module Ristow_salt_mod
