!------------------------------------
program testphase


use sep
implicit none
!-------------------------------------
integer					:: n1, i
real 					:: delta, epsilon, V0, t0, rhovv, rhovh, rhovn, z0m
real                                    :: o1, d1
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:), allocatable         :: gammap
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm
 
!--------------------------------

call sep_init()

call from_param( "n1", n1) ! Angle axis
call from_param( "o1", o1)
call from_param( "d1", d1)

call from_param( "V0", V0) ! Velocity parameters (true and perturbed) 
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "t0", t0)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call sep_close()

! Allocation of the ADCIG where we plot the predicted RMO curve

vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

! Vector of the phase aperture angles
allocate(gammap(n1))
do i=1,n1
   gammap(i)=o1+(i-1)
end do

! COMPUTATION OF THE RESIDUAL MOVEOUT USING THE PHASE ANGLE

! Elliptic group slowness (the input angle is the phase angle - wrong angle)
allocate(Sellg(n1),Svtig(n1))
Sellg=sqrt((1/vvm)**2*cosd(gammap)**2+(1/vhm)**2*sind(gammap)**2)
! VTI group slowness (the input an1le is the phase angle - wrong angle)
Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the phase angle - wrong angle)
allocate(DSellg1(n1), DSellg2(n1))
DSellg1=-(1/vvm)**2*cosd(gammap)**2/Sellg
DSellg2=-(1/vhm)**2*sind(gammap)**2/Sellg

! Computation of the derivatives of the VTI group slownesses (using the phase angle - wrong one)
allocate(DSvtig1(n1),DSvtig2(n1),DSvtig3(n1))
DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))
DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))
DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))

! Computation of the derivative of the depth with respect to the different perturbations
allocate(DZ1(n1),DZ2(n1),DZ3(n1))
DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammap))-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammap))
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammap))

deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

! We compute the theoretical ADCIG
z0m=t0*vvm;
allocate(rmo(n1),zm(n1))
rmo=z0m*(DZ1*(rhovv-1)+DZ2*(rhovh-1)+DZ3*(rhovn-1)) ! Computation of the theoretical RMO
zm=z0m+rmo

call sep_write(zm)

end program testphase
