!------------------------------------
program velocity_spectra


use sep
implicit none
!-------------------------------------
integer					:: nt, ng, n1, n2, n3, it, ig, i1, i2, i3, j, k
real 					:: delta, epsilon, delta2, epsilon2, eta2, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp
real                                    :: ot, og, o1, o2, o3, dt, dg, d1, d2, d3
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:,:), allocatable       :: adcig
real, dimension(:,:,:), allocatable     :: vspectra
real, dimension(:),allocatable          :: gammap, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm
integer, dimension(:), allocatable     :: rmo2

!--------------------------------

call sep_init()

! Parameters of the initial ADCIG
call from_param( "nt", nt)
call from_param( "ot", ot)
call from_param( "dt", dt)
call from_param( "it", it)
call from_param( "ng", ng)
call from_param( "og", og)
call from_param( "dg", dg)

! Parameters of the semblance cube
call from_param( "n1", n1)
call from_param( "o1", o1)
call from_param( "d1", d1)
call from_param( "n2", n2)
call from_param( "o2", o2)
call from_param( "d2", d2)
call from_param( "n3", n3)
call from_param( "o3", o3)
call from_param( "d3", d3)

! Parameters of the velocity model
call from_param( "V0", V0)
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call sep_close()

! Allocation of the ADCIG
allocate(adcig(nt,ng))
call sep_read(adcig,'adcig')

! Allocation of the velocity spectra cube
allocate(vspectra(n1,n2,n3))
vspectra=0

! Definition of the velocities
! True velocities
vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);

! Modified velocities
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

write(0,*) vvm
write(0,*) vhm
write(0,*) vnm

! Definition of the vector of aperture angles
allocate(gammap(n1),gammag(n1))
allocate(betarp(n1),betarg(n1))
allocate(betasp(n1),betasg(n1))

do i1=1,n1
   gammap(i1)=o1+d1*(i1-1)
   betarp(i1)=gammap(i1) ! Only true for flat reflectors
   betasp(i1)=-gammap(i1) ! Only true for flat reflectors
end do

! Computation of the different quantities used to compute the derivatives of the depth with respect to pertubations.

! Elliptical phase velocity at the different propagation angles 
allocate(Vellbetarp(n1),Vellbetasp(n1))
Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)

! VTI phase velocity at the different propagation angles
allocate(Vvtibetarp(n1),Vvtibetasp(n1))
Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)

! Derivative of the elliptic phase velocity with respect to the phase angle
allocate(DVellbetarp(n1),DVellbetasp(n1))
DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)

! Derivative of the VTI phase velocity with respect to the phase angle
allocate(DVvtibetarp(n1), DVvtibetasp(n1))
DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))

! Computation of the group propagation angles
betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))

! Computation of the group aperture angle
gammag=0.5*(betarg-betasg)

deallocate(betarp,betarg,betasp,betasg,Vellbetarp,Vellbetasp,Vvtibetarp,Vvtibetasp,DVellbetarp,DVellbetasp,DVvtibetarp,DVvtibetasp)

! COMPUTATION OF THE RESIDUAL MOVEOUT

! Elliptic group slowness (the input angle is the group angle)
allocate(Sellg(n1),Svtig(n1))
Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)

! VTI group slowness (the input angle is the group angle)
Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
allocate(DSellg1(n1), DSellg2(n1))
DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg

! Computation of the derivatives of the VTI group slownesses
allocate(DSvtig1(n1),DSvtig2(n1),DSvtig3(n1))
DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))

! Computation of the derivative of the depth with respect to the different perturbations
allocate(DZ1(n1),DZ2(n1),DZ3(n1))
DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))

!do j=1,ng2
!   write(0,*) DZ1(j)
!end do
!write(0,*) '************'
!do j=1,ng2
!   write(0,*) DZ2(j)
!end do
!write(0,*) '************'
!do j=1,ng2
!   write(0,*) DZ3(j)
!end do
 
deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

! Actual computation of the velocity spectra
allocate(rmo(n1),rmo2(n1))
!do i1=171,205
   !do i1=150,250 ! Loop over the different traveltimes
   !  write(0,*) i1," out of ",n1
z0m=ot+(it-1)*dt
write(0,*) 'Maximum value: ', adcig(it,1)

do i3=1,n3 ! Loop over the different values of the NMO velocity perturbations
   delta2=(o3+(i3-1)*d3)
   drhovn=vnm/(vvm*sqrt(1+2*delta2))-1
   ! write(0,*) 'drhovn', drhovn
   ! Computation of the residual moveout for the different perturbation values
   
   do i2=1,n2 ! Loop over the different values of the ETA parameter
      ! eta2=(o2+(i2-1)*d2)
      epsilon2=(o2+(i2-1)*d2)
      ! delta2=.5*((vnm/(vvm*(o3+(i3-1)*d3)))**2-1)
      drhovh=vhm/(vvm*sqrt(1+2*epsilon2))-1
      !delta2=0.5*((o3+(i3-1)*d3)**2-1)
      !epsilon2=(1+2*delta2)*eta2+delta2
      !drhovh=vhm/(vvm*sqrt(1+2*epsilon2))-1
      
      rmo=z0m*(DZ2*drhovh+DZ3*drhovn) ! That is a vector 
         !write(0,*) rmo
         rmo2=(rmo-ot)/dt+0.5
         !write(0,*) rmo2
         ! Computation of the semblance values
         vspectra(1,i2,i3)=adcig(it+rmo2(1),1)
         do i1=2,n1
            if (it+rmo2(i1)>0 .AND. it+rmo2(i1)<nt+1) then
               vspectra(i1,i2,i3)=vspectra(i1-1,i2,i3)+adcig(it+rmo2(i1),i1)
            else 
               vspectra(i1,i2,i3)=vspectra(i1-1,i2,i3)
            end if
!            write(0,*) 'RMO:', rmo2(i1),'Adcig:', adcig(it+rmo2(i1),i1)
         end do
      end do
   end do
!end do

!do ig2=1,ng2
!   write(0,*) adcig(179,ig2,1,1,1)
!end do


call sep_write(vspectra)

deallocate(adcig)

end program velocity_spectra
