program Velspec
  use sep
  implicit none

!-------------------------------------

  integer                                 :: nang, n1, n2, n3, nt, nx, n1a, n2a, n3a, n4a, n1v, n2v, n3v
  integer                                 :: nvh, nvn
  real                                    :: ovh, ovn, dvh, dvn
  integer                                 :: i, i1, i2, i3, i4
  real					  :: o1, o2, o3, o1a, o2a, o3a, o4a, o1v, o2v, o3v
  real					  :: d1, d2, d3, d1a, d2a, d3a, d4a, d1v, d2v, d3v
  real                                    :: CMPmin, CMPmax, CMP
  integer                                 :: CMPvmin, CMPvmax, CMPv, CMPamin, CMPamax, CMPa
  real, dimension(:,:,:,:), allocatable   :: adcig
  real, dimension(:,:,:), allocatable     :: velocity
  real, dimension(:,:,:), allocatable     :: velocities
  real, dimension(:,:,:), allocatable     :: vspectra
  real, dimension(:,:,:,:), allocatable   :: vspectras
  real, dimension(:), allocatable         :: gammap, gammag
  real, dimension(:,:), allocatable       :: DZ
  
 !-------------------------------------
 
  call sep_init()               ! Initialize I/O module, "sep".

  call from_par( "n1", n1)
  call from_par( "o1", o1)
  call from_par( "d1", d1)
  
  call from_par( "nvh", nvh)
  call from_par( "nvn", nvn)
  call from_par( "ovh", ovh)
  call from_par( "ovn", ovn)
  call from_par( "dvh", dvh)
  call from_par( "dvn", dvn)
 
  call from_par( "nt", nt)
  call from_par( "nx", nx)

  call from_par( "CMPvmin", CMPvmin)
  call from_par( "CMPvmax", CMPvmax)
  
  call from_aux ( "adcig", "n1", n1a)
  call from_aux ( "adcig", "n2", n2a)
  call from_aux ( "adcig", "n3", n3a)
  call from_aux ( "adcig", "n4", n4a)
  call from_aux ( "adcig", "o2", o2a)
  call from_aux ( "adcig", "d2", d2a)
  call from_aux ( "adcig", "o4", o4a)
  call from_aux ( "adcig", "d4", d4a)

  call from_aux( "velocity", "n1", n1v)
  call from_aux( "velocity", "o1", o1v)
  call from_aux( "velocity", "d1", d1v)
  call from_aux( "velocity", "n2", n2v)
  call from_aux( "velocity", "n3", n3v)
  
!  call sep_close()

  write(0,*) "n1: ", n1
  write(0,*) "nvh: ", nvh
  write(0,*) "nvn: ", nvn

  write(0,*) "nt: ", nt
  write(0,*) "nx: ", nx

  write(0,*) "CMPmin: ", CMPvmin
  write(0,*) "CMPmax: ", CMPvmax

  write(0,*) "n1a: ", n1a
  write(0,*) "n2a: ", n2a
  write(0,*) "n3a: ", n3a
  write(0,*) "n4a: ", n4a

  write(0,*) "o1a: ", o1a
  write(0,*) "o2a: ", o2a

  write(0,*) "n1v: ", n1v
  write(0,*) "n2v: ", n2v
  write(0,*) "n3v: ", n3v
  
  ! Read input data from disk
  allocate( adcig(n1a,n2a,n3a,n4a) )
  allocate( velocity(n1v,n2v,n3v) )

  call sep_read(adcig, 'adcig') 
  call sep_read(velocity,'velocity')
 
  ! Allocation of the different vectors and definition of the phase angle vector
  allocate(gammap(n2a))
  do i=1,n2a
     gammap(i)=o2a+d2a*(i-1)
  end do

  allocate(gammag(n2a),DZ(n2a,3),vspectra(n1,nvh,nvn),vspectras(n1,nvh,nvn,2*nx+1))
  vspectra=0
  vspectras=0
  
  ! Definition of the CMP locations
  CMPmin=o1v+(CMPvmin-1)*d1v
  CMPmax=o1v+(CMPvmax-1)*d1v
  CMPamin=(CMPmin-o4a)/d4a+1.5
  CMPamax=(CMPmax-o4a)/d4a+1.5
  
  n2=CMPvmax-CMPvmin+1
  o2=CMPmin
  d2=d1v
  
  n3=3
    
  allocate(velocities(n1,n2,n3))
  velocities=0
  
  do CMPv=CMPvmin-nx,CMPvmax+nx
     
     CMP=o1v+(CMPv-1)*d1v
     CMPa=(CMP-o4a)/d4a+1.5
     write(0,*) "CMPv :", CMPv, "from ", CMPvmin, "to", CMPvmax
     
     do i1=1,n1-1 ! Computation of the velocity spectras for one CMP - Loop over depth
        
        ! Computation of the group angles from the phase angles (gammag initialized to zero in the routine)
        call Phase2group(gammap, gammag, n2a, velocity(CMPv,1,i1), velocity(CMPv,1,i1), velocity(CMPv,1,i1))
        
        ! Computation of the Residual Moveout (DZ initialized to zero in the routine)
        call Group2derivatives(DZ,gammap, gammag, n2a, velocity(CMPv,1,i1), velocity(CMPv,1,i1), velocity(CMPv,1,i1))
        
        ! Allocation of the velocity spectra cube
        nang=(asind(velocity(CMPv,1,1)/velocity(CMPv,1,i1))-o2a)/d2a+1.5
        call Derivatives2spectra(vspectra, DZ, adcig, i1, CMPa, nang, n1a, n2a, n3a, n4a, n1, o1, d1, nvh, ovh, dvh, nvn, ovn, dvn)
       
     end do
   
     !   write(0,*) "Hello"
     !   write(0,*) vspectra(100,50,50)
  
     ! Smoothing along the depth axis
     call Timesmoothing(vspectra, nt, n1, nvh, nvn)
     
     ! Storage of the velocity spectra
     if (CMPv .LE. CMPvmin+nx) then
        vspectras(:,:,:,CMPv-(CMPvmin-nx)+1)=vspectra
!        write(0,*) "Inf"
!        write(0,*) vspectra(100,50,50)
     else
        ! Reordering of the velocity spectras
        do i=1,2*nx
           vspectras(:,:,:,i)=vspectras(:,:,:,i+1)
        end do
        ! Storage of the velocity spectra
        vspectras(:,:,:,2*nx+1)=vspectra
!         write(0,*) "Sup"
!         write(0,*) vspectra(100,50,50)
     end if
   
     vspectra=0

     if (CMPv .GE. CMPvmin+nx) then
        ! Smoothing of the velocity spectras along CMP-axis
        call CMPsmoothing(vspectras, vspectra, nx, n1, nvh, nvn)
!        write(0,*) "CMP smoothing"
        ! Actual picking
        call Picking(velocities, vspectra, CMPv-(CMPvmin+nx)+1, n1, n2, nvh, ovh, dvh, nvn, ovn, dvn)
     end if
        

  
  end do
  
  call to_history( "n1", n1)
  call to_history( "o1", o1)
  call to_history( "d1", d1)
  call to_history( "n2", n2)
  call to_history( "o2", o2)
  call to_history( "d2", d2)
  call to_history( "n3", 3)
  call to_history( "o3", 0)
  call to_history( "d3", 1)
  
  call sep_write(velocities) 
  
!  call Picking(velocities, vspectra, n1, n2, o2, d2, n3, o3, d3)
   
!  write(0,*) "Picking complete"

!  call to_history( "n2", 3)
!  call to_history( "o2", 0)
!  call to_history( "d2", 1)

!  call sep_write(velocities) 
  


!  deallocate( velocity,adcig)
  
  call exit(0)
  
end program Velspec



!-------------------------------------

subroutine Phase2group(gammap, gammag, n, vvm, vhm, vnm) 
  integer                               :: n, i
  real                                  :: o, d, vvm, vhm, vnm
  real, dimension (n)                   :: gammap
  real, dimension (n)                   :: betarp, betasp, betarg, betasg
  real, dimension (n), intent(OUT)      :: gammag
  real, dimension (n)                   :: Vellbetarp, Vellbetasp, Vvtibetarp, Vvtibetasp
  real, dimension (n)                   :: DVellbetarp,DVellbetasp, DVvtibetarp, DVvtibetasp

  gammag=0

  ! Definition of the phase_angles
  betarp=gammap ! Only true for flat reflectors
  betasp=-gammap ! Only true for flat reflectors
  
  ! Elliptical phase velocity at the different propagation angles 
  Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
  Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)
  
  ! VTI phase velocity at the different propagation angles
  Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
  Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)
  
  ! Derivative of the elliptic phase velocity with respect to the phase angle
  DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
  DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)
  
  ! Derivative of the VTI phase velocity with respect to the phase angle
  DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
  DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))
  
  ! Computation of the group propagation angles
  betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
  betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))
  
  ! Computation of the group aperture angle
  gammag=0.5*(betarg-betasg)
  
end subroutine Phase2group

!-------------------------------------

subroutine Group2derivatives(DZ, gammap, gammag, n, vvm, vhm, vnm) 

  integer                               :: n
  real                                  :: vvm, vhm, vnm
  real, dimension (n)                   :: gammap, gammag, Sellg, Svtig
  real, dimension (n)                   :: DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3
  real, dimension (n,3),intent(OUT)     :: DZ
  
  DZ=0
  
  ! Elliptic group slowness (the input angle is the group angle)
  
  Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)
  
  ! VTI group slowness (the input angle is the group angle)
  Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)
  
  ! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
  
  DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
  DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg
  
  ! Computation of the derivatives of the VTI group slownesses
  DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  
  ! Computation of the derivative of the depth with respect to the different perturbations
  DZ(:,1)=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
  DZ(:,2)=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
  DZ(:,3)=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))

end subroutine Group2derivatives

!-------------------------------------

subroutine Derivatives2spectra(vspectra, DZ, adcig, z0m, CMPa, nang, n1a, n2a, n3a, n4a, n1, o1, d1, nvh, ovh, dvh, nvn, ovn, dvn)

  integer                               :: nang, n1, nvh, nvn, ivh, ivn, i, i1
  real                                  :: o1, ovh, ovn, d1, dvh, dvn
  real                                  :: drhovh, drhovn
  integer                               :: z0m, CMPa
  real                                  :: z0mr
  real, dimension (n2a,3)               :: DZ
  real, dimension (n2a)                 :: rmo
  real, dimension (n1,nvh,nvn),intent(OUT):: vspectra
  real, dimension (:,:), allocatable    :: intvspectra, intdenum
  real, dimension (n1a,n2a,n3a,n4a)     :: adcig
  
  allocate(intvspectra(nvh,nvn), intdenum(nvh,nvn))
  intvspectra=0
  intdenum=0
  
!  write(0,*) "Derivatives2spectra"
  z0mr=(o1+z0m-1)*d1
  
!  write(0,*) 'z0mr', z0mr
!  write(0,*) 'z0m', z0m
  
  do ivh=1,nvh ! Loop over the different values of HORIZONTAL perturbations
     drhovh=(ovh+(ivh-1)*dvh)
!     write(0,*) "ivh :", ivh

     do ivn=1,nvn ! Loop over the different values of the NMO velocity perturbations
        drhovn=(ovn+(ivn-1)*dvn)
!        write(0,*) "ivn :", ivn

        ! Computation of the residual moveout for the different perturbation values
        rmo=z0mr*(DZ(:,2)*drhovh+DZ(:,3)*drhovn) ! That is a vector 
        rmo=(rmo-o1)/d1+1.5      
        
        ! Computation of the semblance values (the fourth parameter in adcig is CMP)
        do i=1,nang
           if (z0m+rmo(i)>0 .AND. z0m+rmo(i)<n1a+1) then
              intvspectra(ivh,ivn)=intvspectra(ivh,ivn)+adcig(z0m+rmo(i),i,1,CMPa)
              intdenum(ivh,ivn)=intdenum(ivh,ivn)+adcig(z0m+rmo(i),i,1,CMPa)**2
           end if
        end do
     end do
  end do
  
  do ivh=1,nvh
     do ivn=1,nvn
        vspectra(z0m,ivh,ivn)=intvspectra(ivh,ivn)**2/(nang*intdenum(ivh,ivn))
!        write(0,*) "test", vspectra(z0m,ivh,ivn)
        
     end do
  end do
  
        !write(0,*) "Hello"
  !   write(0,*) vspectra(z0m,50,50)

end subroutine Derivatives2spectra

!-------------------------------------

subroutine Timesmoothing(vspectra, nt, n1, n2, n3)

  integer                               :: nt, n1, n2, n3
  real, dimension (n1,n2,n3),intent(OUT):: vspectra
  real, dimension (n1,n2,n3)            :: intvspectra

  intvspectra=vspectra

  do i1=nt+1,n1-nt
     do i2=1,n2
        do i3=1,n3
           do it=-nt,nt 
              vspectra(i1,i2,i3)=vspectra(i1,i2,i3)+intvspectra(i1+it,i2,i3)
           end do
           vspectra(i1,i2,i3)=vspectra(i1,i2,i3)/(2*nt+1)
        end do
     end do
  end do
  
end subroutine Timesmoothing

!-------------------------------------

subroutine CMPsmoothing(vspectras, vspectra, nx, n1, nvh, nvn)

  integer                               :: nx, n1, nvh, nvn
  real, dimension (n1,nvh,nvn),intent(OUT):: vspectra
  real, dimension (n1,nvh,nvn,nx)         :: vspectras
  
  do i1=1,n1
     do ivh=1,nvh
        do ivn=1,nvn
           do ix=1,2*nx+1
!              write(0,*) vspectras(i1,ivh,ivn,ix)

              vspectra(i1,ivh,ivn)=vspectra(i1,ivh,ivn)+vspectras(i1,ivh,ivn,ix)
           end do
           vspectra(i1,ivh,ivn)=vspectra(i1,ivh,ivn)/(2*nx+1)
        end do
     end do
  end do

end subroutine CMPsmoothing

!-------------------------------------

subroutine Picking(velocities, vspectra, CMP, n1, n2, nvh, ovh, dvh, nvn, ovn, dvn)

  integer                               :: n1, CMP, nvh, nvn
  real                                  :: ovh, dvh, ovn, dvn
  real                                  :: rhovh, rhovn, semblance
  real, dimension (n1,nvh,nvn)          :: vspectra
  real, dimension (n1,n2,3), intent(OUT):: velocities
  
  do i1=1,n1
     semblance=0
     rhovh=0
     rhovn=0
     do ivh=1,nvh
        do ivn=1,nvn
    !       write(0,*) "Semblance:", vspectra(i1,ivh,ivn)
           if (vspectra(i1,ivh,ivn) .GT. semblance) then
              semblance=vspectra(i1,ivh,ivn)
              rhovh=ovh+(ivh-1)*dvh
              rhovn=ovn+(ivn-1)*dvn
     !         write(0,*) "Semblance:", semblance
           end if
        end do
     end do
!     if ((semblance .GT. 0.5) .AND. (semblance .GT. 1.5*vspectra(i1,n1/2,n1/2)) ) then
     if (semblance .GT. 0.0 ) then
        velocities(i1,CMP,1)=rhovh
        velocities(i1,CMP,2)=rhovn
        velocities(i1,CMP,3)=semblance
     else 
        velocities(i1,CMP,1)=0
        velocities(i1,CMP,2)=0
        velocities(i1,CMP,3)=semblance
     end if
  end do
  
  
end subroutine Picking
