module Pre2D_vti_ssf_mod

  use Pre2D_vti_ssf_types
  use lloyd_mod

  use slow_mod
  use gendown3d_types
  use down_process_types
  use data_types
  use image_types
  use slow_types

  use fft_data_mod

  use bound_mod

  implicit none

  type (Pre2D_vti_ssf_param_type), private :: Pre2D_vti_ssf_par_loc
  logical, private :: debug1,debug2,verb
  logical, private :: time_mig,apply_any_bound
  integer, private :: n_slow_total
  real, private :: slow_const,slow_salt,epsi,min_prop_cos_sq
  real, private :: d_k_norm,max_disp_rel,set_slow
  real, private :: vti_delta_0,vti_epsilon_0,vti_eta_0
  real, private :: vti_delta_const,vti_epsilon_const,vti_eta_const
  real, allocatable, dimension (:), private :: magn,phase,phase_pos
  real, allocatable, dimension (:), private :: k_mx,k_my,k_hx
  real, allocatable, dimension (:,:,:,:), private :: k_sr_sq
   

contains

  subroutine init_Pre2D_vti_ssf_par(Pre2D_vti_ssf_par)
    type (Pre2D_vti_ssf_param_type) Pre2D_vti_ssf_par

    Pre2D_vti_ssf_par_loc=Pre2D_vti_ssf_par

  end subroutine init_Pre2D_vti_ssf_par

  function Pre2D_vti_ssf_init(gen_par,data_par,slow_par)  result(info)
    type (gendown3d_info_type) info
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par

    character(len=128) :: desc_timer

    integer i_k_mx,i_k_my,i_k_hx,ierr,i_k_norm
    real kmx,kmy,khx,pi2,k_norm_sq,k_norm
    real n_cycles_ksq,k_start,k_cutoff,k_decay,sigma_magn_sq,edge_gauss,gauss_magn
    real delta_k

    call init_lloyd()
    verb=gen_par%verb
    debug1=gen_par%debug1
    debug2=gen_par%debug2
!    info=info_default

    if(slow_par%n_comp_slow /= 2) then
       write(0,*)'Reset n_comp_slow to 2'
       slow_par%n_comp_slow = 2
    end if

    if((slow_par%vti_delta_file == SLOW_NO_FILE) .or. &
       (slow_par%vti_eta_file == SLOW_NO_FILE)) then

      if(Pre2D_vti_ssf_par_loc%n_vti_delta /=1) &
        call seperr("n_vti_delta should be 1")
      if(Pre2D_vti_ssf_par_loc%n_vti_eta /=1) &
        call seperr("n_vti_eta should be 1")
      n_slow_total=Pre2D_vti_ssf_par_loc%n_slow_split
         write(0,*)'Computed n_slow_total=',n_slow_total

      vti_delta_const=Pre2D_vti_ssf_par_loc%vti_delta
      vti_epsilon_const=Pre2D_vti_ssf_par_loc%vti_epsilon
      vti_eta_const=Pre2D_vti_ssf_par_loc%vti_eta

      if(vti_eta_const == huge(vti_eta_const)) then
        if((vti_epsilon_const /= huge(vti_epsilon_const)).and.(vti_delta_const /= huge(vti_delta_const))) then
          vti_eta_const=(vti_epsilon_const-vti_delta_const)/(1.+2*vti_delta_const)
          write(0,*)'Computed eta=',vti_eta_const
        else
          call seperr("It needs both epsilon and delta")
        end if
      end if
      if(vti_epsilon_const == huge(vti_epsilon_const)) then
        if((vti_eta_const /= huge(vti_eta_const)).and.(vti_delta_const /= huge(vti_delta_const))) then
          vti_epsilon_const=vti_delta_const+vti_eta_const*(1.+2*vti_delta_const)
          write(0,*)'Computed epsilon=',vti_epsilon_const
        else
          call seperr("It needs both eta and delta")
        end if
      end if
    else
      n_slow_total=Pre2D_vti_ssf_par_loc%n_slow_split* &
                   Pre2D_vti_ssf_par_loc%n_vti_delta* &
                   Pre2D_vti_ssf_par_loc%n_vti_eta
                   write(0,*)'Computed n_slow_total=',n_slow_total
      vti_delta_const=huge(vti_delta_const)
      vti_eta_const=huge(vti_eta_const)
      vti_epsilon_const=huge(vti_epsilon_const)
    end if
         write(0,*)'vti_delta_const=',vti_delta_const
         write(0,*)'vti_eta_const=',vti_eta_const
         write(0,*)'vti_epsilon_const=',vti_epsilon_const
  
    select case (Pre2D_vti_ssf_par_loc%ref_slow(1:3)) 
    case ('min')
      set_slow=1.
    case ('max')
      set_slow=0.
    case ('avg')
      set_slow=.5
    case default
      call seperr('Not supported reference_slowness ')
    end select

    apply_any_bound=gen_par%apply_bound

    time_mig=gen_par%time_mig
    slow_const=gen_par%slow_const
    pi2=2.*acos(-1.)
    min_prop_cos_sq=cos(gen_par%max_propag_angle*pi2/360.)**2
    epsi=epsilon(1.)

    if (slow_par%slowness) then
       if((Pre2D_vti_ssf_par_loc%slow_salt > 0.).and.(slow_const <- 0.)) then
          slow_salt=Pre2D_vti_ssf_par_loc%range_salt* &
                    (Pre2D_vti_ssf_par_loc%slow_salt)
          if(Pre2D_vti_ssf_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
       else 
          slow_salt=-1.
       end if
    else
       if((Pre2D_vti_ssf_par_loc%vel_salt > 0.).and.(slow_const <= 0.)) then
          slow_salt=Pre2D_vti_ssf_par_loc%range_salt* &
                    (1./Pre2D_vti_ssf_par_loc%vel_salt)
          if(Pre2D_vti_ssf_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
       else 
          slow_salt=-1.
       end if
    end if

    if(gen_par%debug1) write(0,*)'Init slow_salt=',slow_salt

    if(gen_par%debug1) write(0,*)'min_prop_cos_sq=',min_prop_cos_sq
    if(gen_par%debug1) write(0,*)'n_mx_pad=',data_par%n_mx_pad
    if(gen_par%debug1) write(0,*)'n_my_pad=',data_par%n_my_pad
    if(gen_par%debug1) write(0,*)'n_hx_pad=',data_par%n_hx_pad

    if(data_par%n_hy_pad /= 1 ) then
       call seperr(" Pre2D: Trying to use Prestack migration with n_hy > 1")
    end if

    allocate (phase(Pre2D_vti_ssf_par_loc%n_disp_rel),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of phase ")
    allocate (magn(Pre2D_vti_ssf_par_loc%n_disp_rel),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of magn ")
    allocate (phase_pos(Pre2D_vti_ssf_par_loc%n_disp_rel),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of phase_pos ")

    allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_my(data_par%n_my_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
    allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")

    allocate (k_sr_sq(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2), &
    stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_sr_sq ")


!    write(0,*)'d_kmx=',data_par%d_kmx
!    write(0,*)'d_kmy=',data_par%d_kmy
!    write(0,*)'kmx_0=',data_par%kmx_0
!    write(0,*)'kmy_0=',data_par%kmy_0

    do i_k_mx=1,data_par%n_mx_pad
       kmx  = i_k_mx - 1.
       if(kmx <= float(data_par%n_mx_pad/2)) then
          k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
       else
          k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
       end if
    end do

    do i_k_my=1,data_par%n_my_pad
       kmy = i_k_my - 1.
       if(kmy <= float(data_par%n_my_pad/2)) then
          k_my(i_k_my) =  data_par%kmy_0+data_par%d_kmy* kmy
       else
          k_my(i_k_my) = -data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad)
       end if
    end do

    if(gen_par%single_hx_dip) then
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
       end do
    else
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          if(khx <= float(data_par%n_hx_pad/2)) then
             k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
          else
             k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
          end if
       end do
    end if
    !write(0,*)'k_hx=',k_hx

    do i_k_hx=1,data_par%n_hx_pad
       do i_k_my=1,data_par%n_my_pad
          do i_k_mx=1,data_par%n_mx_pad
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,SOU_COMP)= &
             .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,REC_COMP)= &
             .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)
             !write(0,*)'k_s_sq(',i_k_hx,') =',k_s_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_r_sq(',i_k_hx,') =',k_r_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_my(',i_k_my,') =',k_mx(i_k_my) 
             !write(0,*)'k_mx(',i_k_mx,') =',k_mx(i_k_mx) 
          end do
       end do
    end do

    n_cycles_ksq=5.
    k_cutoff =0.
    delta_k=sqrt(min_prop_cos_sq)
    k_start=delta_k
    k_decay=-1.*delta_k
    if(k_decay**2 > n_cycles_ksq) call seperr('k_decay^2 > n_cycles_ksq')
    edge_gauss = exp(k_decay)
    sigma_magn_sq=-(k_decay-k_start)**2/(2.*log(edge_gauss))

    !write(0,*)'edge_gauss=',edge_gauss
    !write(0,*)'sigma_magn_sq=',sigma_magn_sq

    d_k_norm=n_cycles_ksq/(Pre2D_vti_ssf_par_loc%n_disp_rel-1)
    do i_k_norm=1,Pre2D_vti_ssf_par_loc%n_disp_rel
       k_norm_sq=1.-(i_k_norm-1)*d_k_norm
       if( k_norm_sq > 0.) then
          k_norm=sqrt(k_norm_sq)
       else
          k_norm=-sqrt(-k_norm_sq)
       end if

       phase(i_k_norm)=k_norm
       phase_pos(i_k_norm)=sqrt(1.+(i_k_norm-1)*d_k_norm)

       if(k_norm > k_start) then
!          magn(i_k_norm)=0.
          magn(i_k_norm)=1.
       else
          !    gauss_magn=exp(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
          !    magn(i_k_norm)=log(gauss_magn)
          magn(i_k_norm)=exp(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
!          magn(i_k_norm)=(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
       end if
    end do
    max_disp_rel=float(Pre2D_vti_ssf_par_loc%n_disp_rel)

    !    write(0,*)'n_disp_rel=',Pre2D_vti_ssf_par_loc%n_disp_rel
    call snap('Phase.h',Pre2D_vti_ssf_par_loc%n_disp_rel,1,phase)
    call snap('Magn.h',Pre2D_vti_ssf_par_loc%n_disp_rel,1,magn)

    desc_timer='Compute diffraction term (loop 1)'
    call SEP_timer_init(31,desc_timer,0.,0)
    desc_timer='Apply diffraction term'
    call SEP_timer_init(32,desc_timer,0.,0)
    desc_timer='Compute correction term'
    call SEP_timer_init(33,desc_timer,0.,0)
    desc_timer='Apply correction term'
    call SEP_timer_init(34,desc_timer,0.,0)
    desc_timer='Interpolate wavefields'
    call SEP_timer_init(35,desc_timer,0.,0)
    desc_timer='Compute diffraction term (loop 2)'
    call SEP_timer_init(36,desc_timer,0.,0)

  end function Pre2D_vti_ssf_init

  function Pre2D_vti_ssf_down_oper(adj,add, &
  n_w_data,w_0_data,d_w_data,i_w_block_par, &
  z_down,d_z_down,data_par, &
  data_block_struct,ddata_block_struct, &
  slow_slice,vti_delta_slice,vti_eta_slice,iz) &
  result(stat)

    integer stat,n_w_data,i_w_block_par,iz
    logical add,adj,found_salt
    real w_0_data,d_w_data,z_down,d_z_down
    type (data_param_type) data_par
    type(data_block_type) :: data_block_struct
    type(data_block_type) :: ddata_block_struct
    real, dimension (:,:,:,:,:) :: slow_slice,vti_delta_slice,vti_eta_slice
    optional data_block_struct,ddata_block_struct
    optional slow_slice,vti_delta_slice,vti_eta_slice

    logical prop_data,prop_ddata,var_slow,var_vti_delta,var_vti_eta
    integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx,i_mx,i_my,i_hx,i_w
    integer i_ksq_norm_num,i_ksq_norm_den
    integer i_slow,n_slow_split_eff,i_surf_comp,surf_comp
    integer, external ::  omp_get_thread_num
    real k0_sq,rphase_num,rphase_den,rphase,rmagn_num,dphase,w_data
    real slow_0
    real sigma,rat,dscat,dz_w,pi2,norm_ksq_num,norm_ksq_den,k0,k0_dz

    integer, allocatable,  dimension (:,:,:,:) :: map_slow
!    integer, allocatable,  dimension (:,:,:,:) :: map_vti_delta,map_vti_eta
    complex, allocatable,  dimension (:,:,:,:,:) :: temp_data
    complex, allocatable,  dimension (:,:,:,:,:) :: temp_ddata

    real,  dimension(4,n_slow_total)      :: slow_prop
    complex,  dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: ph_shift

!    real, dimension (data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad) :: magn_mat,phase_mat

!    real, dimension (Pre2D_vti_ssf_par_loc%n_disp_rel) :: phase_int,phase_pos_int,magn_int

    type(fft_type), dimension (:), pointer  :: fft_data_block_id
    type(fft_type), dimension (:), pointer  :: fft_ddata_block_id
    complex, dimension (:,:,:,:,:), pointer :: data_slice,ddata_slice

!    phase_int=phase
!    phase_pos_int=phase_pos
!    magn_int=magn

!        write(0,*)'Beginning of i_w_block_par=',i_w_block_par

    n_mx=data_par%n_mx_pad
    n_my=data_par%n_my_pad
    n_hx=data_par%n_hx_pad
    pi2=2.*acos(-1.)

    prop_data= present(data_block_struct)
    prop_ddata= present(ddata_block_struct)
    var_slow= present(slow_slice)
    var_vti_delta= present(vti_delta_slice)
    var_vti_eta= present(vti_eta_slice)

    if(var_slow) then
       allocate (map_slow(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of map_slow ")
    end if
!    if(var_vti_delta) then
!       allocate (map_vti_delta(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
!       stat=ierr)  
!       if(ierr .ne. 0) call seperr("Problems with allocation of map_vti_delta ")
!    end if
!    if(var_vti_eta) then
!       allocate (map_vti_eta(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2),&
!       stat=ierr)  
!       if(ierr .ne. 0) call seperr("Problems with allocation of map_vti_eta ")
!    end if
!

    if(prop_data) then
       fft_data_block_id => data_block_struct%fft_data_block_id(:,i_w_block_par)
       data_slice => data_block_struct%data_block(:,:,:,:,:,i_w_block_par)
       allocate(temp_data(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,data_par%n_hy_pad,2),&
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of temp_data")
       temp_data(:,:,:,1,1) = cmplx(0.,0.)
    end if
    if(prop_ddata) then
       fft_ddata_block_id => ddata_block_struct%fft_data_block_id(:,i_w_block_par)
       ddata_slice => ddata_block_struct%data_block(:,:,:,:,:,i_w_block_par)
       allocate(temp_ddata(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,data_par%n_hy_pad,2),&
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of temp_ddata")
       temp_ddata(:,:,:,1,1) = cmplx(0.,0.)
    end if


    if(var_slow) then
       if(Pre2D_vti_ssf_par_loc%lloyd) then
         call Pre2D_vti_slow_prop_lloyd(&
                        n_mx,n_my,n_hx,n_slow_split_eff,   &
                        vti_delta_const,vti_eta_const,vti_epsilon_const,&
                        slow_slice,vti_delta_slice,vti_eta_slice, &
                        slow_prop, map_slow,iz) 
       else 
         call Pre2D_vti_slow_prop(&
                        n_mx,n_my,n_hx,n_slow_split_eff,   &
                        vti_delta_const,vti_eta_const,vti_epsilon_const,&
                        slow_slice,vti_delta_slice,vti_eta_slice, &
                        slow_prop, map_slow) 

       end if
write(0,*) "OUT OUT PUT outout",n_slow_split_eff
       if(debug1) then
         if(omp_get_thread_num() == 0) then
           write(0,*)'n_slow_split_eff=',n_slow_split_eff
           write(0,*)'vel_prop=',1./slow_prop(1,1:n_slow_split_eff)
           write(0,*)'delta_prop=',slow_prop(2,1:n_slow_split_eff)
           write(0,*)'eta_prop=',slow_prop(3,1:n_slow_split_eff)
           write(0,*)'epsilon_prop=',slow_prop(4,1:n_slow_split_eff)
         end if
       end if
    else
       if(Pre2D_vti_ssf_par_loc%n_slow_split /= 1) call seperr('n_slow_split /= 1')
       n_slow_split_eff= 1 
       slow_prop(1,1)=slow_const
       slow_prop(2,1)=vti_delta_const
       slow_prop(3,1)=vti_eta_const
       slow_prop(4,1)=vti_epsilon_const
    end if

    do i_w=1,n_w_data
    !{
      w_data=w_0_data+(i_w-1)*d_w_data
      if( w_data > epsi) then
      !{
       do i_surf_comp=1,2
       !{
          if(adj) then
             !{
             if (i_surf_comp == 1) then
                surf_comp=REC_COMP
             else
                surf_comp=SOU_COMP
             end if
             !}
          else
             !{
             if (i_surf_comp == 1) then
                surf_comp=SOU_COMP
             else
                surf_comp=REC_COMP
             end if
             !}
          end if

          if(adj .and.(var_slow .or. apply_any_bound)) then
          !{
             if(prop_data) then
                call fft_4d_data(&
                  fft_data_block_id(i_w),FFT_BACK, & 
                  n_mx,n_my,n_hx,1,data_slice(:,:,:,:,i_w))

                ! apply absorbing boundary taper
                call apply_bound(data_par,data_slice(:,:,:,:,i_w))
             end if
             if(prop_ddata) then
                call fft_4d_data(&
                  fft_ddata_block_id(i_w),FFT_BACK, & 
                  n_mx,n_my,n_hx,1,ddata_slice(:,:,:,:,i_w))
                ! apply absorbing boundary taper
                call apply_bound(data_par,data_slice(:,:,:,:,i_w))
             end if
          !}
          end if


          if(prop_data) then
             temp_data(:,:,:,1,2) = data_slice(:,:,:,1,i_w) 
          end if
          if(prop_ddata) then
             temp_ddata(:,:,:,1,2) = ddata_slice(:,:,:,1,i_w) 
          end if

                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
!if(map_slow(i_mx,i_my,i_hx,surf_comp) > n_slow_split_eff)write(0,*)'Problem with map_slow=',map_slow(i_mx,i_my,i_hx,surf_comp),'at (ix,iy,ih,icpmp)=',i_mx,i_my,i_hx,surf_comp
!if(map_slow(i_mx,i_my,i_hx,surf_comp) < 1)write(0,*)'Problem with map_slow=',map_slow(i_mx,i_my,i_hx,surf_comp),'at (ix,iy,ih,icpmp)=',i_mx,i_my,i_hx,surf_comp
                            !}
                         end do
                            !}
                      end do
                            !}
                     end do

          do i_slow=1,n_slow_split_eff
          !{
             slow_0=slow_prop(1,i_slow)
             vti_delta_0=slow_prop(2,i_slow)
             vti_eta_0=slow_prop(3,i_slow)
             vti_epsilon_0=slow_prop(4,i_slow)


             k0= (w_data*slow_0)
             k0_sq= k0*k0
             if(time_mig) then
                k0_dz=.5*w_data*d_z_down
             else
                k0_dz=k0*d_z_down
                dz_w=d_z_down*w_data
             end if

             if(k0_sq > epsi) then
                norm_ksq_num=(1.+2*vti_epsilon_0)/(d_k_norm*k0_sq)
                norm_ksq_den=2*vti_eta_0*(1.+2*vti_delta_0)/(d_k_norm*k0_sq)
!write(0,*)'thread number',omp_get_thread_num()
!write(0,*)'w_data',w_data
!write(0,*)'norm_ksq_num',norm_ksq_num
!write(0,*)'norm_ksq_den',norm_ksq_den

             else
                norm_ksq_num=0.
                norm_ksq_den=0.
             end if

             if(adj) then
             !{
               if(var_slow .or. apply_any_bound) then
                !{
                if(.not. time_mig) then
                   !{
                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
                            dphase=(slow_slice(i_mx,i_my,i_hx,1,surf_comp)-slow_0)*dz_w
                            ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end if

                if(prop_data) then
                   !{
                   data_slice(:,:,:,1,i_w) = cmplx(0.,0.) 
                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
                               data_slice(i_mx,i_my,i_hx,1,i_w)=temp_data(i_mx,i_my,i_hx,1,2)
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   if(.not. time_mig) then
                      !{
                      data_slice(:,:,:,1,i_w) = data_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
                      !}
                   end if
                   call fft_4d_data(&
                     fft_data_block_id(i_w),FFT_FORW, & 
                     n_mx,n_my,n_hx,1,data_slice(:,:,:,:,i_w))

!TEMP                   call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,temp_data(:,:,:,1,1))
                     !} 
                  end if
                  if(prop_ddata) then
                     !{
                     ddata_slice(:,:,:,1,i_w) = cmplx(0.,0.) 
                     do i_hx=1,n_hx
                        !{
                        do i_my=1,n_my
                           !{
                           do i_mx=1,n_mx
                              !{
                              if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
                                 ddata_slice(i_mx,i_my,i_hx,1,i_w)=temp_ddata(i_mx,i_my,i_hx,1,2)
                              end if
                              !}
                           end do
                           !}
                        end do
                        !}
                     end do
                     if(.not. time_mig) then
                        !{
                        ddata_slice(:,:,:,1,i_w) = ddata_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
                        !}
                     end if
                     call fft_4d_data(&
                       fft_ddata_block_id(i_w),FFT_FORW, & 
                       n_mx,n_my,n_hx,1,ddata_slice(:,:,:,:,i_w))
!TEMP                     call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,temp_ddata(:,:,:,1,1))
                     !} 
                  end if
                !}
               else
                !{
                 if(prop_data) then
                   data_slice(:,:,:,1,i_w) = temp_data(:,:,:,1,2)
                 end if
                 if(prop_ddata) then
                   ddata_slice(:,:,:,1,i_w) = temp_ddata(:,:,:,1,2)
                 end if
                 !}
               end if
             !}
             else
             !{
               if(prop_data) then
                 data_slice(:,:,:,1,i_w) = temp_data(:,:,:,1,2)
               end if
               if(prop_ddata) then
                 ddata_slice(:,:,:,1,i_w) = temp_ddata(:,:,:,1,2)
               end if
             !}
             end if

             call SEP_timer_start(31)
             do i_k_hx=1,n_hx
                !{
                do i_k_my=1,n_my
                   !{
                   do i_k_mx=1,n_mx
                      !{
                      i_ksq_norm_num=nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)* &
                      norm_ksq_num+1,max_disp_rel))
                      rphase_num=phase(i_ksq_norm_num)
!                      rmagn_num=abs(k0_dz)*magn(i_ksq_norm_num)
                      rmagn_num=magn(i_ksq_norm_num)

                      if(norm_ksq_den > 0.) then
                        i_ksq_norm_den= &
                           nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)* &
                           norm_ksq_den+1,max_disp_rel))
                        rphase_den=phase(i_ksq_norm_den)
                      else
                        i_ksq_norm_den= &
                           nint(min(k_sr_sq(i_k_mx,i_k_my,i_k_hx,surf_comp)* &
                           (-norm_ksq_den)+1,max_disp_rel))
                        rphase_den=phase_pos(i_ksq_norm_den)
                      end if

                      if(abs(rphase_den) > epsi) then
                        rphase=(rphase_num/rphase_den)*k0_dz
!                        phase_mat(i_k_mx,i_k_my,i_k_hx)=(rphase_num/rphase_den)*k0_dz
                      else
                        rphase=0.
!                        phase_mat(i_k_mx,i_k_my,i_k_hx)=0
                      end if
                      if(rmagn_num >= 1) then
                        rmagn_num=1
!                        magn_mat(i_k_mx,i_k_my,i_k_hx)=1
!                      else
!                        rmagn_num=rmagn_num*abs(k0_dz)
!!                        magn_mat(i_k_mx,i_k_my,i_k_hx)=rmagn_num*abs(k0_dz)
                      end if

                      ph_shift(i_k_mx,i_k_my,i_k_hx)= &
                         rmagn_num * cmplx(cos(rphase),sin(rphase))
!                      ph_shift(i_k_mx,i_k_my,i_k_hx)= &
!                         magn_mat(i_k_mx,i_k_my,i_k_hx) * &
!                         cmplx(cos(phase_mat(i_k_mx,i_k_my,i_k_hx)),&
!                               sin(phase_mat(i_k_mx,i_k_my,i_k_hx)))

                      !}
                   end do
                   !}
                end do
                !}
             end do
             call SEP_timer_stop(31)

!              ph_shift= magn_mat* cmplx(cos(phase_mat),sin(phase_mat))

!             call SEP_timer_start(36)
!             do i_k_hx=1,n_hx
!                !{
!                do i_k_my=1,n_my
!                   !{
!                   do i_k_mx=1,n_mx
!                      !{
!                      ph_shift(i_k_mx,i_k_my,i_k_hx)= &
!                         magn_mat(i_k_mx,i_k_my,i_k_hx) * &
!                         cmplx(cos(phase_mat(i_k_mx,i_k_my,i_k_hx)),&
!                               sin(phase_mat(i_k_mx,i_k_my,i_k_hx)))
!                      !}
!                   end do
!                   !}
!                end do
!                !}
!             end do
!             call SEP_timer_stop(36)

             if(prop_data) then
             call SEP_timer_start(32)

                data_slice(:,:,:,1,i_w) = data_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
             call SEP_timer_stop(32)
             end if
             if(prop_ddata) then
                ddata_slice(:,:,:,1,i_w) = ddata_slice(:,:,:,1,i_w) * ph_shift(:,:,:)
             end if

             if(.not. adj) then
             !{
               if(var_slow .or. apply_any_bound) then
               !{
                if(.not. time_mig) then
                   !{
                   call SEP_timer_start(33)
                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
                            dphase=(slow_slice(i_mx,i_my,i_hx,1,surf_comp)-slow_0)* dz_w
                            ph_shift(i_mx,i_my,i_hx)= cmplx(cos(dphase),sin(dphase))
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   call SEP_timer_stop(33)
                   !}
                end if

                if(prop_data) then
                   !{
                   call fft_4d_data(&
                     fft_data_block_id(i_w),FFT_BACK, & 
                     n_mx,n_my,n_hx,1,data_slice(:,:,:,:,i_w))

!TEMP                   call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data(:,:,:,1,1))
                   if(.not. time_mig) then
                      !{
                   call SEP_timer_start(34)
                      data_slice(:,:,:,1,i_w) = data_slice(:,:,:,1,i_w)*ph_shift(:,:,:)
                   call SEP_timer_stop(34)
                      !}
                   end if
                   call SEP_timer_start(35)
                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
                               temp_data(i_mx,i_my,i_hx,1,1)=data_slice(i_mx,i_my,i_hx,1,i_w)
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   call SEP_timer_stop(35)
                   !} 
                end if
                if(prop_ddata) then
                   !{
!TEMP                   call fft_3d_data(fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,temp_ddata(:,:,:,1,1))
                   if(.not. time_mig) then
                      !{
                      ddata_slice(:,:,:,1,i_w) = ddata_slice(:,:,:,1,i_w)*ph_shift(:,:,:)
                      !}
                   end if
                   do i_hx=1,n_hx
                      !{
                      do i_my=1,n_my
                         !{
                         do i_mx=1,n_mx
                            !{
                            if(map_slow(i_mx,i_my,i_hx,surf_comp) == i_slow) then 
                               temp_ddata(i_mx,i_my,i_hx,1,1)=ddata_slice(i_mx,i_my,i_hx,1,i_w)
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !} 
                end if
               !}
               else
               !{
                if(prop_data) then
                  temp_data(:,:,:,1,1)=temp_data(:,:,:,1,1)+data_slice(:,:,:,1,i_w)
                end if
                if(prop_ddata) then
                  temp_ddata(:,:,:,1,1)=temp_ddata(:,:,:,1,1)+ddata_slice(:,:,:,1,i_w)
                end if
               !} 
               end if
             !} 
             else
             !{
              if(prop_data) then
                 temp_data(:,:,:,1,1)=data_slice(:,:,:,1,i_w)
              end if
              if(prop_ddata) then
                 temp_ddata(:,:,:,1,1)=ddata_slice(:,:,:,1,i_w)
              end if
             !} 
             end if

          !} sou slowness loop
          end do

          if(prop_data) then
             data_slice(:,:,:,1,i_w)=temp_data(:,:,:,1,1)
          end if
          if(prop_ddata) then
             ddata_slice(:,:,:,1,i_w)=temp_ddata(:,:,:,1,1)
          end if

          if((.not. adj) .and.(var_slow .or. apply_any_bound)) then
            !{
             if(prop_data) then
                ! apply absorbing boundary taper
                call apply_bound(data_par,data_slice(:,:,:,:,i_w))
                call fft_4d_data(&
                  fft_data_block_id(i_w),FFT_FORW, & 
                  n_mx,n_my,n_hx,1,data_slice(:,:,:,:,i_w))

!TEMP                call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
             end if
             if(prop_ddata) then
                ! apply absorbing boundary taper
                call apply_bound(data_par,ddata_slice(:,:,:,:,i_w))
                call fft_4d_data(&
                  fft_ddata_block_id(i_w),FFT_FORW, & 
                  n_mx,n_my,n_hx,1,ddata_slice(:,:,:,:,i_w))
!TEMP                call fft_3d_data(fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,ddata_slice(:,:,:,1,i_w))
             end if
            !}
          end if

       !} i_surf_comp loop
       end do
      !}
      else
      !{
        if(debug1) write(0,*)'Skip prop because frequency is zero'
        if(prop_data) then
          data_slice(:,:,:,1,i_w)=0.
        end if
        if(prop_ddata) then
          ddata_slice(:,:,:,1,i_w)=0.
        end if
      !}
      end if

    !} i_w loop
    end do


    if(var_slow) then
       deallocate (map_slow,stat=ierr)  
    end if
    if(prop_data) then
       deallocate (temp_data,stat=ierr)  
    end if
    if(prop_ddata) then
       deallocate (temp_ddata,stat=ierr)  
    end if

    stat=0
    return	
  end function Pre2D_vti_ssf_down_oper

  function Pre2D_vti_ssf_scatter_oper(adj,add, &
  n_w_data,w_0_data,d_w_data,i_w_block_par, &
  z_down,d_z_down,data_par, &
  data_block_struct,ddata_block_struct,dslow_slice,&
  slow_slice,vti_delta_slice,vti_eta_slice) &
  result(stat)

    integer stat,n_w_data,i_w_block_par
    logical add,adj
    real w_0_data,d_w_data,z_down,d_z_down
    type (data_param_type) data_par
    type(data_block_type) :: data_block_struct
    type(data_block_type) :: ddata_block_struct
    complex, dimension (:,:,:,:,:) :: dslow_slice
    real, optional, dimension (:,:,:,:,:) :: slow_slice
    real, optional, dimension (:,:,:,:,:)      :: vti_delta_slice
    real, optional, dimension (:,:,:,:,:)      :: vti_eta_slice

    call seperr("Pre2D_vti_ssf_scatter_oper not implemented yet")

    stat=0
    return	
  end function Pre2D_vti_ssf_scatter_oper

  function Pre2D_vti_ssf_clean()  result(stat)
    integer stat,ierr
    stat=0
    if(allocated(phase)) deallocate (phase,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(magn)) deallocate (magn,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_mx)) deallocate (k_mx,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_my)) deallocate (k_my,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_hx)) deallocate (k_hx,stat=ierr)  
    if(ierr /=0) stat=stat+1
    if(allocated(k_sr_sq)) deallocate (k_sr_sq,stat=ierr)  
    if(ierr /=0) stat=stat+1

    return	
  end function Pre2D_vti_ssf_clean

  subroutine Pre2D_vti_slow_prop_lloyd(n_mx,n_my,n_hx, n_slow_split_eff,  &
                              vti_delta_const,vti_eta_const,vti_epsilon_const,&
                              slow_slice,vti_delta_slice,vti_eta_slice, &
                              slow_prop, map_slow,iz)

    integer, intent(inout):: n_slow_split_eff
    integer, intent(in):: n_mx,n_my,n_hx,iz
    real, intent(in):: vti_delta_const,vti_eta_const,vti_epsilon_const
    integer, dimension (:,:,:,:),intent(inout) :: map_slow
    real, dimension (:,:,:,:,:),intent(in)    :: slow_slice
    real, dimension (:,:,:,:,:),intent(in)    :: vti_delta_slice,vti_eta_slice
    real,  dimension(:,:), intent(inout)           :: slow_prop
  integer,allocatable :: iregion(:)
  real ,allocatable :: array(:,:),center(:,:)
  integer :: nstart(3),nn,n1,n2,n3,n4,i,i1,i2,i3,i4,n5,i5,n,nx,ny,ix,iy

write(0,*) "AA 1"

   nstart=(/Pre2D_vti_ssf_par_loc%n_slow_split,Pre2D_vti_ssf_par_loc%n_vti_delta,&
        Pre2D_vti_ssf_par_loc%n_vti_eta/)
   nn=product(nstart)

   nx=maxval(ix_slow)


   allocate(center(nn,3))

   n=nx
   allocate(iregion(n),array(n,3))
   array(:,1)=slow_slice(1,1,1,1,1)
   array(:,2)=vti_delta_slice(1,1,1,1,1)
   array(:,3)=vti_eta_slice(1,1,1,1,1)

    n1=size(slow_slice,1)
    n2=size(slow_slice,2)
    n3=size(slow_slice,3)
    n4=size(slow_slice,4)
    n5=size(slow_slice,5)

   do i5=1,n5
   do i4=1,n4
   do i3=1,n3
   do i2=1,n2
   do i1=1,n1
     ix=ix_slow(i1,i3,i5)
     i=ix
     array(i,1)=slow_slice(i1,i2,i3,i4,i5)
     array(i,2)=vti_delta_slice(i1,i2,i3,i4,i5)/100.
     array(i,3)=vti_eta_slice(i1,i2,i3,i4,i5)/100.
    end do 
    end do 
    end do 
    end do 
    end do 
    n_slow_split_eff=lloyd_go(array,center,nn,iregion,nstart)

    do i=1,n_slow_split_eff
      slow_prop(1:3,i)=center(:,i)
      slow_prop(4,i)=slow_prop(2,i)+slow_prop(3,i)*(1.+2*slow_prop(2,i))
    end do
  
    i=0
   do i5=1,n5
   do i3=1,n3
   do i2=1,n2
   do i1=1,n1
     ix=ix_slow(i1,i3,i5)
     i=ix
     map_slow(i1,i2,i3,i5)=iregion(i)
    end do 
    end do 
    end do 
    end do 
    deallocate(iregion,array,center)
  end subroutine



  subroutine Pre2D_vti_slow_prop(n_mx,n_my,n_hx, n_slow_split_eff,  &
                              vti_delta_const,vti_eta_const,vti_epsilon_const,&
                              slow_slice,vti_delta_slice,vti_eta_slice, &
                              slow_prop, map_slow)

    integer, intent(inout):: n_slow_split_eff
    integer, intent(in):: n_mx,n_my,n_hx
    real, intent(in):: vti_delta_const,vti_eta_const,vti_epsilon_const
    integer, dimension (:,:,:,:),intent(inout) :: map_slow
    real, dimension (:,:,:,:,:),intent(in)    :: slow_slice
    real, dimension (:,:,:,:,:),intent(in)    :: vti_delta_slice,vti_eta_slice
    real,  dimension(:,:), intent(inout)           :: slow_prop

    logical found_salt
    integer i_mx,i_my,i_hx,i_comp,ierr
    integer i_mx_p1,i_mx_m1,i_my_p1,i_my_m1,i_hx_p1,i_hx_m1
    integer n_slow_split_no_salt,n_slow_split_no_salt_all
    integer n_vti_delta_no_salt,n_vti_eta_no_salt
    integer i_slow_eff,i_slow,i_delta,i_eta
    integer i_slow_eff_sou,i_slow_sou,i_delta_sou,i_eta_sou
    integer i_slow_eff_rec,i_slow_rec,i_delta_rec,i_eta_rec
    real min_slow,max_slow,range_slow,delta_slow,mean_slow,inv_delta_slow
    real min_vti_delta,max_vti_delta,range_vti_delta
    real mean_vti_delta,delta_vti_delta,inv_delta_vti_delta
    real min_vti_eta,max_vti_eta,range_vti_eta
    real mean_vti_eta,delta_vti_eta,inv_delta_vti_eta
    real temp_slow_salt,thresh_slow_salt,epsi

    integer, external ::  omp_get_thread_num

    logical, allocatable, dimension (:,:,:,:,:) :: mask_slow 

    allocate (mask_slow(n_mx,n_my,n_hx,1,2),stat=ierr)  
    if(ierr .ne. 0) call seperr("Pre2D_vti_ssf: Problems with allocation of mask_slow ")

    epsi=1.00001 

    if(slow_salt > 0.) then
    !{
       found_salt=any(slow_slice <= slow_salt)
       temp_slow_salt=min(slow_salt,minval(slow_slice))
       thresh_slow_salt=min(slow_salt, &
                          Pre2D_vti_ssf_par_loc%range_salt*minval(slow_slice))
!if((omp_get_thread_num() == 0)) write(0,*)'temp_vel_salt=',1./temp_slow_salt
!if((omp_get_thread_num() == 0)) write(0,*)'thresh_vel_salt=',1./thresh_slow_salt

       if(found_salt) then
       !{
!if((omp_get_thread_num() == 0)) write(0,*)'Found salt'
!if((omp_get_thread_num() == 0)) write(0,*)'shape(slow_slice)',shape(slow_slice)
!if((omp_get_thread_num() == 0)) write(0,*)'shape(mask_slow)',shape(mask_slow)
         mask_slow=.true.
         do i_comp=1,2
           do i_hx=1,n_hx
             do i_my=1,n_my
               do i_mx=1,n_mx
                 if(slow_slice(i_mx,i_my,i_hx,1,i_comp) < thresh_slow_salt) then
                   i_mx_p1=min(n_mx,i_mx+1)
                   i_mx_m1=max(1,i_mx-1)
                   i_my_p1=min(n_my,i_my+1)
                   i_my_m1=max(1,i_my-1)
                   i_hx_p1=min(n_hx,i_hx+1)
                   i_hx_m1=max(1,i_hx-1)
   mask_slow(i_mx_m1:i_mx_p1,i_my_m1:i_my_p1,i_hx_m1:i_hx_p1,1,i_comp)=.false.
                  end if
               end do
             end do
           end do
         end do
         if(any(mask_slow))then ! found a NON salt velocity
           min_slow=minval(slow_slice,mask=mask_slow)
           min_vti_delta=minval(vti_delta_slice,mask=mask_slow)
           min_vti_eta=minval(vti_eta_slice,mask=mask_slow)

           n_slow_split_no_salt=Pre2D_vti_ssf_par_loc%n_slow_split-1
           n_vti_delta_no_salt=max(1,Pre2D_vti_ssf_par_loc%n_vti_delta-1)
           n_vti_eta_no_salt=max(1,Pre2D_vti_ssf_par_loc%n_vti_eta-1)

!           if((omp_get_thread_num() == 0)) write(0,*)'not constant'
!           if((omp_get_thread_num() == 0)) write(0,*)'max_vel=',1./min_slow
         else
!           if((omp_get_thread_num() == 0)) write(0,*)'constant'
           n_slow_split_no_salt=0
           n_vti_delta_no_salt=0
           n_vti_eta_no_salt=0
         end if
       !}
       else
       !{
         mask_slow=.true.
         min_slow=minval(slow_slice)
         min_vti_delta=minval(vti_delta_slice)
         min_vti_eta=minval(vti_eta_slice)
         n_slow_split_no_salt=Pre2D_vti_ssf_par_loc%n_slow_split
         n_vti_delta_no_salt=Pre2D_vti_ssf_par_loc%n_vti_delta
         n_vti_eta_no_salt=Pre2D_vti_ssf_par_loc%n_vti_eta
!         if((omp_get_thread_num() == 0)) write(0,*)'Here: max_vel=',1./min_slow
       !}
       end if
    !}
    else
    !{
       found_salt=.false.
       mask_slow=.true.
       min_slow=minval(slow_slice)
       min_vti_delta=minval(vti_delta_slice)
       min_vti_eta=minval(vti_eta_slice)
       n_slow_split_no_salt=Pre2D_vti_ssf_par_loc%n_slow_split
       n_vti_delta_no_salt=Pre2D_vti_ssf_par_loc%n_vti_delta
       n_vti_eta_no_salt=Pre2D_vti_ssf_par_loc%n_vti_eta
    !}
    end if

    if(n_slow_split_no_salt > 0) then
    !{
      max_slow=maxval(slow_slice)
      range_slow=max_slow-min_slow
      mean_slow=.5*(max_slow+min_slow)
      delta_slow=range_slow/n_slow_split_no_salt

      if(delta_slow <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_slow) then
         if(range_slow <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_slow) then
            n_slow_split_no_salt= 1 
            delta_slow=range_slow
            inv_delta_slow=0.
         else
            n_slow_split_no_salt= & 
            min(n_slow_split_no_salt, &
            max(1,int(range_slow/(Pre2D_vti_ssf_par_loc%min_delta_slow*mean_slow))))
            delta_slow=range_slow/n_slow_split_no_salt
            inv_delta_slow=1./(epsi*delta_slow)
         end if
      else
         inv_delta_slow=1./(epsi*delta_slow)
      end if

    !}
    end if

    if(n_vti_delta_no_salt > 0) then
    !{
      max_vti_delta=maxval(vti_delta_slice)
      range_vti_delta=max_vti_delta-min_vti_delta
      mean_vti_delta=.5*(max_vti_delta+min_vti_delta)
      delta_vti_delta=range_vti_delta/n_vti_delta_no_salt

      if(delta_vti_delta <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_delta) then
         if(range_vti_delta <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_delta) then
            n_vti_delta_no_salt= 1 
            delta_vti_delta=range_vti_delta
            inv_delta_vti_delta=0.
         else
            n_vti_delta_no_salt= & 
            min(n_vti_delta_no_salt, &
            max(1,int(range_vti_delta/(Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_delta))))
            delta_vti_delta=range_vti_delta/n_vti_delta_no_salt
            inv_delta_vti_delta=1./(epsi*delta_vti_delta)
         end if
      else
         inv_delta_vti_delta=1./(epsi*delta_vti_delta)
      end if

    !}
    end if

    if(n_vti_eta_no_salt > 0) then
    !{
      max_vti_eta=maxval(vti_eta_slice)
      range_vti_eta=max_vti_eta-min_vti_eta
      mean_vti_eta=.5*(max_vti_eta+min_vti_eta)
      delta_vti_eta=range_vti_eta/n_vti_eta_no_salt

      if(delta_vti_eta <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_eta) then
         if(range_vti_eta <= Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_eta) then
            n_vti_eta_no_salt= 1 
            delta_vti_eta=range_vti_eta
            inv_delta_vti_eta=0.
         else
            n_vti_eta_no_salt= & 
            min(n_vti_eta_no_salt, &
            max(1,int(range_vti_eta/(Pre2D_vti_ssf_par_loc%min_delta_slow*mean_vti_eta))))
            delta_vti_eta=range_vti_eta/n_vti_eta_no_salt
            inv_delta_vti_eta=1./(epsi*delta_vti_eta)
         end if
      else
         inv_delta_vti_eta=1./(epsi*delta_vti_eta)
      end if

    !}
    end if

    n_slow_split_no_salt_all= n_slow_split_no_salt*n_vti_delta_no_salt*n_vti_eta_no_salt

    ! set velocities outside salt
    if(n_slow_split_no_salt_all > 1) then
    !{
       do i_eta=1,n_vti_eta_no_salt
         do i_delta=1,n_vti_delta_no_salt
           do i_slow=1,n_slow_split_no_salt
              i_slow_eff=i_slow+n_slow_split_no_salt*(i_delta-1)+ &
                          n_slow_split_no_salt*n_vti_delta_no_salt*(i_eta-1)
     slow_prop(1,i_slow_eff)=min_slow+(i_slow-set_slow)*delta_slow
     slow_prop(2,i_slow_eff)=min_vti_delta+(i_delta-set_slow)*delta_vti_delta
     slow_prop(3,i_slow_eff)=min_vti_eta+(i_eta-set_slow)*delta_vti_eta
! next comes from epsilon=delta+eta*(1.+2*delta)
     slow_prop(4,i_slow_eff)=slow_prop(2,i_slow_eff)+ &
                     slow_prop(3,i_slow_eff)*(1.+2*slow_prop(2,i_slow_eff))
           end do
         end do
       end do
    !}
    else
    !{
       if(slow_const > 0.) then  
          slow_prop(1,1)=slow_const
          slow_prop(2,1)=min_vti_delta+(1.-set_slow)*delta_vti_delta
          slow_prop(3,1)=min_vti_eta+(1.-set_slow)*delta_vti_eta
          slow_prop(4,1)=slow_prop(2,1)+slow_prop(3,1)*(1.+2*slow_prop(2,1))
         else
          slow_prop(1,1)=min_slow+(1.-set_slow)*delta_slow
          slow_prop(2,1)=min_vti_delta+(1.-set_slow)*delta_vti_delta
          slow_prop(3,1)=min_vti_eta+(1.-set_slow)*delta_vti_eta
          slow_prop(4,1)=slow_prop(2,1)+slow_prop(3,1)*(1.+2*slow_prop(2,1))
       end if
    !}
    end if



    if(n_slow_split_no_salt_all > 0) then
    !{
! set map_slow outside salt
    do i_hx=1,n_hx
       !{
       do i_my=1,n_my
          !{
          do i_mx=1,n_mx
           !{
            if(mask_slow(i_mx,i_my,i_hx,1,SOU_COMP)) then
             !{
             if(n_slow_split_no_salt >1) then
               i_slow_sou= &
      int((slow_slice(i_mx,i_my,i_hx,1,SOU_COMP)-min_slow)*inv_delta_slow)+1
             else
               i_slow_sou=1
             end if
             if(n_vti_delta_no_salt >1) then
               i_delta_sou= &
       int((vti_delta_slice(i_mx,i_my,i_hx,1,SOU_COMP)-min_vti_delta)*inv_delta_vti_delta)+1
             else
               i_delta_sou=1
             end if
             if(n_vti_eta_no_salt >1) then
               i_eta_sou= &
       int((vti_eta_slice(i_mx,i_my,i_hx,1,SOU_COMP)-min_vti_eta)*inv_delta_vti_eta)+1
             else
               i_eta_sou=1
             end if
             i_slow_eff_sou=i_slow_sou+n_slow_split_no_salt*(i_delta_sou-1)+ &
                        n_slow_split_no_salt*n_vti_delta_no_salt*(i_eta_sou-1)
             map_slow(i_mx,i_my,i_hx,SOU_COMP)= i_slow_eff_sou
             !}
            end if

            if(mask_slow(i_mx,i_my,i_hx,1,REC_COMP)) then
             !{
             if(n_slow_split_no_salt >1) then
               i_slow_rec= &
      int((slow_slice(i_mx,i_my,i_hx,1,REC_COMP)-min_slow)*inv_delta_slow)+1
             else
               i_slow_rec=1
             end if
             if(n_vti_delta_no_salt >1) then
               i_delta_rec= &
       int((vti_delta_slice(i_mx,i_my,i_hx,1,REC_COMP)-min_vti_delta)*inv_delta_vti_delta)+1
             else
               i_delta_rec=1
             end if
             if(n_vti_eta_no_salt >1) then
               i_eta_rec= &
       int((vti_eta_slice(i_mx,i_my,i_hx,1,REC_COMP)-min_vti_eta)*inv_delta_vti_eta)+1
             else
               i_eta_rec=1
             end if
             i_slow_eff_rec=i_slow_rec+n_slow_split_no_salt*(i_delta_rec-1)+ &
                        n_slow_split_no_salt*n_vti_delta_no_salt*(i_eta_rec-1)
             map_slow(i_mx,i_my,i_hx,REC_COMP)= i_slow_eff_rec
             !}
            end if

           !}
          end do
          !}
       end do
       !}
    end do
    !}
    end if


    if(found_salt) then
    !{
       n_slow_split_eff= n_slow_split_no_salt_all+1
       slow_prop(1,n_slow_split_eff)=temp_slow_salt
       slow_prop(2,n_slow_split_eff)=0.
       slow_prop(3,n_slow_split_eff)=0.
       slow_prop(4,n_slow_split_eff)=0.
       do i_hx=1,n_hx
          !{
          do i_my=1,n_my
             !{
             do i_mx=1,n_mx
                !{
!                if(slow_slice(i_mx,i_my,i_hx,1,SOU_COMP) < thresh_slow_salt) then
                if(.not.mask_slow(i_mx,i_my,i_hx,1,SOU_COMP)) then
                   map_slow(i_mx,i_my,i_hx,SOU_COMP)= n_slow_split_eff
                end if
!                if(slow_slice(i_mx,i_my,i_hx,1,REC_COMP) < thresh_slow_salt) then
                if(.not.mask_slow(i_mx,i_my,i_hx,1,REC_COMP)) then
                   map_slow(i_mx,i_my,i_hx,REC_COMP)= n_slow_split_eff
                end if
                !}
             end do
             !}
          end do
          !}
       end do
       deallocate (mask_slow,stat=ierr)  
       if(ierr .ne. 0) call seperr("Pre2D_vti_ssf: Problems with deallocation of mask_slow ")
    !}
    else
    !{
       n_slow_split_eff= n_slow_split_no_salt_all 
    !}
    end if
if((omp_get_thread_num() == 0)) write(0,*)'n_slow_split_no_salt',n_slow_split_no_salt
if((omp_get_thread_num() == 0)) write(0,*)'n_slow_split_no_salt_all',n_slow_split_no_salt_all
if((omp_get_thread_num() == 0)) write(0,*)'n_slow_split_eff',n_slow_split_eff
if((omp_get_thread_num() == 0)) write(0,*)'minval(map_slow)',minval(map_slow)
if((omp_get_thread_num() == 0)) write(0,*)'maxval(map_slow)',maxval(map_slow)
  end subroutine Pre2D_vti_slow_prop

end module Pre2D_vti_ssf_mod

