module data_mod

  use gendown3d_types
  use down_process_types
  use data_types
  use fft_types
  use fft_data_mod
  use restart_types

  implicit none

  integer, external ::  srite,sreed,sseek_block
  integer, private ::  mx_start_inp,my_start_inp,hx_start_inp,hy_start_inp
  integer, private ::  mx_stop_inp,my_stop_inp,hx_stop_inp,hy_stop_inp
  integer, private ::  mx_start_out,my_start_out,hx_start_out,hy_start_out
  integer, private ::  mx_stop_out,my_stop_out,hx_stop_out,hy_stop_out
  real, private :: scale_data

  complex, allocatable, dimension (:), private :: hx_vect
  complex, allocatable, dimension (:,:), private :: hx_shift

contains

  subroutine init_io_data(gen_par,data_par,down_par)
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par

   integer i_k_hx,ierr,i_hx_0_shift
   real khx,new_hx_0

    real, allocatable, dimension (:) :: kkk_hx

    if(gen_par%write_down_data_pad) then
      mx_start_out=1
      mx_stop_out=data_par%n_mx_pad

      my_start_out=1
      my_stop_out=data_par%n_my_pad
  
      hx_start_out=1
      hx_stop_out=data_par%n_hx_pad
  
      hy_start_out=1
      hy_stop_out=data_par%n_hy_pad
    else
      mx_start_out=(int((data_par%n_mx_pad-data_par%n_mx)/2))+1
      mx_stop_out=mx_start_out+data_par%n_mx-1

      my_start_out=(int((data_par%n_my_pad-data_par%n_my)/2))+1
      my_stop_out=my_start_out+data_par%n_my-1
  
      hx_start_out=(int((data_par%n_hx_pad-data_par%n_hx)/2))+1
      hx_stop_out=hx_start_out+data_par%n_hx-1
  
      hy_start_out=(int((data_par%n_hy_pad-data_par%n_hy)/2))+1
      hy_stop_out=hy_start_out+data_par%n_hy-1
    end if

    mx_start_inp=(int((data_par%n_mx_pad-data_par%n_mx)/2))+1
    mx_stop_inp=mx_start_inp+data_par%n_mx-1

    my_start_inp=(int((data_par%n_my_pad-data_par%n_my)/2))+1
    my_stop_inp=my_start_inp+data_par%n_my-1

    hx_start_inp=(int((data_par%n_hx_pad-data_par%n_hx)/2))+1
    hx_stop_inp=hx_start_inp+data_par%n_hx-1

    hy_start_inp=(int((data_par%n_hy_pad-data_par%n_hy)/2))+1
    hy_stop_inp=hy_start_inp+data_par%n_hy-1

!    if(gen_par%debug2) write(0,*)'mx_start_inp=',mx_start_inp
!    if(gen_par%debug2) write(0,*)'mx_stop_inp=',mx_stop_inp
!    if(gen_par%debug2) write(0,*)'my_start_inp=',my_start_inp
!    if(gen_par%debug2) write(0,*)'my_stop_inp=',my_stop_inp
!    if(gen_par%debug2) write(0,*)'hx_start_inp=',hx_start_inp
!    if(gen_par%debug2) write(0,*)'hx_stop_inp=',hx_stop_inp
!    if(gen_par%debug2) write(0,*)'hy_start_inp=',hy_start_inp
!    if(gen_par%debug2) write(0,*)'hy_stop_inp=',hy_stop_inp
!
!    if(gen_par%debug2) write(0,*)'mx_start_out=',mx_start_out
!    if(gen_par%debug2) write(0,*)'mx_stop_out=',mx_stop_out
!    if(gen_par%debug2) write(0,*)'my_start_out=',my_start_out
!    if(gen_par%debug2) write(0,*)'my_stop_out=',my_stop_out
!    if(gen_par%debug2) write(0,*)'hx_start_out=',hx_start_out
!    if(gen_par%debug2) write(0,*)'hx_stop_out=',hx_stop_out
!    if(gen_par%debug2) write(0,*)'hy_start_out=',hy_start_out
!    if(gen_par%debug2) write(0,*)'hy_stop_out=',hy_stop_out

    if (down_par%spike_data .or. down_par%plane_data .or. down_par%diffr_data)  then
       scale_data=1./float(down_par%n_w_block_par*down_par%n_w_block_ser)
       scale_data=sqrt(scale_data)
    else
       scale_data=1.
    end if
    if(gen_par%debug2) write(0,*)'scale_data=',scale_data

    allocate (kkk_hx(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of kkk_hx ")


    if(gen_par%single_hx_dip) then
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          kkk_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
       end do
    else
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          if(khx <= float(data_par%n_hx_pad/2)) then
             kkk_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
          else
             kkk_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
          end if
       end do
    end if

    if(gen_par%hx_0_shift /= 0.) then
      allocate (hx_vect(data_par%n_hx_pad), &
                stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of hx_vect")

      allocate (hx_shift(data_par%n_hx_pad,gen_par%n_hx_0_shift_period), &
                stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of hx_shift")

      do i_hx_0_shift=1,gen_par%n_hx_0_shift_period
        new_hx_0=data_par%hx_0_pad + &
         mod(i_hx_0_shift-gen_par%i_hx_0_shift_start+gen_par%n_hx_0_shift_period, &
              gen_par%n_hx_0_shift_period)* &
                          gen_par%hx_0_shift
  write(0,*)'new_hx_0',new_hx_0
        do i_k_hx=1,data_par%n_hx_pad
          hx_shift(i_k_hx,i_hx_0_shift)=cmplx( &
                                      cos(kkk_hx(i_k_hx)*new_hx_0), &
                                      sin(kkk_hx(i_k_hx)*new_hx_0))
        end do
       end do
     else
      allocate (hx_vect(1), &
                stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of hx_vect")
      allocate (hx_shift(1,1), &
                stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of hx_shift")
     end if

    if(allocated(kkk_hx)) then
       deallocate (kkk_hx,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating kkk_hx')
    end if


    return

  end subroutine init_io_data

  !-------------------------------------------------------
  subroutine input_data(tag_data,i_block_z,i_block_w, &
  gen_par,data_par,down_par,data_block_struct)

    character(len=128) :: tag_data
    integer i_block_z,i_block_w
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par
    type(data_block_type) :: data_block_struct

    type(fft_type), dimension (:,:), pointer  :: fft_data_block_id
    complex, dimension (:,:,:,:,:,:), pointer :: data_block
!    target data_block_struct%data_block

    integer, external :: omp_get_thread_num

    integer i_mx,i_my,i_hx,i_hy,ierr,i_hx_0_shift,i_k_hx
    integer n_bytes,nreed,i_w_block_par,i_w_block_ser
    integer nseek_blocks,ret_sseek,i_z_seek
    integer ihx_spike_0,ihy_spike_0,ihx_spike_1,ihy_spike_1
    integer imx_spike_0,imy_spike_0,imx_spike_1,imy_spike_1
    integer iw_min_block,i_w
    real w_data,nint_w_min_0,mx,my,hx,hy,delta_t_m,t_hyper
    complex j
    complex, allocatable, dimension (:,:,:,:) :: data_slice_nopad

    data_block => data_block_struct%data_block
    fft_data_block_id => data_block_struct%fft_data_block_id

    j=cmplx(0.,1.)


    if (down_par%spike_data)  then

       !{
       if(gen_par%debug1) write(0,*)'spike_data= YES'
       imx_spike_0=1+(gen_par%mx_spike_0-data_par%mx_0_pad)/data_par%d_mx
       imy_spike_0=1+(gen_par%my_spike_0-data_par%my_0_pad)/data_par%d_my
       ihx_spike_0=1+(gen_par%hx_spike_0-data_par%hx_0_pad)/data_par%d_hx
       ihy_spike_0=1+(gen_par%hy_spike_0-data_par%hy_0_pad)/data_par%d_hy
       imx_spike_1=1+(gen_par%mx_spike_1-data_par%mx_0_pad)/data_par%d_mx
       imy_spike_1=1+(gen_par%my_spike_1-data_par%my_0_pad)/data_par%d_my
       ihx_spike_1=1+(gen_par%hx_spike_1-data_par%hx_0_pad)/data_par%d_hx
       ihy_spike_1=1+(gen_par%hy_spike_1-data_par%hy_0_pad)/data_par%d_hy
       if(gen_par%debug2) write(0,*)'t_spike_0=',gen_par%t_spike_0
       if(gen_par%debug2) write(0,*)'imx_spike_0=',imx_spike_0
       if(gen_par%debug2) write(0,*)'imy_spike_0=',imy_spike_0
       if(gen_par%debug2) write(0,*)'ihx_spike_0=',ihx_spike_0
       if(gen_par%debug2) write(0,*)'ihy_spike_0=',ihy_spike_0
       if(gen_par%debug2) write(0,*)'imx_spike_1=',imx_spike_1
       if(gen_par%debug2) write(0,*)'imy_spike_1=',imy_spike_1
       if(gen_par%debug2) write(0,*)'ihx_spike_1=',ihx_spike_1
       if(gen_par%debug2) write(0,*)'ihy_spike_1=',ihy_spike_1
       if(gen_par%debug2) write(0,*)'shape(data_block)=',shape(data_block)


       data_block = cmplx(0.,0.)
       !*$*assert do (serial) 
       do i_w_block_par=1,down_par%n_w_block_par
          !{
          !*$*assert do (serial) 
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             i_w=down_par%n_w_block_par*(i_w_block_ser-1) + i_w_block_par
             w_data=down_par%w_min_block+(i_w-1)*data_par%d_w
             data_block(imx_spike_0,imy_spike_0,ihx_spike_0,ihy_spike_0, &
             i_w_block_ser,i_w_block_par) =  &
             scale_data*cexp(-j*w_data*gen_par%t_spike_0)
             if(gen_par%t_spike_1 >= 0.) then
                data_block(imx_spike_1,imy_spike_1,ihx_spike_1,ihy_spike_1, &
                i_w_block_ser,i_w_block_par) =  &
                data_block(imx_spike_1,imy_spike_1,ihx_spike_1,ihy_spike_1, &
                i_w_block_ser,i_w_block_par) + &
                scale_data*cexp(-j*w_data*gen_par%t_spike_1)
             end if
!write(0,*)'AFter spike  maxval(cabs(data_block))=',maxval(cabs(data_block)),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser,'w_data=',w_data
             !}
          end do
          !}
       end do
       !}
    else
       !{
       if (down_par%diffr_data)  then
          !{
          if(gen_par%debug1) write(0,*)'diffr_data= YES'
          data_block = cmplx(0.,0.)
          !*$*assert do (serial) 
          do i_w_block_par=1,down_par%n_w_block_par
             !{
             !*$*assert do (serial) 
             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                i_w=down_par%n_w_block_par*(i_w_block_ser-1) + i_w_block_par
                w_data=down_par%w_min_block+(i_w-1)*data_par%d_w

                !*$*assert do (serial) 
                do i_hy=1,data_par%n_hy_pad
                   !{
                   hy=data_par%hy_0 + (i_hy-1)*data_par%d_hy
                   !*$*assert do (serial) 
                   do i_hx=1,data_par%n_hx_pad
                      !{
                      hx=data_par%hx_0_pad + (i_hx-1)*data_par%d_hx
                      !*$*assert do (serial) 
                      do i_my=1,data_par%n_my_pad
                         !{
                         !*$*assert do (serial) 
                         do i_mx=1,data_par%n_mx_pad
                            !{
                            mx=data_par%mx_0_pad + (i_mx-1)*data_par%d_mx
                            mx=mx-gen_par%mx_spike_0
                            my=data_par%my_0_pad + (i_my-1)*data_par%d_my
                            my=my-gen_par%my_spike_0

                            t_hyper=sqrt(gen_par%t_spike_0**2+ &
                            (2.*mx*gen_par%slow_const)**2 + &
                            (2.*my*gen_par%slow_const)**2 + &
                            (2.*hx*gen_par%slow_const)**2 + &
                            (2.*hy*gen_par%slow_const)**2)

                            data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par) = &
                            scale_data*cexp(-j*w_data*(t_hyper))

                            if(gen_par%t_spike_1 >= 0.) then
                               mx=data_par%mx_0_pad + (i_mx-1)*data_par%d_mx
                               mx=mx-gen_par%mx_spike_1
                               my=data_par%my_0_pad + (i_my-1)*data_par%d_my
                               my=my-gen_par%my_spike_1
                               t_hyper=sqrt(gen_par%t_spike_1**2+ &
                               (2.*mx*gen_par%slow_const)**2 + &
                               (2.*my*gen_par%slow_const)**2 + &
                               (2.*hx*gen_par%slow_const)**2 + &
                               (2.*hy*gen_par%slow_const)**2)

                               data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par)= &
                               data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par)+ &
                               scale_data*cexp(-j*w_data*(t_hyper))
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       else
          !{
          if (down_par%plane_data)  then
             !{
             if(gen_par%debug1) write(0,*)'plane_data= YES'
             data_block = cmplx(0.,0.)
             !*$*assert do (serial) 
             do i_w_block_par=1,down_par%n_w_block_par
                !{
                !*$*assert do (serial) 
                do i_w_block_ser=1,down_par%n_w_block_ser
                   !{
                   i_w=down_par%n_w_block_par*(i_w_block_ser-1) + i_w_block_par
                   w_data=down_par%w_min_block+(i_w-1)*data_par%d_w

                   !*$*assert do (serial) 
                   do i_hy=1,data_par%n_hy_pad
                      !{
                      hy=data_par%hy_0 + (i_hy-1)*data_par%d_hy
                      !*$*assert do (serial) 
                      do i_hx=1,data_par%n_hx_pad
                         !{
                         hx=data_par%hx_0_pad + (i_hx-1)*data_par%d_hx
                         !*$*assert do (serial) 
                         do i_my=1,data_par%n_my_pad
                            !{
                            my=data_par%my_0_pad + (i_my-1)*data_par%d_my
                            !*$*assert do (serial) 
                            do i_mx=1,data_par%n_mx_pad
                               !{
                               mx=data_par%mx_0_pad + (i_mx-1)*data_par%d_mx
                               t_hyper=sqrt(gen_par%t_spike_0**2+ &
                               (2.*hx*gen_par%slow_const)**2 + &
                               (2.*hy*gen_par%slow_const)**2)
                               delta_t_m= gen_par%px_plane_0*mx + gen_par%py_plane_0*my

                               data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par) = &
                               scale_data*cexp(-j*w_data*(t_hyper+delta_t_m))

                               if(gen_par%t_spike_1 >= 0.) then
                                  t_hyper=sqrt(gen_par%t_spike_1**2+ &
                                  (2.*hx*gen_par%slow_const)**2 + &
                                  (2.*hy*gen_par%slow_const)**2)
                                  delta_t_m= gen_par%px_plane_1*mx + gen_par%py_plane_1*my

                                  data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par)= &
                                  data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par)+ &
                                  scale_data*cexp(-j*w_data*(t_hyper+delta_t_m))
                               end if
                               !}
                            end do
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          else
             !{
             allocate (data_slice_nopad(data_par%n_mx,data_par%n_my, &
             data_par%n_hx,data_par%n_hy), &
             stat=ierr)  
             if(ierr .ne. 0)call seperr("Problems with allocation of data_slice_nopad")


             iw_min_block=nint((down_par%w_min_block-data_par%w_0)/data_par%d_w)+1
             n_bytes= &
             data_par%n_hx*data_par%n_hy*data_par%n_mx*data_par%n_my*8
             i_z_seek = (i_block_z-1)*data_par%n_w
             nseek_blocks=(iw_min_block-1) + i_z_seek

             ret_sseek=sseek_block(trim(tag_data),nseek_blocks,n_bytes,0) 
             if (ret_sseek .ne. nseek_blocks) then
                !{
                write(0,*)'ret_sseek=',ret_sseek
                write(0,*)'tag_data=',trim(tag_data)
                call seperr("Error block seeking data")
                !}
             end if

             !*$*assert do (serial) 
             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                !*$*assert do (serial) 
                do i_w_block_par=1,down_par%n_w_block_par
                   !{
                   nreed= sreed(trim(tag_data),data_slice_nopad,n_bytes)
                   if (nreed .ne. n_bytes) then
                      write(0,*)'tag_data=',trim(tag_data)
                      call seperr("Error reading data")
                   end if
                   if(gen_par%transf_wei) then
                     data_slice_nopad=conjg(data_slice_nopad)
                   end if

                   if(data_par%t_data_0 /= 0.) then
                     i_w=down_par%n_w_block_par*(i_w_block_ser-1) + i_w_block_par
                     w_data=down_par%w_min_block+(i_w-1)*data_par%d_w
                     data_block( &
                     mx_start_inp:mx_stop_inp,my_start_inp:my_stop_inp, &
                     hx_start_inp:hx_stop_inp,hy_start_inp:hy_stop_inp, &
                     i_w_block_ser,i_w_block_par) = &
                     data_slice_nopad(:,:,:,:)*cexp(-j*w_data*data_par%t_data_0)
                   else
                     data_block( &
                     mx_start_inp:mx_stop_inp,my_start_inp:my_stop_inp, &
                     hx_start_inp:hx_stop_inp,hy_start_inp:hy_stop_inp, &
                     i_w_block_ser,i_w_block_par) = &
                     data_slice_nopad(:,:,:,:)
                   end if
                   !}
!write(0,*)'INPUT: after copy maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,i_w_block_ser,i_w_block_par))),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser
                end do
                !}
             end do
             deallocate (data_slice_nopad,stat=ierr)  
             if(ierr .ne. 0) call seperr("Problems with deallocation data_slice_nopad")
             !}
          end if
          !}
       end if
       !}
    end if

    if(gen_par%k_domain) then 
       !{

!write(0,*)'Before ffr in data spike  maxval(cabs(data_block))=',maxval(cabs(data_block))
       do i_w_block_ser=1,down_par%n_w_block_ser
          !{
          !$omp parallel
          !$omp do
!       !$omp master
          do i_w_block_par=1,down_par%n_w_block_par
             !{
!write(0,*)'in  data_input: thread number',omp_get_thread_num()
             if(gen_par%n_data_dim > 4) then
             !{
                call fft_4d_data(&
                  fft_data_block_id(i_w_block_ser,i_w_block_par),FFT_FORW, & 
                  data_par%n_mx_pad,data_par%n_my_pad, &
                  data_par%n_hx_pad,data_par%n_hy_pad, &
                  data_block(:,:,:,:,i_w_block_ser,i_w_block_par))
             !}
             else
             !{
               if(gen_par%hx_0_shift == 0.) then
               !{
!write(0,*)'INPUT: before FFT maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,i_w_block_ser,i_w_block_par))),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser
                call fft_4d_data(&
                  fft_data_block_id(i_w_block_ser,i_w_block_par),FFT_FORW, & 
                  data_par%n_mx_pad,data_par%n_my_pad, &
                  data_par%n_hx_pad,data_par%n_hy_pad, &
                  data_block(:,:,:,:,i_w_block_ser,i_w_block_par))
!write(0,*)'INPUT: after FFT maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,i_w_block_ser,i_w_block_par))),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser
               !}
               else
               !{
 
call seperr('hx_0_shift != 0 in data_mod not implemetend in v14 yet')

                  do i_mx=1,data_par%n_mx_pad
                    i_hx_0_shift= &
             mod(i_mx-gen_par%i_hx_0_shift_start,gen_par%n_hx_0_shift_period)+1
                    do i_my=1,data_par%n_my_pad
                      hx_vect=data_block(i_mx,i_my,:,1,i_w_block_ser,i_w_block_par)
!    THIS WAS NOT COMMENTED    call fft_1d_data(fft_1d_id,FFT_FORW, &
!                           data_par%n_hx_pad, &
!                           hx_vect)
                      data_block(i_mx,i_my,:,1,i_w_block_ser,i_w_block_par) = &
                        hx_vect*hx_shift(:,i_hx_0_shift)
                    end do
                  end do
                  do i_k_hx=1,data_par%n_hx_pad
!   THIS WAS NOT COMMENTED   call fft_2d_data(fft_2d_id,FFT_FORW, &
!                    data_par%n_mx_pad,data_par%n_my_pad, &
!                    data_block(:,:,i_k_hx,1,i_w_block_ser,i_w_block_par))
                  end do
               !}
               end if
             !}
             end if

             !}
          end do
!       !$omp end master
          !$omp end parallel
          !}
       end do

!write(0,*)'After ffr in data spike  maxval(cabs(data_block))=',maxval(cabs(data_block))

       !}
    end if

    return	
  end subroutine input_data


  subroutine output_data(tag_data,i_block_z,gen_par,data_par,down_par,restart_par, &
                         data_block_struct)

    character(len=128) :: tag_data
    integer i_block_z
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) down_par
    type (restart_param_type) :: restart_par
    type(data_block_type) :: data_block_struct

    type(fft_type), dimension (:,:), pointer  :: fft_data_block_id
    complex, dimension (:,:,:,:,:,:), pointer :: data_block

    logical add_to_data,preserve_data
    integer ierr
    integer n_bytes_io,nrite,nreed,i_w_block_par,i_w_block_ser
    integer nseek_blocks,ret_sseek,i_z_seek
    integer iw_min_block,n_w_blocks
    integer i_k_hx
    integer :: i_total_restart,n_total_restart

    complex, allocatable, dimension (:,:,:,:) :: data_slice_io,temp
!    complex, allocatable, dimension (:,:,:,:) :: data_slice_pad

    data_block => data_block_struct%data_block
    fft_data_block_id => data_block_struct%fft_data_block_id

    preserve_data=(data_par%wavefield_file == DATA_OUTPUT) 

 write(0,*)'preserve_data=',preserve_data

    add_to_data=(&
        ((down_par%add).and. &
         gen_par%domain_range(1:9) =='data_imag') .or. &
        ((down_par%add).and. &
        gen_par%domain_range(1:9) =='data_wave') &
                )
!write(0,*)'add_to_data=',add_to_data
!    allocate (data_slice_pad(data_par%n_mx_pad,data_par%n_my_pad, &
!    data_par%n_hx_pad,data_par%n_hy_pad), &
!    stat=ierr)  
!    if(ierr .ne. 0) call seperr("Problems with allocation of data_slice_pad")
!    n_bytes_io=data_par%n_hx_pad*data_par%n_hy_pad* &
!               data_par%n_mx_pad*data_par%n_my_pad*8
!

    if(.not. gen_par%write_down_data_pad) then
      allocate (data_slice_io(data_par%n_mx,data_par%n_my, &
      data_par%n_hx,data_par%n_hy), &
      stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of data_slice_io")
      n_bytes_io= data_par%n_hx*data_par%n_hy*data_par%n_mx*data_par%n_my*8
      write(0,*)'shape data_slice_io=',shape(data_slice_io)
      if(preserve_data) then
        allocate (temp(data_par%n_mx_pad,data_par%n_my_pad, &
        data_par%n_hx_pad,data_par%n_hy_pad), &
        stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp")
 write(0,*)'shape temp=',shape(temp)
      end if
    else
      if(add_to_data) then
        allocate (data_slice_io(data_par%n_mx_pad,data_par%n_my_pad, &
        data_par%n_hx_pad,data_par%n_hy_pad), &
        stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of data_slice_io")
 write(0,*)'shape data_slice_io=',shape(data_slice_io)
        n_bytes_io=data_par%n_hx_pad*data_par%n_hy_pad* &
                   data_par%n_mx_pad*data_par%n_my_pad*8
      end if
    end if

    iw_min_block=nint((down_par%w_min_block-data_par%w_0)/data_par%d_w)+1
    i_z_seek = (i_block_z-1)*data_par%n_w
    nseek_blocks=(iw_min_block-1) + i_z_seek

    if(gen_par%debug1) then
       write(0,*)'tag_data=',trim(tag_data)
   write(0,*)'iw_min_block=',iw_min_block
   write(0,*)'i_block_z=',i_block_z
   write(0,*)'i_z_seek=',i_z_seek
   write(0,*)'nseek_blocks==',nseek_blocks
   write(0,*)'down_par%n_w_block_ser==',down_par%n_w_block_ser
   write(0,*)'down_par%n_w_block_par==',down_par%n_w_block_par
     end if

    ret_sseek=sseek_block(trim(tag_data),nseek_blocks,n_bytes_io,0) 
    if (ret_sseek /= nseek_blocks) then
       !{
       write(0,*)'nseek_blocks=',nseek_blocks
       write(0,*)'ret_sseek=',ret_sseek
       write(0,*)'tag_data=',trim(tag_data)
       call seperr("Error block seeking Data in output")
       !}
    end if


    !*$*assert do (serial) 
    do i_w_block_ser=1,down_par%n_w_block_ser
       !{
!   write(0,*)'i_w_block_ser==',i_w_block_ser
       !*$*assert do (serial) 
       do i_w_block_par=1,down_par%n_w_block_par
          !{
!   write(0,*)'i_w_block_par==',i_w_block_par
          data_block(:,:,:,:,i_w_block_ser,i_w_block_par)=&
            scale_data*data_block(:,:,:,:,i_w_block_ser,i_w_block_par)

          if(gen_par%k_domain) then 
             !{
!             if(gen_par%n_data_dim > 4) then
!                call fft_4d_data(&
!                  fft_data_block_id(i_w_block_ser,i_w_block_par),FFT_BACK, & 
!                  data_par%n_mx_pad,data_par%n_my_pad, &
!                  data_par%n_hx_pad,data_par%n_hy_pad, &
!                  data_block(:,:,:,:,i_w_block_ser,i_w_block_par))
!             else
!write(0,*)'before FFT maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,i_w_block_ser,i_w_block_par))),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser
                if(preserve_data) then
                  temp=data_block(:,:,:,:,i_w_block_ser,i_w_block_par)
                end if
                call fft_4d_data(&
                  fft_data_block_id(i_w_block_ser,i_w_block_par),FFT_BACK, & 
                  data_par%n_mx_pad,data_par%n_my_pad, &
                  data_par%n_hx_pad,data_par%n_hy_pad, &
                  data_block(:,:,:,:,i_w_block_ser,i_w_block_par))

                data_slice_io=&
                data_block( &
                   mx_start_out:mx_stop_out,my_start_out:my_stop_out, &
                   hx_start_out:hx_stop_out,hy_start_out:hy_stop_out, &
                   i_w_block_ser,i_w_block_par)
                if(preserve_data) then
                  data_block(:,:,:,:,i_w_block_ser,i_w_block_par)=temp
                end if
!write(0,*)'after FFT  maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,i_w_block_ser,i_w_block_par))),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser

!             end if
             !}
          end if

          if(add_to_data) then
          !{
            nreed= sreed(trim(tag_data),data_slice_io,n_bytes_io)
            if (nreed .ne. n_bytes_io) then
               write(0,*)'tag_data=',trim(tag_data)
               call seperr("Error reading Data (add==yes)")
            end if
            ret_sseek=sseek_block(trim(tag_data),-1,n_bytes_io,1) 
            n_w_blocks=i_w_block_par-1+(i_w_block_ser-1)*down_par%n_w_block_par 
            if (ret_sseek /= (nseek_blocks +n_w_blocks)) then
               !{
               write(0,*)'nseek_blocks=',nseek_blocks
               write(0,*)'n_w_blocks=',n_w_blocks
               write(0,*)'ret_sseek=',ret_sseek
               write(0,*)'tag_data=',trim(tag_data)
               call seperr("Error block seeking Data in output (add==yes)")
               !}
            end if
            data_slice_io(:,:,:,:) = &
              data_slice_io(:,:,:,:) + &
              data_block( &
                mx_start_out:mx_stop_out,my_start_out:my_stop_out, &
                hx_start_out:hx_stop_out,hy_start_out:hy_stop_out, &
                i_w_block_ser,i_w_block_par)
!write(0,*)'after copy  maxval(cabs(data_slice_io))=',maxval(cabs(data_slice_io)),'with i_wblock_par=',i_w_block_par,'i_w_block_ser=',i_w_block_ser
!              data_slice_pad( &
!                mx_start_out:mx_stop_out,my_start_out:my_stop_out, &
!                hx_start_out:hx_stop_out,hy_start_out:hy_stop_out)

            nrite= srite(trim(tag_data),data_slice_io,n_bytes_io)
            if (nrite .ne. n_bytes_io) then
               write(0,*)'tag_data=',trim(tag_data)
               call seperr("Error writing Data with add_to_data=yes")
            end if
          !}
          else
          !{
            if(.not. gen_par%write_down_data_pad) then
            !{
!              data_slice_io(:,:,:,:) = &
!              data_block( &
!                mx_start_out:mx_stop_out,my_start_out:my_stop_out, &
!                hx_start_out:hx_stop_out,hy_start_out:hy_stop_out, &
!                i_w_block_ser,i_w_block_par)
!                data_slice_pad( &
!                  mx_start_out:mx_stop_out,my_start_out:my_stop_out, &
!                  hx_start_out:hx_stop_out,hy_start_out:hy_stop_out)
              nrite= srite(trim(tag_data),data_slice_io,n_bytes_io)
              if (nrite .ne. n_bytes_io) then
                 write(0,*)'tag_data=',trim(tag_data)
                 call seperr("Error writing Data")
              end if
            !}
            else
            !{
!              nrite= srite(trim(tag_data),data_slice_pad,n_bytes_io)
              nrite= srite(trim(tag_data), data_block(:,:,:,:,i_w_block_ser,i_w_block_par), n_bytes_io)
              if (nrite .ne. n_bytes_io) then
                 write(0,*)'tag_data=',trim(tag_data)
                 call seperr("Error writing Data")
              end if
            !}
            end if
          !}
          end if
          !}
       end do
       !}
    end do

   call sep_prog_stat("Gendown",restart_par%i_total,restart_par%n_total,1)

    if(.not. gen_par%write_down_data_pad) then
      deallocate (data_slice_io,stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with deallocation data_slice_io")
    else
      if(add_to_data .or. preserve_data) then
        deallocate (data_slice_io,stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with deallocation data_slice_io")
        if(preserve_data) then
          deallocate (temp,stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with deallocation data_slice_io")
        end if
      end if
    end if

!    deallocate (data_slice_pad,stat=ierr)  
!    if(ierr .ne. 0) call seperr("Problems with deallocation data_slice_pad")
!
!    if(gen_par%k_domain) then 
!       !{
!       call clean_fft(fft_3d_id)
!       if(gen_par%n_data_dim > 4) call clean_fft(fft_1d_id)
!       !}
!    end if

    return	
  end subroutine output_data




  subroutine clean_io_data(gen_par)
    type (gendown3d_param_type) gen_par

    integer ierr

    if(gen_par%k_domain) then 
!       call clean_fft(fft_3d_id)
!       call clean_fft(fft_2d_id)
!       if(gen_par%n_data_dim > 4) call clean_fft(fft_1d_id)
    end if

    if(allocated(hx_shift)) then
       deallocate (hx_shift,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hx_shift')
    end if
    if(allocated(hx_vect)) then
       deallocate (hx_vect,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hx_vect')
    end if


    return
  end subroutine clean_io_data

end module data_mod
