!*$*OPTIMIZE(5)
!*$*ROUNDOFF(3)
!*$*SCALAROPTIMIZE(3)

module down_gen_par_mod
  use gendown3d_types
  use data_types
  use image_types
  use slow_types
  use down_process_types
  use restart_types
  use bound_types

  use down_gen_ser_mod
  use data_mod
  use image_mod
  use slow_mod
  use filter_data_mod

  implicit none
contains
  function down_gen_par( &
  n_block_w,n_block_z,z_down_0, &
  gen_par,data_par,image_par,down_par,slow_par,restart_par,bound_par, &
  down_oper, scatter_oper, &
  slow_data_block, vti_delta_block, vti_eta_block, dslow_data_block,&
  data_block_struct,ddata_block_struct,&
  image_block_struct,dimage_block_struct,amp_stat_block) &
  result(stat)
    integer stat
    integer n_block_w,n_block_z
    real  z_down_0,z_0_block,inv_pi2
    type(gendown3d_param_type) :: gen_par
    type (data_param_type) :: data_par
    type (image_param_type) :: image_par
    type (down_process_param_type) :: down_par
    type (slow_param_type) :: slow_par
    type (restart_param_type) :: restart_par
    type (bound_param_type) bound_par

    integer down_oper,scatter_oper
    optional down_oper,scatter_oper
    external down_oper,scatter_oper

    real, dimension (:,:,:,:,:,:) :: slow_data_block
    real, dimension (:,:,:,:,:,:) :: vti_delta_block
    real, dimension (:,:,:,:,:,:) :: vti_eta_block
    complex, dimension (:,:,:,:,:,:,:) :: dslow_data_block
    type(data_block_type) :: data_block_struct
    type(data_block_type) :: ddata_block_struct
    type(image_block_type) :: image_block_struct
    type(image_block_type) :: dimage_block_struct
    real, dimension (:,:,:,:,:,:) :: amp_stat_block


    integer stat_par(gen_par%n_threads)

    logical read_back_wave
    integer ierr
    integer i_block_w,i_block_z,i_block_z_read_wave,i_w_block_par
    real nint_w_min_0
!    integer, external :: omp_get_thread_num

    complex, dimension (:,:,:,:,:,:), pointer :: data_block
    complex, dimension (:,:,:,:,:,:), pointer :: ddata_block
    complex, dimension (:,:,:,:,:,:), pointer :: image_block
    complex, dimension (:,:,:,:,:,:), pointer :: dimage_block

    data_block => data_block_struct%data_block
    ddata_block => ddata_block_struct%data_block

    image_block => image_block_struct%image_block
    dimage_block => dimage_block_struct%image_block

    inv_pi2=1./(2.*acos(-1.))
    stat_par=0 ! in case that we never call the inner routines
!    !*$*assert do (serial) 
    do i_block_w=1,n_block_w
       !{
       if(gen_par%verb) write(0,*)'i_block_w=',i_block_w,' of ',n_block_w
       nint_w_min_0=nint(down_par%w_min_0/data_par%d_w)*data_par%d_w
       down_par%w_min_block= &
       nint_w_min_0+(i_block_w-1)*gen_par%n_w_block*data_par%d_w
       down_par%w_max_block= &
       down_par%w_min_block+(gen_par%n_w_block-1)*data_par%d_w
       if(gen_par%verb) write(0,*)'down_par-f_min_0=',down_par%w_min_0*inv_pi2
       if(gen_par%verb) write(0,*)'down_par-f_min_block=',down_par%w_min_block*inv_pi2
       if(gen_par%verb) write(0,*)'down_par-f_max_block=',down_par%w_max_block*inv_pi2


       if(gen_par%stat_shift /= 0.) then
         amp_stat_block=0.
       end if

       if ((data_par%data_file == DATA_INPUT).or. &
       (data_par%data_file == DATA_SYNTH)) then
          !{
          if(gen_par%debug1) write(0,*)'Read Data'
          call SEP_timer_start(14)
          call input_data(data_par%tag_data,1,i_block_w, &
            gen_par,data_par,down_par, &
            data_block_struct)
          if(gen_par%debug1) write(0,*)'Done Read Data'
          call SEP_timer_stop(14)
          if(down_par%filter_data) then
             if(gen_par%debug1) write(0,*)'Filter Data'
             call filter_data(i_block_w,gen_par,data_par,down_par,data_block)
          end if
          !}
       else
          !{
          if (data_par%down_data_file == DATA_INPUT)then
             if(gen_par%debug1) write(0,*)'Read Down_data'
             call input_data(data_par%tag_down_data,1,i_block_w, &
               gen_par,data_par,down_par, &
               data_block_struct)

             if(down_par%filter_data) then
                if(gen_par%debug1) write(0,*)'Filter Down_data'
                call filter_data(i_block_w,gen_par,data_par,down_par,data_block)
             end if
          else
             if(gen_par%debug1) write(0,*)'Init data_block'
             data_block=cmplx(0.,0.)
          end if
          !}
       end if

       if (gen_par%oper_type(1:1) == 'd') then
          if (data_par%ddata_file == DATA_INPUT) then
             if(gen_par%debug1) write(0,*)'Read Ddata'

             call input_data(data_par%tag_ddata,1,i_block_w, &
               gen_par,data_par,down_par, &
               ddata_block_struct)

             if(gen_par%debug1) write(0,*)'After Read Ddata'
          else
             if(gen_par%debug1) write(0,*)'Zero Ddata'
             ddata_block=cmplx(0.,0.)
          end if
       end if

!       !*$*assert do (serial) 
       do i_block_z=1,n_block_z
          !{
          z_0_block=down_par%z_down_0 + &
                    (i_block_z-1)*down_par%n_z_block*down_par%d_z_down
          if(gen_par%verb) write(0,*)'i_block_z=',i_block_z
          if(gen_par%verb) write(0,*)'z_0_block=',z_0_block
          restart_par%i_total=i_block_z+(i_block_w-1)*restart_par%n_block_z
          if(gen_par%verb) write(0,*)'restart_par-i_total=',restart_par%i_total

          ! read appropriate veocity files
          if(slow_par%slow_file == SLOW_INPUT) then
            call SEP_timer_start(11)
              if(gen_par%debug1) write(0,*)'Read Slow'
              call read_slow(slow_par%tag_slow,i_block_w,z_0_block, &
              data_par,down_par,slow_par, &
              slow_data_block,dslow_data_block)
            call SEP_timer_stop(11)
            if(slow_par%vti_delta_file == SLOW_INPUT) then
              if(gen_par%debug1) write(0,*)'Read Vti Delta'
              call read_slow(slow_par%tag_vti_delta,i_block_w,z_0_block, &
              data_par,down_par,slow_par, &
              vti_delta_block,dslow_data_block)
            end if
            if(slow_par%vti_eta_file == SLOW_INPUT) then
              if(gen_par%debug1) write(0,*)'Read Vti Eta'
              call read_slow(slow_par%tag_vti_eta,i_block_w,z_0_block, &
              data_par,down_par,slow_par, &
              vti_eta_block,dslow_data_block)
            end if
          else
             if(gen_par%debug1) write(0,*)'DO NOT read Slow'
          end if

          if(slow_par%dslow_file == SLOW_INPUT) then
             if(gen_par%debug1) write(0,*)'Read Dslow'
             call read_slow(slow_par%tag_dslow,i_block_w,z_0_block, &
             data_par,down_par,slow_par, &
             slow_data_block,dslow_data_block)
             call inp_dslow_block_par(down_par, dslow_data_block)
          else
             if(slow_par%dslow_file == SLOW_OUTPUT) then
                if(gen_par%debug1) write(0,*)'Init Dslow'
                dslow_data_block=cmplx(0.,0.)
             else
                if(gen_par%debug1) write(0,*)'DO NOT read Dslow'
             end if
          end if

          ! read background wavefield 
          if(data_par%wavefield_file == DATA_INPUT) then
             read_back_wave=.true.
             if(down_par%adj) then
                i_block_z_read_wave = (n_block_z - i_block_z +1)
                if(gen_par%debug1) &
                write(0,*)'Read Wavefield i_block_z_read_wave=',i_block_z_read_wave
                call input_data(data_par%tag_wavefield,i_block_z_read_wave, &
                  i_block_w, &
                  gen_par,data_par,down_par, &
                  data_block_struct)

             else
                i_block_z_read_wave = i_block_z 
                if(gen_par%debug1) &
                write(0,*)'Read Wavefield i_block_z_read_wave=',i_block_z_read_wave
                call input_data(data_par%tag_wavefield,i_block_z_read_wave, &
                  i_block_w, &
                  gen_par,data_par,down_par, &
                  data_block_struct)
             end if
          else
             read_back_wave=.false.
          end if
          if(gen_par%debug1) write(0,*)'read_back_wave=',read_back_wave

          ! spread all the image components from the parallel threads
          if(down_par%adj) then
             !{
             if (image_par%image_file == IMAGE_INPUT) then
                !{
                if(gen_par%debug1) write(0,*)'Input Image'
                  call input_image(image_par%tag_image,i_block_w,i_block_z, &
                  gen_par,down_par,image_par,image_block_struct)
                  call inp_image_block_par(down_par,data_par,image_block)
                !}
             end if
             if (image_par%dimage_file == IMAGE_INPUT) then
                !{
                if(gen_par%debug1) write(0,*)'Input Dimage'
                  call input_image(image_par%tag_dimage,i_block_w,i_block_z, &
                  gen_par,down_par,image_par,dimage_block_struct)
                  call inp_image_block_par(down_par,data_par,dimage_block)
                !}
             end if
             !}
          end if

          ! this loop should go in parallel
!          !!*$*assert do (serial) 
!          !*$*assert concurrent call
!          !$ DOACROSS

!$omp parallel
!$omp do
          do i_w_block_par=1,down_par%n_w_block_par
          call SEP_timer_start(15)
             stat_par(i_w_block_par)= down_gen_ser( &
             down_oper, scatter_oper, &
             read_back_wave,i_w_block_par,z_0_block, &
             gen_par,data_par,image_par,down_par,slow_par,bound_par, &
             slow_data_block,vti_delta_block,vti_eta_block,dslow_data_block, &
             data_block_struct,ddata_block_struct, &
             image_block,dimage_block,amp_stat_block)
          call SEP_timer_stop(15)
          end do
!$omp end parallel


          ! summ in all the image components from the parallel threads
          if (image_par%image_file == IMAGE_OUTPUT) then
             !{
             if(gen_par%debug1) write(0,*)'Output Image'
          call SEP_timer_start(12)
!write(0,*)'Before out_image_block_par maxval(cabs(image_block))=',maxval(cabs(image_block))
             call out_image_block_par(down_par, data_par, image_block)
!write(0,*)'AFter out_image_block_par maxval(cabs(image_block))=',maxval(cabs(image_block))
          call SEP_timer_stop(12)
          call SEP_timer_start(13)
             call output_image(image_par%tag_image,i_block_w,i_block_z, &
               gen_par,down_par,image_par,restart_par,image_block_struct)
          call SEP_timer_stop(13)
             !}
          end if
          if (image_par%dimage_file == IMAGE_OUTPUT) then
             !{
             if(gen_par%debug1) write(0,*)'Output Dimage'
             call out_image_block_par(down_par,data_par, dimage_block)
             call output_image(image_par%tag_dimage,i_block_w,i_block_z, &
               gen_par,down_par,image_par,restart_par,dimage_block_struct)
             !}
          end if

          if(slow_par%dslow_file == SLOW_OUTPUT) then
             if(gen_par%debug1) write(0,*)'Write Dslow'
             call out_dslow_block_par(down_par,dslow_data_block)
             call write_slow(slow_par%tag_dslow,i_block_w,z_0_block, &
             data_par,down_par,slow_par, &
             slow_data_block,dslow_data_block)
          end if

          ! write background wavefield 
          if(data_par%wavefield_file == DATA_OUTPUT) then
             if(gen_par%debug1) write(0,*)'Write Wavefield'
             call output_data(data_par%tag_wavefield,i_block_z, &
               gen_par,data_par,down_par,restart_par, &
               data_block_struct)
          end if
          !}
       end do

       if(data_par%data_file == DATA_OUTPUT) then
          if(down_par%filter_data) then
             call filter_data(i_block_w,gen_par,data_par,down_par,data_block)
          end if
          if(gen_par%debug1) write(0,*)'Output Data'
          call output_data(data_par%tag_data,1, &
               gen_par,data_par,down_par,restart_par, &
               data_block_struct)
       end if
       if(data_par%down_data_file == DATA_OUTPUT) then
          if(gen_par%debug1) write(0,*)'Output Down_data'
          call output_data(data_par%tag_down_data,1, &
               gen_par,data_par,down_par,restart_par, &
               data_block_struct)
       end if
       if(data_par%down_ddata_file == DATA_OUTPUT) then
          if(gen_par%debug1) write(0,*)'Output Down_ddata'
          call output_data(data_par%tag_down_ddata,1, &
               gen_par,data_par,down_par,restart_par, &
               ddata_block_struct)
       end if
       !}
    end do
    stat=0
    if(any(stat_par /= 0)) stat =1
    return
  end function down_gen_par
end module down_gen_par_mod

