module down_gen_ser_mod

use gendown3d_types
use data_types
use image_types
use slow_types
use down_process_types
use bound_types
use slow_mod

use image_mod
use data_mod
use bound_mod

implicit none

contains

function down_gen_ser( &
             down_oper,scatter_oper, &
             read_back_wave,i_w_block_par,z_0_block, &
             gen_par,data_par,image_par,down_par,slow_par,bound_par, &
             slow_data_block,vti_delta_block,vti_eta_block,dslow_data_block, &
             data_block_struct,ddata_block_struct, &
             image_block,dimage_block,amp_stat_block) result (stat)


interface 
  function down_oper(adj,add, &
                   n_w_data,w_0_data,d_w_data, i_w_block_par, &
                   z_down,d_z_down,data_par, &
                   data_block_struct,ddata_block_struct, &
                   slow_slice,vti_delta_slice,vti_eta_slice,iz) &
        result(stat)
    use data_types, only: data_param_type,data_block_type
    integer                                   :: stat,n_w_data,i_w_block_par,iz
    logical                                   :: adj,add
    real                                      :: w_0_data,d_w_data
    real                                      :: z_down,d_z_down
    type (data_param_type)                         :: data_par
    type(data_block_type), optional :: data_block_struct
    type(data_block_type), optional :: ddata_block_struct
    real, optional,  dimension (:,:,:,:,:)      :: slow_slice
    real, optional,  dimension (:,:,:,:,:)      :: vti_delta_slice
    real, optional,  dimension (:,:,:,:,:)      :: vti_eta_slice
  end function down_oper

  function scatter_oper(adj,add, &
                 n_w_data,w_0_data,d_w_data, i_w_block_par, &
                 z_down,d_z_down,data_par, &
                 data_block_struct,ddata_block_struct,dslow_slice, &
                 slow_slice,vti_delta_slice,vti_eta_slice) &
                 result(stat)
    use data_types, only: data_param_type,data_block_type
    integer                                   :: stat,n_w_data,i_w_block_par
    logical                                   :: adj,add
    real                                      :: w_0_data,d_w_data
    real                                      :: z_down,d_z_down
    type (data_param_type)                         :: data_par
    type(data_block_type) :: data_block_struct
    type(data_block_type) :: ddata_block_struct
    complex, dimension (:,:,:,:,:)              :: dslow_slice
    real, optional,  dimension (:,:,:,:,:)      :: slow_slice
    real, optional,  dimension (:,:,:,:,:)      :: vti_delta_slice
    real, optional,  dimension (:,:,:,:,:)      :: vti_eta_slice
  end function scatter_oper

end interface

    logical read_back_wave,propagate_back
    integer i_w_block_par,stat
    real z_0_block
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (image_param_type) image_par
    type (down_process_param_type) down_par
    type (slow_param_type) slow_par
    type (bound_param_type) bound_par

    real,  dimension (:,:,:,:,:,:) :: slow_data_block
    real,  dimension (:,:,:,:,:,:) :: vti_delta_block
    real,  dimension (:,:,:,:,:,:) :: vti_eta_block
    real,  dimension (:,:,:,:,:,:) :: amp_stat_block
    complex,  dimension (:,:,:,:,:,:,:) :: dslow_data_block
    type(data_block_type) :: data_block_struct
    type(data_block_type) :: ddata_block_struct
    complex, dimension (:,:,:,:,:,:) :: image_block
    complex, dimension (:,:,:,:,:,:) :: dimage_block
    optional scatter_oper


    integer ierr,i1,i2,i3,i4,i5
    integer i_w_block_ser,n_w_data
    integer i_z_block,i_hx,i_hy,i_my,i_mx
    real w_0_data,d_w_data,z_down,norm,d_z_down

!    integer, external :: omp_get_thread_num

    real, allocatable,  dimension (:) :: norm_w

    complex, dimension (:,:,:,:,:,:), pointer :: data_block
    complex, dimension (:,:,:,:,:,:), pointer :: ddata_block

    data_block => data_block_struct%data_block
    ddata_block => ddata_block_struct%data_block

    if(gen_par%check_stability) then
       allocate (norm_w(down_par%n_w_block_ser),stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of norm_w")
    end if

    if (image_par%image_file == IMAGE_OUTPUT) then
      do i5=1,size(image_block,5)
      do i4=1,size(image_block,4)
      do i3=1,size(image_block,3)
      do i2=1,size(image_block,2)
      do i1=1,size(image_block,1)
       image_block(i1,i2,i3,i4,i5,i_w_block_par)=0.
      end do 
      end do 
      end do 
      end do 
      end do 
    end if
    if (image_par%dimage_file == IMAGE_OUTPUT) then
       dimage_block(:,:,:,:,:,i_w_block_par)=cmplx(0.,0.)
    end if

    do i_z_block=1,down_par%n_z_block
       !{
       z_down=z_0_block+ (i_z_block-1)*down_par%d_z_down
       d_z_down=down_par%d_z_down
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'down_par-adj=',down_par%adj
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'down_par-n_z_block=',down_par%n_z_block

       if(read_back_wave) then
          if (((i_z_block == 1) .and. down_par%adj) .or. &
          ((i_z_block == down_par%n_z_block) .and. (.not. down_par%adj)) &
          ) then
             propagate_back=.false.
                  write(0,*)'Do not propgatae background t i_z_block=',i_z_block
          else
             propagate_back=.true.
          end if
       else
          propagate_back=.true.
       end if

  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'propagate_back=',propagate_back
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'down_par-adj=',down_par%adj
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'down_par-n_z_block=',down_par%n_z_block
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'i_z_block=',i_z_block
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'z_down=',z_down
  if((gen_par%verb) .and. (i_w_block_par ==1)) write(0,*)'d_z_down=',d_z_down
       !if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'down_par-w_min_block=',down_par%w_min_block
       !if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'down_par-d_w=',data_par%d_w


       if(gen_par%check_stability) then
          !{
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             norm_w(i_w_block_ser)=0.
             do i_hy=1,data_par%n_hy
                do i_hx=1,data_par%n_hx
                   do i_my=1,data_par%n_my
                      do i_mx=1,data_par%n_mx
                         if(gen_par%oper_type(1:1) == 'd') then
                            norm_w(i_w_block_ser)= &
                            norm_w(i_w_block_ser) + cabs( &
                            ddata_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par))**2 
                         else
                            norm_w(i_w_block_ser)= &
                            norm_w(i_w_block_ser) + cabs( &
                            data_block(i_mx,i_my,i_hx,i_hy,i_w_block_ser,i_w_block_par))**2 
                         end if
                      end do
                   end do
                end do
             end do
             !}
          end do
          write(0,*)'Max(n_w_ser) norm wavefield at z_down=',z_down,sqrt(maxval(norm_w))
          !}
       end if

       if(.not. down_par%adj) then
          !{
       if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'before image with not adj'
          ! image inside threads
          if (image_par%image_file == IMAGE_OUTPUT) then
             call out_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             data_block,image_block,slow_data_block,amp_stat_block)
          end if
          if (image_par%dimage_file == IMAGE_OUTPUT) then
             call out_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             ddata_block,dimage_block,slow_data_block,amp_stat_block)
          end if
       if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'after image with not adj'
          !}
       end if



       n_w_data=down_par%n_w_block_ser
       d_w_data=down_par%n_w_block_par*data_par%d_w
       w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w

       if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'before calling down_operj'
       ! downward continue
       call SEP_timer_start(17)
       select case (gen_par%oper_type(1:2))
       case ('co')
          stat= down_oper(down_par%adj,down_par%add, &
          n_w_data,w_0_data,d_w_data,i_w_block_par, &
          z_down,d_z_down,data_par, &
          data_block_struct=data_block_struct,&
          iz=i_z_block )
       case ('fu')
          stat= down_oper(down_par%adj,down_par%add,  &
          n_w_data,w_0_data,d_w_data,i_w_block_par, &
          z_down,d_z_down,data_par, &
          data_block_struct=data_block_struct, &
          slow_slice=slow_data_block(:,:,:,:,:,i_z_block), &
          vti_delta_slice=vti_delta_block(:,:,:,:,:,i_z_block), &
          vti_eta_slice=vti_eta_block(:,:,:,:,:,i_z_block), &
          iz=i_z_block &
          )
       case ('dc')
          if(down_par%adj) then
             stat= scatter_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct, &
             dslow_slice=dslow_data_block(:,:,:,:,:,i_z_block,i_w_block_par))
          end if

          if(propagate_back) then
             stat= down_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct,iz=i_z_block)
          else
             stat= down_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             ddata_block_struct=ddata_block_struct,iz=i_z_block)
          end if

          if(.not. down_par%adj) then
             stat= scatter_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct, &
             dslow_slice=dslow_data_block(:,:,:,:,:,i_z_block,i_w_block_par))
          end if

       case ('df')
          if(down_par%adj) then
             stat= scatter_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct, &
             dslow_slice=dslow_data_block(:,:,:,:,:,i_z_block,i_w_block_par), &
             slow_slice=slow_data_block(:,:,:,:,:,i_z_block), &
             vti_delta_slice=vti_delta_block(:,:,:,:,:,i_z_block), &
             vti_eta_slice=vti_eta_block(:,:,:,:,:,i_z_block) &
             )
          end if

          if(propagate_back) then

             stat= down_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct, &
             slow_slice=slow_data_block(:,:,:,:,:,i_z_block), &
             vti_delta_slice=vti_delta_block(:,:,:,:,:,i_z_block), &
             vti_eta_slice=vti_eta_block(:,:,:,:,:,i_z_block),iz=i_z_block &
             )
          else
             stat= down_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             ddata_block_struct=ddata_block_struct, &
             slow_slice=slow_data_block(:,:,:,:,:,i_z_block), &
             vti_delta_slice=vti_delta_block(:,:,:,:,:,i_z_block), &
             vti_eta_slice=vti_eta_block(:,:,:,:,:,i_z_block),iz=i_z_block &

             )
          end if
          if(.not. down_par%adj) then
             stat= scatter_oper(down_par%adj,down_par%add, &
             n_w_data,w_0_data,d_w_data,i_w_block_par, &
             z_down,d_z_down,data_par, &
             data_block_struct=data_block_struct, &
             ddata_block_struct=ddata_block_struct, &
             dslow_slice=dslow_data_block(:,:,:,:,:,i_z_block,i_w_block_par), &
             slow_slice=slow_data_block(:,:,:,:,:,i_z_block), &
             vti_delta_slice=vti_delta_block(:,:,:,:,:,i_z_block), &
             vti_eta_slice=vti_eta_block(:,:,:,:,:,i_z_block) &
             )
          end if

       case default
          write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
          call seperr('Not supported oper_type')
       end select

       if(stat /= 0) call seperr('Problems in down_oper')
       if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'after calling down_operj'

       call SEP_timer_stop(17)

!write(0,*)'AFter down_oper maxval(cabs(data_block))=',maxval(cabs(data_block(:,:,:,:,:,i_w_block_par)))



       if(down_par%adj) then
       if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'before image with adj'
          !{
          ! image inside threads
          if (image_par%image_file == IMAGE_OUTPUT) then
             call out_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             data_block,image_block,slow_data_block,amp_stat_block)
          end if
          if (image_par%dimage_file == IMAGE_OUTPUT) then
             call out_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             ddata_block,dimage_block,slow_data_block,amp_stat_block)
          end if
          if (image_par%image_file == IMAGE_INPUT) then
             call inp_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             data_block,image_block,slow_data_block,amp_stat_block)
          end if
          if (image_par%dimage_file == IMAGE_INPUT) then
             call inp_image_block_ser(i_w_block_par,i_z_block,z_down, &
             gen_par,data_par,down_par,image_par,slow_par, &
             ddata_block,dimage_block,slow_data_block,amp_stat_block)
          end if
      if((gen_par%debug2) .and. (i_w_block_par ==1)) write(0,*)'after image with adj'
          !}
       end if


       !}
    end do

    if(allocated(norm_w)) then
       deallocate (norm_w,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating norm_w')
    end if

    return
  end function down_gen_ser

end module down_gen_ser_mod
