module slow_mod
  use gendown3d_types
  use down_process_types
  use data_types
  use slow_types
  use image_types

  implicit none
  logical, private :: interp_x_slow,interp_y_slow
  integer, external ::  sreed,srite,sseek_block
  integer, private :: nc_bytes,nr_bytes,n_block_z,n_z_half
  integer, private :: sou_slow,rec_slow
  real, private :: scale_dslow

  integer, allocatable, dimension (:,:,:) :: ix_slow,iy_slow
  real, allocatable, dimension (:,:,:), private :: wx_slow,wy_slow
  real, allocatable, dimension (:,:,:,:), private :: slow_block
  real, allocatable, dimension (:,:), private :: rslow_slice
  complex, allocatable, dimension (:,:,:), private :: dslow_block
  complex, allocatable, dimension (:,:), private :: cslow_slice

contains
  subroutine init_io_slow(gen_par,down_par,data_par,slow_par,image_par)
    implicit none
    integer ierr
    type (gendown3d_param_type) gen_par
    type (down_process_param_type) down_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par
    type (image_param_type) image_par
    integer rat_dz

    nr_bytes=slow_par%n_mx_slow*slow_par%n_my_slow*down_par%n_z_block* 4
    nc_bytes=slow_par%n_mx_slow*slow_par%n_my_slow*down_par%n_z_block* 8

    call init_interpolate_slow(gen_par,data_par,slow_par)

    if((slow_par%slow_file /= SLOW_NO_FILE).and. &
       (slow_par%slow_shear_file == SLOW_NO_FILE)) then
       allocate (slow_block(slow_par%n_mx_slow,slow_par%n_my_slow, &
       down_par%n_z_block,1),stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of slow_block (no shear)")
       if(gen_par%debug2) write(0,*)'Allocate slow_block with shape=',shape(slow_block)
    end if

    if((slow_par%slow_file /= SLOW_NO_FILE).and. &
       (slow_par%slow_shear_file /= SLOW_NO_FILE)) then
       allocate (slow_block(slow_par%n_mx_slow,slow_par%n_my_slow, &
       down_par%n_z_block,2),stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of slow_block (yes shear)")
       if(gen_par%debug2) write(0,*)'Allocate slow_block with shape=',shape(slow_block)
    end if


    if(slow_par%dslow_file /= SLOW_NO_FILE) then
       allocate (dslow_block(slow_par%n_mx_slow,slow_par%n_my_slow, &
       down_par%n_z_block),stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of dslow_block")
       if(gen_par%debug2) write(0,*)'Allocate dslow_block with shape=',shape(dslow_block)
    end if

    if(down_par%adj) then
       n_z_half=down_par%n_z_block/2

       if(slow_par%slow_file /= SLOW_NO_FILE) then
          if(gen_par%debug2) write(0,*)'Allocate rslow_slice'
          allocate (rslow_slice(slow_par%n_mx_slow,slow_par%n_my_slow), &
          stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of rslow_slice")
       end if
       if(slow_par%dslow_file /= SLOW_NO_FILE) then
          if(gen_par%debug2) write(0,*)'Allocate cslow_slice'
          allocate (cslow_slice(slow_par%n_mx_slow,slow_par%n_my_slow), &
          stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of cslow_slice")
       end if
    end if

    rat_dz=nint(slow_par%d_z_slow/down_par%d_z_down)
    if((rat_dz /= 1) .and. (rat_dz /= -1)) then
       write(0,*)'slow_par-d_z_slow',slow_par%d_z_slow
       write(0,*)'down_par-d_z_down',down_par%d_z_down
       call seperr('rat_dz /= 1 not implemented yet')
    end if

    n_block_z=image_par%n_z_image/down_par%n_z_block
  end subroutine init_io_slow

  subroutine init_interpolate_slow(gen_par,data_par,slow_par)
    implicit none
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par
    integer ierr
    integer i_mx,i_my,i_hx,i_hy
    real mx,my,hx,hy,rx,ry,sx,sy,inv_d_mx,inv_d_my

    if(slow_par%slow_shear_file /= SLOW_NO_FILE) then
      if(slow_par%shear_receiver) then
        sou_slow=P_SLOW
        rec_slow=S_SLOW
      else
        sou_slow=S_SLOW
        rec_slow=P_SLOW
      end if
    else
        sou_slow=P_SLOW
        rec_slow=P_SLOW
    end if

    interp_x_slow=.false.
    interp_y_slow=.false.

    if (data_par%n_hx_pad /= 1)  interp_x_slow=.true.

    if(((slow_par%n_mx_slow /= data_par%n_mx_pad) .or.  &
        (slow_par%d_mx_slow /= data_par%d_mx) .or.  &
        (slow_par%mx_slow_0 /= data_par%mx_0_pad) .or.  &
        (data_par%n_hx_pad /= 1)) .and. &
        (slow_par%n_mx_slow /= 1) &
      ) interp_x_slow=.true.

    if(((slow_par%n_my_slow /= data_par%n_my_pad) .or.  &
        (slow_par%d_my_slow /= data_par%d_my) .or.  &
        (slow_par%my_slow_0 /= data_par%my_0_pad) .or.  &
        (data_par%n_hy_pad /= 1)) .and. &
        (slow_par%n_my_slow /= 1) &
      ) interp_y_slow=.true.
    if((.not. interp_y_slow) .and. (.not. slow_par%extrapolate_slow)) then
      if(data_par%my_0_pad < slow_par%my_slow_0) then
        call seperr('data_par-my_0_pad < slow_par-my_slow_0')
      end if
      if(data_par%my_0_pad + data_par%d_my*(data_par%n_my_pad-1) > & 
         slow_par%my_slow_0 + slow_par%d_my_slow*(slow_par%n_my_slow-1)) then
        call seperr('Data my larger than max slow my')
      end if
    end if
    
    if(gen_par%debug2) write(0,*)'interp_x_slow=',interp_x_slow
    if(gen_par%debug2) write(0,*)'interp_y_slow=',interp_y_slow

    if(interp_x_slow) then
       !{
       allocate (ix_slow(data_par%n_mx_pad, &
       data_par%n_hx_pad, &
       slow_par%n_comp_slow), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of ix_slow")
       if(gen_par%debug2) write(0,*)'Allocate ix_slow with shape=',shape(ix_slow)
       allocate (wx_slow(data_par%n_mx_pad, &
       data_par%n_hx_pad, &
       slow_par%n_comp_slow), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of wx_slow")
       if(gen_par%debug2) write(0,*)'Allocate wx_slow with shape=',shape(wx_slow)

       inv_d_mx=1./slow_par%d_mx_slow

       do i_hx=1,data_par%n_hx_pad 
          !{
          hx=data_par%hx_0_pad + (i_hx-1)* data_par%d_hx
          do i_mx=1,data_par%n_mx_pad 
             !{
             mx=data_par%mx_0_pad + (i_mx-1)* data_par%d_mx
             sx=mx-hx
             sx=(sx-slow_par%mx_slow_0)*inv_d_mx+1
             if((sx < 1) .or. (sx >= slow_par%n_mx_slow)) then
                if(slow_par%extrapolate_slow) then
                   ix_slow(i_mx,i_hx,SOU_COMP)=min(slow_par%n_mx_slow-1,max(1,nint(sx)))
!                   wx_slow(i_mx,i_hx,SOU_COMP)=.5
                else
                   write(0,*)'sx=',sx
                   call seperr('Try to use velocity/slowness outside range of mx')
                end if
             else
                ix_slow(i_mx,i_hx,SOU_COMP)=nint(sx)
                wx_slow(i_mx,i_hx,SOU_COMP)=sx-int(sx)
             end if

             if(slow_par%n_comp_slow > 1) then
                rx=mx+hx
                rx=(rx-slow_par%mx_slow_0)*inv_d_mx+1
                if((rx < 1) .or. (rx >= slow_par%n_mx_slow)) then
                   if(slow_par%extrapolate_slow) then
                      ix_slow(i_mx,i_hx,REC_COMP)=min(slow_par%n_mx_slow-1,max(1,nint(rx)))
                      wx_slow(i_mx,i_hx,REC_COMP)=.5
                   else
                      write(0,*)'rx=',rx
                      call seperr('Try to use velocity/slowness outside range of my')
                   end if
                else
                   ix_slow(i_mx,i_hx,REC_COMP)=nint(rx)
                   wx_slow(i_mx,i_hx,REC_COMP)=rx-int(rx)
                end if
             end if
             !}
          end do
          !}
       end do
       !}
    end if

    if(interp_y_slow) then
       !{
       allocate (iy_slow(data_par%n_my_pad, &
       data_par%n_hy_pad, &
       slow_par%n_comp_slow), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of iy_slow")
       if(gen_par%debug2) write(0,*)'Allocate iy_slow with shape=',shape(iy_slow)
       allocate (wy_slow(data_par%n_my_pad, &
       data_par%n_hy_pad, &
       slow_par%n_comp_slow), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of wy_slow")
       if(gen_par%debug2) write(0,*)'Allocate wy_slow with shape=',shape(wy_slow)

       inv_d_my=1./slow_par%d_my_slow

       do i_hy=1,data_par%n_hy_pad 
          !{
          hy=data_par%hy_0_pad + (i_hy-1)* data_par%d_hy
          do i_my=1,data_par%n_my_pad 
             !{
             my=data_par%my_0_pad + (i_my-1)* data_par%d_my
             sy=my-hy
             sy=(sy-slow_par%my_slow_0)*inv_d_my+1
             if((sy < 1) .or. (sy >= slow_par%n_my_slow)) then
                if(slow_par%extrapolate_slow) then
                   iy_slow(i_my,i_hy,SOU_COMP)=min(slow_par%n_my_slow-1,max(1,nint(sy)))

                   wy_slow(i_my,i_hy,SOU_COMP)=.5
                else
                   write(0,*)'sx=',sx
                   call seperr('Try to use velocity/slowness outside range of my')
                end if
             else
                iy_slow(i_my,i_hy,SOU_COMP)=nint(sy)
                wy_slow(i_my,i_hy,SOU_COMP)=sy-int(sy)
             end if

             if(slow_par%n_comp_slow > 1) then
                ry=my+hy
                ry=(ry-slow_par%my_slow_0)*inv_d_my+1
                if((ry < 1) .or. (ry >= slow_par%n_my_slow)) then
                   if(slow_par%extrapolate_slow) then
                      iy_slow(i_my,i_hy,REC_COMP)=min(slow_par%n_my_slow-1,max(1,nint(ry)))
                      wy_slow(i_my,i_hy,REC_COMP)=.5
                   else
                      write(0,*)'rx=',rx
                      call seperr('Try to use velocity/slowness outside range of my')
                   end if
                else
                   iy_slow(i_my,i_hy,REC_COMP)=nint(ry)
                   wy_slow(i_my,i_hy,REC_COMP)=ry-int(ry)
                end if
             end if
             !}
          end do
          !}
       end do

       !}
    end if
    return
  end subroutine init_interpolate_slow

  subroutine read_slow(tag_slow,i_block_w,z_0_block, &
  data_par,down_par,slow_par, &
  slow_data_block,dslow_data_block)
    implicit none

    character(len=128) :: tag_slow
    integer :: i_block_w,i_hx,n_bytes_io
    real :: z_0_block
    type (data_param_type) data_par
    type (down_process_param_type) down_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_data_block
    complex, optional, dimension (:,:,:,:,:,:,:) :: dslow_data_block


    character(len=1) :: type_slow
    integer :: ierr,iseek,i_z_block,len_tag,n_bytes,i_block_z

write(0,*) "IN READ",trim(tag_slow)
    len_tag=len_trim(tag_slow)
    if(tag_slow(1:len_tag) == slow_par%tag_dslow(1:len_tag)) then
       type_slow(1:1) = 'c'
       n_bytes=nc_bytes
    else
       if((tag_slow(1:len_tag) == slow_par%tag_slow(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_slow_shear(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_delta(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_eta(1:len_tag))) then
         type_slow(1:1) = 'r'
         n_bytes=nr_bytes
       else
         write(0,*)'tag_slow=',trim(tag_slow)
         call seperr('Not recognized tag_slow')
       end if
    end if

    i_block_z=nint((z_0_block-slow_par%z_slow_0)/ &
                   (down_par%n_z_block*slow_par%d_z_slow))+1

    if(down_par%adj) then
!       iseek=(n_block_z-i_block_z)
       iseek=(i_block_z-1)
    else
       iseek=(i_block_z-1)
    end if
!    write(0,*)'in read_slow z_0_block=',z_0_block
!    write(0,*)'in read_slow slow_par-z_slow_0=',slow_par%z_slow_0
!    write(0,*)'in read_slow slow_par-d_z_slow=',slow_par%d_z_slow
!    write(0,*)'in read_slow down_par-n_z_block=',down_par%n_z_block
!    write(0,*)'in read_slow n_block_z=',n_block_z
!    write(0,*)'in read_slow i_block_z=',i_block_z,' iseek =',iseek

    if(sseek_block(trim(tag_slow),iseek,n_bytes,0) .ne. iseek ) then
       write(0,*)'tag_slow=',trim(tag_slow)
       call seperr('Problems with seeking in tag_slow file (first)')
    end if
    if(slow_par%slow_shear_file /= SLOW_NO_FILE) then
      len_tag=len_trim(slow_par%tag_slow)
      len_tag=len_trim(slow_par%tag_slow_shear)
      if(sseek_block(trim(slow_par%tag_slow_shear(1:len_tag)), &
            iseek,n_bytes,0)/=iseek) then
         call seperr('Problems with seeking in tag_slow_shear file (first)')
      end if
    end if

    if(type_slow(1:1) == 'c') then
       !{
       n_bytes_io=sreed(trim(tag_slow),dslow_block,n_bytes)
       if(n_bytes_io .ne. n_bytes ) then
          write(0,*)'iseek=',iseek
          write(0,*)'n_bytes_io=',n_bytes_io
          write(0,*)'n_bytes=',n_bytes
          write(0,*)'tag_slow=',trim(tag_slow)
          call seperr('Problems with reading tag_slow file (complex)')
       end if

       if(down_par%adj) then
          do i_z_block=1,n_z_half
             cslow_slice=dslow_block(:,:,i_z_block)
             dslow_block(:,:,i_z_block)=dslow_block(:,:,down_par%n_z_block-i_z_block+1)
             dslow_block(:,:,down_par%n_z_block-i_z_block+1)=cslow_slice
          end do
       end if
       !}
    else
       !{
       n_bytes_io=sreed(trim(tag_slow),slow_block(:,:,:,P_SLOW),n_bytes)
       if(n_bytes_io .ne. n_bytes ) then
          write(0,*)'iseek=',iseek
          write(0,*)'n_bytes_io=',n_bytes_io
          write(0,*)'n_bytes=',n_bytes
          write(0,*)'tag_slow=',trim(tag_slow)
          call seperr('Problems with reading tag_slow file (real)')
       end if
       if(slow_par%slow_shear_file /= SLOW_NO_FILE) then
         n_bytes_io=sreed(trim(slow_par%tag_slow_shear), &
                          slow_block(:,:,:,S_SLOW),n_bytes)
         if(n_bytes_io .ne. n_bytes ) then
            write(0,*)'iseek=',iseek
            write(0,*)'n_bytes_io=',n_bytes_io
            write(0,*)'n_bytes=',n_bytes
            write(0,*)'tag_slow_shear=',trim(slow_par%tag_slow_shear)
            call seperr('Problems with reading tag_slow_shear file (real)')
         end if
       end if

       if(down_par%adj) then
          do i_z_block=1,n_z_half
             rslow_slice=slow_block(:,:,i_z_block,P_SLOW)
             slow_block(:,:,i_z_block,P_SLOW)= &
                slow_block(:,:,down_par%n_z_block-i_z_block+1,P_SLOW)
             slow_block(:,:,down_par%n_z_block-i_z_block+1,P_SLOW)=rslow_slice
          end do
          if(slow_par%slow_shear_file /= SLOW_NO_FILE) then
            do i_z_block=1,n_z_half
               rslow_slice=slow_block(:,:,i_z_block,S_SLOW)
               slow_block(:,:,i_z_block,S_SLOW)= &
                  slow_block(:,:,down_par%n_z_block-i_z_block+1,S_SLOW)
               slow_block(:,:,down_par%n_z_block-i_z_block+1,S_SLOW)=rslow_slice
            end do
          end if
       end if
       !}
    end if

!write(0,*)'before interp: min(slow)=',minval(slow_data_block)
write(0,*)'before interp: max(slow)=',maxval(slow_data_block)
    call interpolate_slow(.false.,tag_slow,data_par,down_par,slow_par, &
    slow_data_block,dslow_data_block)
write(0,*)'after interp: min(slow)=',minval(slow_data_block)
!write(0,*)'after interp: max(slow)=',maxval(slow_data_block)

    return	
  end subroutine read_slow

  subroutine write_slow(tag_slow,i_block_w,z_0_block, &
  data_par,down_par,slow_par, &
  slow_data_block,dslow_data_block)
    implicit none

    character(len=128) :: tag_slow
    integer :: i_block_w,i_hx
    real :: z_0_block
    type (down_process_param_type) down_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_data_block
    complex, optional, dimension (:,:,:,:,:,:,:) :: dslow_data_block

    logical add_to_dslow
    character(len=1) :: type_slow
    integer :: ierr,iseek,iseek_slice,i_z_block,len_tag
    integer :: n_bytes_block,n_bytes_slice,i_block_z

    len_tag=len_trim(tag_slow)
    if(tag_slow(1:len_tag) == slow_par%tag_dslow(1:len_tag)) then
       type_slow(1:1) = 'c'
       n_bytes_block=nc_bytes
       if(((i_block_w /= 1) .and. down_par%adj) .or. down_par%add) then
          add_to_dslow=.true.
       else
          add_to_dslow=.false.
       end if
    else
       if((tag_slow(1:len_tag) == slow_par%tag_slow(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_slow_shear(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_delta(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_eta(1:len_tag))) then
         type_slow(1:1) = 'r'
         n_bytes_block=nr_bytes
       else
         write(0,*)'tag_slow=',trim(tag_slow)
         call seperr('Not recognized tag_slow')
       end if
    end if
!    write(0,*)'add_to_dslow=',add_to_dslow
!    write(0,*)'maxval(dslow_block)=',maxval(cabs(dslow_block))
!    write(0,*)'minval(dslow_block)=',minval(cabs(dslow_block))
    n_bytes_slice=n_bytes_block/down_par%n_z_block

    call interpolate_slow(.true.,tag_slow,data_par,down_par,slow_par, &
    slow_data_block,dslow_data_block)
!    write(0,*)'after maxval(dslow_block)=',maxval(cabs(dslow_block))
!    write(0,*)'after minval(dslow_block)=',minval(cabs(dslow_block))

    i_block_z=(z_0_block-slow_par%z_slow_0)/ &
              (down_par%n_z_block*slow_par%d_z_slow)+1

    if(down_par%adj) then
!       iseek=(n_block_z-i_block_z)
       iseek=(i_block_z-1)
    else
       iseek=(i_block_z-1)
    end if
    if(sseek_block(trim(tag_slow),iseek,n_bytes_block,0) .ne. iseek ) then
       write(0,*)'tag_slow=',trim(tag_slow)
       call seperr('Problems with seeking in tag_slow file (first)')
    end if

    if(type_slow(1:1) == 'c') then
       !{
       if(down_par%adj) then
          do i_z_block=1,n_z_half
             cslow_slice=dslow_block(:,:,i_z_block)
             dslow_block(:,:,i_z_block)=dslow_block(:,:,down_par%n_z_block-i_z_block+1)
             dslow_block(:,:,down_par%n_z_block-i_z_block+1)=cslow_slice
          end do
       end if
       if(add_to_dslow) then
          !{
          do i_z_block=1,down_par%n_z_block
             if(sreed(trim(tag_slow),cslow_slice,n_bytes_slice) /= n_bytes_slice ) then
                write(0,*)'tag_slow=',trim(tag_slow)
                call seperr('Problems with reading tag_slow file (complex)')
             end if
             cslow_slice= cslow_slice + dslow_block(:,:,i_z_block)
             iseek_slice=(iseek*down_par%n_z_block) + i_z_block-1
             if(sseek_block(trim(tag_slow),-1,n_bytes_slice,1) /= iseek_slice ) then
                write(0,*)'iseek_slice=',iseek_slice
                write(0,*)'tag_slow=',trim(tag_slow)
                call seperr('Problems with seeking in tag_slow file (add=yes)')
             end if
             if(srite(trim(tag_slow),cslow_slice,n_bytes_slice) /= n_bytes_slice ) then
                write(0,*)'tag_slow=',trim(tag_slow)
                call seperr('Problems with overwriting tag_slow file (complex)')
             end if
          end do
          !}
       else
          !{
          if(srite(trim(tag_slow),dslow_block,n_bytes_block) /= n_bytes_block ) then
             write(0,*)'tag_slow=',trim(tag_slow)
             call seperr('Problems with writing tag_slow file (complex)')
          end if
          !}
       end if
       !}
    else
       !{
       if(down_par%adj) then
          do i_z_block=1,n_z_half
             rslow_slice=slow_block(:,:,i_z_block,P_SLOW)
             slow_block(:,:,i_z_block,P_SLOW)=slow_block(:,:,down_par%n_z_block-i_z_block+1,P_SLOW)
             slow_block(:,:,down_par%n_z_block-i_z_block+1,P_SLOW)=rslow_slice
          end do
       end if
       if(srite(trim(tag_slow),slow_block,n_bytes_block) /= n_bytes_block ) then
          write(0,*)'tag_slow=',trim(tag_slow)
          call seperr('Problems with writing tag_slow file (real)')
       end if
       !}
    end if
    return	
  end subroutine write_slow

  subroutine interpolate_slow(adj,tag_slow,data_par,down_par,slow_par, &
  slow_data_block,dslow_data_block)
    implicit none

    logical adj
    character(len=128) :: tag_slow
    type (data_param_type) data_par
    type (down_process_param_type) down_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_data_block
    complex, optional, dimension (:,:,:,:,:,:,:) :: dslow_data_block

    character(len=1) :: type_slow
    integer :: ierr,i_z_block,len_tag
    integer i_mx,i_my,i_hx,i_hy,i_mx_slow
    real w00_sou,w10_sou,w01_sou,w11_sou
    real w00_rec,w10_rec,w01_rec,w11_rec

    len_tag=len_trim(tag_slow)
    if(tag_slow(1:len_tag) == slow_par%tag_dslow(1:len_tag)) then
       type_slow(1:1) = 'c'
    else
       if((tag_slow(1:len_tag) == slow_par%tag_slow(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_slow_shear(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_delta(1:len_tag)) .or. &
          (tag_slow(1:len_tag) == slow_par%tag_vti_eta(1:len_tag))) then
         type_slow(1:1) = 'r'
         if(adj) call seperr('interpolate_slow: adj=.true. and real slow')
       else
         write(0,*)'tag_slow=',trim(tag_slow)
         call seperr('Not recognized tag_slow')
       end if
    end if

    if(adj) then
       !{
       if(interp_x_slow .and. interp_y_slow) then
          call seperr('interpolate_slow: adj=.true. and 3-D interp')
       else
          !{
          if(interp_x_slow) then
             !{
             if(type_slow(1:1) == 'c') then
                dslow_block=cmplx(0.,0.)
             else
                call seperr('interpolate_slow: adj=.true. and real slow (5)')
             end if

             do i_hy=1,data_par%n_hy_pad 
                !{
                do i_hx=1,data_par%n_hx_pad 
                   !{
                   do i_my=1,data_par%n_my_pad 
                      !{
                      do i_mx=1,data_par%n_mx_pad 
                         !{
                         w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))
                         w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))

                         if(type_slow(1:1) == 'c') then
                            !{
                            i_mx_slow=ix_slow(i_mx,i_hx,SOU_COMP)
                            !write(0,*)'i_mx_slow',i_mx_slow
                            do i_z_block=1,down_par%n_z_block
                               !{
                               dslow_block(i_mx_slow  , &
                               i_my, i_z_block) =             & 
                               dslow_block(i_mx_slow  , &
                               i_my, i_z_block) +             & 
                               dslow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block,1)* &
                               w00_sou 

                               dslow_block(i_mx_slow+1, &
                               i_my, i_z_block) =             &
                               dslow_block(i_mx_slow+1, &
                               i_my, i_z_block) +             &
                               dslow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block,1)* &
                               w10_sou 
                               !}
                            end do
                            !}
                         else
                            if(adj) call seperr('interpolate_slow: adj=.true. and real slow (3)')
                         end if
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do

             if(slow_par%n_comp_slow > 1) then
                !{
                do i_hy=1,data_par%n_hy_pad 
                   !{
                   do i_hx=1,data_par%n_hx_pad 
                      !{
                      do i_my=1,data_par%n_my_pad 
                         !{
                         do i_mx=1,data_par%n_mx_pad 
                            !{
                            w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))
                            w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))

                            if(type_slow(1:1) == 'c') then
                               !{
                               i_mx_slow=ix_slow(i_mx,i_hx,REC_COMP)
                               do i_z_block=1,down_par%n_z_block
                                  !{
                                  dslow_block(i_mx_slow  , &
                                  i_my, i_z_block) =             & 
                                  dslow_block(i_mx_slow  , &
                                  i_my, i_z_block) +             & 
                                  dslow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block,1)* &
                                  w00_rec

                                  dslow_block(i_mx_slow+1, &
                                  i_my, i_z_block) =             &
                                  dslow_block(i_mx_slow+1, &
                                  i_my, i_z_block) +             &
                                  dslow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block,1)* &
                                  w10_rec 
                                  !}
                               end do
                               !}
                            else
                               !{
                               if(adj) call seperr('interpolate_slow: adj=.true. and real slow (4)')
                               !}
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do

                !}
             end if
             !}
          else
             !{
             if(interp_y_slow) then
                call seperr ('Interpolation in y not implemented yet')
             else
                !{
                if(type_slow(1:1) == 'c') then
                   if(slow_par%n_comp_slow > 1) then
                      dslow_block(:,:,:) = &
                      .5*(dslow_data_block(:,:,1,1,SOU_COMP,:,1) + &
                      dslow_data_block(:,:,1,1,REC_COMP,:,1))
                   else
                      dslow_block(:,:,:)=dslow_data_block(:,:,1,1,1,:,1)
                   end if
                else
                   call seperr('interpolate_slow: adj=.true. and real slow (2)')
                end if
                !}
             end if
             !}
          end if
          !}
       end if

       !}
    else
       !{
       if(interp_x_slow .and. interp_y_slow) then
          !{
          do i_hy=1,data_par%n_hy_pad 
             !{
             do i_hx=1,data_par%n_hx_pad 
                !{
                do i_my=1,data_par%n_my_pad 
                   !{
                   do i_mx=1,data_par%n_mx_pad 
                      !{
                      w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (1.-wy_slow(i_my,i_hy,SOU_COMP))
                      w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (1.-wy_slow(i_my,i_hy,SOU_COMP))
                      w01_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (   wy_slow(i_my,i_hy,SOU_COMP))
                      w11_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (   wy_slow(i_my,i_hy,SOU_COMP))
                      if(type_slow(1:1) == 'c') then
                         !{
                         do i_z_block=1,down_par%n_z_block
                            !{
                            dslow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block,1) = &
                            dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)  , &
                            iy_slow(i_my,i_hy,SOU_COMP)  , &
                            i_z_block) * w00_sou +         &
                            dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                            iy_slow(i_my,i_hy,SOU_COMP)  , &
                            i_z_block) * w10_sou +         &
                            dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)  , &
                            iy_slow(i_my,i_hy,SOU_COMP)+1, &
                            i_z_block) * w01_sou +         &
                            dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                            iy_slow(i_my,i_hy,SOU_COMP)+1, &
                            i_z_block) * w11_sou 
                            !}
                         end do
                         !}
                      else
                         !{
                         do i_z_block=1,down_par%n_z_block
                            !{
                            slow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block) = &
                            slow_block(ix_slow(i_mx,i_hx,SOU_COMP)  , &
                            iy_slow(i_my,i_hy,SOU_COMP)  , &
                            i_z_block,sou_slow) 
                            !}
                         end do
                         !}
                      end if
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          end do

          if(slow_par%n_comp_slow > 1) then
             !{
             do i_hy=1,data_par%n_hy_pad 
                !{
                do i_hx=1,data_par%n_hx_pad 
                   !{
                   do i_my=1,data_par%n_my_pad 
                      !{
                      do i_mx=1,data_par%n_mx_pad 
                         !{
                         w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))* &
                         (1.-wy_slow(i_my,i_hy,REC_COMP))
                         w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))* &
                         (1.-wy_slow(i_my,i_hy,REC_COMP))
                         w01_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))* &
                         (   wy_slow(i_my,i_hy,REC_COMP))
                         w11_rec=(   wx_slow(i_mx,i_hx,REC_COMP))* &
                         (   wy_slow(i_my,i_hy,REC_COMP))
                         if(type_slow(1:1) == 'c') then
                            !{
                            do i_z_block=1,down_par%n_z_block
                               !{
                               dslow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block,1) = &
                               dslow_block(ix_slow(i_mx,i_hx,REC_COMP)  , &
                               iy_slow(i_my,i_hy,REC_COMP)  , &
                               i_z_block) * w00_rec +         &
                               dslow_block(ix_slow(i_mx,i_hx,REC_COMP)+1, &
                               iy_slow(i_my,i_hy,REC_COMP)  , &
                               i_z_block) * w10_rec +         &
                               dslow_block(ix_slow(i_mx,i_hx,REC_COMP)  , &
                               iy_slow(i_my,i_hy,REC_COMP)+1, &
                               i_z_block) * w01_rec +         &
                               dslow_block(ix_slow(i_mx,i_hx,REC_COMP)+1, &
                               iy_slow(i_my,i_hy,REC_COMP)+1, &
                               i_z_block) * w11_rec 
                               !}
                            end do
                            !}
                         else
                            !{
                            do i_z_block=1,down_par%n_z_block
                               !{
                               slow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block) = &
                               slow_block(ix_slow(i_mx,i_hx,REC_COMP)  , &
                               iy_slow(i_my,i_hy,REC_COMP)  , &
                               i_z_block,rec_slow)
                               !}
                            end do
                            !}
                         end if
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do

             !}
          end if
          !}
       else
          !{
          if(interp_x_slow) then
             !{
             do i_hy=1,data_par%n_hy_pad 
                !{
                do i_hx=1,data_par%n_hx_pad 
                   !{
                   do i_my=1,data_par%n_my_pad 
                      !{
                      do i_mx=1,data_par%n_mx_pad 
                         !{
                         w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))
                         w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))

                         if(type_slow(1:1) == 'c') then
                            !{
                            do i_z_block=1,down_par%n_z_block
                               !{
                               dslow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block,1) = &
                               dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)  , &
                               i_my,                          &
                               i_z_block) * w00_sou +         &
                               dslow_block(ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                               i_my,                          &
                               i_z_block) * w10_sou 
                               !}
                            end do
                            !}
                         else
                            !{
                            do i_z_block=1,down_par%n_z_block
                               !{
                               slow_data_block(i_mx,i_my,i_hx,i_hy,SOU_COMP,i_z_block) = &
                               slow_block(ix_slow(i_mx,i_hx,SOU_COMP)  , &
                               i_my,                         &
                               i_z_block,sou_slow)

                               !}
                            end do
                            !}
                         end if
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do

             if(slow_par%n_comp_slow > 1) then
                !{
                do i_hy=1,data_par%n_hy_pad 
                   !{
                   do i_hx=1,data_par%n_hx_pad 
                      !{
                      do i_my=1,data_par%n_my_pad 
                         !{
                         do i_mx=1,data_par%n_mx_pad 
                            !{
                            w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))
                            w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))
                            if(type_slow(1:1) == 'c') then
                               !{
                               do i_z_block=1,down_par%n_z_block
                                  !{
                                  dslow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block,1) = &
                                  dslow_block(ix_slow(i_mx,i_hx,REC_COMP)  , &
                                  i_my,                          &
                                  i_z_block) * w00_rec +         &
                                  dslow_block(ix_slow(i_mx,i_hx,REC_COMP)+1, &
                                  i_my,                          &
                                  i_z_block) * w10_rec 
                                  !}
                               end do
                               !}
                            else
                               !{
                               do i_z_block=1,down_par%n_z_block
                                  !{
                                  slow_data_block(i_mx,i_my,i_hx,i_hy,REC_COMP,i_z_block) = &
                                  slow_block(ix_slow(i_mx,i_hx,REC_COMP)  , &
                                  i_my,                          &
                                  i_z_block,rec_slow)
                                  !}
                               end do
                               !}
                            end if
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do

                !}
             end if
             !}
          else
             !{
             if(interp_y_slow) then
                call seperr ('Interpolation in y not implemented yet')
             else
                !{
                if(type_slow(1:1) == 'c') then
                   !{
                   if(slow_par%n_comp_slow > 1) then
                      dslow_data_block(:,:,1,1,SOU_COMP,:,1)=dslow_block(:,:,:)
                      dslow_data_block(:,:,1,1,REC_COMP,:,1)=dslow_block(:,:,:)
                   else
                      dslow_data_block(:,:,1,1,1,:,1)=dslow_block(:,:,:)
                   end if
                   !}
                else
                   !{
                   if(slow_par%n_comp_slow > 1) then
                      slow_data_block(:,:,1,1,SOU_COMP,:)= &
                        slow_block(:,:,:,sou_slow)
                      slow_data_block(:,:,1,1,REC_COMP,:)= &
                        slow_block(:,:,:,rec_slow)
                   else
                      slow_data_block(:,:,1,1,1,:)=slow_block(:,:,:,sou_slow)
                   end if
                   !}
                end if
                !}
             end if
             !}
          end if
          !}
       end if

       if((.not. slow_par%slowness).and. &
          ((tag_slow(1:len_tag) == slow_par%tag_slow(1:len_tag)) .or. &
           (tag_slow(1:len_tag) == slow_par%tag_slow_shear(1:len_tag)))) then
          slow_data_block=1./slow_data_block
          write(0,*)'Convert velocity into slowness'
       end if
       if(slow_par%slow_scale /= 1.) then
          write(0,*)'Scale slowness by slow_scale=',slow_par%slow_scale
          slow_data_block=slow_par%slow_scale*slow_data_block
       end if
       !}
    end if
    return
  end subroutine interpolate_slow


  subroutine out_dslow_block_par(down_par, dslow_data_block)
    type (down_process_param_type) :: down_par
    complex, dimension (:,:,:,:,:,:,:) :: dslow_data_block

    integer i_w_block_par

    do i_w_block_par=2,down_par%n_w_block_par
       dslow_data_block(:,:,:,:,:,:,1) =  &
       dslow_data_block(:,:,:,:,:,:,1) + &
       dslow_data_block(:,:,:,:,:,:,i_w_block_par)
    end do

  end subroutine out_dslow_block_par

  subroutine inp_dslow_block_par(down_par, dslow_data_block)
    type (down_process_param_type) :: down_par
    complex, dimension (:,:,:,:,:,:,:) :: dslow_data_block
    integer i_w_block_par

    do i_w_block_par=2,down_par%n_w_block_par
       dslow_data_block(:,:,:,:,:,:,i_w_block_par) = dslow_data_block(:,:,:,:,:,:,1) 
    end do
  end subroutine inp_dslow_block_par

  subroutine clean_io_slow()
    implicit none
    integer ierr

    if(allocated(slow_block)) then
       deallocate (slow_block,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating slow_block')
    end if
    if(allocated(dslow_block)) then
       deallocate (dslow_block,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating dslow_block')
    end if
    if(allocated(rslow_slice)) then
       deallocate (rslow_slice,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating rslow_slicw')
    end if
    if(allocated(cslow_slice)) then
       deallocate (cslow_slice,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating cslow_slicw')
    end if
    if(allocated(ix_slow)) then
       deallocate (ix_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating ix_slow')
    end if
    if(allocated(iy_slow)) then
       deallocate (iy_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating iy_slow')
    end if
    if(allocated(wx_slow)) then
       deallocate (wx_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating wx_slow')
    end if
    if(allocated(wy_slow)) then
       deallocate (wy_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating wy_slow')
    end if
    return
  end subroutine clean_io_slow
end module slow_mod
