module timers_mod

implicit none

integer, parameter ::  MAX_N_TIMERS = 128
integer, parameter ::  MAX_N_THREADS = 32

logical, private :: timers_on=.false.
integer, private :: max_n_timers_init=0
integer, private :: count_rate,count_max,n_threads_int

logical, private :: init(MAX_N_TIMERS)
character(len=128), private :: desc(MAX_N_TIMERS)
integer, private :: type(MAX_N_TIMERS)
integer, private :: t0_time(3,MAX_N_TIMERS,0:MAX_N_THREADS-1)
integer, private :: t1_time(3,MAX_N_TIMERS,0:MAX_N_THREADS-1)
integer, private :: tot_time(3,MAX_N_TIMERS,0:MAX_N_THREADS-1)
integer, private :: flop(MAX_N_TIMERS)

integer, external :: omp_get_thread_num

contains

subroutine SEP_switch_timers(switch_on,n_threads)
logical switch_on
integer, optional :: n_threads
integer count

if(switch_on) then
   timers_on=.true.
   call system_clock(count,count_rate,count_max)
write(0,*)'count=',count
write(0,*)'count_rate=',count_rate
write(0,*)'count_max=',count_max
if(present(n_threads)) then
  n_threads_int=n_threads
  write(0,*)'n_threads=',n_threads
end if
   if(count_rate == 0) then
     write(0,*)'WARNING: No clock suported'
     write(0,*)'WARNING: Switching off the timers'
   end if
   init(:)=.false.
else
   timers_on=.false.
end if

return 
end subroutine SEP_switch_timers

subroutine SEP_print_timers()
integer i_timer

if (timers_on) then
  do i_timer=1,max_n_timers_init
     if(init(i_timer)) then
       call SEP_timer_print(i_timer)
     end if
  end do
else
  write(0,*) 'Timers were OFF'
end if

return
end subroutine SEP_print_timers


subroutine SEP_timer_init(i_timer,i_desc,i_flop,i_type)
integer i_timer
real i_flop
character(len=128) :: i_desc
integer, optional :: i_type
integer i_thread

if (timers_on) then
  if(i_timer > MAX_N_TIMERS) call seperr('Try to initiliaze too many timers')
  if(i_timer > max_n_timers_init) max_n_timers_init=i_timer
  if(init(i_timer)) then
    write(0,*)'Trying to initilize Timer #',i_timer,' already initialized'
    write(0,*)'Timer #',i_timer,' has NOT been reinitilized'
  else

    init(i_timer)=.true.
    do i_thread=0,n_threads_int-1
      tot_time(1:3,i_timer,i_thread)=0
      t0_time(1:3,i_timer,i_thread)=0
      t1_time(1:3,i_timer,i_thread)=0
    end do
      
    desc(i_timer)=trim(i_desc)
    if(present(i_type)) then
      type(i_timer)=i_type
    else
      type(i_timer)=0
    end if
    flop(i_timer)=i_flop

  end if
end if

return
end subroutine SEP_timer_init

subroutine SEP_timer_start(i_timer)
integer i_timer,i_thread
integer count

i_thread=omp_get_thread_num()
if (timers_on) then
  call system_clock(count=count)
  t0_time(3,i_timer,i_thread)=count
end if

return
end subroutine SEP_timer_start


subroutine SEP_timer_stop(i_timer)
integer i_timer,i_thread
integer count,diff

if (timers_on) then
  call system_clock(count=count)
  i_thread=omp_get_thread_num()
  t1_time(3,i_timer,i_thread)=count
  diff=t1_time(3,i_timer,i_thread)-t0_time(3,i_timer,i_thread)
  if(diff < 0) then
    diff=(count_max-t0_time(3,i_timer,i_thread))+t1_time(3,i_timer,i_thread)
write(0,*)'Timer went over maximum'
write(0,*)'count_max',count_max
write(0,*)'t0_time',t0_time(3,i_timer,i_thread)
write(0,*)'t1_time',t1_time(3,i_timer,i_thread)
write(0,*)'tot_time',tot_time(3,i_timer,i_thread)
write(0,*)'diff',diff
  end if
  tot_time(3,i_timer,i_thread)=tot_time(3,i_timer,i_thread)+diff
end if

return
end subroutine SEP_timer_stop

subroutine SEP_timer_print(i_timer)

integer i_timer,i_thread
real bwidth,norm_flop

if(timers_on) then
  norm_flop=1./1000000.
  if(init(i_timer)) then
    write(0,*)'***** BEG TIMER #',i_timer,trim(desc(i_timer)),'**************'
    do i_thread=0,n_threads_int-1
      if(type(i_timer) == 0) then
      
        if(tot_time(3,i_timer,i_thread) > 0 ) then
          write(0,*)'Thread #',i_thread,' Time = ',float(tot_time(3,i_timer,i_thread))/count_rate
          write(0,*)' Total Mflop ',norm_flop*flop(i_timer)
          write(0,*)' seconds (Elapsed time) ',float(tot_time(3,i_timer,i_thread))/count_rate
          write(0,*)' seconds (User time) '
        
          bwidth=flop(i_timer)*count_rate/(1000000.*float(tot_time(3,i_timer,i_thread)))
          write(0,*)' Measured speed: Mflop/s ',bwidth
          write(0,*)''
        else
          write(0,*)' Thread #',i_thread,'Time too short for computing speed'
        end if
      else
        if(tot_time(3,i_timer,i_thread) > 0 ) then
          write(0,*)'Thread #',i_thread,' Time = ',float(tot_time(3,i_timer,i_thread))/count_rate
          write(0,*)' Total I/O Mbytes',float(flop(i_timer)/count_rate)
          write(0,*)' seconds (Elapsed time) ',float(tot_time(3,i_timer,i_thread)/count_rate)
          write(0,*)' seconds (User time) '
          bwidth=flop(i_timer)*count_rate/(1000000.*float(tot_time(3,i_timer,i_thread)))
          write(0,*)' Measured bandwidth: Mb/s ',bwidth
          write(0,*)''
        else
          write(0,*)'Thread #',i_thread,'Time too short for computing speed'
        end if
      end if
    end do
    write(0,*)'******************** END TIMER #',i_timer,'*********************'
    write(0,*)''
  end if
else
  write(0,*) 'Timers were OFF'
end if
  
return
end subroutine SEP_timer_print

end module timers_mod
