program Velspec
  use sep
  implicit none

!-------------------------------------

  integer                                 :: nang, n1, n2, n3, nt, nx, n1a, n2a, n3a, n4a, n1v, n2v, n3v
  integer                                 :: nvh, nvn, ovh, ovn, dvh, dvn
  integer                                 :: i, i1, i2, i3, i4
  real					  :: o1, o2, o3, o1a, o2a, o3a, o4a, o1v, o2v, o3v
  real					  :: d1, d2, d3, d1a, d2a, d3a, d4a, d1v, d2v, d3v
  real                                    :: CMP 
  integer                                 :: CMPv, CMPa, iCMP
  real, dimension(:,:,:,:), allocatable   :: adcig
  real, dimension(:,:,:), allocatable     :: velocity
  real, dimension(:,:,:), allocatable     :: velocities
  real, dimension(:,:,:), allocatable     :: vspectra
  real, dimension(:,:,:,:), allocatable   :: vspectras
  real, dimension(:), allocatable         :: gammap, gammag
  real, dimension(:,:), allocatable       :: DZ
  
 !-------------------------------------
 
  call sep_init()               ! Initialize I/O module, "sep".

  call from_par( "n1", n1)
  call from_par( "n2", n2)
  call from_par( "n3", n3)
  call from_par( "o1", o1)
  call from_par( "o2", o2)
  call from_par( "o3", o3)
  call from_par( "d1", d1)
  call from_par( "d2", d2)
  call from_par( "d3", d3)
 
  call from_par( "nt", nt)
  call from_par( "nx", nx)
  call from_par( "CMP", CMP)

  
  call from_aux ( "adcig", "n1", n1a)
  call from_aux ( "adcig", "n2", n2a)
  call from_aux ( "adcig", "n3", n3a)
  call from_aux ( "adcig", "n4", n4a)
  call from_aux ( "adcig", "o2", o2a)
  call from_aux ( "adcig", "d2", d2a)
  call from_aux ( "adcig", "o4", o4a)
  call from_aux ( "adcig", "d4", d4a)

  call from_aux( "velocity", "n1", n1v)
  call from_aux( "velocity", "o1", o1v)
  call from_aux( "velocity", "d1", d1v)
  call from_aux( "velocity", "n2", n2v)
  call from_aux( "velocity", "n3", n3v)
  
  call sep_close()

  write(0,*) "n1: ", n1
  write(0,*) "n2: ", n2
  write(0,*) "n3: ", n3

  write(0,*) "nt: ", nt
  write(0,*) "nx: ", nx
  write(0,*) "CMP: ", CMP

  write(0,*) "n1a: ", n1a
  write(0,*) "n2a: ", n2a
  write(0,*) "n3a: ", n3a
  write(0,*) "n4a: ", n4a

  write(0,*) "o1a: ", o1a
  write(0,*) "o2a: ", o2a
  write(0,*) "o4a: ", o4a

  write(0,*) "n1v: ", n1v
  write(0,*) "n2v: ", n2v
  write(0,*) "n3v: ", n3v
  
  ! Read input data from disk
  allocate( adcig(n1a,n2a,n3a,n4a) )
  allocate( velocity(n1v,n2v,n3v) )

  call sep_read(adcig, 'adcig') 
  call sep_read(velocity,'velocity')
 
  ! Allocation of the different vectors and definition of the phase angle vector
  allocate(gammap(n2a))
  do i=1,n2a
     gammap(i)=o2a+d2a*(i-1)
  end do

  allocate(gammag(n2a),DZ(n2a,3),vspectra(n1,n2,n3),vspectras(n1,n2,n3,2*nx+1))
  vspectra=0
  vspectras=0
  
  ! Definition of the depth
  
  CMPa=(CMP-o4a)/d4a+1.5
  write(0,*) "CMPa :", CMPa
  CMPv=(CMP-o1v)/d1v+1.5
  write(0,*) "CMPv :", CMPv  
  
  do iCMP=-nx,nx
     
     do i1=1,n1-1 ! Computation of the velocity spectras for one CMP - Loop over depth
        
        write(0,*) "CMP", CMPv, "+", iCMP, "-", "Depth: ", i1 
        ! Computation of the group angles from the phase angles (gammag initialized to zero in the routine)
        call Phase2group(gammap, gammag, n2a, velocity(CMPv+iCMP,1,i1), velocity(CMPv+iCMP,1,i1), velocity(CMPv+iCMP,1,i1))
        
        ! Computation of the Residual Moveout (DZ initialized to zero in the routine)
        call Group2derivatives(DZ,gammap, gammag, n2a, velocity(CMPv+iCMP,1,i1), velocity(CMPv+iCMP,1,i1), velocity(CMPv+iCMP,1,i1))
        
        ! Allocation of the velocity spectra cube
        nang=(asind(velocity(CMPv+iCMP,1,1)/velocity(CMPv+iCMP,1,i1))-o2a)/d2a+1.5
        call Derivatives2spectra(vspectra, DZ, adcig, i1, CMPa+iCMP, nang, n1a, n2a, n3a, n4a, n1, o1, d1, n2, o2, d2, n3, o3, d3)
        
     end do
     
     ! Smoothing along the depth axis
     call Timesmoothing(vspectra, nt, n1, n2, n3)
     
     vspectras(:,:,:,iCMP+nx+1)=vspectra
     vspectra=0
 
  end do
  
  ! Smoothing along the CMP axis
  call CMPsmoothing(vspectras, vspectra, nx, n1, n2, n3)

  call sep_write(vspectra) 
  
!  call Picking(velocities, vspectra, n1, n2, o2, d2, n3, o3, d3)
   
!  write(0,*) "Picking complete"

!  call to_history( "n2", 3)
!  call to_history( "o2", 0)
!  call to_history( "d2", 1)

!  call sep_write(velocities) 
  


!  deallocate( velocity,adcig)
  
  call exit(0)
  
end program Velspec



!-------------------------------------

subroutine Phase2group(gammap, gammag, n, vvm, vhm, vnm) 
  integer                               :: n, i
  real                                  :: o, d, vvm, vhm, vnm
  real, dimension (n)                   :: gammap
  real, dimension (n)                   :: betarp, betasp, betarg, betasg
  real, dimension (n), intent(OUT)      :: gammag
  real, dimension (n)                   :: Vellbetarp, Vellbetasp, Vvtibetarp, Vvtibetasp
  real, dimension (n)                   :: DVellbetarp,DVellbetasp, DVvtibetarp, DVvtibetasp

  gammag=0

  ! Definition of the phase_angles
  betarp=gammap ! Only true for flat reflectors
  betasp=-gammap ! Only true for flat reflectors
  
  ! Elliptical phase velocity at the different propagation angles 
  Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
  Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)
  
  ! VTI phase velocity at the different propagation angles
  Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
  Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)
  
  ! Derivative of the elliptic phase velocity with respect to the phase angle
  DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
  DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)
  
  ! Derivative of the VTI phase velocity with respect to the phase angle
  DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
  DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))
  
  ! Computation of the group propagation angles
  betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
  betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))
  
  ! Computation of the group aperture angle
  gammag=0.5*(betarg-betasg)
  
end subroutine Phase2group

!-------------------------------------

subroutine Group2derivatives(DZ, gammap, gammag, n, vvm, vhm, vnm) 

  integer                               :: n
  real                                  :: vvm, vhm, vnm
  real, dimension (n)                   :: gammap, gammag, Sellg, Svtig
  real, dimension (n)                   :: DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3
  real, dimension (n,3),intent(OUT)     :: DZ
  
  DZ=0
  
  ! Elliptic group slowness (the input angle is the group angle)
  
  Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)
  
  ! VTI group slowness (the input angle is the group angle)
  Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)
  
  ! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
  
  DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
  DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg
  
  ! Computation of the derivatives of the VTI group slownesses
  DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
  
  ! Computation of the derivative of the depth with respect to the different perturbations
  DZ(:,1)=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
  DZ(:,2)=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
  DZ(:,3)=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))

end subroutine Group2derivatives

!-------------------------------------

subroutine Derivatives2spectra(vspectra, DZ, adcig, z0m, CMPa, nang, n1a, n2a, n3a, n4a, n1, o1, d1, n2, o2, d2, n3, o3, d3)

  integer                               :: nang, n1, n2, n3, i2, i3, i, i1
  real                                  :: o1, o2, o3, d1, d2, d3
  real                                  :: drhovh, drhovn
  integer                               :: z0m, CMPa
  real                                  :: z0mr
  real, dimension (n2a,3)               :: DZ
  real, dimension (n2a)                 :: rmo
  real, dimension (n1,n2,n3),intent(OUT):: vspectra
  real, dimension (:,:), allocatable    :: intvspectra, intdenum
  real, dimension (n1a,n2a,n3a,n4a)     :: adcig
  
  allocate(intvspectra(n2,n3), intdenum(n2,n3))
  intvspectra=0
  intdenum=0
  
!  write(0,*) vspectra
  z0mr=(o1+z0m-1)*d1
  
!  write(0,*) 'z0mr', z0mr
!  write(0,*) 'z0m', z0m
  
  do i2=1,n2 ! Loop over the different values of HORIZONTAL perturbations
     drhovh=(o2+(i2-1)*d2)
!     write(0,*) "i2 :", i2

     do i3=1,n3 ! Loop over the different values of the NMO velocity perturbations
        drhovn=(o3+(i3-1)*d3)
!        write(0,*) "i3 :", i3

        ! Computation of the residual moveout for the different perturbation values
        rmo=z0mr*(DZ(:,2)*drhovh+DZ(:,3)*drhovn) ! That is a vector 
        rmo=(rmo-o1)/d1+1.5      
        
        ! Computation of the semblance values (the fourth parameter in adcig is CMP)
        do i=1,nang
!           if (z0m+rmo(i)>0 .AND. z0m+rmo(i)<n1a+1) then
              intvspectra(i2,i3)=intvspectra(i2,i3)+adcig(z0m+rmo(i),i,1,CMPa)
              intdenum(i2,i3)=intdenum(i2,i3)+adcig(z0m+rmo(i),i,1,CMPa)**2
!           end if
        end do
     end do
  end do
  
  do i2=1,n2
     do i3=1,n3
        vspectra(z0m,i2,i3)=intvspectra(i2,i3)**2/(nang*intdenum(i2,i3))
     end do
  end do

end subroutine Derivatives2spectra

!-------------------------------------

subroutine Timesmoothing(vspectra, nt, n1, n2, n3)

  integer                               :: nt, n1, n2, n3
  real, dimension (n1,n2,n3),intent(OUT):: vspectra
  real, dimension (n1,n2,n3)            :: intvspectra

  intvspectra=vspectra

  do i1=nt+1,n1-nt
     do i2=1,n2
        do i3=1,n3
           do it=-nt,nt 
              vspectra(i1,i2,i3)=vspectra(i1,i2,i3)+intvspectra(i1+it,i2,i3)
           end do
           vspectra(i1,i2,i3)=vspectra(i1,i2,i3)/(2*nt+1)
        end do
     end do
  end do
  
end subroutine Timesmoothing

!-------------------------------------

subroutine CMPsmoothing(vspectras, vspectra, nx, n1, n2, n3)

  integer                               :: nx, n1, n2, n3
  real, dimension (n1,n2,n3),intent(OUT):: vspectra
  real, dimension (n1,n2,n3,nx)         :: vspectras
  
  do i1=1,n1
     do i2=1,n2
        do i3=1,n3
           do ix=1,2*nx+1
              vspectra(i1,i2,i3)=vspectra(i1,i2,i3)+vspectras(i1,i2,i3,ix)
           end do
           vspectra(i1,i2,i3)=vspectra(i1,i2,i3)/(2*nx+1)
        end do
     end do
  end do

end subroutine CMPsmoothing
