!------------------------------------
program velspec_cube


use sep
implicit none
!-------------------------------------
integer					:: n1, n2, n3, nCMP, ng, ng2, nt, nx, i1, i2, i3, iCMP,ig, ig2, it, ix, kCMP
real 					:: delta, epsilon, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp, CMP
real                                    :: o1, o2, o3, oCMP, og, og2, d1, d2, d3, dCMP, dg, dg2
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real                                    :: max,vnmax,vhmax
real, dimension(:,:,:), allocatable     :: adcig
real, dimension(:,:,:), allocatable   :: intvspectra, intdenum ! Is the bunch of nx vspectra will we stack together
real, dimension(:,:,:), allocatable     :: vspectra ! Are the averaged velspec along the time axis - then CMP axis    
real, dimension(:,:,:), allocatable     :: intpicks, picks
real, dimension(:,:),allocatable        :: velocity
real, dimension(:),allocatable          :: gammap, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm

!--------------------------------

call sep_init()

call from_param( "nt", nt) ! Time Windowing
call from_param( "nx", nx) ! CMP Windowing

call from_param( "n1", n1) ! Depth axis
call from_param( "o1", o1)
call from_param( "d1", d1)

call from_param( "nCMP", nCMP) ! CMP axis
call from_param( "oCMP", oCMP)
call from_param( "dCMP", dCMP)

call from_param( "kCMP", kCMP)

!call from_param( "nv", nv) ! Different velocities axis

call from_param( "ng", ng) ! Aperture angle axis
call from_param( "og", og)
call from_param( "dg", dg)

call from_param( "ng2", ng2) ! Aperture angle used for the summation
call from_param( "og2", og2)
call from_param( "dg2", dg2)

call from_param( "n2", n2) ! Perturbation in horizontal velocity (2nd axis)
call from_param( "o2", o2)
call from_param( "d2", d2)

call from_param( "n3", n3) ! Perturbation in NMO velocity (3rd axis)
call from_param( "o3", o3)
call from_param( "d3", d3)

call sep_close()

! Allocation of the ADCIG
allocate(adcig(n1,ng,nCMP))
call sep_read(adcig,'adcig')

! Allocation of the velocity spectra cube
!allocate(intvspectra(n1,n2,n3,2*nx+1))
!allocate(intdenum(n1,n2,n3,2*nx+1))
allocate(intvspectra(n1,n2,n3))
allocate(intdenum(n1,n2,n3))

intvspectra=0
intdenum=0

!allocate(intpicks(n1,nCMP,nv))
allocate(vspectra(n1,n2,n3))
!allocate(denum(n1,n2,n3))
!allocate(picks(n1,nCMP,nv))
vspectra=0
!denum=0
!picks=0

! Allocation of the velocity vector
allocate(velocity(nCMP,n1))
call sep_read(velocity,'velocity')

allocate(gammap(ng2),gammag(ng2))
allocate(betarp(ng2),betarg(ng2))
allocate(betasp(ng2),betasg(ng2))
allocate(Vellbetarp(ng2),Vellbetasp(ng2))
allocate(Vvtibetarp(ng2),Vvtibetasp(ng2))
allocate(DVellbetarp(ng2),DVellbetasp(ng2))
allocate(DVvtibetarp(ng2), DVvtibetasp(ng2))
allocate(Sellg(ng2),Svtig(ng2))
allocate(DSellg1(ng2), DSellg2(ng2))
allocate(DSvtig1(ng2),DSvtig2(ng2),DSvtig3(ng2))
allocate(DZ1(ng2),DZ2(ng2),DZ3(ng2))
allocate(rmo(ng2))


!write(0,*) 'Hello';

do iCMP=1,nx
   
   CMP=kCMP+iCMP-1 ! CMP value 
   
   write(0,*) 'Value of CMP :', CMP ;
   
   intvspectra=0;
   intdenum=0;
   

   ! We change the indices of the velocity spectras in the temporary vector of spectras
   !if(iCMP>nx) then
   !   do ix=1,nx-1
   !      intvspectra(:,:,:,ix)=intvspectra(:,:,:,ix+1)
   !   end do
   !   intvspectra(:,:,:,nx)=0
   !end if
   
   do i1=1,n1
      
      ! Modified velocities - i.e. the velocity we use for the migration of the dataset
      vvm=velocity(CMP,i1);
      vhm=vvm;
      vnm=vvm;

      ! write(0,*) 'Velocity OK'
      
         write(0,*) vvm 
      !   write(0,*) vhm
      !   write(0,*) vnm
      
      ! Definition of the vector of aperture angles
      
      do ig2=1,ng2
         gammap(ig2)=og2+dg2*(ig2-1)
         betarp(ig2)=gammap(ig2) ! Only true for flat reflectors
         betasp(ig2)=-gammap(ig2) ! Only true for flat reflectors
      end do
      
      ! Computation of the different quantities used to compute the derivatives of the depth with respect to pertubations.
      
      ! Elliptical phase velocity at the different propagation angles 
      
      Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
      Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)
      
      ! VTI phase velocity at the different propagation angles
      
      Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
      Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)
      
      ! Derivative of the elliptic phase velocity with respect to the phase angle
      
      DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
      DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)
      
      ! Derivative of the VTI phase velocity with respect to the phase angle
      DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
      DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))
      
      ! Computation of the group propagation angles
      betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
      betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))
      
      ! Computation of the group aperture angle
      gammag=0.5*(betarg-betasg)
      
      !   do ig2=1,ng2
      !   write(0,*) gammag(ig2) 
      !   end do
      
      ! COMPUTATION OF THE RESIDUAL MOVEOUT
      
      ! Elliptic group slowness (the input angle is the group angle)
      
      Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)
      
      ! VTI group slowness (the input angle is the group angle)
      Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)
      
      ! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
      
      DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
      DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg
      
      ! Computation of the derivatives of the VTI group slownesses
      DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
      DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
      DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
      
      ! Computation of the derivative of the depth with respect to the different perturbations
      DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
      DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
      DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))
      
      !      do j=1,ng2
      !         write(0,*) 'DZ1', DZ1(j)
      !      end do
      !      write(0,*) '************'
      !      do j=1,ng2
      !         write(0,*) 'DZ2', DZ2(j)
      !      end do
      !      write(0,*) '************'
      !      do j=1,ng2
      !         write(0,*) 'DZ3', DZ3(j)
      !      end do
      
      ! Actual computation of the velocity spectra
      
      
      ! Definition of the depth: 
      !  write(0,*) i1," out of ",n1
      z0m=o1+(i1-1)*d1
      

      !   write(0,*) 'Computation of the derivatives is OK'

      !   write(0,*) '**************** Horizontal velocity loop *******************' 
      
      do i2=1,n2 ! Loop over the different values of HORIZONTAL perturbations
         drhovh=(o2+(i2-1)*d2)
         
         
         !      write(0,*) '**************** NMOvelocity loop *******************' 
         
         
         do i3=1,n3 ! Loop over the different values of the NMO velocity perturbations
            drhovn=(o3+(i3-1)*d3)
            
            ! Computation of the residual moveout for the different perturbation values
            
            rmo=z0m*(DZ2*drhovh+DZ3*drhovn) ! That is a vector 
            
            !write(0,*) '************'
            ! write(0,*) 'drhovh', drhovh
            ! write(0,*) '************'
            ! write(0,*) 'drhovn', drhovn
            ! write(0,*) '************'
            ! write(0,*) 'z0m', z0m
            ! write(0,*) '************'
            
            ! do j=1,ng2
            !    write(0,*) 'DZ2', DZ2(j)
            ! end do
            ! write(0,*) '************'
            ! do j=1,ng2
            !    write(0,*) 'DZ3', DZ3(j)
            ! end do
            ! write(0,*) '************'
            ! write(0,*) 'rmo(10)', rmo(10)
            !  write(0,*) '************'
            
            rmo=(rmo-o1)/d1+1.5

            !   write(0,*) 'Computation of residual moveout is OK'

            ! Computation of the semblance values - we save it in the last column 

            !   write(0,*) 'Int vel spec',intvspectra(i1,i2,i3,iCMP) 

            do ig2=1,ng2
               if (i1+rmo(ig2)>0 .AND. i1+rmo(ig2)<n1+1) then 
       !           write(0,*) 'ADCIG: ', adcig(i1+rmo(ig2),ig2,CMP)
                 intvspectra(i1,i2,i3)=intvspectra(i1,i2,i3)+adcig(i1+rmo(ig2),ig2,CMP)   
                  intdenum(i1,i2,i3)=intdenum(i1,i2,i3)+adcig(i1+rmo(ig2),ig2,CMP)**2
               end if
            end do
            intvspectra(i1,i2,i3)=intvspectra(i1,i2,i3)**2/(ng2*intdenum(i1,i2,i3))
            !   write(0,*) 'ig loop'
         end do
         !   write(0,*) 'VH loop'
      end do
      !   write(0,*) 'VN loop'
   end do
   
   !   write(0,*) 'Computation of the interval velocity spectras is OK'

!   do i1=1,n1 
!      do i2=1,n2
!         do i3=1,n3
!            intvspectra(i1,i2,i3,iCMP+nx+1)=intvspectra(i1,i2,i3,iCMP+nx+1)**2/(ng2*intdenum(i1,i2,i3,iCMP+nx+1))
!         end do
!      end do
!   end do
end do

!vspectra=intvspectra(:,:,:,2)

! Averaging along the depth and the CMP axes

!do i1=1,n1
!   do i2=1,n2
!      do i3=1,n3
!         do it=-nt,nt 
!            do ix=-nx,nx
!               vspectra(i1,i2,i3)=vspectra(i1,i2,i3)+intvspectra(i1+it,i2,i3,(nx+1)/2+ix)
!            end do
!            vspectra(i1,i2,i3)=vspectra(i1,i2,i3)/((2*nt+1)*(2*nx+1))
!        end do
!      end do
!   end do
!end do

! Averaging along the CMP axis

!iCMP=k2
!do i2=1,n2
!   do i3=1,n3
!      do ix=k2-nx,k2+nx 
!         vspectra(i1,i2,i3,iCMP)=vspectra(i1,i2,i3,iCMP)+vspectra(i1,i2,i3,iCMP+ix)+vspectra(i1,i2,i3,iCMP-ix)
!      end do
!      vspectra(i1,i2,i3,iCMP)=vspectra(i1,i2,i3,iCMP)/(2*nt+1)
!   end do
!end do


   
!         write(0,*) 'velspec value :', vspectra(i1,i2,i3)
!            if (vspectra(i1,i2,i3)>max) then
!               max=vspectra(i1,i2,i3)
!               write(0,*) 'max', max
!               vhmax=o2+(i2-1)*d2
!               vnmax=o3+(i3-1)*d3
!               vhmax=velocity(i1,iCMP)*(1+(o2+(i2-1)*d2))
!               vnmax=velocity(i1,iCMP)*(1+(o3+(i3-1)*d3))
!            end if
!         end do
!      end do
!      picks(i1,iCMP,1)=vhmax
!      picks(i1,iCMP,2)=vnmax
!            write(0,*) 'Vhmax : ', picks(i1,iCMP,1)  
!            write(0,*) 'Vnmax : ', picks(i1,iCMP,2)
!   end do
   
   !do i1=3,n1-2
   !   do iCMP=1,nCMP
   !         do i3=1,n3
   !            vspectra(i1,iCMP,i3)=1/7*(vspectra(i1,iCMP,i3)+vspectra(i1-1,iCMP,i3)+vspectra(i1-2,iCMP,i3)+vspectra(i1-3,iCMP,i3)+vspectra(i1-4,iCMP,i3)+vspectra(i1-5,iCMP,i3)+vspectra(i1-6,iCMP,i3))
!            if (vspectra(i1,iCMP,i3)>max) then
   !               max=vspectra(i1,iCMP,i3)
   !               write(0,*) 'max', max
   !               vhmax=velocity(i1,iCMP)*(oCMP+(iCMP-1)*dCMP)
   !               vnmax=velocity(i1,iCMP)*(o3+(i3-1)*d3)
   !            end if
   !         end do
   !   end do
!   picks(i1,iCMP,1)=vhmax
   !   picks(i1,iCMP,2)=vnmax
   !end do
   
deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

deallocate(betarp,betarg,betasp,betasg,Vellbetarp,Vellbetasp,Vvtibetarp,Vvtibetasp,DVellbetarp,DVellbetasp,DVvtibetarp,DVvtibetasp)

!deallocate(adcig,intvspectra, intdenum)


call sep_write(intvspectra(:,:,:))

end program velspec_cube
