module Slice_mod

  use slice_types

  implicit none

  integer, private :: n_bytes_coord,n_bytes_data,n_bytes_slice
  integer, external ::  sreed,srite,sseek_block

  real, allocatable, dimension (:,:), private :: slice
  real, allocatable, dimension (:,:,:), private :: data_cube
  complex, allocatable, dimension (:), private :: coord

contains

function Slice_init(slice_par)  result(status)
  type (slice_param_type) slice_par
  integer status

  integer ierr

  n_bytes_data=slice_par%n1*slice_par%n2*slice_par%n3*slice_par%esize
  n_bytes_slice=slice_par%n2*slice_par%n3*slice_par%esize
  n_bytes_coord=slice_par%n1_coord*slice_par%esize_coord

  allocate (slice(slice_par%n2,slice_par%n3),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of slice")

  allocate (data_cube(slice_par%n1,slice_par%n2,slice_par%n3),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of data_cube")

  allocate (coord(slice_par%n1_coord),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of coord")

  status=0

end function Slice_init

function Slice_oper(slice_par)  result(status)
  type (slice_param_type) slice_par
  integer status

  logical c2_found
  integer i1,i2,i3,i4,i5,i1_coord,i_c2,i_v1
  real v1,v2,c1,c2,c2_test,w,dist2,min_c2,max_c2

 
  do i5=1,slice_par%n5
  !{ i5
    do i4=1,slice_par%n4
    !{ i4

      if(sreed('in',data_cube,n_bytes_data) /= n_bytes_data) then
        call seperr('Problems with reading input data')
      end if
  
      if(sreed('Slice_coord',coord,n_bytes_coord) /= n_bytes_coord) then
        call seperr('Problems with reading input Slice_coord')
      end if

      min_c2=huge(min_c2)
      max_c2=-huge(max_c2)
      do i1_coord=1,slice_par%n1_coord
        c2_test=real(coord(i1_coord))
        if(c2_test < min_c2 ) min_c2=c2_test
        if(c2_test > max_c2 ) max_c2=c2_test
      end do
      if(slice_par%verb) then
        write(0,*)'$$$$$$'
        write(0,*)'min_c2=',min_c2
        write(0,*)'max_c2=',max_c2
        write(0,*)'$$$$$$'
      end if

      do i2=1,slice_par%n2
      !{ i2

        v2=slice_par%o2+(i2-1)*slice_par%d2
 
        if((v2 <min_c2) .or. (v2 > max_c2)) then
        !{ 
          slice(i2,:)=0.
        !} 
        else
        !{ 

          c2 = huge(c2)
          c2_found=.false.
          do i1_coord=1,slice_par%n1_coord
            dist2=abs(c2_test - v2)
            c2_test=real(coord(i1_coord))
            if(abs(c2_test - v2) < dist2) then
              c2_found=.true.
              c2=c2_test
              i_c2=i1_coord
            end if
          end do
          if(.not. c2_found) call seperr("Problem with founding correct trace")
          if(c2 < slice_par%o2) then
            write(0,*)'c2=',c2
            write(0,*)'min c2=',slice_par%o2
            call seperr("c2 out of range (too small)")
          end if
          if(c2 > slice_par%o2 + (slice_par%n2-1)*slice_par%d2) then
            write(0,*)'c2=',c2
            write(0,*)'max c2=',slice_par%o2 + (slice_par%n2-1)*slice_par%d2
            call seperr("c2 out of range (too large)")
          end if
          c1=aimag(coord(i_c2))
          if(c1 < slice_par%o1) then
            write(0,*)'c1=',c1
            write(0,*)'min c1=',slice_par%o1
            call seperr("c1 out of range (too small)")
          end if
          if(c1 > slice_par%o1 + (slice_par%n1-2)*slice_par%d1) then
            write(0,*)'c1=',c1
            write(0,*)'max c1=',slice_par%o1 + (slice_par%n1-2)*slice_par%d1
            call seperr("c1 out of range (too large)")
          end if
            
          v1=((c1-slice_par%o1)/slice_par%d1)+1
          i_v1=int(v1)
          w=(v1-i_v1)
  !         i_v2=nint((c2-slice_par%o2)/slice_par%d2)+1
  
          if(slice_par%verb) then
            write(0,*)'****'
            write(0,*)'v2=',v2
            write(0,*)'c2=',c2
            write(0,*)'abs(v2-c2)=',abs(v2-c2)
            write(0,*)'c1=',c1
            write(0,*)'v1=',v1
            write(0,*)'i_v1=',i_v1
            write(0,*)'w=',w
            write(0,*)'****'
          end if
  
          do i3=1,slice_par%n3
          !{ i3
  
            slice(i2,i3)=data_cube(i_v1  ,i2,i3)*(1.-w) + &
                         data_cube(i_v1+1,i2,i3)*(w) 
  
          !} i3
          end do
  
        !} 
        end if

      !} i2
      end do

      if(srite('out',slice,n_bytes_slice) /= n_bytes_slice) then
        call seperr('Problems with writing output')
      end if

        !} 
    !} i4
    end do
  !} i5
  end do

   
  status=0


end function Slice_oper

end module Slice_mod
