!------------------------------------
program velocity_spectra


use sep
implicit none
!-------------------------------------
integer					:: nt, n2, ng, n1, it, ig, i1, i2, i3, j, k
real 					:: delta, epsilon, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp
real                                    :: ot, og, o1, o2, o3, dt, dg, d1, d2, d3
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:,:), allocatable       :: adcig
real, dimension(:,:,:), allocatable     :: intvspectra, vspectra, denum
real, dimension(:),allocatable          :: gammap, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, zm
real, dimension(:,:), allocatable       :: z

!--------------------------------

call sep_init()

! Parameters of the initial ADCIG
call from_aux( "adcig", "n1", nt)
call from_aux( "adcig", "o1", ot)
call from_aux( "adcig", "d1", dt)

call from_aux( "adcig", "n2", n1)
call from_aux( "adcig", "o2", o1)
call from_aux( "adcig", "d2", d1)

call from_param( "it", it)

! Parameters of the semblance cube

! Parameters of the velocity model
call from_param( "V0", V0)
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call to_history("n1", n1)
call to_history("o1", o1)
call to_history("d1", d1)

n2=3

call to_history("n2", n2)
call to_history("o2", 0)
call to_history("d2", 1)

call sep_close()

write(0,*) "n1", n1
write(0,*) "o1", o1
write(0,*) "d1", d1


vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);

! Modified velocities
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

write(0,*) vvm
write(0,*) vhm
write(0,*) vnm

allocate(gammap(n1),gammag(n1))
allocate(betarp(n1),betarg(n1))
allocate(betasp(n1),betasg(n1))
allocate(Vellbetarp(n1),Vellbetasp(n1))
allocate(Vvtibetarp(n1),Vvtibetasp(n1))
allocate(DVellbetarp(n1),DVellbetasp(n1))
allocate(DVvtibetarp(n1), DVvtibetasp(n1))
allocate(Sellg(n1),Svtig(n1))
allocate(DSellg1(n1), DSellg2(n1))
allocate(DSvtig1(n1),DSvtig2(n1),DSvtig3(n1))
allocate(DZ1(n1),DZ2(n1),DZ3(n1))

! Definition of the vector of aperture angles
do i1=1,n1
   gammap(i1)=o1+d1*(i1-1)
   betarp(i1)=gammap(i1) ! Only true for flat reflectors
   betasp(i1)=-gammap(i1) ! Only true for flat reflectors
end do

! Computation of the different quantities used to compute the derivatives of the depth with respect to pertubations.

! Elliptical phase velocity at the different propagation angles 
Vellbetarp=sqrt(vvt**2*cosd(betarp)**2+vht**2*sind(betarp)**2)
Vellbetasp=sqrt(vvt**2*cosd(betasp)**2+vht**2*sind(betasp)**2)

! VTI phase velocity at the different propagation angles
Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvt**2*(vnt**2-vht**2)*sind(2*betarp)**2))/2)
Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvt**2*(vnt**2-vht**2)*sind(2*betasp)**2))/2)

! Derivative of the elliptic phase velocity with respect to the phase angle
DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vht**2-vvt**2)
DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vht**2-vvt**2)

! Derivative of the VTI phase velocity with respect to the phase angle
DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vht**2-vvt**2)+1/sqrt(Vellbetarp**4+vvt**2*(vnt**2-vht**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vht**2-vvt**2)+2*vvt**2*(vnt**2-vht**2)*sind(2*betarp)*cosd(2*betarp)))
DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vht**2-vvt**2)+1/sqrt(Vellbetasp**4+vvt**2*(vnt**2-vht**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vht**2-vvt**2)+2*vvt**2*(vnt**2-vht**2)*sind(2*betasp)*cosd(2*betasp)))

! Computation of the group propagation angles
betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))

! Computation of the group aperture angle
gammag=0.5*(betarg-betasg)
write(0,*) "Gammag:",  gammag
write(0,*) "Gammap:",  gammap

deallocate(betarp,betarg,betasp,betasg,Vellbetarp,Vellbetasp,Vvtibetarp,Vvtibetasp,DVellbetarp,DVellbetasp,DVvtibetarp,DVvtibetasp)

! COMPUTATION OF THE RESIDUAL MOVEOUT

! Elliptic group slowness (the input angle is the group angle)
Sellg=sqrt((1/vvt)**2*cosd(gammag)**2+(1/vht)**2*sind(gammag)**2)

! VTI group slowness (the input angle is the group angle)
Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvt)**2*((1/vnt)**2-(1/vht)**2)*sind(2*gammag)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
DSellg1=-(1/vvt)**2*cosd(gammag)**2/Sellg
DSellg2=-(1/vht)**2*sind(gammag)**2/Sellg

! Computation of the derivatives of the VTI group slownesses
DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvt)**2*((1/vnt)**2-(1/vht)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvt)**2*((1/vnt)**2-(1/vht)**2)*sind(2*gammag)**2))
DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvt)**2*(1/vht)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvt)**2*((1/vnt)**2-(1/vht)**2)*sind(2*gammag)**2))
DSvtig3=-((1/vvt)**2*(1/vht)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvt)**2*((1/vnt)**2-(1/vht)**2)*sind(2*gammag)**2))

! Computation of the derivative of the depth with respect to the different perturbations
DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))
 
! Actual computation of the residual moveout
allocate(z(n1,n2))

write(0,*) n2
z=0
!write(0,*) rmo
z0m=ot+(it-1)*dt
write(0,*) "z0m :", z0m

!write(0,*) DZ2*rhovh
!write(0,*) DZ3*rhovn
z(:,1)=gammap
!write(0,*) "z(:,1):", z(:,1)

z(:,2)=z0m+z0m*(DZ2*(rhovh-1)+DZ3*(rhovn-1)) ! That is a vector


DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)**2)-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)**2)
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)**2)

z(:,3)=z0m+z0m*(DZ2*(rhovh-1)+DZ3*(rhovn-1))

!write(0,*) "rmo: ", z(:,1)

!z(:,2)=gammap
write(0,*) "rmo: ", z(:,3)

call sep_write(z)

end program velocity_spectra

