% Computes derivatives along tangent plane starting from group angles
% and group velocities

function [dz_dhx,dz_dmx]=gen_tan_plane_ang(s_func,s_der_func,s_par,t_data_inp,mx_data_inp,hx_data_inp,gamma,alpha,aniso,flag)

[n_gamma,n_alpha]=size(gamma);

ts=size(t_data_inp);
if((ts(1) == 1) & (ts(2) == 1))
  for i_alpha=1:n_alpha
    for i_gamma=1:n_gamma
      t_data(i_gamma,i_alpha)=t_data_inp;
    end
  end
else
  if((ts(1) ~= n_gamma) & (ts(2) ~= n_alpha))
    disp('gen_surf_ang: error with mismatching sizes')
    dbquit
  end
  t_data=t_data_inp;
end

for i_alpha=1:n_alpha
  for i_gamma=1:n_gamma
    gam=gamma(i_gamma,i_alpha);
    alp=alpha(i_gamma,i_alpha);
    if(abs(cosd(alp)) > abs(sind(gam)))
      % The different quantities we will need are the following
      S_s=s_func(s_par,alp-gam,'group','s');
      S_r=s_func(s_par,alp+gam,'group','r');
      cos_s=cosd(alp-gam);
      cos_r=cosd(alp+gam);
      den=1./((S_s+S_r) + (S_r-S_s)*tand(alp)*tand(gam));
      L=t_data(i_gamma,i_alpha)*den;
      
      % The derivatives we will need are the following 
      dz_dalp(i_gamma,i_alpha)= ...
        -L*tand(alp)*(cosd(alp)^2 + sind(gam)^2)/(cosd(alp)*cosd(gam));

      dz_dgam(i_gamma,i_alpha)= ...
        -L*tand(gam)*(cosd(gam)^2 + sind(alp)^2)/(cosd(alp)*cosd(gam));

      dmx_dalp(i_gamma,i_alpha)= ...  
        -L*cosd(alp)/cosd(gam);

      dmx_dgam(i_gamma,i_alpha)= ...
        -L*(sind(gam)*sind(alp))/cosd(gam)^2;

      dhx_dalp(i_gamma,i_alpha)= ...
        -L*(sind(gam)*sind(alp))/cosd(alp)^2;

      dhx_dgam(i_gamma,i_alpha)= ...
        -L*cosd(gam)/cosd(alp);

      if(aniso)
        % In case the medium is aniostropic, we have to add the following
        % quantities
        den_sq=den^2;
        dS_s_dalp=s_der_func(s_par,alp-gam,'group','s');
        dS_s_dgam=-dS_s_dalp;

        dS_r_dalp=s_der_func(s_par,alp+gam,'group','r');
        dS_r_dgam=dS_r_dalp;

        dL_dalp=-t_data(i_gamma,i_alpha)*den_sq*((dS_r_dalp + dS_s_dalp) + ...
                                (dS_r_dalp - dS_s_dalp)*tand(alp)*tand(gam) +...
                                (S_r - S_s)*tand(gam)/cosd(alp)^2);
        dL_dgam=-t_data(i_gamma,i_alpha)*den_sq*((dS_r_dgam + dS_s_dgam) + ...
                                (dS_r_dgam - dS_s_dgam)*tand(alp)*tand(gam) +...
                                (S_r - S_s)*tand(alp)/cosd(gam)^2);

        z_sc=(cosd(alp)^2-sind(gam)^2)/(cosd(alp)*cosd(gam));
        mx_sc=-sind(alp)/cosd(gam);
        hx_sc=-sind(gam)/cosd(alp);

        dz_dalp(i_gamma,i_alpha)= dz_dalp(i_gamma,i_alpha) + dL_dalp*z_sc;
        dz_dgam(i_gamma,i_alpha)= dz_dgam(i_gamma,i_alpha) + dL_dgam*z_sc;

        dmx_dalp(i_gamma,i_alpha)= dmx_dalp(i_gamma,i_alpha) + dL_dalp*mx_sc;
        dmx_dgam(i_gamma,i_alpha)= dmx_dgam(i_gamma,i_alpha) + dL_dgam*mx_sc;

        dhx_dalp(i_gamma,i_alpha)= dhx_dalp(i_gamma,i_alpha) + dL_dalp*hx_sc;
        dhx_dgam(i_gamma,i_alpha)= dhx_dgam(i_gamma,i_alpha) + dL_dgam*hx_sc;

      end 

      % Eventually, we estimate the partial following derivatives - that
      % are the tangent planes of the impulse response
      
      dz_dhx(i_gamma,i_alpha)= - ...
         (dz_dalp(i_gamma,i_alpha)*dmx_dgam(i_gamma,i_alpha) - ...
          dz_dgam(i_gamma,i_alpha)*dmx_dalp(i_gamma,i_alpha))/ ...
         (dmx_dalp(i_gamma,i_alpha)*dhx_dgam(i_gamma,i_alpha) - ...
          dmx_dgam(i_gamma,i_alpha)*dhx_dalp(i_gamma,i_alpha));

      dz_dmx(i_gamma,i_alpha)= - ...
         (-dz_dalp(i_gamma,i_alpha)*dhx_dgam(i_gamma,i_alpha) + ...
           dz_dgam(i_gamma,i_alpha)*dhx_dalp(i_gamma,i_alpha))/ ...
         ( dmx_dalp(i_gamma,i_alpha)*dhx_dgam(i_gamma,i_alpha) - ...
           dmx_dgam(i_gamma,i_alpha)*dhx_dalp(i_gamma,i_alpha));
    else
      z(i_gamma,i_alpha)=flag;
      dz_dhx(i_gamma,i_alpha)= NaN;
      dz_dmx(i_gamma,i_alpha)= NaN;
    end
  end
end

