module fft_data_mod

!  use down_process_types
  use fft_types
  use timers_mod

  implicit none

  ! GOOD For SGI
  integer, parameter ::  FFT_FORW=1
  integer, parameter ::  FFT_BACK=-1

  ! GOOD For LINUX
  !integer, parameter ::  FFT_FORW=-1
  !integer, parameter ::  FFT_BACK=1


  ! Parameters for FFTW package
  integer FFTW_FORWARD,FFTW_BACKWARD
  parameter (FFTW_FORWARD=1,FFTW_BACKWARD=-1)

  integer FFTW_REAL_TO_COMPLEX,FFTW_COMPLEX_TO_REAL
  parameter (FFTW_REAL_TO_COMPLEX=-1,FFTW_COMPLEX_TO_REAL=1)

  integer FFTW_ESTIMATE,FFTW_MEASURE
  parameter (FFTW_ESTIMATE=0,FFTW_MEASURE=1)

  integer FFTW_OUT_OF_PLACE,FFTW_IN_PLACE,FFTW_USE_WISDOM
  parameter (FFTW_OUT_OF_PLACE=0)
  parameter (FFTW_IN_PLACE=8,FFTW_USE_WISDOM=16)

  integer FFTW_THREADSAFE
  parameter (FFTW_THREADSAFE=128)


!  type fft_type 
!     integer   :: ierr,ndim
!     integer   :: plan_forw,plan_back
!     real :: scale
!  end type fft_type

  integer, external :: omp_get_thread_num

contains

  !----------------------------------------------------------------------
  subroutine  init_fft_1d(fft_id,n1,fast_init)
    type (fft_type), intent(out) :: fft_id
!  function  init_fft_1d(n1,fast_init) result (fft_id)
!    type (fft_type) fft_id
    integer n1,stat,flag
    logical, optional :: fast_init

    if(present(fast_init)) then
       if(fast_init) then
          flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       else
          flag=ior(ior(FFTW_MEASURE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       end if
    else
       flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
    end if

    write(0,*)'flag with ior',flag

    fft_id%scale=sqrt(1./(n1))


    !write(0,*)'fft_id',n1

    call fftw_f77_create_plan(fft_id%plan_forw,n1,FFTW_FORWARD,flag)
   !write(0,*)'fft1d: plan_forw=',fft_id%plan_forw

    call fftw_f77_create_plan(fft_id%plan_back,n1,FFTW_BACKWARD,flag)
   !write(0,*)'fft1d: plan_back=',fft_id%plan_back

    if((fft_id%plan_forw == 0 ).or.(fft_id%plan_back == 0)) then
       fft_id%ierr=1
    else
       fft_id%ierr=0
    end if

    fft_id%ndim=1

!write(0,*)'AFTER PLAN plan_forw=',fft_id%plan_forw
!write(0,*)'AFTER PLAN plan_back=',fft_id%plan_back
!
!write(0,*)'AFTER PLAN scale=',fft_id%scale
!write(0,*)'AFTER PLAN ierr=',fft_id%ierr
!write(0,*)'AFTER PLAN ndim=',fft_id%ndim

!$omp flush (fft_id)

!  end function init_fft_1d
  end subroutine init_fft_1d

  !----------------------------------------------------------------------
  subroutine  init_fft_2d(fft_id,n1,n2,fast_init) 
    type (fft_type), intent(out) :: fft_id
!  function  init_fft_2d(n1,n2,fast_init) result (fft_id)
!    type (fft_type) fft_id
    integer n1,n2,stat,flag
    logical, optional :: fast_init

    if(present(fast_init)) then
       if(fast_init) then
          flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       else
          flag=ior(ior(FFTW_MEASURE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       end if
    else
       flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
    end if


    fft_id%scale=sqrt(1./(n1*n2))

    call fftw2d_f77_create_plan(fft_id%plan_forw,n1,n2,FFTW_FORWARD,flag)
    !write(0,*)'fft2d: plan_forw=',fft_id%plan_forw

    call fftw2d_f77_create_plan(fft_id%plan_back,n1,n2,FFTW_BACKWARD,flag)
    !write(0,*)'fft2d: plan_back=',fft_id%plan_back

    if((fft_id%plan_forw ==0 ).or.(fft_id%plan_back == 0)) then
       fft_id%ierr=1
    else
       fft_id%ierr=0
    end if

    fft_id%ndim=2

    !write(0,*)'End of init fft2d'

  end subroutine init_fft_2d
!  end function init_fft_2d

  !----------------------------------------------------------------------
  subroutine init_fft_3d(fft_id,n1,n2,n3,fast_init) 
    type (fft_type), intent(out) :: fft_id
!  function init_fft_3d(n1,n2,n3,fast_init) result (fft_id)
!    type (fft_type) fft_id
    integer n1,n2,n3,stat,flag
    logical, optional :: fast_init
    character (len=128) :: desc_timer


    if(present(fast_init)) then
       if(fast_init) then
          desc_timer='Init FFTW 3-D (fast_init)'
          call SEP_timer_init(20,desc_timer,0.,0)
          flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       else
          desc_timer='Init FFTW 3-D (no fast_init)'
          call SEP_timer_init(20,desc_timer,0.,0)
          flag=ior(ior(FFTW_MEASURE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
       end if
    else
       flag=ior(ior(FFTW_ESTIMATE,FFTW_USE_WISDOM),FFTW_THREADSAFE)
    end if

    fft_id%scale=sqrt(1./(n1*n2*n3))

    call SEP_timer_start(20)
    call fftw3d_f77_create_plan(fft_id%plan_forw,n1,n2,n3,FFTW_FORWARD,flag)
    call fftw3d_f77_create_plan(fft_id%plan_back,n1,n2,n3,FFTW_BACKWARD,flag)
    call SEP_timer_stop(20)

    if((fft_id%plan_forw ==0 ).or.(fft_id%plan_back == 0)) then
       fft_id%ierr=1
    else
       fft_id%ierr=0
    end if

    fft_id%ndim=3

    desc_timer='Computing FFTW 3-D (no fast_init)'
    call SEP_timer_init(21,desc_timer,0.,0)

  end subroutine init_fft_3d

  !----------------------------------------------------------------------
  subroutine fft_1d_data(fft_id,sign,n1,data_vec)
    type (fft_type) fft_id
    integer :: n1
    integer :: sign
    complex, dimension (:) :: data_vec
    complex, dimension (n1) :: temp_1d
   

!    if(omp_get_thread_num() == 0) write(0,*)'Beg fft 1d'
    temp_1d=data_vec*fft_id%scale

    if(sign == FFT_FORW) then
       !write(0,*)'Forw 1d'
       call fftw_f77_one(fft_id%plan_forw, temp_1d,data_vec)
    else
!       if(omp_get_thread_num() == 0) write(0,*)'Back 1d'
       call fftw_f77_one(fft_id%plan_back, temp_1d,data_vec)
    end if


    return	
  end subroutine fft_1d_data

  !----------------------------------------------------------------------
  subroutine fft_2d_data(fft_id,sign,n1,n2,data_slice)
    type (fft_type) fft_id
    integer :: n1,n2
    integer :: sign
    complex, dimension (:,:) :: data_slice
    complex, dimension (n1,n2) :: temp_2d

    !write(0,*)'Beg fft 2d'
    temp_2d=data_slice*fft_id%scale

    if(sign == FFT_FORW) then
       !write(0,*)'Forw 2d'
       call fftwnd_f77_one(fft_id%plan_forw, temp_2d,data_slice)
    else
       !write(0,*)'Back 2d'
       call fftwnd_f77_one(fft_id%plan_back, temp_2d,data_slice)
    end if

    return	
  end subroutine fft_2d_data


  !----------------------------------------------------------------------
  subroutine fft_3d_data(fft_id,sign,n1,n2,n3,data_slice)
    type (fft_type) fft_id
    integer :: n1,n2,n3
    integer :: sign
    complex,  dimension (:,:,:) :: data_slice
    complex,  dimension (n1,n2,n3) :: temp_3d

    !write(0,*)'Beg fft 3d'

    temp_3d=data_slice*fft_id%scale

    ! write(0,*)'fft_3d shape(data_slice)=',shape(data_slice)
!write(0,*)'Inside ffft fft_3d maxval(cabs(data_slice))=',maxval(cabs(data_slice))

!    if(omp_get_thread_num() == 0) then
!!if(omp_get_thread_num() == 0) write(0,*)'#31'
!       call SEP_timer_start(21)
!    end if
    if(sign == FFT_FORW) then
       !write(0,*)'Forw 3d'
!if(omp_get_thread_num() == 0) write(0,*)'#32'
       call fftwnd_f77_one(fft_id%plan_forw, temp_3d,data_slice)
!if(omp_get_thread_num() == 0) write(0,*)'#33'
    else
       !write(0,*)'Back 3d'
       call fftwnd_f77_one(fft_id%plan_back, temp_3d,data_slice)
    end if
!    if(omp_get_thread_num() == 0) then
!!if(omp_get_thread_num() == 0) write(0,*)'#34'
!       call SEP_timer_stop(21)
!    end if

    ! write(0,*)'after fft_3d maxval(cabs(data_slice))=',maxval(cabs(data_slice))

    return	
  end subroutine fft_3d_data

  !----------------------------------------------------------------------
  subroutine fft_3d_two_pass_data(fft_1d_id,fft_2d_id,sign,n1,n2,n3,data_slice)
    type (fft_type) fft_1d_id,fft_2d_id
    integer :: n1,n2,n3,i1,i2,i3
    integer :: sign
    real    :: scale
    complex,  dimension (:,:,:) :: data_slice
    complex,  dimension (n1,n2) :: temp_2d
    complex,  dimension (n3)    :: temp_1d

    scale=fft_2d_id%scale*fft_1d_id%scale

    !temp_2d(:,:) = data_slice(:,:,1)
    !call fftwnd_f77_one(fft_2d_id%plan_back, temp_2d(:,:), data_slice(:,:,i3))
    !write(0,*)'Passed it 2D !'

    !temp_1d(:) = data_slice(1,1,:)
    !call fftw_f77_one(fft_1d_id%plan_back, temp_1d(:), temp_data(:))
    !write(0,*)'Passed it 1D !'

    if(sign == FFT_FORW) then

       do i3=1,n3
          temp_2d = data_slice(:,:,i3)*scale
          call fftwnd_f77_one(fft_2d_id%plan_forw, temp_2d, data_slice(:,:,i3))
       end do
       if(n3 > 1) then
          do i2=1,n2
             do i1=1,n1
                temp_1d = data_slice(i1,i2,:)
                call fftw_f77_one(fft_1d_id%plan_forw, temp_1d, &
                data_slice(i1,i2,:))
             end do
          end do
       end if

    else

       do i3=1,n3
          temp_2d = data_slice(:,:,i3)*scale
          call fftwnd_f77_one(fft_2d_id%plan_back, temp_2d, data_slice(:,:,i3))
       end do
       if(n3 > 1) then
          do i2=1,n2
             do i1=1,n1
                temp_1d = data_slice(i1,i2,:)
                call fftw_f77_one(fft_1d_id%plan_back, temp_1d, &
                data_slice(i1,i2,:))
             end do
          end do
       end if

    end if

    return	
  end subroutine fft_3d_two_pass_data

  !----------------------------------------------------------------------
  subroutine fft_4d_two_pass_data(fft_1d_id,fft_3d_id,sign, &
  n1,n2,n3,n4,data_slice)
    type (fft_type), intent(in) :: fft_1d_id,fft_3d_id
    integer, intent(in) :: n1,n2,n3,n4
    integer, intent(in) :: sign
    complex, intent(inout),  dimension (:,:,:,:)  :: data_slice

    integer, automatic :: i1,i2,i3,i4
    real, automatic    :: scale
    complex, automatic,  dimension (n1,n2,n3) :: temp_3d
    complex, automatic,  dimension (n4)       :: temp_1d

!write(0,*)'n1=',n1
!write(0,*)'n2=',n2
!write(0,*)'n3=',n3
!write(0,*)'n4=',n4
!write(0,*)'INSIDE'
!write(0,*)'shape(temp_3d)=',shape(temp_3d)
!write(0,*)'shape(temp_1d)=',shape(temp_1d)

    scale=fft_3d_id%scale*fft_1d_id%scale
!write(0,*)'#31 scale=',scale,'thread#',omp_get_thread_num()

    if(sign == FFT_FORW) then

!if(omp_get_thread_num() == 0) write(0,*)'#31'

       do i4=1,n4
          temp_3d = data_slice(:,:,:,i4)*scale
!write(0,*)'#331 scale=',scale,'thread#',omp_get_thread_num()
          call fftwnd_f77_one(fft_3d_id%plan_forw, temp_3d, data_slice(:,:,:,i4))
       end do

!if(omp_get_thread_num() == 0) write(0,*)'#32'
!write(0,*)'#32'

       if(n4 > 1) then
          do i3=1,n3
             do i2=1,n2
                do i1=1,n1
                   temp_1d = data_slice(i1,i2,i3,:)
                   call fftw_f77_one(fft_1d_id%plan_forw, temp_1d, &
                   data_slice(i1,i2,i3,:))
                end do
             end do
          end do
       end if
!if(omp_get_thread_num() == 0) write(0,*)'#33'
!write(0,*)'#33'

    else   

       do i4=1,n4
          temp_3d = data_slice(:,:,:,i4)*scale
          call fftwnd_f77_one(fft_3d_id%plan_back, temp_3d, data_slice(:,:,:,i4))
       end do
       if(n4 > 1) then
          do i3=1,n3
             do i2=1,n2
                do i1=1,n1
                   temp_1d = data_slice(i1,i2,i3,:)
                   call fftw_f77_one(fft_1d_id%plan_back, temp_1d, &
                   data_slice(i1,i2,i3,:))
                end do
             end do
          end do
       end if

    end if

    return	
  end subroutine fft_4d_two_pass_data


  !----------------------------------------------------------------------
  subroutine clean_fft(fft_id) 
    type (fft_type) fft_id
    integer ierr

    select case (fft_id%ndim)
    case (1)
       !write(0,*)'Beg clean fft 1d'
       call fftw_f77_destroy_plan(fft_id%plan_forw)
       call fftw_f77_destroy_plan(fft_id%plan_back)
       !write(0,*)'End clean fft 1d'
    case (2)
       !write(0,*)'Beg clean fft 2d'
       call fftwnd_f77_destroy_plan(fft_id%plan_forw)
       call fftwnd_f77_destroy_plan(fft_id%plan_back)
       !write(0,*)'End clean fft 2d'
    case (3)
       !write(0,*)'Beg clean fft 3d'
       call fftwnd_f77_destroy_plan(fft_id%plan_forw)
       call fftwnd_f77_destroy_plan(fft_id%plan_back)
       !write(0,*)'End clean fft 3d'
    case default
       write(0,*)'fft_id%ndim=',fft_id%ndim
       call seperr('Not supported fft_id%ndim')
    end select


    return	
  end subroutine clean_fft

end module fft_data_mod
