module image_mod

  use gendown3d_types
  use down_process_types
  use data_types
  use image_types
  use slow_types
  use restart_types

  use fft_data_mod

  implicit none

  logical, private,save ::  stat_phase,save_memory
  integer, external ::  sreed,srite,sseek_block
  integer, private :: nc_bytes,nr_bytes,n_block_z,n_z_half
  real, private,save :: scale_image,epsi_kz,epsi_w,epsi_px,slow_const_image_surf
  real, private,save :: k_hx_min,k_hx_max,freq_pow_scale
  
  real, allocatable, dimension (:), private,save :: k_mx,k_my,k_hx,k_hy
  real, allocatable, dimension (:,:,:,:), private,save :: k_sr_sq
  real, allocatable, dimension (:,:), private,save :: rimage_slice_2d
  real, allocatable, dimension (:,:,:,:), private,save :: rimage_slice_4d
  complex, allocatable, dimension (:), private,save :: hx_shift_forw
  complex, allocatable, dimension (:), private,save :: hx_shift_back
  complex, allocatable, dimension (:,:), private,save :: cimage_slice_2d
  complex, allocatable, dimension (:,:,:,:), private,save :: cimage_slice_4d

contains

  !-----------------------------------------------------------------------------
  subroutine init_io_image(gen_par,data_par,image_par,down_par)

    integer ierr,i_k_hx,i_k_hy,i_k_mx,i_k_my
    real inv_back_wave_length,band_width,khx,khy,kmx,kmy,pi
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (image_param_type) image_par
    type (down_process_param_type) down_par

    !+++
!    if(image_par%n_hy_image > 1) call seperr('n_hy_image > 1 : not supported yet')

    save_memory=gen_par%save_memory

    epsi_px=1*epsilon(1.)
    epsi_kz=10000*epsilon(1.)
    epsi_w=100*epsilon(1.)
    stat_phase=gen_par%stat_shift /= 0.
    write(0,*)'stat_phase=',stat_phase

    freq_pow_scale=gen_par%freq_pow_scale
    write(0,*)'freq_pow_scale=',freq_pow_scale

    if((gen_par%amp_weights(1:1) == 'W') .and. &
    (gen_par%slow_const_image_z0 <=0.)) then
       call seperr('slow_const_image_z0 is <= 0')
    else
       slow_const_image_surf=gen_par%slow_const_image_z0
    end if

    if(.not. save_memory) then 
       !{
      if(gen_par%image_zero_offset .or. gen_par%image_cig) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_2d'
            allocate (rimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_2d")
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_2d'
            allocate (cimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_2d")
         end if
         !}
      end if
  
      if( .not. gen_par%image_zero_offset) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_4d'
            allocate (rimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_4d")
            if(gen_par%debug2) write(0,*)'Allocated rimage_slice_4d with shape',shape(rimage_slice_4d)
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_4d'
            allocate (cimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(gen_par%debug2) write(0,*)'Allocated cimage_slice_4d with shape',shape(cimage_slice_4d)
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_4d")
         end if
         !}
      end if
       !}
    end if
  
    allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")
    allocate (k_hy(data_par%n_hy_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hy ")

    allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_my(data_par%n_my_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")

    allocate (k_sr_sq(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,2), &
    stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_sr_sq ")

    !+++

    allocate (hx_shift_forw(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of hx_shift_forw ")
    allocate (hx_shift_back(data_par%n_hx_pad),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of hx_shift_back ")

    if(gen_par%single_hx_dip) then
       k_hx_min=data_par%khx_0
       k_hx_max=data_par%khx_0 + (data_par%n_hx_pad-1  )*data_par%d_khx
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
       end do
    else
       k_hx_min=-(data_par%n_hx_pad/2-1)*data_par%d_khx
       k_hx_max= (data_par%n_hx_pad/2  )*data_par%d_khx
!write(0,*)'k_hx_min',k_hx_min
!write(0,*)'k_hx_max',k_hx_max
       do i_k_hx=1,data_par%n_hx_pad
          khx = i_k_hx - 1.
          if(khx <= float(data_par%n_hx_pad/2)) then
             k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
          else
             k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
          end if
       end do
    end if
    !+++
    do i_k_hy=1,data_par%n_hy_pad
       khy = i_k_hy - 1.
       if(khy <= float(data_par%n_hy_pad/2)) then
          k_hy(i_k_hy) =  data_par%khy_0+data_par%d_khy* khy
       else
          k_hy(i_k_hy) = -data_par%khy_0+data_par%d_khy*(khy - data_par%n_hy_pad)
       end if
    end do

    pi=acos(-1.)
    do i_k_hx=1,data_par%n_hx_pad
       if(gen_par%image_cig) then
         hx_shift_forw(i_k_hx)=cmplx( &
           cos(k_hx(i_k_hx)*data_par%hx_0_pad+gen_par%stat_shift*pi), &
           sin(k_hx(i_k_hx)*data_par%hx_0_pad+gen_par%stat_shift*pi))
         hx_shift_back(i_k_hx)=cmplx( &
           cos(-k_hx(i_k_hx)*data_par%hx_0_pad-gen_par%stat_shift*pi), &
           sin(-k_hx(i_k_hx)*data_par%hx_0_pad-gen_par%stat_shift*pi))
        else
         hx_shift_forw(i_k_hx)=cmplx( &
           cos(gen_par%stat_shift*pi), &
           sin(gen_par%stat_shift*pi))
         hx_shift_back(i_k_hx)=cmplx( &
           cos(-gen_par%stat_shift*pi), &
           sin(-gen_par%stat_shift*pi))
        end if
       !      hx_shift_forw(i_k_hx)=1.
       !      hx_shift_back(i_k_hx)=1.
    end do

    do i_k_mx=1,data_par%n_mx_pad
       kmx  = i_k_mx - 1.
       if(kmx <= float(data_par%n_mx_pad/2)) then
          k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
       else
          k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
       end if
    end do

    do i_k_my=1,data_par%n_my_pad
       kmy = i_k_my - 1.
       if(kmy <= float(data_par%n_my_pad/2)) then
          k_my(i_k_my) =  data_par%kmy_0+data_par%d_kmy* kmy
       else
          k_my(i_k_my) = -data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad)
       end if
    end do

    do i_k_hx=1,data_par%n_hx_pad
       do i_k_my=1,data_par%n_my_pad
          do i_k_mx=1,data_par%n_mx_pad
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,1)= &
             .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
             k_sr_sq(i_k_mx,i_k_my,i_k_hx,2)= &
             .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)

             !write(0,*)'k_s_sq(',i_k_hx,') =',k_s_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_r_sq(',i_k_hx,') =',k_r_sq(i_k_mx,i_k_my,i_k_hx) 
             !write(0,*)'k_my(',i_k_my,') =',k_mx(i_k_my) 
             !write(0,*)'k_mx(',i_k_mx,') =',k_mx(i_k_mx) 
          end do
       end do
    end do




    !write(0,*)'hx_shift_back',hx_shift_back


    nr_bytes=image_par%n_mx_image*image_par%n_my_image* &
    image_par%n_hx_image*image_par%n_hy_image* 4

    nc_bytes=image_par%n_mx_image*image_par%n_my_image* &
    image_par%n_hx_image*image_par%n_hy_image* 8

    n_z_half=down_par%n_z_block/2
    n_block_z=image_par%n_z_image/down_par%n_z_block
    if(gen_par%debug2) write(0,*)'n_block_z=',n_block_z

    band_width=(.5*(down_par%w_max_0+down_par%w_max_1) - &
                .5*(down_par%w_min_0+down_par%w_min_1))/(2.*acos(-1.))
    if(band_width == 0) band_width=data_par%d_w
    if(gen_par%debug2) write(0,*)'band_width=',band_width

    if(gen_par%slow_const > 0.) then
       inv_back_wave_length=2.*gen_par%slow_const*band_width
    else
       inv_back_wave_length=2.*band_width
    end if
    if(gen_par%debug2) write(0,*)'inv_back_wave_length=',inv_back_wave_length

    scale_image=abs(image_par%d_z_image)*inv_back_wave_length
    scale_image=scale_image/float(down_par%n_w_block_par*down_par%n_w_block_ser)
    if((gen_par%image_zero_offset) .or. (gen_par%image_cig)) then 
       scale_image=scale_image/float(image_par%n_hx_image*image_par%n_hy_image)
       scale_image=sqrt(scale_image)
    else
       scale_image=sqrt(scale_image)
    end if
    if(gen_par%debug2) write(0,*)'scale_image=',scale_image

  end subroutine init_io_image


  !-----------------------------------------------------------------------------
  subroutine output_image(tag_image,i_block_w,i_block_z, &
  gen_par,down_par,image_par,restart_par,image_block)

    character(len=128) :: tag_image
    integer :: i_block_w,i_block_z
    type (gendown3d_param_type) gen_par
    type (down_process_param_type) down_par
    type (image_param_type) image_par
    type (restart_param_type) :: restart_par
    complex, dimension (:,:,:,:,:,:) :: image_block

    logical write_block,add_to_image
    character(len=1) :: type_image
    integer :: i_z_block,i_hx,i_hy,len_tag
    integer :: ierr,iseek,iseek_0,n_bytes
    integer :: i_total_restart,n_total_restart
    type (fft_type) fft_1d_id,fft_2d_id,fft_3d_id

    if(gen_par%debug2) write(0,*)'gen_par%k_domain',gen_par%k_domain
    if(gen_par%debug2) write(0,*)'gen_par%image_zero_offset',gen_par%image_zero_offset
    if(gen_par%debug2) write(0,*)'gen_par%image_cig',gen_par%image_cig
    if(gen_par%debug2) write(0,*)'save_memory',save_memory


    if(save_memory) then 
       !{
      if(gen_par%image_zero_offset .or. gen_par%image_cig) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_2d'
            allocate (rimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_2d")
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_2d'
            allocate (cimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_2d")
         end if
         !}
      end if
  
      if( .not. gen_par%image_zero_offset) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_4d'
            allocate (rimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_4d")
            if(gen_par%debug2) write(0,*)'Allocated rimage_slice_4d with shape',shape(rimage_slice_4d)
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_4d'
            allocate (cimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(gen_par%debug2) write(0,*)'Allocated cimage_slice_4d with shape',shape(cimage_slice_4d)
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_4d")
         end if
         !}
      end if
       !}
    end if

    len_tag=len_trim(tag_image)
    if(tag_image(1:len_tag) == image_par%tag_dimage(1:len_tag)) then
       if(image_par%esize_dimage == 4) then
          type_image(1:1) = 'r'
          n_bytes=nr_bytes
       else
          type_image(1:1) = 'c'
          n_bytes=nc_bytes
       end if
    else
       if(tag_image(1:len_tag) == image_par%tag_image(1:len_tag)) then
          if(image_par%esize_image == 4) then
             type_image(1:1) = 'r'
             n_bytes=nr_bytes
          else
             type_image(1:1) = 'c'
             n_bytes=nc_bytes
          end if
       else
          write(0,*)'tag_image=',trim(tag_image)
          call seperr('Not recognized tag_image')
       end if
    end if

    if(gen_par%k_domain) then 
       !{
       if(gen_par%image_zero_offset .or. gen_par%image_cig) then 
          !{
          !if(gen_par%debug2) write(0,*)'output_image: inv fft n1=',n1
          !if(gen_par%debug2) write(0,*)'output_image: inv fft n2=',n2
          call init_fft_2d(fft_2d_id,image_par%n_mx_image,image_par%n_my_image)
!          fft_2d_id=init_fft_2d(image_par%n_mx_image,image_par%n_my_image)
          do i_z_block=1,down_par%n_z_block
             !{
             do i_hy=1,image_par%n_hy_image
                !{
                do i_hx=1,image_par%n_hx_image
                   !{
                   cimage_slice_2d=image_block(:,:,i_hx,i_hy,i_z_block,1)
                   call fft_2d_data(fft_2d_id,FFT_BACK, &
                   image_par%n_mx_image,image_par%n_my_image, &
                   cimage_slice_2d)
                   image_block(:,:,i_hx,i_hy,i_z_block,1)=cimage_slice_2d
                   !}
                end do
                !}
             end do
             !}
          end do
          call clean_fft(fft_2d_id)
          !}
       else 
          !{
          !if(gen_par%debug2) write(0,*)'output_image: inv fft n1=',n1
          !if(gen_par%debug2) write(0,*)'output_image: inv fft n2=',n2
          !if(gen_par%debug2) write(0,*)'output_image: inv fft n3=',n3

          call init_fft_1d(fft_1d_id,image_par%n_hy_image) 
          call init_fft_3d(fft_3d_id,image_par%n_mx_image, &
             image_par%n_my_image,image_par%n_hx_image)
          do i_z_block=1,down_par%n_z_block
              !{
              cimage_slice_4d=image_block(:,:,:,:,i_z_block,1)
!write(0,*)'Outside fft_3d maxval(cabs(cimage_slice_4d))=',maxval(cabs(cimage_slice_4d)),'with i_z_block=',i_z_block,'and i_hy=',i_hy
              call fft_4d_two_pass_data(fft_1d_id,fft_3d_id,FFT_BACK, &
              image_par%n_mx_image,image_par%n_my_image, &
              image_par%n_hx_image,image_par%n_hy_image,cimage_slice_4d)
              image_block(:,:,:,:,i_z_block,1)=cimage_slice_4d
              !}
          end do
          call clean_fft(fft_1d_id)
          call clean_fft(fft_3d_id)
          !}
       end if

       !}
    end if

    write_block=.true.
    if( restart_par%restart .and. &
        (&
         (i_block_w <  restart_par%i_block_w) .or. &
         ((i_block_w <= restart_par%i_block_w) .and. &
          (i_block_z <  restart_par%i_block_z)) &
        ) &
      ) then 
       write_block=.false.
       if(gen_par%debug2) write(0,*)'Write_block = false'
    end if

    !       (tag_image(1:len_tag)==image_par%tag_dimage(1:len_tag))) .or. &

    add_to_image= ( &
    (i_block_w /= 1) .or. &
    ((down_par%add).and. gen_par%domain_range(1:9) =='data_imag') .or. &
    ((down_par%add).and. gen_par%domain_range(1:9) =='dslow_dim') .or. &
    ((i_block_w == 1) .and. (i_block_z >= restart_par%i_block_z) .and. &
    (restart_par%restart)) &
    )
    !write(0,*)'add_to_image=',add_to_image

    if(write_block) then
       !{
       if(add_to_image) then
          !{
          ! add to the image
          !          write(0,*)'add_to_image seeek in image'
          if(gen_par%debug2) write(0,*)'Adding to ',tag_image(1:len_tag)
          if(down_par%adj) then
             !{
             iseek_0=(n_block_z - i_block_z)*down_par%n_z_block
             !}
          else
             !{
             iseek_0=(i_block_z-1)*down_par%n_z_block
             !}
          end if
          do i_z_block=1,down_par%n_z_block
             !{

             if(down_par%adj) then
                !{
                iseek= iseek_0 + (down_par%n_z_block-i_z_block)
                !}
             else
                !{
                iseek= iseek_0 + (i_z_block -1)
                !}
             end if

             if(sseek_block(trim(tag_image),iseek,n_bytes,0) .ne. iseek ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with seeking in tag_image file (first)')
             end if

             if(gen_par%image_zero_offset) then 
                !{
                if(type_image(1:1) == 'c') then
                   !{
                   if(sreed(trim(tag_image),cimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with reading tag_image file')
                   end if
                   cimage_slice_2d=cimage_slice_2d+ &
                   image_block(:,:,1,1,i_z_block,1)
                   !}
                else
                   !{
                   if(sreed(trim(tag_image),rimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with reading tag_image file')
                   end if
                   rimage_slice_2d=rimage_slice_2d + &
                   real(image_block(:,:,1,1,i_z_block,1))
                   !}
                end if

                if(sseek_block(trim(tag_image),-1,n_bytes,1) .ne. iseek ) then
                   write(0,*)'tag_image=',trim(tag_image)
                   call seperr('Problems with seeking in tag_image file (second)')
                end if

                if(type_image(1:1) == 'c') then
                   !{
                   if(srite(trim(tag_image),cimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                else
                   !{
                   if(srite(trim(tag_image),rimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                end if
                call sep_prog_stat("Gendown",restart_par%i_total,restart_par%n_total,1)
                !}
             else
                !{
                if(type_image(1:1) == 'c') then
                   !{
                   if(sreed(trim(tag_image),cimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with reading tag_image file')
                   end if
                   cimage_slice_4d=cimage_slice_4d+image_block(:,:,:,:,i_z_block,1)
                   !}
                else
                   !{
                   if(sreed(trim(tag_image),rimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with reading tag_image file')
                   end if
                   rimage_slice_4d=rimage_slice_4d+real(image_block(:,:,:,:,i_z_block,1))
                   !}
                end if

                if(sseek_block(trim(tag_image),-1,n_bytes,1) .ne. iseek ) then
                   write(0,*)'tag_image=',trim(tag_image)
                   call seperr('Problems with seeking in tag_image file (second)')
                end if

                if(type_image(1:1) == 'c') then
                   !{
                   if(srite(trim(tag_image),cimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                else
                   !{
                   if(srite(trim(tag_image),rimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                end if
                call sep_prog_stat("Gendown",restart_par%i_total,restart_par%n_total,1)
                !}
             end if
             !}
          end do
          !}
       else
          !{
          ! do not add to the image
          if(down_par%adj) then
             iseek_0=(n_block_z - i_block_z)*down_par%n_z_block
          else
             iseek= (i_block_z - 1)*down_par%n_z_block
             if(sseek_block(trim(tag_image),iseek,n_bytes,0) .ne. iseek ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with seeking in tag_image file (first)')
             end if
          end if

          do i_z_block=1,down_par%n_z_block
             !{
             if(down_par%adj) then
                !{
                iseek= iseek_0 + (down_par%n_z_block-i_z_block)
                if(sseek_block(trim(tag_image),iseek,n_bytes,0) .ne. iseek ) then
                   write(0,*)'tag_image=',trim(tag_image)
                   call seperr('Problems with seeking in tag_image file (first)')
                end if
                !}
             end if

             if(gen_par%image_zero_offset) then 
                !{

                if(type_image(1:1) == 'c') then
                   !{
                   cimage_slice_2d=image_block(:,:,1,1,i_z_block,1)
                   if(srite(trim(tag_image),cimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                else
                   !{
                   rimage_slice_2d=real(image_block(:,:,1,1,i_z_block,1))
                   if(srite(trim(tag_image),rimage_slice_2d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                end if
                call sep_prog_stat("Gendown",restart_par%i_total,restart_par%n_total,1)
                !}
             else
                !{
                if(type_image(1:1) == 'c') then
                   !{
                   cimage_slice_4d=image_block(:,:,:,:,i_z_block,1)
                   if(srite(trim(tag_image),cimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                else
                   !{
                   rimage_slice_4d=real(image_block(:,:,:,:,i_z_block,1))
                   if(srite(trim(tag_image),rimage_slice_4d,n_bytes) .ne. n_bytes ) then
                      write(0,*)'tag_image=',trim(tag_image)
                      call seperr('Problems with writing tag_image file')
                   end if
                   !}
                end if
                call sep_prog_stat("Gendown",restart_par%i_total,restart_par%n_total,1)
                !}
             end if
             !}
          end do
          !}
       end if
       !} write_block condition
    end if

    if(save_memory) then 
      if(allocated(rimage_slice_2d)) then
         deallocate (rimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_2d')
      end if
      if(allocated(rimage_slice_4d)) then
         deallocate (rimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
      if(allocated(cimage_slice_2d)) then
         deallocate (cimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating cimage_slice_2d')
      end if
      if(allocated(cimage_slice_4d)) then
         deallocate (cimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
    end if

    return	
  end subroutine output_image


  !-----------------------------------------------------------------------------
  subroutine input_image(tag_image,i_block_w,i_block_z, &
  gen_par,down_par,image_par,image_block)

    character(len=128) :: tag_image
    integer :: i_block_w,i_block_z
    type (gendown3d_param_type) gen_par
    type (down_process_param_type) down_par
    type (image_param_type) image_par
    complex, dimension (:,:,:,:,:,:) :: image_block

    character(len=1) :: type_image
    integer :: i_z_block,i_hx,i_hy,len_tag
    integer :: ierr,iseek,iseek_0,n_bytes
    type (fft_type) fft_1d_id,fft_2d_id,fft_3d_id

    if(save_memory) then 
       !{
      if(gen_par%image_zero_offset .or. gen_par%image_cig) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_2d'
            allocate (rimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_2d")
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_2d'
            allocate (cimage_slice_2d(image_par%n_mx_image,image_par%n_my_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_2d")
         end if
         !}
      end if
  
      if( .not. gen_par%image_zero_offset) then 
         !{
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 4)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 4))) then
            if(gen_par%debug2) write(0,*)'Allocate rimage_slice_4d'
            allocate (rimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of rimage_slice_4d")
            if(gen_par%debug2) write(0,*)'Allocated rimage_slice_4d with shape',shape(rimage_slice_4d)
         end if
  
         if(((image_par%image_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_image == 8)) .or. & 
         ((image_par%dimage_file /= IMAGE_NO_FILE) .and. &
         ( image_par%esize_dimage == 8)) .or. &
         (gen_par%k_domain)) then 
            if(gen_par%debug2) write(0,*)'Allocate cimage_slice_4d'
            allocate (cimage_slice_4d(image_par%n_mx_image,image_par%n_my_image, &
            image_par%n_hx_image,image_par%n_hy_image), &
            stat=ierr)  
            if(gen_par%debug2) write(0,*)'Allocated cimage_slice_4d with shape',shape(cimage_slice_4d)
            if(ierr .ne. 0) call seperr("Problems with allocation of cimage_slice_4d")
         end if
         !}
      end if
       !}
    end if

    len_tag=len_trim(tag_image)
    if(tag_image(1:len_tag) == image_par%tag_dimage(1:len_tag)) then
       if(image_par%esize_dimage == 4) then
          type_image(1:1) = 'r'
          n_bytes=nr_bytes
       else
          type_image(1:1) = 'c'
          n_bytes=nc_bytes
       end if
    else
       if(tag_image(1:len_tag) == image_par%tag_image(1:len_tag)) then
          if(image_par%esize_image == 4) then
             type_image(1:1) = 'r'
             n_bytes=nr_bytes
          else
             type_image(1:1) = 'c'
             n_bytes=nc_bytes
          end if
       else
          write(0,*)'tag_image=',trim(tag_image)
          call seperr('Not recognized tag_image')
       end if
    end if

    if(gen_par%debug2) write(0,*)'gen_par%k_domain',gen_par%k_domain
    if(gen_par%debug2) write(0,*)'gen_par%image_zero_offset',gen_par%image_zero_offset
    if(gen_par%debug2) write(0,*)'gen_par%image_cig',gen_par%image_cig
    if(gen_par%debug2) write(0,*)'type_image',type_image


    if(down_par%adj) then
       !{
       iseek_0=(n_block_z - i_block_z)*down_par%n_z_block
       !}
    else
       !{
       iseek_0=(i_block_z-1)*down_par%n_z_block
       !}
    end if

    !write(-1,*)'input iseek_0=',iseek_0
    !write(0,*)'n_bytes=',n_bytes
    do i_z_block=1,down_par%n_z_block
       !{

       if(down_par%adj) then
          !{
          iseek= iseek_0 + (down_par%n_z_block-i_z_block)
          !}
       else
          !{
          iseek= iseek_0 + (i_z_block -1)
          !}
       end if

       if(sseek_block(trim(tag_image),iseek,n_bytes,0) .ne. iseek ) then
          write(0,*)'tag_image=',trim(tag_image)
          call seperr('Problems with seeking in tag_image file (first)')
       end if

       if(gen_par%image_zero_offset) then 
          !{

          if(type_image(1:1) == 'c') then
             !{
             if(sreed(trim(tag_image),cimage_slice_2d,n_bytes) .ne. n_bytes ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with reading tag_image file')
             end if
             image_block(:,:,1,1,i_z_block,1)=  cimage_slice_2d
             !}
          else
             !{
             if(sreed(trim(tag_image),rimage_slice_2d,n_bytes) .ne. n_bytes ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with reading tag_image file')
             end if
             image_block(:,:,1,1,i_z_block,1)=  cmplx(rimage_slice_2d,0.)
             !write(0,*)'iseek=',iseek
             !}
          end if

          !}
       else
          !{
          if(type_image(1:1) == 'c') then
             !{
             if(sreed(trim(tag_image),cimage_slice_4d,n_bytes) .ne. n_bytes ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with reading tag_image file')
             end if
             image_block(:,:,:,:,i_z_block,1)= cimage_slice_4d
             !}
          else
             !{
             if(sreed(trim(tag_image),rimage_slice_4d,n_bytes) .ne. n_bytes ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with reading tag_image file')
             end if
             image_block(:,:,:,:,i_z_block,1)=  cmplx(rimage_slice_4d,0.)
             !}
          end if
          !}
       end if
       !}
    end do


    if(gen_par%k_domain) then 
       !{
       if(gen_par%image_zero_offset .or. gen_par%image_cig) then 
          !{
          !if(gen_par%debug2) write(0,*)'input_image: inv fft n1=',n1
          !if(gen_par%debug2) write(0,*)'input_image: inv fft n2=',n2
          call init_fft_2d(fft_2d_id,image_par%n_mx_image,image_par%n_my_image)
!          fft_2d_id=init_fft_2d(image_par%n_mx_image,image_par%n_my_image)
          do i_z_block=1,down_par%n_z_block
             !{
             do i_hy=1,image_par%n_hy_image
                !{
                do i_hx=1,image_par%n_hx_image
                   !{
                   cimage_slice_2d=image_block(:,:,i_hx,i_hy,i_z_block,1)
                   call fft_2d_data(fft_2d_id,FFT_FORW, &
                   image_par%n_mx_image,image_par%n_my_image, &
                   cimage_slice_2d)
                   image_block(:,:,i_hx,i_hy,i_z_block,1)=cimage_slice_2d
                   !}
                end do
                !}
             end do
             !}
          end do
          call clean_fft(fft_2d_id)
          !}
       else 
          !{
          call init_fft_1d(fft_1d_id,image_par%n_hy_image)
          call init_fft_3d(fft_3d_id,image_par%n_mx_image, &
             image_par%n_my_image,image_par%n_hx_image)
           do i_z_block=1,down_par%n_z_block
              !{
              cimage_slice_4d=image_block(:,:,:,:,i_z_block,1)
              call fft_4d_two_pass_data(fft_1d_id,fft_3d_id,FFT_FORW, &
              image_par%n_mx_image,image_par%n_my_image, &
              image_par%n_hx_image,image_par%n_hy_image,cimage_slice_4d)
              image_block(:,:,:,:,i_z_block,1)=cimage_slice_4d
              !}
           end do
          call clean_fft(fft_1d_id)
          call clean_fft(fft_3d_id)
          !}
       end if

       !}
    end if

    if(save_memory) then 
      if(allocated(rimage_slice_2d)) then
         deallocate (rimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_2d')
      end if
      if(allocated(rimage_slice_4d)) then
         deallocate (rimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
      if(allocated(cimage_slice_2d)) then
         deallocate (cimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating cimage_slice_2d')
      end if
      if(allocated(cimage_slice_4d)) then
         deallocate (cimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
    end if

    return	
  end subroutine input_image

  !-----------------------------------------------------------------------------
  recursive subroutine out_image_block_ser(i_w_block_par,i_z_block,z_down, &
  gen_par,data_par,down_par,image_par,slow_par, &
  data_block,image_block,slow_data_block,amp_stat_block)

    integer i_z_block,i_w_block_par
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) :: down_par
    type (image_param_type) image_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_data_block
    real, dimension (:,:,:,:,:,:) :: amp_stat_block
    complex, dimension (:,:,:,:,:,:) :: data_block
    complex, dimension (:,:,:,:,:,:) :: image_block

    integer i_k_mx,i_k_my,i_k_hx,i_k_hy,i_k_hx_p1
    integer i_px,i_px_p1,i_pp,i_pp_p1,i_w_block_ser
    real px,py,pp,inv_w_data,d_w_data,w_0_data,w_data,inv_d_k_hx
    real z_down
    real inv_d_px,inv_d_py,w_px,w_pp
    real k0,k0_sq,kz_s,kz_r,kz_s_sq,kz_r_sq,w_kz,k_s,k_r
    real kk_hx,w_k_hx,kz_y_sq,kz_x,k_hy_alpha,alpha,k_mx_sq,k_hx_sq
    real k_hy_alpha_surf,alpha_surf,k0_surf,k0_surf_sq
    real kz_s_surf,kz_r_surf,kz_s_surf_sq,kz_r_surf_sq
    real slow_const_image,vel_const_image
    real dz_stat,num1_stat,num2_stat
    real kz_min_amp_sq,kz_max_amp_sq,w_p_min_amp,w_p_max_amp
    complex hx_shift

!if((omp_get_thread_num() == 0)) write(0,*)'INSIDE k_hx_min',k_hx_min
!if((omp_get_thread_num() == 0)) write(0,*)'INSIDE k_hx_max',k_hx_max
 

    if(gen_par%slow_const_image > 0) then
       slow_const_image=gen_par%slow_const_image
       vel_const_image=(1./slow_const_image)
    else
       slow_const_image=slow_data_block(1,1,1,1,1,i_z_block)
!             if(omp_get_thread_num() == 0) write(0,*)'slow_const_image=',slow_const_image
       vel_const_image=(1./slow_const_image)
    end if

    kz_min_amp_sq=(2*image_par%w_min_amp*slow_const_image)**2
    kz_max_amp_sq=(2*image_par%w_max_amp*slow_const_image)**2

!if((omp_get_thread_num() == 0)) write(0,*)'kz_min_amp_sq=',kz_min_amp_sq
!if((omp_get_thread_num() == 0)) write(0,*)'kz_max_amp_sq=',kz_max_amp_sq

    if(stat_phase) then
       dz_stat=down_par%d_z_down
    end if

    if(gen_par%image_zero_offset) then
       !{
       do i_w_block_ser=1,down_par%n_w_block_ser
          !{
          do i_k_hy=1,data_par%n_hy_pad
             do i_k_hx=1,data_par%n_hx_pad
                image_block(:,:,1,1,i_z_block,i_w_block_par) = &
                image_block(:,:,1,1,i_z_block,i_w_block_par) + &
                data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)* &
                scale_image*hx_shift_forw(i_k_hx)
             end do
          end do
          !}
       end do
       !}
    else
       !{
       if(gen_par%image_cig) then
!if((omp_get_thread_num() == 0)) write(0,*)'gen_par%cig_forw_transf(1:3)',gen_par%cig_forw_transf(1:3)
          !{
          select case (gen_par%cig_forw_transf(1:3))

          case('pus')
             !{

             inv_d_px=1./image_par%d_px_image
             !+++
             inv_d_py=1./image_par%d_py_image

             d_w_data=down_par%n_w_block_par*data_par%d_w
             w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w

             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                inv_w_data=1./(w_0_data+(i_w_block_ser-1)*d_w_data)
                do i_k_hy=1,data_par%n_hy_pad
                   !{
                   do i_k_hx=1,data_par%n_hx_pad
                      !{
                      px=(k_hx(i_k_hx)*inv_w_data-image_par%px_image_0)*inv_d_px+1
                      py=(k_hy(i_k_hy)*inv_w_data-image_par%py_image_0)*inv_d_py+1
                      !+++
                      if( data_par%n_hy_pad > 1 .and. image_par%cig_p_for_px) then
                         !{
                         pp = sqrt( px**2 + py**2 )
                         i_pp=int(pp)
                         w_pp=pp-i_pp
                         if((i_pp > 0) .and. (i_pp <= image_par%n_px_image)) then
                            !if(gen_par%debug2) write(0,*) 'Use real norm for p'
                            image_block(:,:,i_pp,1,i_z_block,i_w_block_par) = &
                            image_block(:,:,i_pp,1,i_z_block,i_w_block_par) + (1.-w_pp)* &
                            data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)* &
                            scale_image*hx_shift_forw(i_k_hx)
                            i_pp_p1=i_pp+1
                            if((i_pp_p1) <= image_par%n_px_image) then
                               image_block(:,:,i_pp_p1,1,i_z_block,i_w_block_par)=&
                               image_block(:,:,i_pp_p1,1,i_z_block,i_w_block_par) + (w_pp)* &
                               data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)* &
                               scale_image*hx_shift_forw(i_k_hx)
                            end if
                         end if
                         !}
                      else
                         !{
                         w_data=w_0_data+(i_w_block_ser-1)*d_w_data
                         k0_sq= (w_data*slow_const_image)**2
                         do i_k_my=1,data_par%n_my_pad
                            !{
                            do i_k_mx=1,data_par%n_mx_pad
                               !{

                               select case (gen_par%amp_weights(1:1))
                               case ('n')
                                  !{
                                  w_kz=1.
                                  !}
                               case ('j')
                                  !{
                                  if(w_data > epsi_w) then
                                     w_kz=1./w_data
                                  else
                                     w_kz=0.
                                  end if
                                  !}
                               case ('i')
                                  !{
                                  kz_s_sq=k0_sq - k_sr_sq(i_k_mx,i_k_my,i_k_hx,1)
                                  kz_r_sq=k0_sq - k_sr_sq(i_k_mx,i_k_my,i_k_hx,2)

                                  if((w_data > epsi_w) .and. &
                                  (kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then 
                                     ! Jacobian
                                     w_kz=(sqrt(kz_s_sq)+sqrt(kz_r_sq))/(w_data*(w_data))
                                     !     w_kz=w_data*(sqrt(kz_s_sq)+sqrt(kz_r_sq))/sqrt((kz_s_sq*kz_r_sq))
                                  else
                                     w_kz=0.
                                  end if

                                  !}
                               case default
                                  call seperr('Unknown value for amp_weights')
                               end select

                               i_px=int(px)
                               if( (i_px > 0) .and. (i_px <= image_par%n_px_image)) then
                                  w_px=px-i_px
                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) = &
                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) + &
                                  (1.-w_px)* &
                                  data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)* &
                                  scale_image*hx_shift_forw(i_k_hx)*w_kz

                                  i_px_p1=i_px+1
                                  if(i_px_p1 <= image_par%n_px_image) then
                                     image_block(i_k_mx,i_k_my,i_px_p1,1,i_z_block,i_w_block_par)=&
                                     image_block(i_k_mx,i_k_my,i_px_p1,1,i_z_block,i_w_block_par)+&
                                     (w_px)* &
                                     data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)* &
                                     scale_image*hx_shift_forw(i_k_hx)*w_kz
                                  end if
                               end if
                               !}
                            end do
                            !}
                         end do
                         !}
                      end if
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          case('pul')
             !{
!if((omp_get_thread_num() == 0)) write(0,*)'Aft:: maxval(data_block)',maxval(cabs(data_block))
             d_w_data=down_par%n_w_block_par*data_par%d_w
             w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w
             inv_d_k_hx=1./data_par%d_khx
             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                w_data=w_0_data+(i_w_block_ser-1)*d_w_data
                k0= (w_data*slow_const_image)
                k0_sq= (k0)**2
                k0_surf= (w_data*slow_const_image_surf)
                k0_surf_sq= (k0_surf)**2
                do i_k_my=1,data_par%n_my_pad
                   !{
                   do i_k_mx=1,data_par%n_mx_pad
                      !{
                      k_mx_sq=k_mx(i_k_mx)**2

                      do i_k_hy=1,data_par%n_hy_pad
                         !{
                         do i_px=1,image_par%n_px_image
                            !{
!if((omp_get_thread_num() == 1)) write(0,*)'HERE #11'
                            px=image_par%px_image_0+(i_px-1)*image_par%d_px_image + &
                            epsi_px
                            kk_hx=px*w_data
                            k_hx_sq=kk_hx**2
                            if((kk_hx >=0.).and.(kk_hx < k_hx_max)) then
                               kk_hx=kk_hx*inv_d_k_hx+1
                               i_k_hx=int(kk_hx)
                               w_k_hx=kk_hx-i_k_hx
                            else
                               if((kk_hx <0.).and.(kk_hx >= k_hx_min)) then
                                  kk_hx=data_par%n_hx_pad+kk_hx*inv_d_k_hx+1
                                  i_k_hx=int(kk_hx)
                                  w_k_hx=kk_hx-i_k_hx
                               else
                                  i_k_hx=0
                               end if
                            end if
!if((omp_get_thread_num() == 1)) write(0,*)'i_k_hx ',i_k_hx
!if((omp_get_thread_num() == 1)) write(0,*)'w_data ',w_data
!if((omp_get_thread_num() == 1)) write(0,*)'px ',px
                            if(image_par%w_min_amp >0.) then
                              w_p_min_amp= .5*vel_const_image* &
 sqrt(kz_min_amp_sq + k_mx_sq+k_hx_sq + (k_mx_sq*k_hx_sq)/kz_min_amp_sq)
! sqrt(kz_min_amp_sq + k_hx_sq)
                              w_p_max_amp= .5*vel_const_image* &
 sqrt(kz_max_amp_sq + k_mx_sq+k_hx_sq + (k_mx_sq*k_hx_sq)/kz_max_amp_sq)
! sqrt(kz_max_amp_sq + k_hx_sq)
                            else
                              w_p_min_amp=-1
                              w_p_max_amp=huge(1.)
                            end if

!                            if(i_k_hx > 0) then
                            if((i_k_hx > 0) .and. &
                               (w_data >= w_p_min_amp) .and. &
                               (w_data <= w_p_max_amp) ) then
                               !{
                               i_k_hx_p1=i_k_hx+1
                               if(i_k_hx == data_par%n_hx_pad) i_k_hx_p1=1

!if((omp_get_thread_num() == 1)) write(0,*)'gen_par%amp_weights(1:1)',gen_par%amp_weights(1:1)

                               select case (gen_par%amp_weights(1:1))
                               case ('n')
                                  !{
!if((omp_get_thread_num() == 1)) write(0,*)'HERE #1'
                                  w_kz=1.
                                  !}
                               case ('j')
                                  !{
!                                  w_kz=1.
                                  w_kz=w_data
                                  !}
                               case ('u')
                                  !{
                                  k_s=.5*(k_mx(i_k_mx) - px*w_data)
                                  k_r=.5*(k_mx(i_k_mx) + px*w_data)
                                  kz_s_sq=k0_sq - k_s**2
                                  kz_r_sq=k0_sq - k_r**2

                                  if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then 
                                     kz_s=sqrt(kz_s_sq)
                                     kz_r=sqrt(kz_r_sq)

                                     w_kz=sqrt( &
                                     w_data* &
                                     (kz_s*(slow_const_image*k0 - .5*px*k_r) + &
                                     kz_r*(slow_const_image*k0 + .5*px*k_s))/&
                                     (kz_s*kz_r) &
                                     )
                                  else
                                     w_kz=0.
                                  end if
                                  !}
                               case ('i')
                                  !{
                                  k_s=.5*(k_mx(i_k_mx) - px*w_data)
                                  k_r=.5*(k_mx(i_k_mx) + px*w_data)
                                  kz_s_sq=k0_sq - k_s**2
                                  kz_r_sq=k0_sq - k_r**2

                                  if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then 
                                     kz_s=sqrt(kz_s_sq)
                                     kz_r=sqrt(kz_r_sq)

                                     w_kz=(kz_s*(slow_const_image*k0 - .5*px*k_r) + &
                                     kz_r*(slow_const_image*k0 + .5*px*k_s))/&
                                     (kz_s*kz_r)
                                  else
                                     w_kz=0.
                                  end if
                                  !}
                               case ('W')
                                  !{

                                  k_s=.5*(k_mx(i_k_mx) - px*w_data)
                                  k_r=.5*(k_mx(i_k_mx) + px*w_data)

                                  kz_s_sq=k0_sq - k_s**2
                                  kz_r_sq=k0_sq - k_r**2
                                  kz_s_surf_sq=k0_surf_sq - k_s**2
                                  kz_r_surf_sq=k0_surf_sq - k_r**2

                                  if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)&
                                  .and. (w_data > epsi_w)) then 
                                     !{
                                     kz_s=sqrt(kz_s_sq)
                                     kz_r=sqrt(kz_r_sq)

                                     kz_s_surf=sqrt(kz_s_surf_sq)
                                     kz_r_surf=sqrt(kz_r_surf_sq)

                                     w_kz=(kz_s*(slow_const_image*k0 - .5*px*k_r) + &
                                     kz_r*(slow_const_image*k0 + .5*px*k_s))/&
                                     (kz_s*kz_r)

                                     !                           if(i_k_my /= 1) then
                                     !                           !{ 
                                     kz_x=kz_s+kz_r
                                     kz_y_sq=kz_x**2- k_my(i_k_my)**2

                                     if(kz_y_sq > epsi_kz) then
                                        !{
                                        w_kz=w_kz * kz_x/sqrt(kz_y_sq)

                                        alpha=(kz_r-kz_s)/(kz_r+kz_s)
                                        alpha_surf=(kz_r_surf-kz_s_surf)/ &
                                        (kz_r_surf+kz_s_surf)
                                        k_hy_alpha=alpha*k_my(i_k_my)
                                        k_hy_alpha_surf=alpha_surf*k_my(i_k_my)

                                        if(stat_phase) then
                                           num1_stat=kz_s_sq
                                           num2_stat=kz_r_sq
                                        end if

                                        kz_s_sq=kz_s_sq - &
                                        .25*(k_my(i_k_my) - k_hy_alpha)**2
                                        kz_r_sq=kz_r_sq - &
                                        .25*(k_my(i_k_my) + k_hy_alpha)**2

                                        kz_s_surf_sq=kz_s_surf_sq - &
                                        .25*(k_my(i_k_my) - k_hy_alpha_surf)**2
                                        kz_r_surf_sq=kz_r_surf_sq - &
                                        .25*(k_my(i_k_my) + k_hy_alpha_surf)**2

                                        if((kz_s_sq > epsi_kz).and. (kz_r_sq > epsi_kz)) then
                                           !{

                                           kz_s=sqrt(kz_s_sq)
                                           kz_r=sqrt(kz_r_sq)
                                           kz_s_surf=sqrt(kz_s_surf_sq)
                                           kz_r_surf=sqrt(kz_r_surf_sq)

!                                           ! inverse of modeling scattering factor 
!                                           w_kz=w_kz*4*kz_s*kz_r/k0_sq
                                            w_kz=w_data*w_kz*4*kz_s*kz_r/k0_sq
!                                           ! slant stack jacobian  
!                                           w_kz=w_kz*w_data
                                           ! WKBJ  

                                           w_kz=w_kz*sqrt((kz_s_surf* kz_r_surf)/(kz_s* kz_r))

                                           ! stationary phase xamplitude correction
                                           if(stat_phase) then
                                              amp_stat_block(i_k_mx,i_k_my,i_px,i_k_hy,i_w_block_ser,i_w_block_par)= &
                                              amp_stat_block(i_k_mx,i_k_my,i_px,i_k_hy,i_w_block_ser,i_w_block_par)+ &
                                              dz_stat*(num1_stat/((kz_s)**3) + num2_stat/((kz_r)**3))
                                              w_kz=w_kz* &
                                              sqrt(amp_stat_block(i_k_mx,i_k_my,i_px,i_k_hy,i_w_block_ser,i_w_block_par))
                                           end if
                                           !}
                                        else
                                           !{
                                           w_kz=0
                                           !}
                                        end if
                                        !}
                                     else
                                        w_kz=0
                                     end if
                                     !                           !} 
                                     !                           else
                                     !                            w_kz=w_kz*4*kz_s*kz_r/k0_sq
                                     !                           end if
                                     !}
                                  else
                                     w_kz=0.
                                  end if
                                  !}
                               case ('t')
                                  !{

                                  k_s=.5*(k_mx(i_k_mx) - px*w_data)
                                  k_r=.5*(k_mx(i_k_mx) + px*w_data)
                                  kz_s_sq=k0_sq - k_s**2
                                  kz_r_sq=k0_sq - k_r**2

                                  if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)&
                                  .and. (w_data > epsi_w)) then 
                                  !{
                                     kz_s=sqrt(kz_s_sq)
                                     kz_r=sqrt(kz_r_sq)

                                     w_kz=(kz_s*(slow_const_image*k0 - .5*px*k_r) + &
                                     kz_r*(slow_const_image*k0 + .5*px*k_s))/&
                                     (kz_s*kz_r)

                                     if(i_k_my /= 1) then
                                        kz_x=kz_s+kz_r
                                        kz_y_sq=kz_x**2- k_my(i_k_my)**2

                                        if(kz_y_sq > epsi_kz) then
                                           w_kz=w_kz * kz_x/sqrt(kz_y_sq)

                                           alpha=(kz_r-kz_s)/(kz_r+kz_s)
                                           k_hy_alpha=alpha*k_my(i_k_my)
                                           kz_s_sq=kz_s_sq - &
                                           .25*(k_my(i_k_my) - k_hy_alpha)**2
                                           kz_r_sq=kz_r_sq - &
                                           .25*(k_my(i_k_my) + k_hy_alpha)**2

                                           if((kz_s_sq > epsi_kz).and. & 
                                           (kz_r_sq > epsi_kz))then
                                              kz_s=sqrt(kz_s_sq)
                                              kz_r=sqrt(kz_r_sq)
                                              w_kz=w_kz*4*kz_s*kz_r/k0_sq
                                           else
                                              w_kz=0
                                           end if
                                        else
                                           w_kz=0
                                        end if
                                     else
! BB commented in June 2002
!                                        w_kz=w_kz*4*kz_s*kz_r/k0_sq
! BB inverted in June 2002
!                                        w_kz=w_kz*k0_sq/(4*kz_s*kz_r)
! BB scaled by w in June 2002
!                                        w_kz=w_kz*w_data
! BB scaled by w in August 2002
!                                        w_kz=w_kz*w_kz
                                        if(freq_pow_scale .ne. 0.) then
                                          w_kz=w_kz*w_data**freq_pow_scale
                                        end if
                                     end if
                                  !}
                                  else
                                  !{
                                     w_kz=0.
                                  !}
                                  end if
                                  !}
                               case default
                                  call seperr('Unknown value for amp_weights')
                               end select

                               if(i_k_hx_p1 <= data_par%n_hx_pad) then

                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) = &
                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) + &
                                  ((1.-w_k_hx)*hx_shift_forw(i_k_hx)* &
                                  data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)+ &
                                  (   w_k_hx)*hx_shift_forw(i_k_hx_p1)* &
                                  data_block(i_k_mx,i_k_my,i_k_hx_p1,i_k_hy,i_w_block_ser,i_w_block_par))* &
                                  scale_image*w_kz
                               else
!if((omp_get_thread_num() == 0)) write(0,*)'HERE #3'
                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) = &
                                  image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) + &
                                  ((1.-w_k_hx)* &
                                  data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par))* &
                                  scale_image*hx_shift_forw(i_k_hx)*w_kz
                               end if

                               !}
                            end if

                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                ! if(omp_get_thread_num() == 0) write(0,*)'#1'
                !}
             end do
             !}
          case default
             call seperr('Unknown value for cig_forw_transf')
          end select
          !}
       else
          !{


          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             d_w_data=down_par%n_w_block_par*data_par%d_w
             w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w
             w_data=w_0_data+(i_w_block_ser-1)*d_w_data

             do i_k_my=1,data_par%n_my_pad
                !{
                do i_k_mx=1,data_par%n_mx_pad
                   !{
                      k_mx_sq=k_mx(i_k_mx)**2
             do i_k_hy=1,data_par%n_hy_pad
                !{
                do i_k_hx=1,data_par%n_hx_pad
                   !{
                      k_hx_sq=k_hx(i_k_hx)**2

                      if(image_par%w_min_amp >0.) then
                         w_p_min_amp= .5*vel_const_image* &
 sqrt(kz_min_amp_sq + k_mx_sq+k_hx_sq + (k_mx_sq*k_hx_sq)/kz_min_amp_sq)
                         w_p_max_amp= .5*vel_const_image* &
 sqrt(kz_max_amp_sq + k_mx_sq+k_hx_sq + (k_mx_sq*k_hx_sq)/kz_max_amp_sq)

!                         w_p_min_amp= .5*vel_const_image* &
! sqrt(kz_min_amp_sq)
!                         w_p_max_amp= .5*vel_const_image* &
! sqrt(kz_max_amp_sq)
                       else
                         w_p_min_amp=-1
                         w_p_max_amp=huge(1.)
                       end if

                       if((w_data >= w_p_min_amp) .and. &
                          (w_data <= w_p_max_amp) ) then

                           if(freq_pow_scale .ne. 0.) then
                               w_kz=w_data**freq_pow_scale
                           else
                               w_kz=1.
                           end if

   image_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_z_block,i_w_block_par) = &
      image_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_z_block,i_w_block_par) + &
      data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par)*&
       scale_image*hx_shift_forw(i_k_hx)*w_kz
!                    scale_image*hx_shift_forw(i_k_hx)

                       end if
                   !}
                end do
                !}
             end do
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       end if
       !}
    end if

  end subroutine out_image_block_ser

  !-----------------------------------------------------------------------------
  recursive subroutine inp_image_block_ser(i_w_block_par,i_z_block,z_down, &
  gen_par,data_par,down_par,image_par,slow_par, &
  data_block, image_block,slow_data_block,amp_stat_block)

    !+++ Not modified yet for full-3D CRA

    integer i_z_block,i_w_block_par
    type (gendown3d_param_type) gen_par
    type (data_param_type) data_par
    type (down_process_param_type) :: down_par
    type (image_param_type) image_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_data_block
    real, dimension (:,:,:,:,:,:) :: amp_stat_block
    complex, dimension (:,:,:,:,:,:) :: data_block
    complex, dimension (:,:,:,:,:,:) :: image_block

    integer i_k_mx,i_k_my,i_k_hx,i_k_hx_p1,i_k_hy,i_px,i_px_p1,i_w_block_ser
    real px,py,inv_w_data,d_w_data,w_0_data,w_data,inv_d_px,w_px
    real k0,k0_sq,kz_s_sq,kz_r_sq,w_kz,z_down
    real kk_hx,inv_d_k_hx,w_k_hx,k_s,k_r,kz_r,kz_s
    real slow_const_image

    if(gen_par%slow_const_image > 0) then
       slow_const_image=gen_par%slow_const_image
    else
       slow_const_image=slow_data_block(1,1,1,1,1,i_z_block)
    end if

    if(gen_par%image_zero_offset) then
       !{
       select case (gen_par%amp_weights(1:1))
       case ('n')
          !{
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             do i_k_hy=1,data_par%n_hy_pad
                !{
                do i_k_hx=1,data_par%n_hx_pad
                   !{
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                   image_block(:,:,1,1,i_z_block,i_w_block_par)* &
                   scale_image*hx_shift_back(i_k_hx)
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       case ('j')
          !{
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             do i_k_hy=1,data_par%n_hy_pad
                !{
                do i_k_hx=1,data_par%n_hx_pad
                   !{
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                   image_block(:,:,1,1,i_z_block,i_w_block_par)* &
                   scale_image*hx_shift_back(i_k_hx)
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       case ('t')
          !{
          d_w_data=down_par%n_w_block_par*data_par%d_w
          w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             w_data=w_0_data+(i_w_block_ser-1)*d_w_data
             k0_sq= (w_data*slow_const_image)**2
             do i_k_hy=1,data_par%n_hy_pad
                !{
                do i_k_hx=1,data_par%n_hx_pad
                   !{
                   do i_k_my=1,data_par%n_my_pad
                      !{
                      do i_k_mx=1,data_par%n_mx_pad
                         !{
                         kz_s_sq=k0_sq - k_sr_sq(i_k_mx,i_k_my,i_k_hx,1)
                         kz_r_sq=k0_sq - k_sr_sq(i_k_mx,i_k_my,i_k_hx,2)
                         ! Jacobian
                         if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then 
                            w_kz=w_data*(sqrt(kz_s_sq)+sqrt(kz_r_sq))/sqrt((kz_s_sq*kz_r_sq))
                         else
                            w_kz=1.
                         end if
                         data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                         data_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                         image_block(i_k_mx,i_k_my,1,1,i_z_block,i_w_block_par)* &
                         scale_image*hx_shift_back(i_k_hx)*w_kz
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       case default
          call seperr('Unknown value for amp_weights')
       end select

       !}
    else
       !{
       if(gen_par%image_cig) then
          !{
          select case (gen_par%cig_forw_transf(1:3))

          case('pus')
             !{
             select case (gen_par%amp_weights(1:1))
             case ('n')
                !{
                !}
             case ('j')
                !{
                call seperr('Not implemented value for amp_weights')
                !}
             case ('t')
                !{
                call seperr('Not implemented value for amp_weights')
                !}
             case default
                call seperr('Unknown value for amp_weights')
             end select

             inv_d_px=1./image_par%d_px_image
             d_w_data=down_par%n_w_block_par*data_par%d_w
             w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w
             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                inv_w_data=1./(w_0_data+(i_w_block_ser-1)*d_w_data)
                do i_k_hy=1,data_par%n_hy_pad
                   !{
                   do i_k_hx=1,data_par%n_hx_pad
                      !{
                      px=(k_hx(i_k_hx)*inv_w_data-image_par%px_image_0)*inv_d_px+1
                      i_px=int(px)
                      w_px=px-i_px
                      if( (i_px > 0) .and. (i_px <= image_par%n_px_image)) then
                         !            i_px_p1=min(i_px+1,image_par%n_px_image);
                         i_px_p1=i_px+1

                         if(i_px_p1 <= image_par%n_px_image) then
                            data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                            data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                            (1.-w_px)*image_block(:,:,i_px   ,1,i_z_block,i_w_block_par)* &
                            scale_image*hx_shift_back(i_k_hx)+ &
                            (   w_px)*image_block(:,:,i_px_p1,1,i_z_block,i_w_block_par)* &
                            scale_image*hx_shift_back(i_k_hx)
                         else
                            data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                            data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                            (1.-w_px)*image_block(:,:,i_px   ,1,i_z_block,i_w_block_par)* &
                            scale_image*hx_shift_back(i_k_hx)
                         end if
                      end if
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          case('pul')
             !{
             d_w_data=down_par%n_w_block_par*data_par%d_w
             w_0_data=down_par%w_min_block+(i_w_block_par-1)*data_par%d_w
             inv_d_k_hx=1./data_par%d_khx
             do i_w_block_ser=1,down_par%n_w_block_ser
                !{
                w_data=w_0_data+(i_w_block_ser-1)*d_w_data
                k0= (w_data*slow_const_image)
                k0_sq= (k0)**2
                do i_k_my=1,data_par%n_my_pad
                   !{
                   do i_k_mx=1,data_par%n_mx_pad
                      !{
                      do i_px=1,image_par%n_px_image
                         !{
                         px=image_par%px_image_0+(i_px-1)*image_par%d_px_image + &
                         epsi_px
                         kk_hx=px*w_data
                         if((kk_hx >=0.).and.(kk_hx < k_hx_max)) then
                            kk_hx=kk_hx*inv_d_k_hx+1
                            i_k_hx=int(kk_hx)
                            w_k_hx=kk_hx-i_k_hx
                         else
                            if((kk_hx <0.).and.(kk_hx >= k_hx_min)) then
                               kk_hx=data_par%n_hx_pad+kk_hx*inv_d_k_hx+1
                               i_k_hx=int(kk_hx)
                               w_k_hx=kk_hx-i_k_hx
                            else
                               i_k_hx=0
                            end if
                         end if
                         if(i_k_hx > 0) then
                            !{
                            i_k_hx_p1=i_k_hx+1
                            if(i_k_hx == data_par%n_hx_pad) i_k_hx_p1=1

                            select case (gen_par%amp_weights(1:1))
                            case ('n')
                               !{
                               w_kz=1.
                               !}
                            case ('j')
                               !{
                               w_kz=w_data
                               !}
                            case ('u')
                               !{
                               k_s=.5*(k_mx(i_k_mx) - px*w_data)
                               k_r=.5*(k_mx(i_k_mx) + px*w_data)
                               kz_s_sq=k0_sq - k_s**2
                               kz_r_sq=k0_sq - k_r**2

                               if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then 
                                  kz_s=sqrt(kz_s_sq)
                                  kz_r=sqrt(kz_r_sq)

                                  w_kz=sqrt( &
                                  w_data* &
                                  (kz_s*(slow_const_image*k0 - .5*px*k_r) + &
                                  kz_r*(slow_const_image*k0 + .5*px*k_s))/&
                                  (kz_s*kz_r) &
                                  )
                               else
                                  w_kz=0.
                               end if
                               !}
                            case ('i')
                               !{
                               w_kz=w_data
                               !}
                            case ('t')
                               !{
                               k_s=.5*(k_mx(i_k_mx) - px*w_data)
                               k_r=.5*(k_mx(i_k_mx) + px*w_data)
                               kz_s_sq=k0_sq - k_s**2
                               kz_r_sq=k0_sq - k_r**2

                               if((kz_s_sq > epsi_kz) .and. (kz_r_sq > epsi_kz)) then
                                  kz_s=sqrt(kz_s_sq)
                                  kz_r=sqrt(kz_r_sq)

                                  w_kz=(w_data*k0_sq)/(4.*kz_s*kz_r)
                               else
                                  w_kz=0.
                               end if
                               !}
                            case default
                               call seperr('amp_weights == inverse does not make sense for modeling')
                            end select

                            data_block(i_k_mx,i_k_my,i_k_hx,1,i_w_block_ser,i_w_block_par)= &
                            data_block(i_k_mx,i_k_my,i_k_hx,1,i_w_block_ser,i_w_block_par)+ &
                            (1.-w_k_hx)*hx_shift_back(i_k_hx)*w_kz*scale_image* &
                            image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par)

                            if(i_k_hx_p1 <= data_par%n_hx_pad) then

                               data_block(i_k_mx,i_k_my,i_k_hx_p1,1,i_w_block_ser,i_w_block_par)= &
                               data_block(i_k_mx,i_k_my,i_k_hx_p1,1,i_w_block_ser,i_w_block_par)+ &
                               (   w_k_hx)*hx_shift_back(i_k_hx_p1)*w_kz*scale_image* &
                               image_block(i_k_mx,i_k_my,i_px,1,i_z_block,i_w_block_par) 
                            end if
                            !}
                         end if

                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          case default
             call seperr('Unknown value for cig_forw_transf')
          end select
          !}
       else
          !{
          do i_w_block_ser=1,down_par%n_w_block_ser
             !{
             do i_k_hy=1,data_par%n_hy_pad
                !{
                do i_k_hx=1,data_par%n_hx_pad
                   !{
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) = &
                   data_block(:,:,i_k_hx,i_k_hy,i_w_block_ser,i_w_block_par) + &
                   image_block(:,:,i_k_hx,i_k_hy,i_z_block,i_w_block_par)* &
                   scale_image
                   !                  scale_image*hx_shift_back(i_k_hx)
                   !}
                end do
                !}
             end do
             !}
          end do
          !}
       end if
       !}
    end if

  end subroutine inp_image_block_ser


  !-----------------------------------------------------------------------------
  subroutine out_image_block_par(down_par,data_par,image_block)

    type (down_process_param_type) :: down_par
    type (data_param_type) data_par
    complex, dimension (:,:,:,:,:,:) :: image_block

    integer i_w_block_par,i_z_block,i_k_hy,i_k_hx,i_k_my,i_k_mx

    if(save_memory) then 
      !{

      do i_w_block_par=2,down_par%n_w_block_par
         !{
        do i_z_block=1,down_par%n_z_block
           !{
         do i_k_hy=1,data_par%n_hy_pad
            !{
            do i_k_hx=1,data_par%n_hx_pad
               !{
             do i_k_my=1,data_par%n_my_pad
                !{
                do i_k_mx=1,data_par%n_mx_pad
                   !{
         image_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_z_block,1) =  &
         image_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_z_block,1) + &
         image_block(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_z_block,i_w_block_par)
                   !}
                end do
                !}
             end do
               !}
            end do
            !}
         end do
         !}
        end do
       !}
      end do
      !}
    else
      !{
 
      do i_w_block_par=2,down_par%n_w_block_par
         !{
  !write(0,*)'in  out_image_block: thread number',omp_get_thread_num()
  !write(0,*)'in  out_image_block: i_w_block_par',i_w_block_par
         image_block(:,:,:,:,:,1) =  &
         image_block(:,:,:,:,:,1) + &
         image_block(:,:,:,:,:,i_w_block_par)
         !}
      end do
      !}
    end if

  end subroutine out_image_block_par


  !-----------------------------------------------------------------------------
  subroutine inp_image_block_par(down_par,data_par,image_block)

    type (down_process_param_type) :: down_par
    type (data_param_type) data_par
    complex, dimension (:,:,:,:,:,:) :: image_block

    integer i_w_block_par

    do i_w_block_par=2,down_par%n_w_block_par
       !{
       image_block(:,:,:,:,:,i_w_block_par) = image_block(:,:,:,:,:,1) 
       !}
    end do

  end subroutine inp_image_block_par



  !-----------------------------------------------------------------------------
  subroutine clean_io_image()
    integer ierr

    if(allocated(hx_shift_forw)) then
       deallocate (hx_shift_forw,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hx_shift_forw')
    end if
    if(allocated(hx_shift_back)) then
       deallocate (hx_shift_back,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hx_shift_back')
    end if
    if(allocated(k_hx)) then
       deallocate (k_hx,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating k_hx')
    end if
    if(allocated(k_hy)) then
       deallocate (k_hy,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating k_hy')
    end if
    if(.not. save_memory) then 
      if(allocated(rimage_slice_2d)) then
         deallocate (rimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_2d')
      end if
      if(allocated(rimage_slice_4d)) then
         deallocate (rimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
      if(allocated(cimage_slice_2d)) then
         deallocate (cimage_slice_2d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating cimage_slice_2d')
      end if
      if(allocated(cimage_slice_4d)) then
         deallocate (cimage_slice_4d,stat=ierr)  
         if(ierr /= 0) call seperr('Problems with deallocating rimage_slice_4d')
      end if
    end if


    return
  end subroutine clean_io_image


end module image_mod
