module sep_90_mod 
!<
!sep_90_mod
!
!Description
! This module contains the basic building blocks  of the sepf90 library.
!Comments
! The basic sep_90 type is made up if
!   a. marker (data_type) indicating data type (real,integer,complex,
!       v_real,v_integer, or v_complex)
!   b. three arrays (rtraces, itraces, and ctraces) only one of which
!      will be allocated for a given dataset
!    i. the _traces structures contain 1-D pointer vectors of type real,
!        complex, or integer
!   c. a pointer to a sep_space structure (which is somewhat analagous to
!      a vector space, hence the name. Contains:
!    i. a 1-D array of sep_hdr, containing 1-D arrays representing int and
!      real headers (rhdr and ihdr)
!    ii. an array of type key_info which contains :
!     a)  the location within the real or int header array
!     b)  the key name,type,fmt (format)
!   iii. n,o,d, label, and unit arrays that contain the gridding params
!    iv. n_keys,n_int - total number of keys and number of integer keys
!     v. sparse_grid (defaults to true). Indicates whether grid has holes
!
!type rsep_tr {
!  real, pointer, dimension(:)   :: data }
!type csep_tr {
!  complex, pointer, dimension(:)   :: data }
!type isep_tr {
!  integer, pointer, dimension(:)   :: data }
!
!type sep_hdr {
!  real, dimension(:), pointer   :: rhdr
!  integer, dimension(:), pointer   :: ihdr }
!
!type sep_key_info {
!  integer       :: location
!  character(len=128)    :: name,type,fmt  }
!
!type sep_space{
!  type(sep_hdr), dimension(:), pointer   :: headers
!  integer, dimension(:), pointer     :: n
!  real, dimension(:), pointer     :: o,d
!  character(len=128),dimension(:),pointer  :: label,unit
!  type(sep_key_info),dimension(:),pointer :: keys
!  logical :: sparse_grid
!  integer         :: n_keys,n_int }
!
!
!type sep_90 {
!	character(len=128) :: data_type
!  type(rsep_tr), dimension(:), pointer   :: rtraces
!  type(csep_tr), dimension(:), pointer   :: ctraces
!  type(isep_tr), dimension(:), pointer   :: itraces
!	type(sep_space), pointer               :: space }
!
!
! Category
! Lib:Sepf90:General
!>
    implicit none
  type rsep_tr  
    real, pointer, dimension(:)   :: data
  end type
  type csep_tr  
    complex, pointer, dimension(:)   :: data
  end type
  type isep_tr  
    integer, pointer, dimension(:)   :: data
  end type
  type sep_hdr  
    real, dimension(:), pointer   :: rhdr
    integer, dimension(:), pointer   :: ihdr
  end type
  type sep_key_info  
    integer       :: location
    character(len=128)    :: name,type,fmt
  end type
  type sep_space 
    type(sep_hdr), dimension(:), pointer   :: headers
    integer, dimension(:), pointer     :: n
    real, dimension(:), pointer     :: o,d
    character(len=128),dimension(:),pointer  :: label,unit
    type(sep_key_info),dimension(:),pointer :: keys
    logical :: sparse_grid
    integer         :: n_keys,n_int
  end type
  type sep_90  
    character(len=128) :: data_type
    type(rsep_tr), dimension(:), pointer   :: rtraces
    type(csep_tr), dimension(:), pointer   :: ctraces
    type(isep_tr), dimension(:), pointer   :: itraces
    type(sep_space), pointer               :: space
  end type
end module 
