module ComAz_salt_ssf_mod

! Type definitions for this module
use ComAz_salt_ssf_types

! Types definition for Gendown package
use gendown3d_types
use down_process_types
use data_types
use image_types
use slow_types

! FFT's module
use fft_data_mod

use bound_mod

implicit none

type (ComAz_salt_ssf_param_type), private :: ComAz_salt_ssf_par_loc
type (fft_type), private :: fft_1d_id,fft_2d_id
real, private :: slow_const,slow_salt,epsi,min_prop_cos_sq
real, private :: d_k_norm,min_lap_sq,max_disp_rel
real, allocatable, dimension (:), private :: magn,phase
real, allocatable, dimension (:), private :: k_mx,k_my_sq,k_hx
real, allocatable, dimension (:,:), private :: k_r_sq,k_s_sq

contains

! copies parameters into a structure private to the module 
subroutine init_ComAz_salt_ssf_par(ComAz_salt_ssf_par)
type (ComAz_salt_ssf_param_type) ComAz_salt_ssf_par

ComAz_salt_ssf_par_loc=ComAz_salt_ssf_par

end subroutine init_ComAz_salt_ssf_par

! Initiliazes parameters and allocate and initialize arrays
function ComAz_salt_ssf_init(gen_par,data_par,slow_par)  result(info)
type (gendown3d_info_type) info
type (gendown3d_param_type) gen_par
type (data_param_type) data_par
type (slow_param_type) slow_par

integer i_k_mx,i_k_my,i_k_hx,ierr,i_k_norm
real kmx,kmy,khx,pi2,k_norm_sq,k_norm
real n_cycles_ksq,k_start,k_cutoff,k_decay,sigma_magn_sq,edge_gauss,gauss_magn
real delta_k

!info=info_default already initialized in declaration of type

if(slow_par%n_comp_slow /= 2) then
  write(0,*)'Reset n_comp_slow to 2 (It should be changed!)'
  slow_par%n_comp_slow = 2
end if

! constant slowness
slow_const=gen_par%slow_const

! set pi and epsi
pi2=2.*acos(-1.)
epsi=epsilon(1.)

! set maximum propagation angle 
min_prop_cos_sq=cos(gen_par%max_propag_angle*pi2/360.)**2

! set slow_salt to positive if defined, otherwise to negative
if (slow_par%slowness) then
  if(ComAz_salt_ssf_par_loc%slow_salt > 0.) then
    slow_salt=1.01*(ComAz_salt_ssf_par_loc%slow_salt)
    if(ComAz_salt_ssf_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
  else 
    slow_salt=-1.
  end if
else
  if(ComAz_salt_ssf_par_loc%vel_salt > 0.) then
    slow_salt=1.01*(1./ComAz_salt_ssf_par_loc%vel_salt)
    if(ComAz_salt_ssf_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
  else 
    slow_salt=-1.
  end if
end if

! initialize FFTs
!fft_1d_id=init_fft_1d(data_par%n_hx_pad)
call init_fft_1d(fft_1d_id,data_par%n_hx_pad)
!fft_2d_id=init_fft_2d(data_par%n_mx_pad,data_par%n_my_pad)
call init_fft_2d(fft_2d_id,data_par%n_mx_pad,data_par%n_my_pad)
info%status=fft_2d_id%ierr

! check that there are no cross-line offsets
if(data_par%n_hy_pad /= 1 ) then
  call seperr(" ComAz_salt: Trying to use ComAz_saltack migration with n_hy > 1")
end if

! allocate tables for square-root operators 
allocate (phase(ComAz_salt_ssf_par_loc%n_disp_rel),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of phase ")
allocate (magn(ComAz_salt_ssf_par_loc%n_disp_rel),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of magn ")

! allocate tables for wavenumbers
allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
allocate (k_my_sq(data_par%n_my_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_my_sq ")
allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")

allocate (k_r_sq(data_par%n_mx_pad,data_par%n_hx_pad), &
         stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_r_sq ")
allocate (k_s_sq(data_par%n_mx_pad,data_par%n_hx_pad), &
         stat=ierr)  
if(ierr .ne. 0) call seperr("Problems with allocation of k_s_sq ")


! init k_mx table
do i_k_mx=1,data_par%n_mx_pad
  kmx  = i_k_mx - 1.
  if(kmx <= float(data_par%n_mx_pad/2)) then
    k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
  else
    k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
  end if
end do

! init k_my table
do i_k_my=1,data_par%n_my_pad
  kmy = i_k_my - 1.
  if(kmy <= float(data_par%n_my_pad/2)) then
    k_my_sq(i_k_my) =  (data_par%kmy_0+data_par%d_kmy* kmy)**2
  else
    k_my_sq(i_k_my) = (-data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad))**2
  end if
end do

! init k_hx table
if(gen_par%single_hx_dip) then
  do i_k_hx=1,data_par%n_hx_pad
    khx = i_k_hx - 1.
    k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
  end do
else
  do i_k_hx=1,data_par%n_hx_pad
    khx = i_k_hx - 1.
    if(khx <= float(data_par%n_hx_pad/2)) then
      k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
    else
      k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
    end if
  end do
end if

! init source and geophone components of wavenumbers
do i_k_hx=1,data_par%n_hx_pad
  do i_k_mx=1,data_par%n_mx_pad
    k_s_sq(i_k_mx,i_k_hx)= &
          .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
    k_r_sq(i_k_mx,i_k_hx)= &
          .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)
  end do
end do


! now initialize square-root operator tables
n_cycles_ksq=5.
k_cutoff =0.
delta_k=sqrt(min_prop_cos_sq)
k_start=delta_k
k_decay=-1.*delta_k
if(k_decay**2 > n_cycles_ksq) call seperr('k_decay^2 > n_cycles_ksq')
edge_gauss = exp(k_decay)
sigma_magn_sq=-(k_decay-k_start)**2/(2.*log(edge_gauss))

d_k_norm=n_cycles_ksq/(ComAz_salt_ssf_par_loc%n_disp_rel-1)
do i_k_norm=1,ComAz_salt_ssf_par_loc%n_disp_rel
  k_norm_sq=1.-(i_k_norm-1)*d_k_norm
  if( k_norm_sq > 0.) then
    k_norm=sqrt(k_norm_sq)
  else
    k_norm=-sqrt(-k_norm_sq)
  end if

  phase(i_k_norm)=k_norm

  if(k_norm > k_start) then
    magn(i_k_norm)=0.
  else
    magn(i_k_norm)=(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
  end if
end do
max_disp_rel=float(ComAz_salt_ssf_par_loc%n_disp_rel)

!call snap('Phase.h',ComAz_salt_ssf_par_loc%n_disp_rel,1,phase)
!call snap('Magn.h',ComAz_salt_ssf_par_loc%n_disp_rel,1,magn)

end function ComAz_salt_ssf_init

! downward continue wavefield by one depth step
function ComAz_salt_ssf_down_oper(adj,add, &
n_w_data,w_0_data,d_w_data, &
z_down,d_z_down,data_par, &
data_slice,ddata_slice,slow_slice) &
result(stat)

  integer stat,n_w_data
  logical add,adj,found_salt
  real w_0_data,d_w_data,z_down,d_z_down
  type (data_param_type) data_par
  real, dimension (:,:,:,:,:) :: slow_slice
  complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
  optional data_slice,ddata_slice,slow_slice

  logical prop_data,prop_ddata,var_slow,prop_salt
  integer i_k_mx,i_k_my,i_k_hx,ierr,n_mx,n_my,n_hx,i_mx,i_my,i_hx,i_ksq_norm,i_w
  integer i_slow,n_slow_split_eff,n_slow_split_no_salt,map_slow
  real k0_sq,rphase,rphase_s,rphase_r,rmagn,dphase,w_data,k_sq_comaz
  real min_slow,max_slow,delta_slow,slow_0,inv_delta_slow,range_slow,mean_slow
  real sigma,rat,dscat,dz_w,pi2,norm_ksq,k0,k0_dz,slow
  complex im_dz_w,ddscat

  complex, allocatable,  dimension (:,:,:,:) :: temp_data_1
  complex, allocatable,  dimension (:,:,:,:) :: temp_data_2

  real,  dimension(ComAz_salt_ssf_par_loc%n_slow_split)      :: slow_prop
  real,  dimension (data_par%n_mx_pad,data_par%n_hx_pad) :: alpha
  complex, dimension (ComAz_salt_ssf_par_loc%n_disp_rel) :: ph_shift_tab

  ! set constants
  n_mx=data_par%n_mx_pad
  n_my=data_par%n_my_pad
  n_hx=data_par%n_hx_pad
  pi2=2.*acos(-1.)

  ! if data_slice is present propagate background wavefield
  prop_data= present(data_slice)

  ! if ddata_slice is present propagate scattered wavefield
  prop_ddata= present(ddata_slice)

  ! if slow_slice is present use variable velocity, otherwise use slow_const
  var_slow= present(slow_slice)


  ! allocate temporary arrays for wavefield copies
  if(prop_data) then
     allocate(temp_data_1(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,1),&
     stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of temp_data_1")
     allocate(temp_data_2(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,1),&
     stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of temp_data_2")
  end if
  if(prop_ddata) then
     call seperr('ComAz_salt does not support scattering yet')
  end if


  if(var_slow) then
     !{
     ! set  min_slow, max_slow  and n_slow_split_no_salt 
     if(slow_salt > 0.) then 
        ! salt slowness was defined 
        min_slow=minval(slow_slice,mask=(slow_slice > slow_salt))
        found_salt=any(slow_slice <= slow_salt)
        if(found_salt) then 
           ! salt found in this depth slice
           n_slow_split_no_salt=ComAz_salt_ssf_par_loc%n_slow_split-1
        else
           ! salt NOT found in this depth slice
           n_slow_split_no_salt=ComAz_salt_ssf_par_loc%n_slow_split
        end if
     else
        ! salt slowness was NOT defined 
        found_salt=.false.
        min_slow=minval(slow_slice)
        n_slow_split_no_salt=ComAz_salt_ssf_par_loc%n_slow_split
     end if
     max_slow=maxval(slow_slice)

     ! set  inv_delta_slow and n_slow_split_no_salt 
     range_slow=max_slow-min_slow
     mean_slow=.5*(max_slow+min_slow)
     delta_slow=range_slow/n_slow_split_no_salt
     if(delta_slow < ComAz_salt_ssf_par_loc%min_delta_slow*mean_slow) then
        if(range_slow < ComAz_salt_ssf_par_loc%min_delta_slow*mean_slow) then
           n_slow_split_no_salt= 1 
           delta_slow=range_slow
           inv_delta_slow=0.
        else
           n_slow_split_no_salt= & 
           min(n_slow_split_no_salt, &
           max(1,int(range_slow/(ComAz_salt_ssf_par_loc%min_delta_slow*mean_slow))))
           delta_slow=range_slow/n_slow_split_no_salt
           inv_delta_slow=1./delta_slow
        end if
     else
        inv_delta_slow=1./delta_slow
     end if

     ! set velocitues outside salt
     !  write(0,*)'delta_slow=',delta_slow
     !  write(0,*)'min_slow=',min_slow
     do i_slow=1,n_slow_split_no_salt
        slow_prop(i_slow)=min_slow+(i_slow-.5)*delta_slow
        !  write(0,*)'slow_prop(i_slow)=',slow_prop(i_slow)
     end do

     ! set  n_slow_split_eff 
     if(found_salt) then
        n_slow_split_eff= n_slow_split_no_salt + 1
        slow_prop(n_slow_split_eff)=slow_salt
        !  write(0,*)'slow_prop(salt)=',slow_prop(n_slow_split_eff)
     else
        n_slow_split_eff= n_slow_split_no_salt 
     end if
     !}
  else
     !{
     if(ComAz_salt_ssf_par_loc%n_slow_split /= 1) call seperr('n_slow_split /= 1')
     n_slow_split_eff= 1 
     slow_prop(1)=slow_const
     !}
  end if

  do i_w=1,n_w_data
     !{ beginning loop over frequencies
     w_data=w_0_data+(i_w-1)*d_w_data
     do i_k_hx=1,n_hx
        !{
        do i_k_my=1,n_my
           !{
           do i_k_mx=1,n_mx
              !{
              temp_data_2(i_k_mx,i_k_my,i_k_hx,1) = &
              data_slice(i_k_mx,i_k_my,i_k_hx,1,i_w) 
              !}
           end do
           !}
        end do
        !}
     end do

     do i_slow=1,n_slow_split_eff
        !{ beginning loop over slowness for sources
        ! set current slowness
        slow_0=slow_prop(i_slow)

        ! determine if in salt
        if(found_salt .and. i_slow == n_slow_split_eff) then
           prop_salt=.true.
        else
           prop_salt=.false.
        end if

        do i_k_hx=1,n_hx
           !{
           do i_k_my=1,n_my
              !{
              do i_k_mx=1,n_mx
                 !{
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1) = &
                 temp_data_2(i_k_mx,i_k_my,i_k_hx,1) 
                 !}
              end do
              !}
           end do
           !}
        end do

        k0= (w_data*slow_0)
        k0_sq= k0*k0
        k0_dz=k0*d_z_down
        dz_w=d_z_down*w_data
        im_dz_w=cmplx(0.,dz_w)

        ! compute wavenumber normalization factor to look-up into square-root table
        if(k0_sq > epsi) then
           norm_ksq=1./(d_k_norm*k0_sq)
        else
           write(0,*)'I should not be here k0_sq=',k0_sq
           write(0,*)'k0_sq=',k0_sq
           write(0,*)'slow_0=',slow_0
           write(0,*)'w_data=',w_data
           call seperr('ComAz_salt_ssf: Zero frequency')
           norm_ksq=2.*ComAz_salt_ssf_par_loc%n_disp_rel/min_lap_sq
        end if

       if(k0_dz >0) then
          ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),phase(:)))
       else
          ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),-phase(:)))
       end if

        if((var_slow) .and. (adj)) then
           !{
           ! not needed for migration
           call seperr('wrong option  1')
           !}
        end if

        ! compute phase shift at the sources for current slowness using tables
        do i_k_hx=1,n_hx
           !{
           do i_k_mx=1,n_mx
              !{
              ! in-line source wavenumber
              i_ksq_norm=nint(min(k_s_sq(i_k_mx,i_k_hx)*norm_ksq+1,max_disp_rel))
              rphase_s=(phase(i_ksq_norm))
              ! in-line receiver wavenumber
              i_ksq_norm=nint(min(k_r_sq(i_k_mx,i_k_hx)*norm_ksq+1,max_disp_rel))
              rphase_r=(phase(i_ksq_norm))
              if((rphase_s > 0.) .and. (rphase_r > 0.)) then
                 ! apply common-azimuth equation if propgating waves
                 alpha(i_k_mx,i_k_hx)=(rphase_s/(rphase_r+rphase_s))**2
              else
                 ! set k_hy to 0 if evanescent.
                 alpha(i_k_mx,i_k_hx)=.25
              end if
              !}
           end do
           if(i_k_hx == data_par%n_hx_pad/2 + 1) alpha(:,i_k_hx)=.25
           !}
        end do

        do i_k_hx=1,n_hx
           !{
           do i_k_my=1,n_my
              !{
              do i_k_mx=1,n_mx
                 !{
                 ! evaluate DSR with common-azimuth k_hy
                 k_sq_comaz=k_s_sq(i_k_mx,i_k_hx)+k_my_sq(i_k_my)*alpha(i_k_mx,i_k_hx)
                 i_ksq_norm=nint(min(k_sq_comaz*norm_ksq+1,max_disp_rel))
!                 rphase=phase(i_ksq_norm)*k0_dz
!                 rmagn=abs(k0_dz)*magn(i_ksq_norm)

                 ! apply phase shift
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1) =  &
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1)*ph_shift_tab(i_ksq_norm)
!                 temp_data_1(i_k_mx,i_k_my,i_k_hx)*cexp(cmplx(rmagn,rphase))
                 !}
              end do
              !}
           end do
           !}
        end do

        if((var_slow) .and. (.not. adj)) then
           !{
           ! call FFTs (two passes to use less memoy)
           !write(0,*)'#1 Before FFT i_w=',i_w
           call fft_3d_two_pass_data( &
           fft_1d_id,fft_2d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data_1(:,:,:,1))
           !write(0,*)'#1 After FFT i_w=',i_w

           ! apply split_step correction and nearest neighborhood interpolation 
           ! according to local velocity
           do i_hx=1,n_hx
              !{
              do i_my=1,n_my
                 !{
                 do i_mx=1,n_mx
                    !{
                    slow=slow_slice(i_mx,i_my,i_hx,1,SOU_COMP)
                    if(prop_salt) then
                       !{
                       if(slow <= slow_salt) then
                          !{
                          dphase=(slow-slow_0)* dz_w
                          data_slice(i_mx,i_my,i_hx,1,i_w)= &
                          temp_data_1(i_mx,i_my,i_hx,1) * cmplx(cos(dphase),sin(dphase))
                          !}
                       end if
                       !}
                    else
                       !{
                       map_slow=min(n_slow_split_no_salt, &
                       int((slow - min_slow)*inv_delta_slow)+1)
                       if(map_slow == i_slow) then 
                          !{
                          dphase=(slow - slow_0)* dz_w
                          data_slice(i_mx,i_my,i_hx,1,i_w)= &
                          temp_data_1(i_mx,i_my,i_hx,1) * cmplx(cos(dphase),sin(dphase))

                          !}
                       end if
                       !}
                    end if
                    !}
                 end do
                 !}
              end do
              !}
           end do
           !} 
        else
           !{
           do i_hx=1,n_hx
              !{
              do i_my=1,n_my
                 !{
                 do i_mx=1,n_mx
                    !{
                    data_slice(i_mx,i_my,i_hx,1,i_w)= temp_data_1(i_mx,i_my,i_hx,1) 
                    !}
                 end do
                 !}
              end do
              !}
           end do
           !} 
        end if

        !} end loop over slowness for sources
     end do

     if((var_slow) .and. (.not. adj)) then
        ! if we applied the split-step correction tranform back to wavenumbers
        !{

        ! apply absorbing boundary taper
        call apply_bound(data_par,data_slice(:,:,:,:,i_w))

        !write(0,*)'#2 Before FFT i_w=',i_w
        call fft_3d_two_pass_data( &
        fft_1d_id,fft_2d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
        !write(0,*)'#2 After FFT i_w=',i_w
        !}
     end if

     do i_k_hx=1,n_hx
        !{
        do i_k_my=1,n_my
           !{
           do i_k_mx=1,n_mx
              !{
              temp_data_2(i_k_mx,i_k_my,i_k_hx,1) = &
              data_slice(i_k_mx,i_k_my,i_k_hx,1,i_w) 
              !}
           end do
           !}
        end do
        !}
     end do

     do i_slow=1,n_slow_split_eff
        !{ beginning loop over slowness for receivers

        ! set current slowness
        slow_0=slow_prop(i_slow)

        ! determine if in salt
        if(found_salt .and. i_slow == n_slow_split_eff) then
           prop_salt=.true.
        else
           prop_salt=.false.
        end if

        do i_k_hx=1,n_hx
           !{
           do i_k_my=1,n_my
              !{
              do i_k_mx=1,n_mx
                 !{
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1) = &
                 temp_data_2(i_k_mx,i_k_my,i_k_hx,1) 
                 !}
              end do
              !}
           end do
           !}
        end do

        k0= (w_data*slow_0)
        k0_sq= k0*k0
        k0_dz=k0*d_z_down
        dz_w=d_z_down*w_data
        im_dz_w=cmplx(0.,dz_w)

        ! compute wavenumber normalization factor to look-up into square-root table
        if(k0_sq > epsi) then
           norm_ksq=1./(d_k_norm*k0_sq)
        else
           write(0,*)'I should not be here k0_sq=',k0_sq
           write(0,*)'k0_sq=',k0_sq
           write(0,*)'slow_0=',slow_0
           write(0,*)'w_data=',w_data
           call seperr('ComAz_salt_ssf: Zero frequency')
           norm_ksq=2.*ComAz_salt_ssf_par_loc%n_disp_rel/min_lap_sq
        end if

         if(k0_dz >0) then
            ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),phase(:)))
         else
            ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),-phase(:)))
         end if

        if((var_slow) .and. (adj)) then
           !{
           ! not needed for migration
           call seperr('wrong option  1')
           !}
        end if

        ! compute phase shift at the receivers for current slowness using tables
        do i_k_hx=1,n_hx
           !{
           do i_k_mx=1,n_mx
              !{
              ! in-line source wavenumber
              i_ksq_norm=nint(min(k_s_sq(i_k_mx,i_k_hx)*norm_ksq+1,max_disp_rel))
              rphase_s=(phase(i_ksq_norm))
              ! in-line receiver wavenumber
              i_ksq_norm=nint(min(k_r_sq(i_k_mx,i_k_hx)*norm_ksq+1,max_disp_rel))
              rphase_r=(phase(i_ksq_norm))
              if((rphase_s > 0.) .and. (rphase_r > 0.)) then
                 ! apply common-azimuth equation if propagating waves
                 alpha(i_k_mx,i_k_hx)=(rphase_r/(rphase_r+rphase_s))**2
              else
                 ! set k_hy to 0 if evanescent.
                 alpha(i_k_mx,i_k_hx)=.25
              end if
              !}
           end do
           if(i_k_hx == data_par%n_hx_pad/2 + 1) alpha(:,i_k_hx)=.25
           !}
        end do

        do i_k_hx=1,n_hx
           !{
           do i_k_my=1,n_my
              !{
              do i_k_mx=1,n_mx
                 !{
                 ! evaluate DSR with common-azimuth k_hy
                 k_sq_comaz=k_r_sq(i_k_mx,i_k_hx)+k_my_sq(i_k_my)*alpha(i_k_mx,i_k_hx)
                 i_ksq_norm=nint(min(k_sq_comaz*norm_ksq+1,max_disp_rel))
!                 rphase=phase(i_ksq_norm)*k0_dz
!                 rmagn=abs(k0_dz)*magn(i_ksq_norm)

                 ! apply phase shift
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1) =  &
                 temp_data_1(i_k_mx,i_k_my,i_k_hx,1)*ph_shift_tab(i_ksq_norm)
!                 temp_data_1(i_k_mx,i_k_my,i_k_hx)*cexp(cmplx(rmagn,rphase))

                 !}
              end do
              !}
           end do
           !}
        end do

        if((var_slow) .and. (.not. adj)) then
           !{
           ! call FFTs (two passes to use less memoy)
           !write(0,*)'#3 Before FFT i_w=',i_w
           call fft_3d_two_pass_data( &
           fft_1d_id,fft_2d_id,FFT_BACK,n_mx,n_my,n_hx,temp_data_1(:,:,:,1))
           !write(0,*)'#3 After FFT i_w=',i_w

           ! apply split_step correction and nearest neighborhood interpolation 
           ! according to local velocity
           do i_hx=1,n_hx
              !{
              do i_my=1,n_my
                 !{
                 do i_mx=1,n_mx
                    !{
                    slow=slow_slice(i_mx,i_my,i_hx,1,REC_COMP)
                    if(prop_salt) then
                       !{
                       if(slow <= slow_salt) then
                          !{
                          dphase=(slow-slow_0)* dz_w
                          data_slice(i_mx,i_my,i_hx,1,i_w)= &
                          temp_data_1(i_mx,i_my,i_hx,1) * cmplx(cos(dphase),sin(dphase))
                          !}
                       end if
                       !}
                    else
                       !{
                       map_slow=min(n_slow_split_no_salt, &
                       int((slow - min_slow)*inv_delta_slow)+1)
                       if(map_slow == i_slow) then 
                          !{
                          dphase=(slow - slow_0)* dz_w
                          data_slice(i_mx,i_my,i_hx,1,i_w)= &
                          temp_data_1(i_mx,i_my,i_hx,1) * cmplx(cos(dphase),sin(dphase))

                          !}
                       end if
                       !}
                    end if
                    !}
                 end do
                 !}
              end do
              !}
           end do

           !} 
        else
           !{
           do i_hx=1,n_hx
              !{
              do i_my=1,n_my
                 !{
                 do i_mx=1,n_mx
                    !{
                    data_slice(i_mx,i_my,i_hx,1,i_w)= temp_data_1(i_mx,i_my,i_hx,1) 
                    !}
                 end do
                 !}
              end do
              !}
           end do
           !} 
        end if

        !} end loop over slowness for receivers
     end do

     if((var_slow) .and. (.not. adj)) then
        !{
        ! if we applied the split-step correction tranform back to wavenumbers
        !write(0,*)'#4 Before FFT i_w=',i_w

        ! apply absorbing boundary taper
        call apply_bound(data_par,data_slice(:,:,:,:,i_w))

        call fft_3d_two_pass_data( &
        fft_1d_id,fft_2d_id,FFT_FORW,n_mx,n_my,n_hx,data_slice(:,:,:,1,i_w))
        !write(0,*)'#4 After FFT i_w=',i_w
        !}
     end if

     !}
  end do

  if(prop_data) then
     deallocate (temp_data_1,stat=ierr)  
     deallocate (temp_data_2,stat=ierr)  
  end if

  stat=0

  return	
end function ComAz_salt_ssf_down_oper


function ComAz_salt_ssf_scatter_oper(adj,add, &
                     n_w_data,w_0_data,d_w_data, &
                     z_down,d_z_down,data_par, &
                     data_slice,ddata_slice,dslow_slice,slow_slice) &
         result(stat)

integer stat,n_w_data
logical add,adj
real w_0_data,d_w_data,z_down,d_z_down
type (data_param_type) data_par
real, dimension (:,:,:,:,:) :: slow_slice
complex, dimension (:,:,:,:,:) :: dslow_slice
complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
optional slow_slice

call seperr('ComAz_salt_ssf_scatter_oper not supported yet!')

stat=0

return	
end function ComAz_salt_ssf_scatter_oper

end module ComAz_salt_ssf_mod
