module Off2Ang3D_mod

  use off2ang3d_types

  use fft_data_mod

  implicit none

  logical, private :: ang_2d,reciprocal_hx,reciprocal_hx_wrap,nam_adcig
  logical, private :: apply_jac_gamma,apply_jac_azim_1,apply_jac_azim_2
  logical, private :: apply_jac_both,apply_jac_none
  logical, private :: jac_nowrap,space_domain,comaz_data
  character(len=128),private :: tag_output
  integer, external ::  sreed,srite,sseek_block

  integer, private :: n_kz,n_kmx,n_kmy,n_khx,n_khy
  integer, private :: n_hx,n_hy,n_hx_pad,n_hy_pad
  integer, private :: n_bytes_hx_hy,n_bytes_gamma_azim
  integer, private :: n_threads,n_gamma,n_azim
  integer, private :: i_hx_0
  integer, private :: n_khy_bias

  real, private :: sign_khy,sign_khy_nam,number_aliases
  real, private :: hx_0,hy_0,d_hx,d_hy
  real, private :: d_kz,d_kmx,d_kmy,d_khx,d_khy
  real, private :: kz_0,kmx_0,kmy_0,khx_0,khy_0
  real, private :: azim_0,azim_0_max,gamma_0,d_azim,d_gamma
  real, private :: pi,inv_2pi,inv_pi,delta_azim_gamma_0
  real, private :: k_hx_min,k_hx_max,k_hy_min,k_hy_max
  real, private :: vel_khy_mean,freq_khy_mean,max_mult_khy_mean
  real, private :: khy_bias_min,hx_max,hy_max

  integer, allocatable, dimension (:,:), private :: azim_bound
  real, allocatable, dimension (:), private :: k_mx,k_my,k_z,k_hx,k_hy
  real, allocatable, dimension (:), private :: khy_bias_min_vec,khy_bias_max_vec
  real, allocatable, dimension (:,:), private :: khy_mean,khy_bias_mat
  real, allocatable, dimension (:,:,:), private :: jac_mat
  complex, allocatable, dimension (:,:,:), private :: hx_hy,gamma_azim
  type (fft_type), private :: fft_nhx_nhy_id

contains

  function Off2Ang3D_init(off2ang3d_par)  result(status)
    type (off2ang3d_param_type) off2ang3d_par
    integer status

    integer i_kmx,i_kmy,i_kz,i_khx,i_khy,ierr,i_z
    integer i_khy_bias,i_khy_nint,i_khy_bias_cycles
    real kz,kmx,kmy,khx,khy
    real wavenumb_khy_mean_sq,inv_d_khy,kz_s,kz_r,k_my_min,k_my_max
    real khy_range,khy_mean_min,khy_mean_max
    real khy_bias,khy_bias_max

    status=0

    pi=acos(-1.)
    inv_2pi=1./(2.*pi)
    inv_pi=1./(pi)

    ang_2d=off2ang3d_par%ang_2d
    reciprocal_hx=off2ang3d_par%reciprocal_hx
    reciprocal_hx_wrap=off2ang3d_par%reciprocal_hx_wrap
    nam_adcig=off2ang3d_par%nam_adcig
    space_domain=off2ang3d_par%space_domain
    comaz_data=off2ang3d_par%comaz_data

    apply_jac_gamma=off2ang3d_par%apply_jac_gamma
    apply_jac_azim_1=off2ang3d_par%apply_jac_azim_1
    apply_jac_azim_2=off2ang3d_par%apply_jac_azim_2
    apply_jac_both=apply_jac_gamma .and. (apply_jac_azim_1.or.apply_jac_azim_2)
    apply_jac_none=.not.(apply_jac_gamma.or.apply_jac_azim_1.or.apply_jac_azim_2)
    jac_nowrap=off2ang3d_par%jac_nowrap
      write(0,*)'apply_jac_both=',apply_jac_both
      write(0,*)'apply_jac_none=',apply_jac_none
      write(0,*)'jac_nowrap=',jac_nowrap

    n_kz=off2ang3d_par%n_kz
    n_kmx=off2ang3d_par%n_kmx
    n_kmy=off2ang3d_par%n_kmy

    n_hx=off2ang3d_par%n_hx
    n_hx_pad=off2ang3d_par%n_hx_pad
    if(n_hx_pad < n_hx) then
      write(0,*)'n_hx_pad=',n_hx_pad
      write(0,*)'n_hx=',n_hx
      call seperr ("n_hx_pad must > n_hx")
    end if
    if((.not. space_domain) .and. (mod(n_hx_pad,2) /=0)) then
      write(0,*)'n_hx_pad=',n_hx_pad
      call seperr ("n_hx_pad must be even")
    end if
    n_khx=n_hx_pad

    n_hy=off2ang3d_par%n_hy
    n_hy_pad=off2ang3d_par%n_hy_pad
    if(n_hy_pad < n_hy) then
      write(0,*)'n_hy_pad=',n_hy_pad
      write(0,*)'n_hy=',n_hy
      call seperr ("n_hy_pad must > n_hy")
    end if
    if((.not.space_domain) .and. (mod(n_hy_pad,2) /=0).and. (n_hy_pad /=1)) then
      write(0,*)'n_hy_pad=',n_hy_pad
      call seperr ("n_hy_pad must be even")
    end if
    n_khy=n_hy_pad
write(0,*)'n_khy=',n_khy

    hx_0=off2ang3d_par%hx_0
    hy_0=off2ang3d_par%hy_0
!    hy_0=0.
    d_hx=off2ang3d_par%d_hx
    d_hy=off2ang3d_par%d_hy

    d_kz=2*pi*off2ang3d_par%d_kz
    d_kmx=2*pi*off2ang3d_par%d_kmx
    d_kmy=2*pi*off2ang3d_par%d_kmy

    d_khx=2.*pi/(d_hx*FLOAT(n_hx_pad))
    d_khy=2.*pi/(d_hy*FLOAT(n_hy_pad))

    kz_0=0.
    kmx_0=0.
    kmy_0=0.
    khx_0=0.
    khy_0=0.

    n_azim=off2ang3d_par%n_azim_max
    azim_0=(pi*off2ang3d_par%azim_0/180.)
    d_azim=(pi*off2ang3d_par%d_azim/180.)
    azim_0_max=(pi*off2ang3d_par%azim_0_max/180.)

    n_gamma=off2ang3d_par%n_gamma
    gamma_0=(pi*off2ang3d_par%gamma_0/180.)
    d_gamma=(pi*off2ang3d_par%d_gamma/180.)


    n_threads=off2ang3d_par%n_threads
!    if(off2ang3d_par%create_coplanar_stack) then
      if(mod(n_kz,n_threads) /=0) then
        write(0,*)'n_threads=',n_threads
        write(0,*)'n_kz=',n_kz
        call seperr ("n_kz be multiple of n_threads")
      end if
!    end if

!    n_bytes_hx=n_hx*n_threads*8
    n_bytes_hx_hy=n_hx*n_hy*n_threads*8
    if(off2ang3d_par%create_coplanar_ang) then
      n_bytes_gamma_azim=n_gamma*1*n_threads*8
    else
      n_bytes_gamma_azim=n_gamma*n_azim*n_threads*8
    end if

    allocate (azim_bound(2,n_gamma),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of azim_bound")

    allocate (k_z(n_kz),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
    allocate (k_mx(n_kmx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_my(n_kmy),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
    allocate (k_hx(n_khx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")
    allocate (k_hy(n_khy),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of k_hy ")

    allocate (jac_mat(n_hx_pad,n_hy_pad,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of jac_mat ")
write(0,*)'after alloc shape jac_mat',shape(jac_mat)
    jac_mat=0.


    allocate (hx_hy(n_hx_pad,n_hy_pad,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of hx_khy ")

write(0,*)'after alloc shape hx_hy',shape(hx_hy)


    allocate (gamma_azim(n_gamma,n_azim,n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of gamma_azim ")

    number_aliases=off2ang3d_par%number_aliases
    sign_khy=off2ang3d_par%sign_khy

! Notice the negative signs because of the signs in Ft3d
! (depends on FFT conventions)
    do i_kz=1,n_kz
       kz  = i_kz - 1.
       if(kz <= float(n_kz/2)) then
!          k_z(i_kz) =  kz_0+d_kz* kz
          k_z(i_kz) =  -(kz_0+d_kz* kz)
       else
!          k_z(i_kz) = -kz_0+d_kz*(kz-n_kz)
          k_z(i_kz) = -(kz_0+d_kz*(kz-n_kz))
       end if
    end do

! Notice the negative signs because of the signs in Ft3d
! (depends on FFT conventions)
    do i_kmx=1,n_kmx
       kmx  = i_kmx - 1.
       if(kmx <= float(n_kmx/2)) then
!          k_mx(i_kmx) =  kmx_0+d_kmx* kmx
          k_mx(i_kmx) =  -(kmx_0+d_kmx* kmx)
       else
!          k_mx(i_kmx) = -kmx_0+d_kmx*(kmx-n_kmx)
          k_mx(i_kmx) = -(kmx_0+d_kmx*(kmx-n_kmx))
       end if
    end do

! Notice the negative signs because of the signs in Ft3d
! (depends on FFT conventions)
    do i_kmy=1,n_kmy
       kmy  = i_kmy - 1.
       if(kmy <= float(n_kmy/2)) then
!          k_my(i_kmy) =  kmy_0+d_kmy* kmy
          k_my(i_kmy) =  -(kmy_0+d_kmy* kmy)
       else
!          k_my(i_kmy) = -kmy_0+d_kmy*(kmy-n_kmy)
          k_my(i_kmy) = -(kmy_0+d_kmy*(kmy-n_kmy))
       end if
    end do

    if(space_domain) then
    !{

      hx_max=off2ang3d_par%hx_max
      hy_max=off2ang3d_par%hy_max
write(0,*)'hx_max',hx_max
write(0,*)'hy_max',hy_max

      do i_khx=1,n_khx
         k_hx(i_khx) =  khx_0+(i_khx-n_khx/2)*d_khx
      end do

!      do i_khy=1,n_khy
!         k_hy(i_khy) =  khy_0+(i_khy-n_khy/2)*d_khy
!      end do

      do i_khy=1,n_khy
         khy  = i_khy - 1.
         if(khy <= float(n_khy/2)) then
            k_hy(i_khy) =  khy_0+d_khy* khy
         else
            k_hy(i_khy) = -khy_0+d_khy*(khy-n_khy)
         end if
      end do

!      k_hx_min = khx_0-(number_aliases+1.)*(d_khx*(n_khx-1)/2)
!      k_hx_max = khx_0+(number_aliases+1.)*(d_khx*(n_khx-1)/2.) ! eliminate positive nyquist
      k_hx_min = khx_0-(number_aliases+1.)*(d_khx*(n_khx/2-1))
      k_hx_max = khx_0+(number_aliases+1.)*(d_khx*(n_khx/2-1)) ! eliminate positive nyquist
      if(comaz_data) then
!        k_hy_min = -HUGE(kh_y_min) bug
!        k_hy_max = HUGE(kh_y_max)
      else
!        k_hy_min = khy_0-d_khy*(n_khy-1)/2
!        k_hy_max = khy_0+d_khy*(n_khy-1)/2. ! eliminate positive nyquist
      k_hy_min = khy_0-(number_aliases+1.)*(d_khy*(n_khy/2-1))
      k_hy_max = khy_0+(number_aliases+1.)*(d_khy*(n_khy/2-1)) ! eliminate positive nyquist
      end if
    !}
    else
    !{
      do i_khx=1,n_khx
         khx  = i_khx - 1.
         if(khx <= float(n_khx/2)) then
            k_hx(i_khx) =  khx_0+d_khx* khx
         else
              k_hx(i_khx) = -khx_0+d_khx*(khx-n_khx)
         end if
      end do

      call reorder_fft_wavenumber(FFT_FORW,n_khx,k_hx)
      khx_0=k_hx(1)

      do i_khy=1,n_khy
         khy  = i_khy - 1.
         if(khy <= float(n_khy/2)) then
            k_hy(i_khy) =  khy_0+d_khy* khy
         else
            k_hy(i_khy) = -khy_0+d_khy*(khy-n_khy)
         end if
      end do
  
      call reorder_fft_wavenumber(FFT_FORW,n_khy,k_hy)
      khy_0=k_hy(1)
  
      fft_nhx_nhy_id=init_fft_2d(n_hx_pad,n_hy_pad,fast_init=.false.)
      if(off2ang3d_par%debug1) write(0,*)'scale (nhx_nhy)=',fft_nhx_nhy_id%scale
      status=fft_nhx_nhy_id%ierr
    !}
    end if


    if(reciprocal_hx) then
      i_hx_0=-nint(hx_0/d_hx)+1
    end if
    if(reciprocal_hx_wrap) then
      i_hx_0=-nint(hx_0/d_hx)+1
      if(i_hx_0 .ne. 1) call seperr("i_hx_0 should be 1 with reciprocal_hx_wrap")
    end if


    if(off2ang3d_par%nam_adcig) then
    !{
    sign_khy_nam=off2ang3d_par%sign_khy_nam
    vel_khy_mean=off2ang3d_par%vel_khy_mean
    freq_khy_mean=off2ang3d_par%freq_khy_mean
    max_mult_khy_mean=off2ang3d_par%max_mult_khy_mean

    allocate(khy_mean(n_kmx,n_khx),stat=ierr)
    if(ierr/=0) call seperr('cant allocate khy_mean')

    wavenumb_khy_mean_sq=4*(2*pi*freq_khy_mean/vel_khy_mean)**2

    inv_d_khy=1./d_khy

    k_my_min=minval(k_my)
    k_my_max=maxval(k_my)

    k_hy_min=minval(k_hy)
    k_hy_max=maxval(k_hy)

write(0,*)'wavenumb_khy_mean_sq=',wavenumb_khy_mean_sq
!write(0,*)'k_my=',k_my
!write(0,*)'k_my_min=',k_my_min
!write(0,*)'k_my_max=',k_my_max
!write(0,*)'k_hy=',k_hy
!write(0,*)'k_hy_min=',k_hy_min
!write(0,*)'k_hy_max=',k_hy_max


    do i_khx=1,n_khx
    !{
      do i_kmx=1,n_kmx
      !{
        if(vel_khy_mean <= 0.) then 
          khy_mean(i_kmx,i_khx)=0.
        else
          kz_s=wavenumb_khy_mean_sq - (k_mx(i_kmx)-k_hx(i_khx))**2
          kz_r=wavenumb_khy_mean_sq - (k_mx(i_kmx)+k_hx(i_khx))**2

!     write(0,*)'kz_s(',i_kmx,',',i_khx,')=',kz_s
!     write(0,*)'kz_r(',i_kmx,',',i_khx,')=',kz_r

          if((kz_s > 0.).and.(kz_r > 0.)) then

            khy_mean(i_kmx,i_khx)=((sqrt(kz_r)-sqrt(kz_s))/(sqrt(kz_r)+sqrt(kz_s)))
            if (abs(khy_mean(i_kmx,i_khx)) > max_mult_khy_mean) then
              if (khy_mean(i_kmx,i_khx) > 0. ) then
                khy_mean(i_kmx,i_khx)=max_mult_khy_mean
              else
                khy_mean(i_kmx,i_khx)=-max_mult_khy_mean
              end if
            end if
  
          else
            if(kz_r >=  kz_s) then
!              if(abs(kz_r-kz_s) < abs(kmap%krx(2,2)**2))  then
!              if(abs(kz_r-kz_s) < sqrt((d_kmx**2+d_khx**2)) ) then
              if(i_khx == (n_khx/2)) then
                khy_mean(i_kmx,i_khx)=0
              else
                khy_mean(i_kmx,i_khx)=max_mult_khy_mean
              end if
            else
              if(i_khx == (n_khx/2)) then
                khy_mean(i_kmx,i_khx)=0
              else
                khy_mean(i_kmx,i_khx)=-max_mult_khy_mean
              end if
            end if

          end if
        end if

! possible BUG
!        khy_mean(ikmx,ikhx)=-khy_mean(ikmx,ikhx)

      !}
      end do
    !}
    end do

!    do i_khx=1,n_khx
!      do i_kmx=1,n_kmx
!     write(0,*)'khy_mean(',i_kmx,',',i_khx,')=',khy_mean(i_kmx,i_khx)
!      end do
!    end do

 call snap('khy_mean.h',n_kmx,n_khx,khy_mean)

    khy_range=k_hy_max-k_hy_min+d_khy
  write(0,*)'before rounding khy_range=',khy_range
    khy_range=nint(khy_range/d_khy)*d_khy
  write(0,*)'after rounding khy_range=',khy_range
    khy_range=n_khy*d_khy
  write(0,*)'after second rounding khy_range=',khy_range

    khy_mean_min=minval(khy_mean)
    khy_mean_max=maxval(khy_mean)

    khy_bias_min=min(k_my_min*khy_mean_max, &
                     k_my_max*khy_mean_min ) - &
                     d_khy

    khy_bias_max=max(k_my_min*khy_mean_min, &
                     k_my_max*khy_mean_max ) + &
                     d_khy

  write(0,*)'before rounding khy_bias_min=',khy_bias_min
   write(0,*)'before rounding khy_bias_max=',khy_bias_max

    khy_bias_min=floor(khy_bias_min/d_khy)*d_khy
    khy_bias_max=floor(khy_bias_max/d_khy)*d_khy

  write(0,*)'after rounding khy_bias_min=',khy_bias_min
   write(0,*)'after rounding khy_bias_max=',khy_bias_max

    n_khy_bias=(khy_bias_max-khy_bias_min)*inv_d_khy +1

    allocate(khy_bias_mat(n_khy_bias,n_khy),stat=ierr)
    if(ierr/=0) call seperr('cant allocate khy_bias_mat')
    allocate(khy_bias_min_vec(n_khy_bias),stat=ierr)
    if(ierr/=0) call seperr('cant allocate khy_bias_min_vec')
    allocate(khy_bias_max_vec(n_khy_bias),stat=ierr)
    if(ierr/=0) call seperr('cant allocate khy_bias_max_vec')


    !}
    end if

    do i_khy_bias=1,n_khy_bias
      do i_khy=1,n_khy
         khy_bias_mat(i_khy_bias,i_khy)=-1
      end do
    end do

    do i_khy_bias=1,n_khy_bias
    !{
      khy_bias=khy_bias_min + (i_khy_bias-1)*d_khy
      do i_khy=1,n_khy
        khy=khy_bias+k_hy(i_khy)
!        khy=nint(khy/d_khy)*d_khy

        i_khy_nint=nint(khy/d_khy)

!        if(khy >= 0.)  then
        if(i_khy_nint >= 0)  then

          i_khy_nint=i_khy_nint
!          i_khy_nint=nint(khy/d_khy)
!          i_khy_nint=floor(khy/d_khy)
          i_khy_bias_cycles=mod(i_khy_nint,n_khy)+1
!           khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=khy_range*floor((khy-khy_min)/khy_range)
    khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=khy_range*floor((nint((khy-k_hy_min)/d_khy)*d_khy)/(khy_range))+k_hy(i_khy_bias_cycles)
!    khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=khy_range*nint((khy-k_hy_min)/khy_range)+k_hy(i_khy_bias_cycles)
        else
!          i_khy_nint=nint(khy/d_khy)+1
          i_khy_nint=i_khy_nint+1
          i_khy_bias_cycles=n_khy-mod(-i_khy_nint,n_khy)
!           khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=-khy_range*floor((khy_max-khy)/khy_range)
    khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=-khy_range*floor((nint((k_hy_max-khy)/d_khy)*d_khy)/khy_range)+k_hy(i_khy_bias_cycles)

!    khy_bias_mat(i_khy_bias,i_khy_bias_cycles)=-khy_range*nint((k_hy_max-khy)/khy_range)+k_hy(i_khy_bias_cycles)
        end if

      end do
      khy_bias_min_vec(i_khy_bias)=minval(khy_bias_mat(i_khy_bias,:))
      khy_bias_max_vec(i_khy_bias)=maxval(khy_bias_mat(i_khy_bias,:))
    !}
    end do

 call snap('khy_bias_mat.h',n_khy_bias,n_khy,khy_bias_mat)
 call snap('khy_bias_min_vec.h',n_khy_bias,1,khy_bias_min_vec)
 call snap('khy_bias_max_vec.h',n_khy_bias,1,khy_bias_max_vec)

     if(off2ang3d_par%debug1) then
write(0,*)'status=',status
write(0,*)'number_aliases=',number_aliases
write(0,*)'n_kz=',n_kz
write(0,*)'n_kmx=',n_kmx
write(0,*)'n_kmy=',n_kmy
write(0,*)'n_hx=',n_hx
write(0,*)'n_hy=',n_hy
write(0,*)'n_hx_pad=',n_hx_pad
write(0,*)'n_hy_pad=',n_hy_pad
write(0,*)'hx_0=',hx_0
write(0,*)'hy_0=',hy_0
write(0,*)'d_hx=',d_hx
write(0,*)'d_hy=',d_hy
write(0,*)'khx_0=',khx_0
write(0,*)'khy_0=',khy_0
write(0,*)'n_bytes_hx_hy=',n_bytes_hx_hy
write(0,*)'n_bytes_gamma_azim=',n_bytes_gamma_azim
       if(space_domain) then
write(0,*)'k_hx=',k_hx
write(0,*)'k_hy=',k_hy
write(0,*)'k_hx_min=',k_hx_min
write(0,*)'k_hx_max=',k_hx_max
write(0,*)'k_hy_min=',k_hy_min
write(0,*)'k_hy_max=',k_hy_max
       else
write(0,*)'k_hx=',k_hx
write(0,*)'k_hy=',k_hy
       end if
write(0,*)'k_mx=',k_mx
write(0,*)'k_my=',k_my
write(0,*)'k_my_min=',k_my_min
write(0,*)'k_my_max=',k_my_max
write(0,*)'k_z=',k_z
write(0,*)'i_hx_0=',i_hx_0
     end if


  end function Off2Ang3D_init

  function Off2Ang3D_oper(off2ang3d_par) result(status)

  type (off2ang3d_param_type) off2ang3d_par
  integer status

  logical in_bounds,in_kh_bands,in_khx_band,in_khy_band
  integer i_hx,i_hy,i_kmy,i_kmx,i_kz,i_kz_thread
  integer i_khx,i_khy,i_azim,i_gamma,i_thread
  integer i_khx_copl,i_khy_copl,i_khx_copl_p1,i_khy_copl_p1
!  integer i_hy_0,i_hx_rec,i_hy_rec
  integer i_hx_wrap,i_khy_bias
  real inv_d_khx,kkz,kkz_den,const,khx_copl,khx_copl_int,w_khx_copl
  real dkhx_copl,dkhy_copl
  real inv_d_khy,khy_copl,khy_copl_int,w_khy_copl,hy,hx
  real azim,sin_azim,cos_azim,k_mx_azim,k_my_azim,khx_azim,khy_azim
  real gamma,tan_gamma,cos_delta,sin_delta,tan_alpha,jac,inv_cos_gamma_sq
  real dkmx_dazim,dkmy_dazim,dkhx_dazim,dkhy_dazim,dkhx_dgamma,dkhy_dgamma
  real tan_gamma_jac,jac_azim_norm,jac_samp
  real weight_wrap_azim_1,weight_wrap_azim_2
  real norm_jac_azim_1,norm_jac_azim_2
  real khy_bias,khy_diff,khy_diff_new,khy_copl_bias

  inv_d_khx=1./d_khx
  inv_d_khy=1./d_khy

  if(off2ang3d_par%create_coplanar_ang .or. &
     off2ang3d_par%create_coplanar_azim_ang) then
  write(0,*)'Here #1'
    azim_bound=comp_azim_bound(off2ang3d_par)
  write(0,*)'Here #2'
    if(off2ang3d_par%create_coplanar_ang) then
      tag_output=trim(off2ang3d_par%tag_coplanar_ang)
    else
      tag_output=trim(off2ang3d_par%tag_coplanar_azim_ang)
    end if
  end if

  if(delta_azim_gamma_0 /= 0.) then
    weight_wrap_azim_1=1.
    weight_wrap_azim_2=inv_pi*(180./delta_azim_gamma_0)
  else
    weight_wrap_azim_1=1.
    weight_wrap_azim_2=inv_pi
  end if
  write(0,*)'weight_wrap_azim_1',weight_wrap_azim_1
  write(0,*)'weight_wrap_azim_2',weight_wrap_azim_2
  norm_jac_azim_1=(pi*delta_azim_gamma_0/180.)/(2.*sqrt(d_khx**2 + d_khy**2))

!  norm_jac_azim_2=1./(sqrt(2.)*sqrt(d_khx**2 + d_khy**2))
  norm_jac_azim_2=.5/(2.*sqrt(d_khx**2 + d_khy**2))
!  norm_jac_azim_2=.1/(2.*sqrt(d_khx**2 + d_khy**2))
  write(0,*)'norm_jac_azim_1',norm_jac_azim_1
  write(0,*)'norm_jac_azim_2',norm_jac_azim_2

  do i_kmy=1,n_kmy
  !{
  do i_kmx=1,n_kmx
  !{

  do i_kz=1,n_kz,n_threads
  !{
    i_kz_thread=i_kz+i_thread-1
    jac_mat=0.
    if(sreed('in',hx_hy,n_bytes_hx_hy) /= n_bytes_hx_hy) then
       call seperr('Problems with reading input file')
    end if

    do i_thread=1,n_threads
    !{
    do i_hy=1,n_hy

!!write(0,*)'i_hy=',i_hy
!!write(0,*)'shape hax_hy',shape(hx_hy)
!      if(sreed('in',hx_hy(1,i_hy,i_thread),n_bytes_hx) /= n_bytes_hx) then
!          call seperr('Problems with reading input file')
!      end if
!!write(0,*)'after n_hx=',n_hx
!!write(0,*)'after n_hx_pad=',n_hx_pad

      do i_hx=n_hx+1,n_hx_pad
write(0,*)'i_hx=',i_hx
        hx_hy(i_hx,i_hy,i_thread)=cmplx(0.,0.)
      end do
    end do
    do i_hy=n_hy+1,n_hy_pad
!write(0,*)'i_hy=',i_hy
      do i_hx=1,n_hx_pad
!write(0,*)'i_hx=',i_hx
        hx_hy(i_hx,i_hy,i_thread)=cmplx(0.,0.)
      end do
    end do

!   if((reciprocal_hx) .and. (n_hx_pad > n_hx)) then
!   !{
!write(0,*)'recirpcal'
!!write(0,*)'shape hx_hy',shape(hx_hy)
!    i_hy_0=(-hy_0/d_hy)+1
!!write(0,*)'i_hy_0=',i_hy_0
!    do i_hy=1,min(2*i_hy_0-1,n_hy)
!      i_hy_rec=2*i_hy_0-i_hy
!      do i_hx=n_hx_pad-n_hx+2,n_hx_pad
!        i_hx_rec=n_hx_pad-i_hx+2
!  if((i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(i_kz_thread == n_kz/4)) then
!write(0,*)'i_hy=',i_hy
!write(0,*)'i_hy_rec=',i_hy_rec
!write(0,*)'i_hx=',i_hx
!write(0,*)'i_hx_rec=',i_hx_rec
!end if
!        hx_hy(i_hx,i_hy,i_thread)= &
!           hx_hy(i_hx_rec,i_hy_rec,i_thread)
!      end do
!!write(0,*)'finished i_hx'
!    end do
!!write(0,*)'finished i_hy'
!
!   !}
!   end if

    !}
    end do

    gamma_azim=0.
    !$OMP PARALLEL DEFAULT (SHARED), &
    !$OMP&  PRIVATE (in_bounds,in_kh_bands,in_khx_band,in_khy_band,&
    !$OMP&   i_thread,i_kz_thread,i_gamma,i_azim,i_khy,i_khx,i_hx,i_hy,&
    !$OMP&   i_khx_copl,i_khy_copl,i_khx_copl_p1,i_khy_copl_p1,&
    !$OMP&   i_hx_wrap,i_khy_bias,&
    !$OMP&   kkz,kkz_den,gamma,tan_gamma,tan_gamma_jac,azim,sin_azim,cos_azim,&
    !$OMP&   k_mx_azim,k_my_azim,dkmx_dazim,dkmy_dazim,const,khx_azim,khy_azim,&
    !$OMP&   khx_copl,khy_copl,jac,sin_delta,cos_delta,tan_alpha,&
    !$OMP&   inv_cos_gamma_sq,dkhx_dazim,dkhy_dazim,dkhx_dgamma,dkhy_dgamma,&
    !$OMP&   jac_samp,dkhx_copl,dkhy_copl,hx,hy,&
    !$OMP&   khx_copl_int,w_khx_copl,khy_copl_int,w_khy_copl,&
    !$OMP&   khy_bias,khy_diff,khy_diff_new,khy_copl_bias)
    !$OMP DO
    do i_thread=1,n_threads
    !{
     
      i_kz_thread=i_kz+i_thread-1

      kkz=k_z(i_kz_thread)

      
      if(.not. space_domain) then
        call fft_2d_data_par(fft_nhx_nhy_id,FFT_FORW,n_hx_pad,n_hy_pad,i_thread,hx_hy)
        call reorder_fft_2d_data_par(FFT_FORW,n_khx,n_khy,i_thread,hx_hy)
      end if

      if(off2ang3d_par%create_coplanar_ang .or. &
         off2ang3d_par%create_coplanar_azim_ang) then
      !{

        if((i_kz_thread == 1)) then
          kkz_den=k_z(2)
        else
          kkz_den=kkz
        end if


        do i_gamma=1,n_gamma
        !{

          gamma=gamma_0 + (i_gamma-1)*d_gamma
          tan_gamma=tan(gamma)
!          tan_gamma_jac=.5*( &
!                abs(tan(gamma-.5*d_gamma))+abs(tan(gamma+.5*d_gamma)) )
          tan_gamma_jac=tan_gamma

          if(off2ang3d_par%create_coplanar_ang) then
            gamma_azim(i_gamma,1,i_thread)= 0.
          end if

          do i_azim=azim_bound(1,i_gamma),azim_bound(2,i_gamma)
          !{
            azim=azim_0_max + (i_azim-1)*d_azim
            sin_azim=sin(azim)
            cos_azim=cos(azim)
      
      
            k_mx_azim=   k_mx(i_kmx)*cos_azim + k_my(i_kmy)*sin_azim
            k_my_azim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim

! notice differrent than my notes because change in sign in beta
            dkmx_dazim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
            dkmy_dazim= - k_mx(i_kmx)*cos_azim - k_my(i_kmy)*sin_azim

! inverse is
!      k_mx= k_mx_azim*cos_azim + k_my_azim*sin_azim
!      k_my=-k_mx_azim*sin_azim + k_my_azim*cos_azim


!            const=-k_my_azim*k_mx_azim/(kkz_den**2 + k_my_azim**2)
            const=k_my_azim*k_mx_azim/(kkz_den**2 + k_my_azim**2)

!            if(.not. apply_jac_none) then
!              jac_azim_norm=d_azim/sqrt(d_khx**2 + d_khy**2)
!            end if

            if(ang_2d) then
              khx_azim= -tan_gamma*sign(sqrt(kkz**2),kkz)
  !            khx_azim= tan_gamma*(sqrt(kkz**2))
            else
              khx_azim= -tan_gamma*sign(sqrt(kkz**2 + k_my_azim**2),kkz)
  !            khx_azim= tan_gamma*(sqrt(kkz**2 + k_my_azim**2))
            end if

            khy_azim=-sign_khy*khx_azim*const
    
            khx_copl= khx_azim*cos_azim - khy_azim*sin_azim
            khy_copl= khx_azim*sin_azim + khy_azim*cos_azim

            if(apply_jac_none) then
            !{
              jac=1
            !}
            else
            !{

              sin_delta=k_my_azim/sqrt(k_my_azim**2 + kkz_den**2) 
!              cos_delta=kkz/sqrt(k_my_azim**2 + kkz_den**2) 
              cos_delta=abs(kkz_den)/sqrt(k_my_azim**2 + kkz_den**2) 
              tan_alpha=k_mx_azim/kkz_den


              if(apply_jac_both) then
                inv_cos_gamma_sq=(1.+tan_gamma_jac**2)
    
                dkhx_dazim=-tan_gamma_jac*sin_delta*dkmy_dazim
                dkhy_dazim= tan_gamma_jac*(           sin_delta*dkmx_dazim+ &
                                      tan_alpha*(cos_delta**3)*dkmy_dazim)
    
                dkhx_dgamma=kkz_den*inv_cos_gamma_sq/cos_delta
                dkhy_dgamma=kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta
    
                jac_samp=(dkhx_dgamma*dkhy_dazim + dkhy_dgamma*dkhx_dazim)*&
                          (d_azim/(d_khx*d_khy))
!                jac=max(.05,jac_samp)
                jac=jac_samp


              else
                if(apply_jac_azim_1 .or. apply_jac_azim_2) then
  
                  dkhx_dazim=-tan_gamma_jac*sin_delta*dkmy_dazim
                  dkhy_dazim= tan_gamma_jac*(         sin_delta*dkmx_dazim + &
                                      tan_alpha*(cos_delta**3)*dkmy_dazim)

                  dkhx_copl = dkhx_dazim*cos_azim - dkhy_dazim*sin_azim + &
                              (- khx_azim*sin_azim - khy_azim*cos_azim)
                  dkhy_copl = dkhx_dazim*sin_azim + dkhy_dazim*cos_azim + &
                              (  khx_azim*cos_azim - khy_azim*sin_azim)

! matlab code
!	dkhx(i) = dkhx_dazim(i)*cos(beta) - dkhy_dazim(i)*sin(beta) + ...
!                          (- khx_azim(i)*sin(beta) - khy_azim(i)*cos(beta));
!	dkhy(i) = dkhx_dazim(i)*sin(beta) + dkhy_dazim(i)*cos(beta) + ...
!                          (  khx_azim(i)*cos(beta) - khy_azim(i)*sin(beta));


                  if(apply_jac_azim_1) then
                    if(jac_nowrap) then
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_1, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_1)

!                        jac=max(0.,abs(dkhy_dazim*dkhx_dazim)/ &
!                          (abs(d_khx*dkhy_dazim) + abs(d_khy*dkhx_dazim)) )
                      else
                        jac=0.
                      end if
                    else
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_1, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_1)

!                        jac=max(weight_wrap_azim,abs(dkhy_dazim*dkhx_dazim)/ &
!                          (abs(d_khx*dkhy_dazim) + abs(d_khy*dkhx_dazim)) )
    
                      else
                        jac=1.
                      end if
                    end if
                  else
                    if(jac_nowrap) then
                      if(gamma /= 0.) then
                        jac=max(0., &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_2)
    
                      else
                        jac=0.
                      end if
                    else
                      if(gamma /= 0.) then
                        jac=max(weight_wrap_azim_2, &
                        sqrt(dkhx_copl**2 + dkhy_copl**2)*norm_jac_azim_2)
    
                      else
                        jac=weight_wrap_azim_2
                      end if
                    end if
                  end if

!                  jac=abs((abs(d_khx*dkhy_dazim) + abs(d_khy*dkhx_dazim))/ &
!                       (dkhy_dazim*dkhx_dazim))
      
!                  jac_samp=sqrt((dkhx_dazim*d_azim/d_khx)**2 + &
!                                (dkhy_dazim*d_azim/d_khy)**2 )
!                  jac=jac_samp
!    !              jac=max(1.,sqrt(dkhx_dazim**2+dkhy_dazim**2)*jac_azim_norm)
!    !              jac=sqrt(dkhx_dazim**2 + dkhy_dazim**2)


                else
                  if(apply_jac_gamma) then
                    inv_cos_gamma_sq=(1.+tan_gamma_jac**2)
    
                    dkhx_dgamma=kkz_den*inv_cos_gamma_sq/cos_delta
                    dkhy_dgamma=kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta
        
                    jac=sqrt(dkhx_dgamma**2 + dkhy_dgamma**2)
                  end if
                end if
              end if
            !}
            end if

            if(space_domain) then
            !{
              if(nam_adcig) then
              !{ nam_adcig
                in_khx_band=(khx_copl >= k_hx_min) .and. &
                            (khx_copl <= k_hx_max)
                if(in_khx_band) then
                !{
                   i_khx_copl= nint((khx_copl-khx_0)/d_khx) + n_khx/2

                   khy_bias=sign_khy_nam*(khy_mean(i_kmx,i_khx_copl)*k_my(i_kmy))
                   i_khy_bias=(khy_bias-khy_bias_min)*inv_d_khy +1

                   in_khy_band=(khy_copl >= khy_bias_min_vec(i_khy_bias)).and. &
                               (khy_copl <= khy_bias_max_vec(i_khy_bias))

                   if(in_khy_band) then
                   !{
                     khy_diff=HUGE(khy_diff)
                     do i_khy=1,n_khy
                       khy_diff_new=(khy_bias_mat(i_khy_bias,i_khy)-khy_copl)**2
                       if(khy_diff_new < khy_diff) then
                         khy_diff=khy_diff_new
                         i_khy_copl=i_khy
                       end if
                     end do
                     khy_copl_bias=k_hy(i_khy_copl)
                   !}
                   end if
                !}
                end if
                in_kh_bands=in_khx_band .and. in_khy_band
              !} nam_adcig
              else
              !{ nam_adcig
                in_kh_bands=(khx_copl >= k_hx_min) .and. &
                            (khx_copl <= k_hx_max) .and.  &
                            (khy_copl >= k_hy_min) .and. &
                            (khy_copl <= k_hy_max)
                 khy_copl_bias=khy_copl
              !} nam_adcig
             end if
             if(in_kh_bands) then
              !{
                gamma_azim(i_gamma,i_azim,i_thread)=0.
                if(reciprocal_hx .or. reciprocal_hx_wrap) then
                  !{
                  if(reciprocal_hx) then
                  !{
                do i_hy=1,n_hy_pad
                  hy=hy_0+(i_hy-1)*d_hy
                  if(abs(hy) < hy_max) then 
                  !{
                  gamma_azim(i_gamma,i_azim,i_thread)= &
                       gamma_azim(i_gamma,i_azim,i_thread) +  &
                           hx_hy(i_hx_0,i_hy,i_thread)* &
                           cexp(cmplx(0,khy_copl_bias*hy))
                  do i_hx=i_hx_0+1,n_hx_pad
                    hx=hx_0+(i_hx-1)*d_hx
                    if(abs(hx) < hx_max) then 
                    !{
                    gamma_azim(i_gamma,i_azim,i_thread)= &
                         gamma_azim(i_gamma,i_azim,i_thread) +  &
                           hx_hy(i_hx,i_hy,i_thread)* &
                           2.*jac*cos(khx_copl*hx+khy_copl_bias*hy)
                    !}
                    end if
                  end do
                  !}
                  end if
                end do
                  !}
                  else
                  !{
                do i_hy=1,n_hy_pad
                  hy=hy_0+(i_hy-1)*d_hy
                  if(abs(hy) < hy_max) then 
                  !{
                  gamma_azim(i_gamma,i_azim,i_thread)= &
                       gamma_azim(i_gamma,i_azim,i_thread) +  &
                           hx_hy(i_hx_0,i_hy,i_thread)* &
                           cexp(cmplx(0,khy_copl_bias*hy))
                  do i_hx=i_hx_0+1,n_hx_pad/2
                    i_hx_wrap=n_hx_pad-i_hx+2
                    hx=hx_0+(i_hx-1)*d_hx
                    if(abs(hx) < hx_max) then 
                    !{
                    gamma_azim(i_gamma,i_azim,i_thread)= &
                         gamma_azim(i_gamma,i_azim,i_thread) +  &
                           (hx_hy(i_hx,i_hy,i_thread)* &
                          jac*cexp(cmplx(0.,khx_copl*hx+khy_copl_bias*hy))) &
                            + &
                           (hx_hy(i_hx_wrap,i_hy,i_thread)* &
                          jac*cexp(cmplx(0.,-khx_copl*hx+khy_copl_bias*hy)))

!                    gamma_azim(i_gamma,i_azim,i_thread)= &
!                         gamma_azim(i_gamma,i_azim,i_thread) +  &
!                           (hx_hy(i_hx,i_hy,i_thread) + &
!                            hx_hy(i_hx_wrap,i_hy,i_thread))* &
!                           2.*jac*cos(khx_copl*hx+khy_copl*hy)
                    !}
                    end if
                  end do
                  !}
                  end if
                end do
                  !}
                  end if
                !}
                else
                !{
                do i_hy=1,n_hy_pad
                  hy=hy_0+(i_hy-1)*d_hy
                  if(abs(hy) < hy_max) then 
                  !{
                  do i_hx=1,n_hx_pad
                    hx=hx_0+(i_hx-1)*d_hx
                    if(abs(hx) < hx_max) then 
                    !{
                    gamma_azim(i_gamma,i_azim,i_thread)= &
                         gamma_azim(i_gamma,i_azim,i_thread) +  &
                           hx_hy(i_hx,i_hy,i_thread)* &
                           jac*cexp(cmplx(0,khx_copl*hx+khy_copl_bias*hy))
                    !}
                    end if
                  end do
                  !}
                  end if
                end do
                !}
                end if
              !}
              else
              !{
!              write(0,*)'At gamma=',gamma*inv_pi*180.,'  azim=',azim*inv_pi*180.
!              if(khx_copl < k_hx_min) write(0,*)'khx < k_hx_min'
!              if(khx_copl > k_hx_max) write(0,*)'khx > k_hx_max'
!              if(khy_copl_bias < k_hy_min) write(0,*)'khy < k_hy_min'
!              if(khy_copl_bias > k_hy_max) write(0,*)'khy > k_hy_max'
              !}
              end if
            !} space_domain
            else
            !{
            khx_copl_int=(khx_copl-khx_0)*inv_d_khx+1
            i_khx_copl=floor(khx_copl_int)
            w_khx_copl=khx_copl_int-i_khx_copl
  
            khy_copl_int=(khy_copl-khy_0)*inv_d_khy+1
            i_khy_copl=floor(khy_copl_int)
            w_khy_copl=khy_copl_int-i_khy_copl
  
!           if((i_khx_copl > -nint(n_khx*(number_aliases))) .and. &
!               (i_khx_copl <= (nint(n_khx*(number_aliases+1 ))) ) .and.  &
!               (i_khy_copl > -nint(n_khy*(number_aliases-1.))) .and. &
!               (i_khy_copl <= (nint(n_khy*(number_aliases  )))) &

            if((i_khx_copl > -n_khx*number_aliases) .and. &
               (i_khx_copl < (n_khx*(number_aliases+1))) .and.  &
               (i_khy_copl > -n_khy*number_aliases) .and. &
               (i_khy_copl < (n_khy*(number_aliases+1))) &
              ) then
              in_bounds=.true.
              i_khx_copl=mod((number_aliases+1)*n_khx+i_khx_copl-1,float(n_khx))+1
              i_khx_copl_p1=mod((number_aliases+1)*n_khx+i_khx_copl,float(n_khx))+1
              i_khy_copl=mod((number_aliases+1)*n_khy+i_khy_copl-1,float(n_khy))+1
              i_khy_copl_p1=mod((number_aliases+1)*n_khy+i_khy_copl,float(n_khy))+1

!              i_khx_copl=   imod(nint((number_aliases)*n_khx)+i_khx_copl-1,n_khx)+1
!              i_khx_copl_p1=imod(nint((number_aliases)*n_khx)+i_khx_copl  ,n_khx)+1
!  
!              i_khy_copl=   imod(nint((number_aliases)*n_khy)+i_khy_copl-1,n_khy)+1
!              i_khy_copl_p1=imod(nint((number_aliases)*n_khy)+i_khy_copl  ,n_khy)+1

              jac_mat(i_khx_copl,i_khy_copl,i_thread) = jac

            else
  !            if((n_khy == 1).and.(i_khy_copl == 1) then
  !              in_bounds=.true.
  !              i_khx_copl=imod((number_aliases+1)*n_khx+i_khx_copl-1,n_khx)+1
  !            else
                in_bounds=.false.
  !            end if
            end if


!  if((i_kmy == 1).and. (i_kmx == 1).and.(i_kz_thread == 2)) then
!  if((i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(i_kz_thread == n_kz/4)) then
!  if((i_kmy == n_kmy/16).and. (i_kmx == n_kmx/16).and.(i_kz_thread == n_kz/16)) then
!  if((i_kmy == 1).and. (i_kmx == n_kmx/16)) then
  if((i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(i_kz_thread == n_kz/4)) then
  write(0,*)'***'
  write(0,*)'i_gamma=',i_gamma
  write(0,*)'gamma=',gamma*(180/pi)
  write(0,*)'tan_gamma=',tan_gamma
  write(0,*)'tan_gamma_jac=',tan_gamma_jac
  write(0,*)'i_azim=',i_azim
  write(0,*)'azim=',azim*(180/pi)
  write(0,*)'in_bounds=',in_bounds
  write(0,*)'khx_copl=',khx_copl
  write(0,*)'khx_copl_int=',khx_copl_int
  write(0,*)'i_khx_copl=',i_khx_copl
  write(0,*)'w_khx_copl=',w_khx_copl
  write(0,*)'khy_copl=',khy_copl
  write(0,*)'khy_copl_int=',khy_copl_int
  write(0,*)'i_khy_copl=',i_khy_copl
  write(0,*)'w_khy_copl=',w_khy_copl
  write(0,*)'jac_samp=',jac_samp
  write(0,*)'jac=',jac
  write(0,*)'***'
  end if

            if(in_bounds) then
            !{
!  (hx_hy(i_khx_copl  ,i_khy_copl  ,i_thread)*(1.-w_khx_copl)+ &
!   hx_hy(i_khx_copl+1,i_khy_copl  ,i_thread)*(   w_khx_copl) &
!   ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0))

if(off2ang3d_par%create_coplanar_ang) then

  gamma_azim(i_gamma,1,i_thread)=  gamma_azim(i_gamma,1,i_thread) +  &
  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0))* jac

!  gamma_azim(i_gamma,1,i_thread)=  gamma_azim(i_gamma,1,i_thread) +  &
!  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
!   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
!   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
!   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
!  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac

else
  gamma_azim(i_gamma,i_azim,i_thread)=   &
  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac

!  gamma_azim(i_gamma,i_azim,i_thread)=  &
!  (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+ &
!   hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+ &
!   hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+ &
!   hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
!  ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac

end if

            !}
            end if

            !} space_domain
            end if
          !} azim
          end do
        !} gamma
        end do
      !}
      else
      !{
        do i_azim=1,n_azim
        !{

          azim=azim_0 + (i_azim-1)*d_azim
          sin_azim=sin(azim)
          cos_azim=cos(azim)


          k_mx_azim=   k_mx(i_kmx)*cos_azim + k_my(i_kmy)*sin_azim
          k_my_azim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
  
! notice differrent than my notes because change in sign in beta
          dkmx_dazim= - k_mx(i_kmx)*sin_azim + k_my(i_kmy)*cos_azim
          dkmy_dazim= - k_mx(i_kmx)*cos_azim - k_my(i_kmy)*sin_azim

! inverse is
!      k_mx= k_mx_azim*cos_azim + k_my_azim*sin_azim
!      k_my=-k_mx_azim*sin_azim + k_my_azim*cos_azim


        tag_output=trim(off2ang3d_par%tag_coplanar_stack)
        if(off2ang3d_par%create_stack) then
        !{
          gamma_azim(1,i_azim,i_thread)=cmplx(0.,0.)
          do i_khy=1,n_khy
          !{
  
            do i_khx=1,n_khx
            !{
              gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread) + &
                  hx_hy(i_khx,i_khy,i_thread)* &
                  cexp(cmplx(0,k_hx(i_khx)*hx_0+k_hy(i_khy)*hy_0))* &
                  fft_nhx_nhy_id%scale
  !                cmplx(0.,-sign(1.,kkz)) *&
  !                cmplx(0.,-sign(sqrt(abs(kkz)),kkz)) *&
  
            !}
            end do
  
          !}
          end do
  
        !}
        else
        !{
          gamma_azim(1,i_azim,i_thread)=cmplx(0.,0.)
  
  !        do i_khy=1,n_khy
  !        !{
  
            if((i_kz_thread == 1)) then
              kkz_den=k_z(2)
            else
              kkz_den=kkz
            end if

            if(.not. apply_jac_none) then
              sin_delta=k_my_azim/sqrt(k_my_azim**2 + kkz_den**2) 
              cos_delta=abs(kkz_den)/sqrt(k_my_azim**2 + kkz_den**2) 
              tan_alpha=k_mx_azim/kkz_den
            end if

!            const=-k_my_azim*k_mx_azim/(abs(kkz_den)*sqrt(kkz_den**2 + k_my_azim**2))
            const=-k_my_azim*k_mx_azim/(kkz_den**2 + k_my_azim**2)
  
            do i_khx=1,n_khx
            !{

              khx_azim=k_hx(i_khx)
              khy_azim=sign_khy*khx_azim*const

              if(apply_jac_none) then
                jac=1
              else
                tan_gamma_jac=(khx_azim*cos_delta)/kkz_den
                if(apply_jac_azim_1 .or. apply_jac_azim_2) then
    
                  dkhx_dazim=tan_gamma_jac*sin_delta*dkmy_dazim
                  dkhy_dazim=tan_gamma_jac*(           sin_delta*dkmx_dazim -&
                                        tan_alpha*(cos_delta**3)*dkmy_dazim)
      
                  jac_samp=sqrt((dkhx_dazim*d_azim/d_khx)**2 + &
                                (dkhy_dazim*d_azim/d_khy)**2 )
!                  jac=max(.05,jac_samp)
                  jac=jac_samp
!                  jac=sqrt(dkhx_dazim**2 + dkhy_dazim**2)
                else
                  if(apply_jac_gamma) then
                    call seperr('Not possible to apply_jac_gamma in this case')
                  else
                    if(apply_jac_both) then
                    call seperr('Not possible to apply_jac_both in this case')
                    end if
                  end if
                end if
              end if

  !            khy_copl=sign_khy*k_hx(i_khx)*const
  

              khx_copl= khx_azim*cos_azim - khy_azim*sin_azim
              khy_copl= khx_azim*sin_azim + khy_azim*cos_azim
  
              khx_copl_int=(khx_copl-khx_0)*inv_d_khx+1
              i_khx_copl=floor(khx_copl_int)
              w_khx_copl=khx_copl_int-i_khx_copl
  
              khy_copl_int=(khy_copl-khy_0)*inv_d_khy+1
              i_khy_copl=floor(khy_copl_int)
              w_khy_copl=khy_copl_int-i_khy_copl
  
  !            if((i_khx_copl > -n_khx*number_aliases) .and. &
  !               (i_khx_copl < (n_khx*(number_aliases+1))) .and.  &
  !               (i_khy_copl > -n_khy*number_aliases) .and. &
  !               (i_khy_copl < (n_khy*(number_aliases+1))) &
  !
!            if((i_khx_copl > -nint(n_khx*(number_aliases-1.))) .and. &
!               (i_khx_copl <= (nint(n_khx*(number_aliases  ))) ) .and.  &
!               (i_khy_copl > -nint(n_khy*(number_aliases-1.))) .and. &
!               (i_khy_copl <= (nint(n_khy*(number_aliases  )))) &
!                ) then
!                in_bounds=.true.
!           i_khx_copl=   imod(nint((number_aliases)*n_khx)+i_khx_copl-1,n_khx)+1
!           i_khx_copl_p1=imod(nint((number_aliases)*n_khx)+i_khx_copl  ,n_khx)+1
!           i_khy_copl=   imod(nint((number_aliases)*n_khy)+i_khy_copl-1,n_khy)+1
!           i_khy_copl_p1=imod(nint((number_aliases)*n_khy)+i_khy_copl  ,n_khy)+1

            if((i_khx_copl > -n_khx*number_aliases) .and. &
               (i_khx_copl < (n_khx*(number_aliases+1))) .and.  &
               (i_khy_copl > -n_khy*number_aliases) .and. &
               (i_khy_copl < (n_khy*(number_aliases+1))) &
              ) then
              in_bounds=.true.
              i_khx_copl=mod((number_aliases+1)*n_khx+i_khx_copl-1,float(n_khx))+1
              i_khx_copl_p1=mod((number_aliases+1)*n_khx+i_khx_copl,float(n_khx))+1
              i_khy_copl=mod((number_aliases+1)*n_khy+i_khy_copl-1,float(n_khy))+1
              i_khy_copl_p1=mod((number_aliases+1)*n_khy+i_khy_copl,float(n_khy))+1


              else
                in_bounds=.false.
              end if
  
!  if((i_khx == n_khx).and. (i_kmx == n_kmx/4).and.(i_kz_thread == n_kz/4)) then
!  if((i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(i_kz_thread == n_kz/4)) then
!  if((i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(i_azim == (n_azim+1)/2)) then
!  if((i_kmy == n_kmy/8).and. (i_kmx == n_kmx/2).and.(i_azim == (n_azim+1)/2)) then
  if((i_kmy == 1).and. (i_kmx == n_kmx/2).and.(i_azim == (n_azim+1)/2)) then
  write(0,*)'***'
  write(0,*)'azim=',azim
  write(0,*)'in_bounds=',in_bounds
  write(0,*)'kkz_den=',kkz_den
  write(0,*)'kmy=',k_my(i_kmy)
  write(0,*)'kmx=',k_mx(i_kmx)
!  write(0,*)'khx_copl=',khx_copl
!  write(0,*)'khx_copl_int=',khx_copl_int
!  write(0,*)'i_khx_copl=',i_khx_copl
!  write(0,*)'w_khx_copl=',w_khx_copl
!  write(0,*)'khy_copl=',khy_copl
!  write(0,*)'khy_copl_int=',khy_copl_int
!  write(0,*)'i_khy_copl=',i_khy_copl
!  write(0,*)'w_khy_copl=',w_khy_copl
  write(0,*)'sin_delta=',sin_delta
  write(0,*)'cos_delta=',cos_delta
  write(0,*)'tan_alpha=',tan_alpha
  write(0,*)'tan_gamma_jac=',tan_gamma_jac
  write(0,*)'jac_samp=',jac_samp
  write(0,*)'jac=',jac
  write(0,*)'***'
  end if
  !            if(i_khy_copl ==0) then
  !              gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread) + &
  !                (hx_hy(i_khx,n_khy,i_thread)*(1.-w_khy_copl) + &
  !                 hx_hy(i_khx,1    ,i_thread)*(   w_khy_copl)) * &
  !                cexp(cmplx(0,k_hx(i_khx)*hx_0+khy_copl*hy_0))
  !            end if
  !
  !            if((i_khy_copl > 0) .and. (i_khy_copl < n_khy)) then
              if(in_bounds) then
              !{

  !              if(i_khy_copl == n_khy) then
  !                gamma_azim(1,i_azim,i_thread)=gamma_azim(1,i_azim,i_thread) +&
  !                  (hx_hy(i_khx,n_khy,i_thread)*(1.-w_khy_copl) + &
  !                   hx_hy(i_khx,1    ,i_thread)*(   w_khy_copl)) * &
  !                  cexp(cmplx(0,k_hx(i_khx)*hx_0+khy_copl*hy_0))
  !              else
  !              gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread) + &
  !                (hx_hy(i_khx,i_khy_copl  ,i_thread)*(1.-w_khy_copl) + &
  !                 hx_hy(i_khx,i_khy_copl+1,i_thread)*(   w_khy_copl)) * &
  !                cexp(cmplx(0,k_hx(i_khx)*hx_0+khy_copl*hy_0))
  !                fft_nhx_nhy_id%scale
  !              end if
  
                gamma_azim(1,i_azim,i_thread)= gamma_azim(1,i_azim,i_thread)+&
 (hx_hy(i_khx_copl   ,i_khy_copl   ,i_thread)*(1.-w_khx_copl)*(1-w_khy_copl)+&
  hx_hy(i_khx_copl   ,i_khy_copl_p1,i_thread)*(1.-w_khx_copl)*(  w_khy_copl)+&
  hx_hy(i_khx_copl_p1,i_khy_copl   ,i_thread)*(   w_khx_copl)*(1-w_khy_copl)+&
  hx_hy(i_khx_copl_p1,i_khy_copl_p1,i_thread)*(   w_khx_copl)*(  w_khy_copl) &
     ) * cexp(cmplx(0,khx_copl*hx_0+khy_copl*hy_0)) * jac
              !}
              end if
  
            !} khx
            end do
  
  !        !} khy
  !        end do
          !}
          end if

        !} azim
        end do

        !}
        end if

!!     (i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and.(mod(i_kz-1,n_kz/8)==0)) then
!!     (i_kmy == 1 ).and. (i_kmx == 1).and. &
  if((.not. apply_jac_none).and. &
     (i_kmy == n_kmy/4).and. (i_kmx == n_kmx/4).and. &
     ((i_kz_thread==2).or. &
      (i_kz_thread==10).or. &
      (i_kz_thread==18).or. &
      (i_kz_thread==26).or. &
      (i_kz_thread==40).or. &
      (i_kz_thread==48).or. &
      (i_kz_thread==56).or. &
      (i_kz_thread==64) &
     )) then
!               call snap('jac_mat.h',n_khx,n_khy,jac_mat(:,:,i_thread))
!  write(0,*)'i_kz=',i_kz
!  write(0,*)'kkz_den=',kkz_den
!               call snap('jac_mat.h',n_khx,n_khy,jac_mat)
  end if

      !} thread
      end do
      !$OMP END PARALLEL


    if(apply_jac_none) then
      if(srite(trim(tag_output), &
           gamma_azim,n_bytes_gamma_azim)/= n_bytes_gamma_azim) then
          write(0,*)'tag_output=',trim(tag_output)
          call seperr('Problems with writing tag_output file')
      end if
    else
! possibel BUG with n_thread > 1
      if((i_kz >= 3).and.(i_kz <= n_kz-1)) then 
        if(srite(trim(tag_output), &
             gamma_azim,n_bytes_gamma_azim)/= n_bytes_gamma_azim) then
            write(0,*)'tag_output=',trim(tag_output)
            call seperr('Problems with writing tag_output file')
        end if
      else
        gamma_azim=0.
        if(srite(trim(tag_output), &
             gamma_azim,n_bytes_gamma_azim)/= n_bytes_gamma_azim) then
            write(0,*)'tag_output=',trim(tag_output)
            call seperr('Problems with writing tag_output file')
        end if
      end if
    end if

  !} kz
  end do

  !} kmx
  end do
  !} kmy
  end do
        

  status=0
      
  end function Off2Ang3D_oper


subroutine fft_2d_data_par(fft_id,sign,n1,n2,i_thread,data_par)
type (fft_type) fft_id
integer :: n1,n2,i_thread
integer :: sign
complex, dimension (:,:,:) :: data_par

call fft_2d_data(fft_id,sign,n1,n2,data_par(:,:,i_thread))
!
end subroutine fft_2d_data_par

subroutine reorder_fft_2d_data_par(sign,n1,n2,i_thread,data_par)
integer :: n1,n2,i_thread
integer :: sign
complex, dimension (:,:,:) :: data_par

integer n1d2,i1,n2d2,i2
complex, dimension (n1) :: temp1
complex, dimension (n2) :: temp2

if(n1 /= 1) then
  n1d2=n1/2
else
  n1d2=1
end if

if(n2 /= 1) then
  n2d2=n2/2
else
  n2d2=1
end if

if(sign==FFT_FORW) then

do i2=1,n2
  do i1=n1d2,n1
    temp1(i1)=data_par(i1-n1d2+1,i2,i_thread)
  end do
  do i1=1,n1d2-1
    temp1(i1)=data_par(n1d2+i1+1,i2,i_thread)
  end do

  do i1=1,n1
    data_par(i1,i2,i_thread)=temp1(i1)
  end do
end do

do i1=1,n1
  do i2=n2d2,n2
    temp2(i2)=data_par(i1,i2-n2d2+1,i_thread)
  end do
  do i2=1,n2d2-1
    temp2(i2)=data_par(i1,n2d2+i2+1,i_thread)
  end do

  do i2=1,n2
    data_par(i1,i2,i_thread)=temp2(i2)
  end do
end do

end if


end subroutine reorder_fft_2d_data_par

subroutine reorder_fft_wavenumber(sign,n1,k)
integer :: n1
integer :: sign
real, dimension (:) :: k

integer n1d2,i1
real, dimension (n1) :: temp

if(n1 /= 1) then
  n1d2=n1/2
else
  n1d2=1
end if

if(sign==FFT_FORW) then

  do i1=n1d2,n1
    temp(i1)=k(i1-n1d2+1)
  end do
  do i1=1,n1d2-1
    temp(i1)=k(n1d2+i1+1)
  end do

  do i1=1,n1
    k(i1)=temp(i1)
  end do


end if

end subroutine reorder_fft_wavenumber

function  comp_azim_bound(off2ang3d_par) result (azim_bound)
  type (off2ang3d_param_type) off2ang3d_par
  integer azim_bound(2,off2ang3d_par%n_gamma)


  integer power,i_gamma
!  real pi,azim_min,azim_max,azim_min_0,azim_max_0,d_azim_loc,gamma_max_max,conv
!  real azim_min,azim_max,azim_min_0,azim_max_0,azim_min_gamma_max,azim_max_gamma_max
  real azim_min,azim_max,azim_min_0,azim_max_0
  real inv_d_azim,gamma_max_max,conv
  real azim_bound_min,azim_bound_max,delta_gamma,gamma


  pi=acos(-1.)
  if(off2ang3d_par%azim_min_gamma_max /= 0.) then
    azim_min=off2ang3d_par%azim_min_gamma_max
  else
    azim_min=off2ang3d_par%azim_0
  end if
  if(off2ang3d_par%azim_max_gamma_max /= 0.) then
    azim_max=off2ang3d_par%azim_max_gamma_max
  else
    azim_max=off2ang3d_par%azim_0+(off2ang3d_par%n_azim-1)*off2ang3d_par%d_azim
  end if

  if(off2ang3d_par%azim_gamma_0 /= 0.) then
    azim_min_0=-off2ang3d_par%azim_gamma_0
    azim_max_0=off2ang3d_par%azim_gamma_0
  else
    azim_min_0=azim_min
    azim_max_0=azim_max
  end if
  inv_d_azim=1./off2ang3d_par%d_azim
  power=off2ang3d_par%sin_pow
  delta_azim_gamma_0=azim_max_0-azim_min_0

  write(0,*)'azim_min=',azim_min
  write(0,*)'azim_max=',azim_max
  write(0,*)'delta_azim_gamma_0=',delta_azim_gamma_0
  write(0,*)'inv_d_azim=',inv_d_azim

  gamma_max_max=90
  delta_gamma=gamma_max_max
  conv=pi/2
  do i_gamma=1,off2ang3d_par%n_gamma
    gamma=off2ang3d_par%gamma_0+(i_gamma-1)*off2ang3d_par%d_gamma

!    if(gamma >0.) then
      azim_bound_min=azim_min + &
        (azim_min_0-azim_min)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
      azim_bound_max=azim_max + &
        (azim_max_0-azim_max)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
      azim_bound(1,i_gamma)=(azim_bound_min-off2ang3d_par%azim_0_max)*inv_d_azim+1
      azim_bound(2,i_gamma)=(azim_bound_max-off2ang3d_par%azim_0_max)*inv_d_azim+1

if(off2ang3d_par%debug1) then
  write(0,*)'gamma=',gamma
  write(0,*)'azim_0_max=',off2ang3d_par%azim_0_max
  write(0,*)'azim_bound_min=',azim_bound_min
  write(0,*)'azim_bound_max=',azim_bound_max
  write(0,*)'azim_bound(1)=',azim_bound(1,i_gamma)
  write(0,*)'azim_bound(2)=',azim_bound(2,i_gamma)
end if

!    else
!      azim_bound_min=azim_min + &
!        (azim_max_0-azim_min)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
!      azim_bound_max=azim_max + &
!        (-azim_max_0-azim_max)*(sin(conv*(gamma_max_max-gamma)/delta_gamma)**power)
!      azim_bound(1,i_gamma)=(azim_bound_min-azim_0_max)/d_azim+1
!      azim_bound(2,i_gamma)=(azim_bound_max-azim_0_max)/d_azim+1
!    end if

   end do


end function comp_azim_bound


end module Off2Ang3D_mod
