!------------------------------------
program testgroup


use sep
implicit none
!-------------------------------------
integer					:: n1, i
real 					:: ax, delta, epsilon, V0, t0, rhovv, rhovh, rhovn, z0m
real                                    :: diff1, diff2
real                                    :: o1, d1
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:),allocatable          :: hs, hr, gammap, gammarp, gammarp2, gammasp, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm
 
!--------------------------------

call sep_init()

call from_param( "n1", n1) ! Angle axis
call from_param( "o1", o1)
call from_param( "d1", d1)

call from_param( "ax", ax) ! Geological dip

call from_param( "V0", V0) ! Velocity parameters (true and perturbed) 
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "t0", t0)

call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call sep_close()

! Allocation of the ADCIG where we plot the predicted RMO curve

vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

! Vector of the phase aperture angles
allocate(gammap(n1),gammarp(n1),gammasp(n1),gammag(n1))
allocate(betarp(n1),betarg(n1))
allocate(betasp(n1),betasg(n1))

do i=1,n1
   gammap(i)=o1+(i-1)
end do

! We first initialize the different aperture angles as equal
gammasp=gammap 
gammarp=gammap
betarp=ax+gammarp 
betasp=ax-gammasp

write(0,*) 'Initialization'
write(0,*) 'Gamma: ',gammarp(40), gammasp(40)

! To test whether the reflection angles are OK, we need the phase velocities (or slownesses). 

! Elliptical phase velocity at the different propagation angles 
allocate(Vellbetarp(n1),Vellbetasp(n1))
Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)

! VTI phase velocity at the different propagation angles
allocate(Vvtibetarp(n1),Vvtibetasp(n1))
Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)
write(0,*) dot_product(Vvtibetarp-Vvtibetasp,Vvtibetarp-Vvtibetasp)

allocate(hr(n1),hs(n1))
hr=sind(gammarp)/Vvtibetarp
hs=sind(gammasp)/Vvtibetasp

diff1=dot_product(hr-hs,hr-hs)/dot_product(hr,hr)
diff2=diff1
i=0
allocate(gammarp2(n1))
do while((diff2 .le. diff1) .or. diff2>1e-4)

!   gammarp2=asind(sind(gammasp)*Vvtibetarp/Vvtibetasp)
!   if (dot_product(gammarp2-gammarp,gammarp2-gammarp)< dot_product(90-gammarp2-gammarp,90-gammarp2-gammarp)) then 
!      gammarp=gammarp2
!   else 
!      gammarp=90-gammarp2
!   endif

   gammarp=asind(sind(gammasp)*Vvtibetarp/Vvtibetasp)
   gammasp=2*gammap-gammarp
    write(0,*) 'Gamma: ', gammap(40), gammarp(40), gammasp(40)
   betarp=ax+gammarp
   betasp=ax-gammasp
    write(0,*) 'Beta: ', betarp(40), betasp(40)
   hr=sind(gammarp)/Vvtibetarp
   hs=sind(gammasp)/Vvtibetasp
   diff1=diff2
   diff2=dot_product(hr-hs,hr-hs)/dot_product(hr,hr)
   Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
   Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)
   Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
   Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)
    write(0,*) Vvtibetasp(40), Vvtibetarp(40)  
    write(0,*) 'Loop ', i, ' done, norm equals ', diff2
   i=i+1
end do

! Once the two aperture angles have been computed, we have to compute the group aperture angles

! Derivative of the elliptic phase velocity with respect to the phase angle
allocate(DVellbetarp(n1),DVellbetasp(n1))
DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)

! Derivative of the VTI phase velocity with respect to the phase angle
allocate(DVvtibetarp(n1), DVvtibetasp(n1))
DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))

! Computation of the group propagation angles
betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))

! Computation of the group aperture angle
gammag=0.5*(betarg-betasg)

!deallocate(betarp,betarg,betasp,betasg,Vellbetarp,Vellbetasp,Vvtibetarp,Vvtibetasp,DVellbetarp,DVellbetasp,DVvtibetarp,DVvtibetasp)

! COMPUTATION OF THE RESIDUAL MOVEOUT

! Elliptic group slowness (the input angle is the group angle)
!allocate(Sellg(n1),Svtig(n1))
!Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)
! VTI group slowness (the input an1le is the group angle)
!Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
!allocate(DSellg1(n1), DSellg2(n1))
!DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
!DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg

! Computation of the derivatives of the VTI group slownesses
!allocate(DSvtig1(n1),DSvtig2(n1),DSvtig3(n1))
!DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**2+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
!DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**2+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
!DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**2+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))

! Computation of the derivative of the depth with respect to the different perturbations
!allocate(DZ1(n1),DZ2(n1),DZ3(n1))
!DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
!DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
!DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))

!deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

! We compute the theoretical ADCIG
!z0m=t0*vvm;
!allocate(rmo(n1),zm(n1))
!rmo=z0m*(DZ1*(rhovv-1)+DZ2*(rhovh-1)+DZ3*(rhovn-1)) ! Computation of the theoretical RMO
!zm=z0m+rmo
!do i=1,n1
!   write(0,*) zm(i)
!end do
 
!call sep_write(zm)

end program testgroup
