!------------------------------------
program velocity_spectra


use sep
implicit none
!-------------------------------------
integer					:: n1, n2, n3, n4, ng, ng2, i1, i2, i3, i4, ig, ig2, j, k, j1 ,j2, j3, j4
real 					:: delta, epsilon, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp, vv, vh, vn
real                                    :: o1, o2, o3, o4, og, og2, d1, d2, d3, d4, dg, dg2
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:,:,:,:,:), allocatable :: adcig
real, dimension(:,:,:,:), allocatable   :: vspectra
real, dimension(:),allocatable          :: gammap, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm

!--------------------------------

call sep_init()

call from_param( "n1", n1)
call from_param( "o1", o1)
call from_param( "d1", d1)
call from_param( "ng", ng)
call from_param( "og", og)
call from_param( "dg", dg)
call from_param( "ng2", ng2)
call from_param( "og2", og2)
call from_param( "dg2", dg2)
call from_param( "n2", n2)
call from_param( "o2", o2)
call from_param( "d2", d2)
call from_param( "n3", n3)
call from_param( "o3", o3)
call from_param( "d3", d3)
call from_param( "n4", n4)
call from_param( "o4", o4)
call from_param( "d4", d4)
call from_param( "V0", V0)
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call sep_close()

! Allocation of the ADCIG
allocate(adcig(n1,ng,1,1,1))
call sep_read(adcig,'adcig')

! Allocation of the velocity spectra cube
allocate(vspectra(n1,n2,n3,n4))
vspectra=0

! Definition of the velocities
! True velocities
vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);

! Modified velocities
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

write(0,*) vvm
write(0,*) vhm
write(0,*) vnm

! Definition of the vector of aperture angles
allocate(gammap(ng2),gammag(ng2))
allocate(betarp(ng2),betarg(ng2))
allocate(betasp(ng2),betasg(ng2))

do ig2=1,ng2
   gammap(ig2)=og2+dg2*(ig2-1)
   betarp(ig2)=gammap(ig2) ! Only true for flat reflectors
   betasp(ig2)=-gammap(ig2) ! Only true for flat reflectors
end do

! Computation of the different quantities used to compute the derivatives of the depth with respect to pertubations.

! Elliptical phase velocity at the different propagation angles 
allocate(Vellbetarp(ng2),Vellbetasp(ng2))
Vellbetarp=sqrt(vvm**2*cosd(betarp)**2+vhm**2*sind(betarp)**2)
Vellbetasp=sqrt(vvm**2*cosd(betasp)**2+vhm**2*sind(betasp)**2)

! VTI phase velocity at the different propagation angles
allocate(Vvtibetarp(ng2),Vvtibetasp(ng2))
Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2))/2)
Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2))/2)

! Derivative of the elliptic phase velocity with respect to the phase angle
allocate(DVellbetarp(ng2),DVellbetasp(ng2))
DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhm**2-vvm**2)
DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhm**2-vvm**2)

! Derivative of the VTI phase velocity with respect to the phase angle
allocate(DVvtibetarp(ng2), DVvtibetasp(ng2))
DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhm**2-vvm**2)+1/sqrt(Vellbetarp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betarp)*cosd(2*betarp)))
DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhm**2-vvm**2)+1/sqrt(Vellbetasp**4+vvm**2*(vnm**2-vhm**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhm**2-vvm**2)+2*vvm**2*(vnm**2-vhm**2)*sind(2*betasp)*cosd(2*betasp)))

! Computation of the group propagation angles
betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))

! Computation of the group aperture angle
gammag=0.5*(betarg-betasg)

deallocate(betarp,betarg,betasp,betasg,Vellbetarp,Vellbetasp,Vvtibetarp,Vvtibetasp,DVellbetarp,DVellbetasp,DVvtibetarp,DVvtibetasp)

! COMPUTATION OF THE RESIDUAL MOVEOUT

! Elliptic group slowness (the input angle is the group angle)
allocate(Sellg(ng2),Svtig(ng2))
Sellg=sqrt((1/vvm)**2*cosd(gammag)**2+(1/vhm)**2*sind(gammag)**2)

! VTI group slowness (the input angle is the group angle)
Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
allocate(DSellg1(ng2), DSellg2(ng2))
DSellg1=-(1/vvm)**2*cosd(gammag)**2/Sellg
DSellg2=-(1/vhm)**2*sind(gammag)**2/Sellg

! Computation of the derivatives of the VTI group slownesses
allocate(DSvtig1(ng2),DSvtig2(ng2),DSvtig3(ng2))
DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))
DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammag)**2))

! Computation of the derivative of the depth with respect to the different perturbations
allocate(DZ1(ng2),DZ2(ng2),DZ3(ng2))
DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))

do j=1,ng2
   write(0,*) DZ1(j)
end do
write(0,*) '************'
do j=1,ng2
   write(0,*) DZ2(j)
end do
write(0,*) '************'
do j=1,ng2
   write(0,*) DZ3(j)
end do
 
deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

! Actual computation of the velocity spectra
allocate(rmo(ng2))
do i1=171,205
   !do i1=150,250 ! Loop over the different traveltimes
   !  write(0,*) i1," out of ",n1
   z0m=o1+(i1-1)*d1
   
   do i2=1,n2 ! Loop over the different values of VERTICAL perturbations
      vv=(o2+(i2-1)*d2)
      do i3=1,n3 ! Loop over the different values of the HORIZONTAL perturbations
         vn=(o3+(i3-1)*d3)
         do i4=1,n4 ! Loop over the different values of the NMO velocity perturbations
            vh=vn*sqrt(1+2*(o4+(i4-1)*d4))
            drhovv=vv/vvm-1
            drhovh=vh/vhm-1
            drhovn=vn/vnm-1
            
            ! Computation of the residual moveout for the different perturbation values
            rmo=z0m*(DZ1*drhovv+DZ2*drhovh+DZ3*drhovn) ! That is a vector 
            rmo=(rmo-o1)/d1+1.5
            
            ! Computation of the semblance values
            do ig2=1,ng2
               if (i1+rmo(ig2)>0 .AND. i1+rmo(ig2)<n1+1) vspectra(i1,i2,i3,i4)=vspectra(i1,i2,i3,i4)+adcig(i1+rmo(ig2),ig2,1,1,1)
            end do
         end do
      end do
   end do
end do

call sep_write(vspectra)

deallocate(adcig)

end program velocity_spectra
