!------------------------------------
program velspecphase


use sep
implicit none
!-------------------------------------
integer					:: n1, n2, n3, n4, ng, i1, i2, i3, i4, ig, j, k
real 					:: delta, epsilon, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp
real                                    :: o1, o2, o3, o4, og, d1, d2, d3, d4, dg
real                                    :: vvt, vht, vnt, vvm, vhm, vnm
real, dimension(:,:,:,:,:), allocatable :: adcig
real, dimension(:,:,:,:), allocatable   :: vspectra
real, dimension(:), allocatable         :: gammap
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm

!--------------------------------

call sep_init()

call from_param( "n1", n1)
call from_param( "o1", o1)
call from_param( "d1", d1)
call from_param( "ng", ng)
call from_param( "og", og)
call from_param( "dg", dg)
call from_param( "n2", n2)
call from_param( "o2", o2)
call from_param( "d2", d2)
call from_param( "n3", n3)
call from_param( "o3", o3)
call from_param( "d3", d3)
call from_param( "n4", n4)
call from_param( "o4", o4)
call from_param( "d4", d4)
call from_param( "V0", V0)
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call sep_close()

allocate(adcig(n1,ng,1,1,1))
call sep_read(adcig,'adcig')

allocate(vspectra(n1,n2,n3,n4))
vspectra=0

vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

allocate(gammap(ng))
do ig=1,ng
   gammap(ig)=og+dg*(ig-1)
end do


! COMPUTATION OF THE RESIDUAL MOVEOUT USING THE PHASE ANGLE

! Elliptic group slowness (the input angle is the phase angle - wrong angle)
allocate(Sellg(ng),Svtig(ng))
Sellg=sqrt((1/vvm)**2*cosd(gammap)**2+(1/vhm)**2*sind(gammap)**2)
! VTI group slowness (the input angle is the phase angle - wrong angle)
Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))/2)

! Computation of the derivatives of the elliptic group slownesses (input angle is still the phase angle - wrong angle)
allocate(DSellg1(ng), DSellg2(ng))
DSellg1=-(1/vvm)**2*cosd(gammap)**2/Sellg
DSellg2=-(1/vhm)**2*sind(gammap)**2/Sellg

! Computation of the derivatives of the VTI group slownesses (using the phase angle - wrong one)
allocate(DSvtig1(ng),DSvtig2(ng),DSvtig3(ng))
DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))
DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvm)**2*(1/vhm)**2*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))
DSvtig3=-((1/vvm)**2*(1/vhm)**2*sind(2*gammap)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvm)**2*((1/vnm)**2-(1/vhm)**2)*sind(2*gammap)**2))

! Computation of the derivative of the depth with respect to the different perturbations
allocate(DZ1(ng),DZ2(ng),DZ3(ng))
DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammap))-1
DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammap))
DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammap))

deallocate(Sellg,Svtig,DSellg1,DSellg2,DSvtig1,DSvtig2,DSvtig3)

! Actual computation of the velocity spectra
allocate(rmo(ng))
do i1=179,205
!do i1=150,250 ! Loop over the different traveltimes
 !  write(0,*) i1," out of ",n1
   z0m=o1+(i1-1)*d1
   
   do i2=1,n2 ! Loop over the different values of VERTICAL perturbations
     drhovv=(o2+(i2-1)*d2)
      
     do i3=1,n3 ! Loop over the different values of the HORIZONTAL perturbations
         drhovh=(o3+(i3-1)*d3)
         
         do i4=1,n4 ! Loop over the different values of the NMO velocity perturbations
            drhovn=(o4+(i4-1)*d4)
            
            ! Computation of the residual moveout for the different perturbation values
            rmo=z0m*(DZ1*drhovv+DZ2*drhovh+DZ3*drhovn) ! That is a vector 
            rmo=(rmo-o1)/d1+1.5
            
            ! Computation of the semblance values
            do ig=1,ng
             !  write(0,*) rmo(ig)
            if (i1+rmo(ig)>0 .AND. i1+rmo(ig)<n1+1) vspectra(i1,i2,i3,i4)=vspectra(i1,i2,i3,i4)+adcig(i1+rmo(ig),ig,1,1,1)
            end do 
         end do
      end do
   end do
end do

call sep_write(vspectra)

deallocate(adcig)

end program velspecphase
