!------------------------------------
program velocity_spectra


use sep
implicit none
!-------------------------------------
integer					:: nt, ng, n1, n2, n3, it, ig, i1, i2, i3, j, k
real 					:: delta, epsilon, V0, rhovv, rhovh, rhovn, z0m, drhovv, drhovh, drhovn, temp
real                                    :: ot, og, o1, o2, o3, dt, dg, d1, d2, d3
real                                    :: vvt, vht, vnt, vvm, vhm, vnm, vvtest, vhtest, vntest
real, dimension(:,:), allocatable       :: adcig
real, dimension(:,:,:), allocatable     :: intvspectra, vspectra, denum
real, dimension(:),allocatable          :: gammap, gammag, betarp, betarg, betasp, betasg 
real, dimension(:), allocatable         :: Vellbetarp, Vellbetasp, Vvtibetarp,Vvtibetasp, DVellbetarp, DVellbetasp, DVvtibetarp, DVvtibetasp
real, dimension(:), allocatable         :: Sellg, Svtig, DSellg1, DSellg2, DSvtig1, DSvtig2, DSvtig3, DZ1, DZ2, DZ3, rmo, zm
integer, dimension(:), allocatable     :: rmo2

!--------------------------------

call sep_init()

! Parameters of the initial ADCIG
call from_aux( "adcig", "n1", nt)
call from_aux( "adcig", "o1", ot)
call from_aux( "adcig", "d1", dt)
!call from_aux( "adcig", "n2", ng)
!call from_aux( "adcig", "o2", og)
!call from_aux( "adcig", "d2", dg)

call from_aux( "adcig", "n2", n1)
call from_aux( "adcig", "o2", o1)
call from_aux( "adcig", "d2", d1)

call from_param( "it", it)

! Parameters of the semblance cube
!call from_param( "n1", n1) ! Additional
!call from_param( "o1", o1) ! Additional
!call from_param( "d1", d1) ! Additional

call from_param( "n2", n2)
call from_param( "o2", o2)
call from_param( "d2", d2)
call from_param( "n3", n3)
call from_param( "o3", o3)
call from_param( "d3", d3)

! Parameters of the velocity model
call from_param( "V0", V0)
call from_param( "delta", delta)
call from_param( "epsilon", epsilon)
call from_param( "rhovv", rhovv)
call from_param( "rhovh", rhovh)
call from_param( "rhovn", rhovn)

call to_history("n1", n1)
call to_history("o1", o1)
call to_history("d1", d1)

write(0,*) "n1", n1
write(0,*) "o1", o1
write(0,*) "d1", d1

write(0,*) "n2", n2
write(0,*) "n3", n3

call sep_close()

! Allocation of the ADCIG
allocate(adcig(nt,n1))
!allocate(adcig(nt,ng))
call sep_read(adcig,'adcig')

! Allocation of the velocity spectra cube

allocate(vspectra(n1,n2,n3))
vspectra=0
allocate(denum(n1,n2,n3))
denum=0
allocate(intvspectra(n1,n2,n3))
intvspectra=0

! Definition of the velocities
! True velocities
vvt=V0;
vht=V0*sqrt(1+2*epsilon);
vnt=V0*sqrt(1+2*delta);

! Modified velocities
vvm=vvt*rhovv;
vhm=vht*rhovh;
vnm=vnt*rhovn;

write(0,*) vvm
write(0,*) vhm
write(0,*) vnm

! Definition of the vector of aperture angles
allocate(gammap(n1),gammag(n1))
allocate(betarp(n1),betarg(n1))
allocate(betasp(n1),betasg(n1))

do i1=1,n1
   gammap(i1)=o1+d1*(i1-1)
   betarp(i1)=gammap(i1) ! Only true for flat reflectors
   betasp(i1)=-gammap(i1) ! Only true for flat reflectors
end do

! Computation of the different quantities used to compute the derivatives of the depth with respect to pertubations.

! Elliptical phase velocity at the different propagation angles 
allocate(Vellbetarp(n1),Vellbetasp(n1))
allocate(Vvtibetarp(n1),Vvtibetasp(n1))
allocate(DVellbetarp(n1),DVellbetasp(n1))
allocate(DVvtibetarp(n1), DVvtibetasp(n1))
allocate(Sellg(n1),Svtig(n1))
allocate(DSellg1(n1), DSellg2(n1))
allocate(DSvtig1(n1),DSvtig2(n1),DSvtig3(n1))
allocate(DZ1(n1),DZ2(n1),DZ3(n1))


!write(0,*) "Gammag:",  gammag
!write(0,*) "Gammap:",  gammap

! Actual computation of the velocity spectra
allocate(rmo(n1),rmo2(n1))
!do i1=171,205
   !do i1=150,250 ! Loop over the different traveltimes
   !  write(0,*) i1," out of ",n1
z0m=ot+(it-1)*dt

vvtest=vvm

write(0,*) 'Maximum value: ', adcig(it,1)

do i2=1,n2 ! Loop over the different values of the HORIZONTAL perturbations
   !drhovh=vht/(o2+(i2-1)*d2)-1
   ! Last one
   vhtest=vhm/((o2+(i2-1)*d2)*rhovh)
   drhovh=(o2+(i2-1)*d2)*rhovh-1
   !drhovh=(o2+(i2-1)*d2)-1
!            write(0,*) 'drhovh', drhovh       
   !         i4=31
   do i3=1,n3 ! Loop over the different values of the NMO velocity perturbations
      !drhovn=vnt/(o3+(i3-1)*d3)-1
      ! Last one
      vntest=vnm/((o3+(i3-1)*d3)*rhovn)
      drhovn=(o3+(i3-1)*d3)*rhovn-1
      !drhovn=(o3+(i3-1)*d3)-1
!                  write(0,*) 'drhovn', drhovn
      
      
      ! **************************** Computation of the group angle **************************************************************
      Vellbetarp=sqrt(vvtest**2*cosd(betarp)**2+vhtest**2*sind(betarp)**2)
      Vellbetasp=sqrt(vvtest**2*cosd(betasp)**2+vhtest**2*sind(betasp)**2)
      
      ! VTI phase velocity at the different propagation angles
      Vvtibetarp=sqrt((Vellbetarp**2+sqrt(Vellbetarp**4+vvtest**2*(vntest**2-vhtest**2)*sind(2*betarp)**2))/2)
      Vvtibetasp=sqrt((Vellbetasp**2+sqrt(Vellbetasp**4+vvtest**2*(vntest**2-vhtest**2)*sind(2*betasp)**2))/2)
      
      ! Derivative of the elliptic phase velocity with respect to the phase angle
      DVellbetarp=sind(2*betarp)/(2*Vellbetarp)*(vhtest**2-vvtest**2)
      DVellbetasp=sind(2*betasp)/(2*Vellbetasp)*(vhtest**2-vvtest**2)
      
      ! Derivative of the VTI phase velocity with respect to the phase angle
      DVvtibetarp=1/(4*Vvtibetarp)*(sind(2*betarp)*(vhtest**2-vvtest**2)+1/sqrt(Vellbetarp**4+vvtest**2*(vntest**2-vhtest**2)*sind(2*betarp)**2)*(Vellbetarp**2*sind(2*betarp)*(vhtest**2-vvtest**2)+2*vvtest**2*(vntest**2-vhtest**2)*sind(2*betarp)*cosd(2*betarp)))
      DVvtibetasp=1/(4*Vvtibetasp)*(sind(2*betasp)*(vhtest**2-vvtest**2)+1/sqrt(Vellbetasp**4+vvtest**2*(vntest**2-vhtest**2)*sind(2*betasp)**2)*(Vellbetasp**2*sind(2*betasp)*(vhtest**2-vvtest**2)+2*vvtest**2*(vntest**2-vhtest**2)*sind(2*betasp)*cosd(2*betasp)))
      
      ! Computation of the group propagation angles
      betarg=atand((tand(betarp)+DVvtibetarp/Vvtibetarp)/(1-tand(betarp)*DVvtibetarp/Vvtibetarp))
      betasg=atand((tand(betasp)+DVvtibetasp/Vvtibetasp)/(1-tand(betasp)*DVvtibetasp/Vvtibetasp))
      
      ! Computation of the group aperture angle
      gammag=0.5*(betarg-betasg)

      ! *************************** Computation of the derivatives **********************************************************
      
      ! Elliptic group slowness (the input angle is the group angle)
      Sellg=sqrt((1/vvtest)**2*cosd(gammag)**2+(1/vhtest)**2*sind(gammag)**2)

      ! VTI group slowness (the input angle is the group angle)
      Svtig=sqrt((Sellg**2+sqrt(Sellg**4+(1/vvtest)**2*((1/vntest)**2-(1/vhtest)**2)*sind(2*gammag)**2))/2)

      ! Computation of the derivatives of the elliptic group slownesses (input angle is still the group angle)
      DSellg1=-(1/vvtest)**2*cosd(gammag)**2/Sellg
      DSellg2=-(1/vhtest)**2*sind(gammag)**2/Sellg

      ! Computation of the derivatives of the VTI group slownesses
      DSvtig1=Sellg/(2*Svtig)*DSellg1+(2*DSellg1*Sellg**3-(1/vvtest)**2*((1/vntest)**2-(1/vhtest)**2)*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvtest)**2*((1/vntest)**2-(1/vhtest)**2)*sind(2*gammag)**2))
      DSvtig2=Sellg/(2*Svtig)*DSellg2+(2*DSellg2*Sellg**3+(1/vvtest)**2*(1/vhtest)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvtest)**2*((1/vntest)**2-(1/vhtest)**2)*sind(2*gammag)**2))
      DSvtig3=-((1/vvtest)**2*(1/vhtest)**2*sind(2*gammag)**2)/(4*Svtig*sqrt(Sellg**4+(1/vvtest)**2*((1/vntest)**2-(1/vhtest)**2)*sind(2*gammag)**2))

      ! Computation of the derivative of the depth with respect to the different perturbations
      DZ1=-1/Svtig*DSvtig1*(1+tand(gammap)*tand(gammag))-1
      DZ2=-1/Svtig*DSvtig2*(1+tand(gammap)*tand(gammag))
      DZ3=-1/Svtig*DSvtig3*(1+tand(gammap)*tand(gammag))
      
      ! Computation of the residual moveout for the different perturbation values
      
      !                  write(0,*) '************'
      !                  write(0,*) "DZ2:", DZ2
      !                  write(0,*) '************'
      !                  write(0,*) "DZ3:", DZ3
      
      rmo=z0m*(DZ2*drhovh+DZ3*drhovn) ! That is a vector 
      !         write(0,*) "rmo:", rmo
      rmo2=(rmo-ot)/dt+0.5
      !write(0,*) rmo2
      ! Computation of the semblance values
      ! vspectra(1,i2,i3)=adcig(it+rmo2(1),1) !Initial
      intvspectra(1,i2,i3)=adcig(it+rmo2(1),1)
      denum(1,i2,i3)=adcig(it+rmo2(1),1)**2
      vspectra(1,i2,i3)=intvspectra(1,i2,i3)**2/denum(1,i2,i3)
      
         do i1=2,n1
!            if (it+rmo2(i1)>0 .AND. it+rmo2(i1)<nt+1) then
               intvspectra(i1,i2,i3)=intvspectra(i1-1,i2,i3)+adcig(it+rmo2(i1),i1)!Initial
               denum(i1,i2,i3)=denum(i1-1,i2,i3)+adcig(it+rmo2(i1),i1)**2
!            end if
            vspectra(i1,i2,i3)=intvspectra(i1,i2,i3)**2/(i1*denum(i1,i2,i3))
!            write(0,*) "Value: ", vspectra(i1,i2,i3) 
!            write(0,*) 'RMO:', rmo2(i1),'Adcig:', adcig(it+rmo2(i1),i1)
         end do
      end do
   end do
!end do

!do ig2=1,ng2
!   write(0,*) adcig(179,ig2,1,1,1)
!end do


call sep_write(vspectra)

deallocate(adcig)

end program velocity_spectra
